/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017 .. 2019,2021,2022 Harald Foerster	*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#ifndef CFF_FFMPEG__H
#define CFF_FFMPEG__H

#include <stdint.h>
#include <libavcodec/avcodec.h>
#include <libavcodec/bsf.h>
#include <libavformat/avformat.h>


/*
	Callback for 'cff_ffmpeg_file_open' or 'cff_ffmpeg_program_open'
	call from your function 'cff_ffmpeg_read_keyframes' till
	the return value is zero.
*/

typedef int (func_cff_callback_t)(void *user_data);


/*
	Callback for 'cff_ffmpeg_log_callback'
	If the return value is not zero, the callback will be stopped.
*/

typedef int (func_log_callback_t)(const char *str, int len, void *user_data);


/*
	Callback for 'cff_ffmpeg_copy_softcut'
	If the return value is not zero, copying will be stopped.
*/

typedef struct
{
	int	decoded;	/* Number of decoded frames (-1 => text off) */
	int	encoded;	/* Like above for encoded frames */
	void	*func_data;	/* Data for progressbar, mutex and so on */

} cff_softcut_pbar_t;

typedef int (func_softcut_callback_t)(cff_softcut_pbar_t *softcut_data);


#define kCFFMediaVideo	0
#define kCFFMediaAudio	1
#define kCFFMediaOther	2
#define kCFFMediaTotal	3
#define kCFFMediaAll	3


typedef struct
{
	const char	*name;
	char		*codecs;

} cff_fmtlist_t;

typedef struct
{
	const char	*name;
	char		*codecs;

} cff_bsflist_t;

enum cff_FrameFlags
{
	kKeyFrameFlag	= 1 << 0,
	kIFrameFlag	= 1 << 1,
	kPFrameFlag	= 1 << 2
};

enum cff_VideoFrame
{
	kVideoFramePrevCut,
	kVideoFramePrev05Sec,
	kVideoFramePrevHalfSec,
	kVideoFramePrevSingle,
	kVideoFrameNextSingle,
	kVideoFrameNextHalfSec,
	kVideoFrameNext05Sec,
	kVideoFrameNextCut,

	kVideoFrameStart,
	kVideoFramePrev10Min,
	kVideoFramePrev05Min,
	kVideoFramePrev01Min,
	kVideoFrameNext01Min,
	kVideoFrameNext05Min,
	kVideoFrameNext10Min,
	kVideoFrameEnd,

	kVideoFrameShiftAdd = kVideoFrameStart
};


#define kCFFStreamFlagGotStart		(1 << 0)
#define kCFFStreamFlagGotOffset		(1 << 1)
#define kCFFStreamFlagTsFromDuration	(1 << 2)
#define kCFFStreamFlagTsRescale		(1 << 3)
#define kCFFStreamFlagBsfInit		(1 << 4)
#define kCFFStreamFlagBsfFlush		(1 << 5)
#define kCFFStreamFlagBsfDone		(1 << 6)
#define kCFFStreamFlagDvbSub		(1 << 7)
/* Stream-sync */
#define kCFFStreamFlagCopy		(1 << 8)
#define kCFFStreamFlagCopyDone		(1 << 9)

struct StreamData
{
	AVStream	*iStream;
	AVStream	*oStream;
	AVBSFContext	*bsf_ctx;
	unsigned int	flags;
	unsigned int	nbPackets;
	uint64_t	nbBytesWritten;
	int64_t		duration;
	int64_t		pkt_dts;
	int64_t		dts_offset;
	int64_t		delta_last;
	/* Stream-sync */
	int64_t		dts_start;
	int64_t		dts_stop;
};

typedef struct
{
	unsigned int	max;		/* Max. entries */
	unsigned int	num;		/* Number of entries */
	unsigned int	key_frames;	/* Number of key frames */
	AVPacket	**packets;	/* Array of entries */

} cff_pktqueue_t;


#define kCFFSoftcutFlagLibx264	(1 << 0)
#define kCFFSoftcutFlagLibx265	(1 << 1)

typedef struct
{
	AVCodecContext		*oc;		/* Encoder needed for soft cutting */
	const AVCodec		*codec;		/* Codec needed for encoding */
	func_softcut_callback_t	*func;		/* Callback */
	void			*func_data;	/* And the data for it */
	struct StreamData	*vst;		/* Pointer to the video data */
	AVFrame			*prev_frame;	/* Previous decoded frame */
	int64_t			dts_next;	/* Next start sequence */
	int64_t			pts_lastdec;	/* Pts of last decoded frame */
	int64_t			dts_stop;	/* Last packet in queue 3 */
	int64_t			duration_last;	/* Last duration */
	int64_t			pktdts_last;	/* Last dts of written packet */
	unsigned int		flags;		/* Encoder */
	unsigned int		gap_num;	/* Gap currently in work */
	unsigned int		queue_idx;	/* Queue in use */
	cff_pktqueue_t		pktqueue[4];	/* Stored video packets */

} cff_softcut_t;


#define kCFFCopyFlagTsKeepStart		(1 << 0) /* Keep original start timestamps */
#define kCFFCopyFlagStreamSync		(1 << 1) /* Try better synchron */
#define kCFFCopyFlagSeek		(1 << 2) /* Goto next sequence */
#define kCFFCopyFlagEOF			(1 << 3) /* End of file or copying */

typedef struct
{
	AVFormatContext		*outFormatCtx;
	int64_t			dts_start;	/* DTS of first video packet */
	int64_t			dts_beg;	/* Video start sequence */
	int64_t			dts_end;	/* Video end sequence */
	int64_t			dts_next;	/* Next stop */
	int64_t			dts_finish;	/* Start of next sequence (Stream-sync) */
	int64_t			gap_offset;	/* Offset of skipped video */
	unsigned int		cutidx;		/* Current index of cuttings */
	int			nb_streams;	/* Number of streams */
	unsigned int		flags;
	cff_softcut_t		softcut;
	struct StreamData	**ast;		/* Array of pointers */
	struct StreamData	*streamData;

} cff_copy_t;

/*
	Packet and frame dts may differ, find the packet
	dts by the matching packet and frame pos.
*/

#define kCFFPacketArrayEntries	50

typedef struct
{
	int64_t	dts;
	int64_t	pos;

} cff_packet_t;

typedef struct
{
	int64_t frame;	/* Decoded frame */
	int64_t packet;	/* Undecoded frame */
	int64_t	pktpos;	/* Position */	

} cff_dts_t;

typedef struct
{
	unsigned int	num;		/* Number of entries */
	unsigned int	idx;		/* Index of next dts (plus one) */
	int64_t		pkt_dts[1];	/* Timestamp of flushed frames */

} cff_drain_t;

#define kCFFUndoArrayEntries	20

enum cff_UndoAction
{
	kCutpointAdd,
	kCutpointRemove
};

typedef struct
{
	enum cff_UndoAction	action;
	unsigned int		num;	/* Number of cuttings */
	cff_dts_t		dts[1];	/* Undo or redo */

} cff_undo_t;

typedef struct
{
	uint8_t		*data;
	int		width;
	int		height;

} cff_rgb24_t;

typedef struct
{
	AVCodecContext	*ic;
	const char	*bsf_mp4toannexb; /* Softcut: enable bitstream filter */
	AVPacket	*tmppkt;	/* Packet used in funcs */
	AVFrame		*avframe;	/* Decoded frame */
	cff_rgb24_t	rgb24;		/* Rgb24 picture */
	int64_t		io_seek_start;	/* First frame (avio_tell) */
	int64_t		io_seek_end;	/* End of file (avio_tell) */
	int64_t		pktpos_start;	/* Position of first packet */
	unsigned int	stream_index;	/* Index of video stream */
	unsigned int	min_gop_size;	/* Min gop size (dts duration) */
	unsigned int	max_gop_size;	/* Max gop size (dts duration) */
	int64_t		stream_duration;/* Duration of stream */
	int64_t		frame_duration;	/* Duration per frame */
	cff_dts_t	current;	/* Current position */
	int64_t		dts_start;	/* Start time */
	int64_t		dts_stop;	/* Stop time */
	int64_t		dts_lastgop;	/* Start time of the last gop */
	cff_drain_t	*drain;		/* Timestamps in drain mode */

	/* Packet 'dts' and 'pos' values */
	cff_packet_t	packet[kCFFPacketArrayEntries];

	cff_undo_t	*undo[kCFFUndoArrayEntries];
	cff_undo_t	*redo[kCFFUndoArrayEntries];

	const char	*pict_type;	/* 'I', 'P', 'B' .. */
	unsigned int	frame_width;	/* Width of decoded frame */
	unsigned int	frame_height;	/* Height of decoded frame */
	unsigned int	frame_aspect[2];/* Aspect ratio of decoded frame */
	unsigned int	frame_total;	/* Number of frames */
	unsigned int	frame_current;	/* Current decoded frame */
	unsigned int	time_total[4];	/* Duration (h:m:s.us) */
	unsigned int	time_current[4];/* Time of decoded frame (h:m:s.us) */

} cff_vframe_t;

typedef struct
{
	cff_dts_t	*list;	/* Array of cuttings */
	unsigned int	max;	/* Max entries of the list */
	unsigned int	num;	/* Number of cuttings */

} cff_cut_t;

typedef struct
{
	unsigned int	*index;		/* Index list of video programs */
	unsigned int	num;		/* Number of programs */
	unsigned int	selected;	/* Current program */

} cff_program_t;

typedef struct
{
	AVFormatContext	*inFormatCtx;
	char		*fname;		/* Name of video file */
	int64_t		fsize;		/* Video file size */
	int		nb_streams;	/* Number of streams */
	int		dirty;		/* Video or cuttings not saved */
	cff_program_t	program;	/* Programs */
	cff_vframe_t	vframe;		/* The video frame */
	cff_cut_t	cut;		/* Cuttings */
	cff_copy_t	copy;		/* Copy stream(s) */

} cff_video_t;

typedef struct
{
	const char	*bsf_name;		/* Filter to use */
	int		ts_from_duration;	/* Compute timestamps from duration */

} cff_media_t;

typedef struct
{
	const char		*fname;		/* Name of output file */
	const char		*format_name;	/* Muxer to use */
	const char		*encoder_name;	/* Softcut => decode / encode */
	int			*stream_idx;	/* List of stream id's */
	int			nb_streams;	/* Number of streams */
	int			shortest;	/* Finish with shortest stream duration */
	int			ts_keep_start;	/* Keep original start timestamps */
	int			stream_sync;	/* For better synchron */
	cff_media_t		media[kCFFMediaTotal];
	char			**info;		/* Storage for result messsage */
	func_cff_callback_t	*func;		/* Callback ... */
	void			*user_data;	/* ... and the data for it */

	struct
	{
		int			both_gops; /* Encode both (end/start) GOPs */
		func_softcut_callback_t	*func;	/* Callback ... */
		void			*data;	/* ... and the data for it */

	} softcut;

} cff_copy_param_t;

typedef struct
{
	const char		*fname;		/* Name of image file */
	const char		*format_name;	/* The image format */
	int			width;		/* New width of the image */
	int			height;		/* New height of the image */
	int			keep_dimension;	/* Keep width and height */
	int			keep_ratio;	/* Keep aspect ratio */
	int			gray;		/* Gray */
	char			**info;		/* Storage for result messsage */

} cff_image_param_t;

extern cff_video_t gVideo;

char *cff_ffmpeg_get_libs_info(void);
char *cff_ffmpeg_get_formats_info(void);
char *cff_ffmpeg_get_codecs_info(void);
char *cff_ffmpeg_get_bitstream_filters_info(void);

cff_fmtlist_t  *cff_ffmpeg_get_format_names(cff_video_t *video);
cff_bsflist_t  *cff_ffmpeg_get_bitstream_filter_names(cff_video_t *video, int media_type);

/* Returns number of formats and codecs, '0' if none was found */
int   cff_ffmpeg_get_image_format_names(char ***long_name, char ***format_name);
/* Returns number of encoders, '0' if none was found */
int   cff_ffmpeg_get_encoder_names(cff_video_t *video, char ***encoder_names);
/* Returns extension (e.g. '.ts') and length (e.g. '3'), NULL if none was found */
char* cff_ffmpeg_get_outformat_extension(const char *format_name, int *slen);

int         cff_ffmpeg_get_current_loglevel_index(void);
const char *cff_ffmpeg_get_loglevel_name(int logindex);
int         cff_ffmpeg_set_loglevel(const char *logname, int lock, func_log_callback_t func, void *user_data);

int cff_ffmpeg_init(const char *logname);

int         cff_ffmpeg_file_open_cb(cff_video_t *video, double *progress, const char **error);
const char *cff_ffmpeg_file_open(cff_video_t *video, const char *fname, func_cff_callback_t func, void *user_data);
void        cff_ffmpeg_file_close(cff_video_t *video);
char       *cff_ffmpeg_file_info(cff_video_t *video);

const char *cff_ffmpeg_program_open(cff_video_t *video, unsigned int index, func_cff_callback_t func, void *user_data);
const char *cff_ffmpeg_get_program_name(cff_video_t *video, unsigned int index, int *prgnum);

int cff_ffmpeg_frame2rgb24(cff_video_t *video, int width, int height);
int cff_ffmpeg_get_frame_next(cff_video_t *video, enum cff_VideoFrame next, int width, int height, enum cff_FrameFlags frame_flags);
int cff_ffmpeg_get_frame_cutpoint(cff_video_t *video, unsigned int next, int width, int height, enum cff_FrameFlags frame_flags);
int cff_ffmpeg_get_frame_slider_click(cff_video_t *video, double next, int width, int height, enum cff_FrameFlags frame_flags);
int cff_ffmpeg_get_frame_slider_move(cff_video_t *video, unsigned int frame_nr, int width, int height, enum cff_FrameFlags frame_flags);

int cff_ffmpeg_video_undo(cff_video_t *video);
int cff_ffmpeg_video_redo(cff_video_t *video);

/* Returns '(unsigned int) -1' on error */
unsigned int cff_ffmpeg_get_cut_frame_number(const cff_video_t *video, unsigned int index);

/* Returns 1 if distance between prev cutpoint or start is greater 'video->vframe.max_gop_size' */
int cff_ffmpeg_video_cut_list_distance(const cff_video_t *video);

int  cff_ffmpeg_video_cut_find_current(const cff_video_t *video);
int  cff_ffmpeg_video_cut_add(cff_video_t *video);
int  cff_ffmpeg_video_cut_remove(cff_video_t *video);
void cff_ffmpeg_video_cut_list_clear(cff_video_t *video);
int  cff_ffmpeg_video_cut_list_load(cff_video_t *video, const char *fname);
int  cff_ffmpeg_video_cut_list_save(cff_video_t *video, const char *fname);

int cff_ffmpeg_get_stream_info(cff_video_t *video, int index, int *stream_idx, char *buf, int buf_sz);

int         cff_ffmpeg_copy_cb(cff_video_t *video, double *progress, const char **error);
const char *cff_ffmpeg_copy_open(cff_video_t *video, const cff_copy_param_t *param);
const char *cff_ffmpeg_save_image(cff_video_t *video, const cff_image_param_t *param);

#endif /* CFF_FFMPEG__H */
