/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017 .. 2019 Harald Foerster		*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#include <glib.h>
#include <string.h>
#include <stdlib.h>

#include "cff_gprop.h"

enum
{
	GPROP_FIRST,
	GPROP_STR  = GPROP_FIRST,
	GPROP_INT,
	GPROP_BOOL,
	GPROP_LAST = GPROP_BOOL
};

typedef struct
{
	int	item;
	void	*value;

} cff_gprop_t;

static GSList *cff_gprop_list[3] = {NULL};

static const char* cff_gprop_type_tab[] =
{
	"str.",
	"int.",
	"bool."
};

static const char* cff_gprop_item_str[] =
{
	"FileOpenPath",
	"FileSavePath",
	"FileImagePath",
	"FileVideoFilter",
	"FileImageFilter",
	"EditPath",
	"OptLoglevel",
	"OptLogWinText",
	"OptLogWinBackground",

	NULL
};

static const char* cff_gprop_item_int[] =
{
	"WindowWidth",
	"WindowHeight",

	"FileSaveWidth",
	"FileSaveHeight",

	"FileImageWidth",
	"FileImageHeight",

	"FileInfoWidth",
	"FileInfoHeight",

	"HelpFFmpegWidth",
	"HelpFFmpegHeight",

	"HelpCutterFFWidth",
	"HelpCutterFFHeight",

	"OptLogWinWidth",
	"OptLogWinHeight",

	"FileFilterIndex",

	NULL
};

static const char* cff_gprop_item_bool[] =
{
	"OptKeyframe",
	"OptIframe",
	"OptPframe",
	"OptCursor",
	"OptOverwrite",
	"OptDelete",
	"OptBeep",
	"OptSaveConfig",

	NULL
};

static const char* const* cff_gprop_item_tab[] =
{
	cff_gprop_item_str,
	cff_gprop_item_int,
	cff_gprop_item_bool
};

static const char cff_gprop_false_str[] = "false";
static const char cff_gprop_true_str[]  = "true";

static int		cff_gprop_find(const cff_gprop_t *gp, gconstpointer data);
static int		cff_gprop_sort(const cff_gprop_t *gp, const cff_gprop_t *data);
static cff_gprop_t*	cff_gprop_get_entry(int item, int type);
static void 		cff_gprop_set_value(int item, void *val, int type);
static int		cff_gprop_get_value(int item, void *valp, int type);

static int cff_gprop_find(const cff_gprop_t *gp, gconstpointer data)
{
	int item = GPOINTER_TO_INT(data);

	return gp->item - item;
}

static int cff_gprop_sort(const cff_gprop_t *gp, const cff_gprop_t *data)
{
	return gp->item - data->item;
}

static cff_gprop_t* cff_gprop_get_entry(int item, int type)
{
	GSList *node = g_slist_find_custom(cff_gprop_list[type],
			GINT_TO_POINTER(item), (GCompareFunc) cff_gprop_find);

	return (node == NULL) ? (cff_gprop_t *) node : (cff_gprop_t *) node->data;
}

static void cff_gprop_set_value(int item, void *val, int type)
{
	cff_gprop_t *gp = cff_gprop_get_entry(item, type);

	if (gp)
	{
		if (type == GPROP_STR)
		{
			g_free(gp->value);
		}
	}

	else
	{
		GSList **lst;

		gp = g_malloc(sizeof(cff_gprop_t));

		gp->item = item;

		 lst = &cff_gprop_list[type];
		*lst = g_slist_insert_sorted(*lst, gp, (GCompareFunc) cff_gprop_sort);
	}

	gp->value = (type == GPROP_STR) ? g_strdup(val) : val;

} /* static void cff_gprop_set_value(int, void*, int) */

void cff_gprop_set_str(int item, const char *sval)
{
	if ((unsigned int) item <= kPropStrItemLast)
	{
		cff_gprop_set_value(item, (void *) sval, GPROP_STR);
	}
}

void cff_gprop_set_int(int item, int ival)
{
	if ((unsigned int) item <= kPropIntItemLast)
	{
		cff_gprop_set_value(item, GINT_TO_POINTER(ival), GPROP_INT);
	}
}

void cff_gprop_set_bool(int item, int bval)
{
	if ((unsigned int) item <= kPropBoolItemLast)
	{
		cff_gprop_set_value(item, GINT_TO_POINTER(bval), GPROP_BOOL);
	}
}

static int cff_gprop_get_value(int item, void *valp, int type)
{
	cff_gprop_t *gp = cff_gprop_get_entry(item, type);

	if (gp == NULL)
	{
		return FALSE;
	}

	if (sizeof(char*) == sizeof(int))
	{
		int *p = (int *) valp;

		*p = GPOINTER_TO_INT(gp->value);
	}

	else
	{
		if (type == GPROP_STR)
		{
			char **s = (char **) valp;

			*s = (char *) gp->value;
		}

		else
		{
			int *p = (int *) valp;

			*p = GPOINTER_TO_INT(gp->value);
		}
	}

	return TRUE;

} /* static int cff_gprop_get_value(int, void*, int) */

char *cff_gprop_get_str(int item)
{
	char *str;

	if (cff_gprop_get_value(item, &str, GPROP_STR) &&
				str != NULL && str[0] != '\0')
	{
		return g_strdup(str);
	}

	return NULL;

} /* char *cff_gprop_get_str(int) */

int cff_gprop_get_int(int item, int *ivalp)
{
	return cff_gprop_get_value(item, (void *) ivalp, GPROP_INT);
}

int cff_gprop_get_bool(int item, int *bvalp)
{
	return cff_gprop_get_value(item, (void *) bvalp, GPROP_BOOL);
}

int cff_gprop_parse_str(char *str)
{
	int type = GPROP_LAST;

	do
	{
		const char	*ptr;
		int		len;

		ptr	= cff_gprop_type_tab[type];
		len	= strlen(ptr);

		if (strncmp(str, ptr, len) == 0)
		{
			int			item;
			const char* const*	itab;
			void			*value;

			str += len;

			len = strcspn(str, "=");

			if (len == 0)
			{
				break;
			}

			itab = cff_gprop_item_tab[type];
			item = 0;

			while (*itab)
			{
				if (strncmp(str, *itab, len) == 0)
				{
					break;
				}

				itab++;
				item++;
			}

			if (*itab == NULL)
			{
				break;
			}

			str += len + 1;

			if (*str == '\0')
			{
				break;
			}

			if (type == GPROP_STR)
			{
				if (*str != '"')
				{
					break;
				}

				value = (void *) &str[1];

				str = strrchr(value, '"');

				if (str == NULL)
				{
					break;
				}

				*str = '\0';
			}

			else if (type == GPROP_INT)
			{
				value = GINT_TO_POINTER(atoi(str));
			}

			else
			{
				if (strcmp(str, cff_gprop_false_str) == 0)
				{
					value = GUINT_TO_POINTER(FALSE);
				}

				else if (strcmp(str, cff_gprop_true_str) == 0)
				{
					value = GUINT_TO_POINTER(TRUE);
				}

				else
				{
					break;
				}
			}

			cff_gprop_set_value(item, value, type);

			return TRUE;

		} /* if (strncmp(str, ptr, len) == 0) */
	}
	while (--type >= GPROP_FIRST);

	return FALSE;

} /* int cff_gprop_parse_str(char*) */

static void cff_gprop_dump(cff_gprop_t *gp, gpointer data)
{
	struct
	{
		FILE		*fd;
		const char	*id;
		int		type;

	} *s;

	int			i;
	const char* const*	p;

	s = data;

	if (s->type == GPROP_STR)
	{
		char *str = gp->value;

		if (str == NULL || str[0] == '\0')
		{
			return;
		}
	}

	p = cff_gprop_item_tab[s->type];

	fprintf(s->fd, "%s %s%s=", s->id, cff_gprop_type_tab[s->type], p[gp->item]);

	switch (s->type)
	{
		case GPROP_STR:
			if (gp->value)
			{
				fprintf(s->fd, "\"%s\"\n", (char *) gp->value);
			}
			break;

		case GPROP_INT:
			i = GPOINTER_TO_INT(gp->value);
			fprintf(s->fd, "%d\n", i);
			break;

		default: /* GPROP_BOOL */
			fprintf(s->fd, "%s\n", (gp->value ?
					cff_gprop_true_str : cff_gprop_false_str));
			break;
	}

} /* static void cff_gprop_dump(cff_gprop_t*, gpointer) */

void cff_gprop_save_data(FILE *fd, const char *id)
{
	struct
	{
		FILE		*fd;
		const char	*id;
		int		type;
	} s;

	s.fd   = fd;
	s.id   = id;
	s.type = GPROP_FIRST;

	do
	{
		g_slist_foreach(cff_gprop_list[s.type], (GFunc) cff_gprop_dump, &s);
	}
	while (++s.type <= GPROP_LAST);

} /* void cff_gprop_save_data(FILE*, const char*) */

/* EOF */
