/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017,2018,2021 Harald Foerster		*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#else
#include "getopt.h"
#endif

#ifdef BUILD_MINGW
#include <fcntl.h>
#include <windows.h>
#endif

#include "cff_macros-gtk.h"
#include "cff_accel-gtk.h"
#include "cff_gprop.h"
#include "cff_nls.h"
#include "cff_charconv.h"

/*
#include <gtk/gtk.h>
#include "cff_tools-gtk.h"
*/


#define kPrintVersion	(1 << 0)
#define kPrintHelp	(1 << 1)

const gchar *cutterff_package_str = PACKAGE;
const gchar *cutterff_version_str = VERSION;

#ifdef BUILD_MINGW
/* Prototyp */
static int cff_main_open_console(void);
#endif

static const char cff_main_cutterffrc[] =
#ifdef BUILD_MINGW
			"cutterff.rc";
#else
			".cutterffrc";
#endif


enum
{
	CFF_FIRST,
	CFF_MENUACCEL = CFF_FIRST,
	CFF_PROP,
	CFF_LAST = CFF_PROP
};

static const char *cff_main_cfgrc_name[] =
{
	"MenuAccel:",
	"Property:"
};


static const char cff_main_parse_msg[] = gettext_nop("Unable to parse");
static const char cff_main_parse_fmt[] = "%s : '%s'\n";

#if GTK_MAJOR_VERSION > 2
static GSList *cff_main_gaccel_slist = NULL;
#endif

static void cff_main_fprn_stderr(const char *format, const char *mesg, const char *detail)
{

#ifdef BUILD_MINGW
	if ( ! cff_main_open_console())
	{
		return;
	}
#endif
	if (format == NULL)
	{
		format = "%s : %s\n";
	}

	fprintf(stderr, format, gettext(mesg), detail);

} /* static void cff_main_fprn_stderr(const char*, const char*, const char*) */

static FILE *cff_main_open_rc(const char *modus)
{
	FILE		*fd;
	char		*rc;
	const char	*config, *fname;
#ifdef BUILD_MINGW
	const char	*config_utf8;
#endif

	fname = cff_main_cutterffrc;

#if HAVE_GCONFIGDIR
	config = g_get_user_config_dir();
	if (config != NULL)
	{
#ifndef BUILD_MINGW
		/* skip '.' */
		fname++;
#endif
	}
	else
#endif
	{
		config = g_get_home_dir();

		if (config == NULL)
		{
			return NULL;
		}
	}

#ifdef BUILD_MINGW
	config_utf8	= config;
	config		= cff_charconv_utf8_to_filename(config_utf8);
	if (config == NULL)
	{
		config = config_utf8;
	}
#endif
	rc = g_strconcat(config, G_DIR_SEPARATOR_S, fname, NULL);

#ifdef BUILD_MINGW
	CFF_CHARCONV_UTF8_FREE((char *) config, config_utf8);
#endif
	fd = fopen(rc, modus);
	g_free(rc);

	return fd;

} /* static FILE *cff_main_open_rc(const char*) */

char *cff_main_save_rc(void)
{
	FILE *fd = cff_main_open_rc("w");

	if (fd == NULL)
	{
		return gettext_nop("Unable to open the cutterffrc file");
	}

	fprintf(fd, "# %s\n# %s\n\n",
			gettext("This resource file was generated by 'CutterFF'"),
					gettext("You may edit it if you're careful!"));

	cff_gaccel_save_data(fd, cff_main_cfgrc_name[CFF_MENUACCEL]);
	cff_gprop_save_data(fd, cff_main_cfgrc_name[CFF_PROP]);

	fclose(fd);

	return NULL;

} /* char *cff_main_save_rc(void) */

static void cff_main_load_rc(void)
{
	FILE *fd;
	char *line, pom[2048];

	fd = cff_main_open_rc("r");

	if (fd == NULL)
	{
		return;
	}

	while ((line = fgets(pom, sizeof(pom), fd)) != NULL)
	{
		static const char white_spaces[] = " \t";

		int i;

		line += strspn(line, white_spaces);

		if (*line == '#' || (i = strcspn(line, "\n")) == 0)
		{
			continue;
		}

		line[i]	= '\0';
		i	= CFF_LAST;

		do
		{
			int		len;
			const char	*str;

			str = cff_main_cfgrc_name[i];
			len = strlen(str);

			if (g_ascii_strncasecmp(line, str, len) == 0)
			{
				line += len;
				line += strspn(line, white_spaces);
				break;
			}
		}
		while (--i >= CFF_FIRST);

		switch (i)
		{
			case CFF_MENUACCEL:
			{

#if GTK_MAJOR_VERSION > 2
				cff_main_gaccel_slist = g_slist_append(cff_main_gaccel_slist, g_strdup(line));
#else
				if (cff_gaccel_parse_str(line) == FALSE)
				{
					cff_main_fprn_stderr(cff_main_parse_fmt, cff_main_parse_msg, line);
				}
#endif
				break;
			}

			case CFF_PROP:
			{
				if (cff_gprop_parse_str(line) == FALSE)
				{
					cff_main_fprn_stderr(cff_main_parse_fmt, cff_main_parse_msg, line);
				}

				break;
			}

			default:
			{
				cff_main_fprn_stderr(cff_main_parse_fmt, cff_main_parse_msg, line);

				break;
			}

		} /* switch (i) */

	} /* while ((line = fgets(pom, sizeof(pom), fd)) != NULL) */

	fclose(fd);

} /* static void cff_main_load_rc(void) */

static void cff_main_print_version_help(int what, const char *prgname)
{

#ifdef BUILD_MINGW
	if ( ! cff_main_open_console())
	{
		return;
	}
#endif
	if (what & kPrintVersion)
	{
		printf("%s-%s\n", cutterff_package_str, cutterff_version_str);
	}

	if ( ! (what & kPrintHelp))
	{
		return;
	}

	printf("%s %s %s %s\n", gettext("Usage:"), prgname,
				gettext("[OPTION]"), gettext("[FILENAME]"));

	printf(
		"  -l,  --loglevel=         %s\n"
		"                             quiet (%s)\n"
		"                             panic\n"
		"                             fatal\n"
		"                             error\n"
		"                             warning\n"
		"                             info\n"
		"                             verbose\n"
		"                             debug\n"
#ifdef HAVE_GETTEXT
		"       --locale=de_DE\n"
		"       --nls-dir=%s %s\n"
#endif
		"  -v,  --version           %s\n"
		"  -h,  --help              %s\n",
		gettext("Loglevel for printing messages to the console"),
		gettext("default"),
#ifdef HAVE_GETTEXT
		gettext("DIR      "),
			gettext("set the nlsdir (eg. /usr/share/locale)"),
#endif
		gettext("display version information and exit"),
		gettext("display this help and exit")

	); /* printf() */

	return;

} /* static void cff_main_print_version_help(int, const char*) */

#ifdef BUILD_MINGW

static int cff_main_window_console = 0x00;

static int cff_main_connect_console(void)
{
	int	newline;
	DWORD	which;

	newline	= 0;
	which	= STD_OUTPUT_HANDLE;

	for (;;)
	{
		int		fnr;
		HANDLE		stdHandle;
		FILE		*stdFp;

		stdFp = (which == STD_OUTPUT_HANDLE) ? stdout : stderr;

		if ((fnr = fileno(stdFp)) == -1 ||
			(stdHandle = (HANDLE) _get_osfhandle(fnr)) ==
							INVALID_HANDLE_VALUE ||
				GetFileType(stdHandle) == FILE_TYPE_UNKNOWN)
		{
			int	osfHandle;
			FILE	*fp;

			stdHandle = GetStdHandle(which);

			if (stdHandle == INVALID_HANDLE_VALUE)
			{
				break;
			}

			osfHandle = _open_osfhandle((long) stdHandle, _O_TEXT);

			if (osfHandle == -1)
			{
				break;
			}

			fp = _fdopen(osfHandle, "w" );

			if (fp == NULL)
			{
				break;
			}

			*stdFp = *fp;

			if (newline == FALSE)
			{
				newline = TRUE;
				fputs("\n", stdFp);
			}
		}

		if (which != STD_OUTPUT_HANDLE)
		{
			return TRUE;
		}

		which = STD_ERROR_HANDLE;

	} /* for (;;) */

	return FALSE;

} /* static int cff_main_connect_console(void) */

static int cff_main_attach_console(void)
{
	if (AttachConsole(ATTACH_PARENT_PROCESS))
	{
		cff_main_window_console = 0x10;

		if (cff_main_connect_console())
		{
			cff_main_window_console = 0x11;
			return 0x01;
		}
	}

	return 0x00;

} /* static int cff_main_attach_console(void) */

static int cff_main_open_console(void)
{
	if ( ! (cff_main_window_console & 0x10) && ! cff_main_attach_console())
	{
		cff_main_window_console = 0x10;

		if ( ! AllocConsole())
		{
			return 0x00;
		}

		if (cff_main_connect_console())
		{
			cff_main_window_console = 0x11;
		}

	} /*  if ( ! (cff_main_window_console & 0x10) && ! cff_main_attach_console()) */

	return cff_main_window_console & 0x01;

} /* static int cff_main_open_console(int) */

#endif /* BUILD_MINGW */

static int cff_main_cmd_setup(int argc, char **argv, const char **fname, const char **logname)
{
	const struct option options[] =
	{
		{"loglevel", 1, NULL, 'l'},
#ifdef HAVE_GETTEXT
		{"locale", 1, NULL, 'g'},
		{"nls-dir", 1, NULL, 'g'},
#endif
		{"version", 0, NULL, 'v'},
		{"help", 0, NULL, 'h'},
		/* !!! GTK&GDK OPTIONS !!! */
		{"gtk-module", 1, NULL, 'g'},
		{"g-fatal-warnings", 0, NULL, 'g'},
		{"gtk-debug", 2, NULL, 'g'},
		{"gtk-no-debug", 2, NULL, 'g'},
		{"gdk-debug", 2, NULL, 'g'},
		{"gdk-no-debug", 2, NULL, 'g'},
		{"display", 1, NULL, 'g'},
		{"sync", 0, NULL, 'g'},
		{"no-xshm", 0, NULL, 'g'},
		{"name", 1, NULL, 'g'},
		{"class", 1, NULL, 'g'},
		{"gxid_host", 1, NULL, 'g'},
		{"gxid_port", 1, NULL, 'g'},
		{"xim-preedit", 1, NULL, 'g'},
		{"xim-status", 1, NULL, 'g'},
		{NULL, 0, NULL, 0},
	};

	int c, index, print_info;

#ifdef BUILD_MINGW
	if (argc <= 1 || ! cff_main_attach_console())
	{
		/* disable error messages - no 'stderr' */
		opterr = 0;
	}
#endif

	print_info = 0;

	while ((c = getopt_long(argc, argv, "l:vh",
					options, &index)) != - 1)
	{
		switch (c)
		{
			case 'l':
				*logname = optarg;
				break;
			case 'v':
				print_info |= kPrintVersion;
				break;
			case 'h':
				print_info |= kPrintHelp;
				break;
			case 'g':
				break;

			default:
				return EINVAL;

		} /* switch (c) */

	} /* while ((c = getopt_long(argc, argv, ... )) != - 1) */

	if (print_info)
	{
		cff_main_print_version_help(print_info, argv[0]);
		exit(0);
	}

	if (optind < argc)
	{
		if (argc > optind + 1)
		{
			cff_main_fprn_stderr(NULL, gettext_nop("Illegal argument"), argv[optind]);
			return EINVAL;
		}

		*fname = argv[optind];
	}

	cff_main_load_rc();

	return 0;

} /* static int cff_main_cmd_setup(int, char**, const char**, const char**) */

static void cff_main_sigterm(int signum)
{
	int save_flag;

	/* store settings */
	if ( ! cff_gprop_get_bool(kPropBoolOptSaveConfig, &save_flag) || save_flag)
	{
		cff_main_save_rc();
	}

	if (signum == SIGTERM)
	{
		_exit(0);
	}

} /* static void cff_main_sigterm(int) */

static void cff_main_atexit(void)
{
	/* user lost interrest */
	cff_main_sigterm(SIGQUIT);
}

static void cff_main_signal_init(void)
{

#ifdef HAVE_SIGACTION

	struct sigaction sigact;

	/* block other signals, while executing 'sa_handler' */
	sigfillset(&sigact.sa_mask);
	sigact.sa_flags = 0;

	/* 'cutterff' was terminated */
	sigact.sa_handler = cff_main_sigterm;
	sigaction(SIGTERM, &sigact, NULL);
#else
	signal(SIGTERM, cff_main_sigterm);
#endif

} /* static void cff_main_signal_init(void) */

#ifdef HAVE_GSETENV
#define CFF_SETENV(name, value) g_setenv((name), (value), TRUE)
#else
#define CFF_SETENV(name, value) putenv(g_strconcat((name), "=", (value), NULL))
#endif

static void cff_main_init_locale(int argc, const char **argv)
{
	const char *arg, *locale, *nlsdir;

	locale = nlsdir = NULL;

	while ((arg = *(argv++)) != NULL)
	{
		if (strncmp(arg, "--locale", 8) == 0)
		{
			if (arg[8] == '=')
			{
				/* --locale=LOCALE */
				locale = &arg[9];
			}

			else
			{
				/* --locale LOCALE */
				locale = *(argv++);
			}
		}

		else if (strncmp(arg, "--nls-dir", 9) == 0)
		{
			if (arg[9] == '=')
			{
				/* --nls-dir=DIR */
				nlsdir = &arg[10];
			}

			else
			{
				/* --nls-dir DIR */
				nlsdir = *(argv++);
			}
		}
	}

	if (locale)
	{
		const char *s;

		CFF_SETENV("LANG", locale);

		s = strchr(locale, '.');

		if (s)
		{
			s = g_strndup(locale, s - locale);
		}

		else
		{
			s = locale;
		}

		CFF_SETENV("LANGUAGE", s);
		CFF_SETENV("LC_ALL", locale);

		if (s != locale)
		{
			g_free((char *) s);
		}
	}

	CFF_NLS_INIT_LOCALE(nlsdir);

} /* static void cff_main_init_locale(int, const char**) */

int main(int argc, char **argv)
{
	/* cff_ffmpeg.c */
	extern int cff_ffmpeg_init(const char*);

	int		error;
	const char	*fname, *logname;

	cff_main_init_locale(argc, (const char **) argv);

	fname   = NULL;
	logname = NULL;
	error   = cff_main_cmd_setup(argc, argv, &fname, &logname);

	if (error)
	{
		return error;
	}

	if (cff_ffmpeg_init(logname) < 0)
	{
		cff_main_fprn_stderr(NULL, gettext_nop("Illegal argument"), logname);
		return EINVAL;
	}

	gtk_init(&argc, &argv);

#if glib_major_version == 2 && glib_minor_version < 32
	if ( ! g_thread_supported())
	{
		g_thread_init(NULL);
	}
#endif

#if GTK_MAJOR_VERSION > 2
	while (cff_main_gaccel_slist != NULL)
	{
		gchar *line = cff_main_gaccel_slist->data;

		if (cff_gaccel_parse_str(line) == FALSE)
		{
			cff_main_fprn_stderr(cff_main_parse_fmt, cff_main_parse_msg, line);
		}

		g_free(line);
		cff_main_gaccel_slist = g_slist_remove(cff_main_gaccel_slist, line);
	}
#endif
	cff_gui_init(fname, logname);

	cff_main_signal_init();
	atexit(cff_main_atexit);

	gtk_main();

	return 0;

} /* int main(int, char**) */

/* EOF */
