/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017 .. 2019,2021,2022 Harald Foerster	*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/


#include <string.h>

#ifdef HAVE_SEMAPHORE
#include <semaphore.h>
#endif

#include <libswscale/swscale.h>
#include <libavutil/imgutils.h>
#include <libavutil/parseutils.h>
/*  AV_RB24 AV_RB32 */
#include <libavutil/intreadwrite.h>

#include "cff_ffmpeg.h"
#include "cff_nls.h"

#if !defined PRId64 || PRI_MACROS_BROKEN
#undef PRId64
#define PRId64 "lld"
#endif


/*
#include <libavcodec/avcodec.h>
#include <libavcodec/bsf.h>
#include <libavformat/avformat.h>
#include <stdint.h>
#include <stdio.h>
#include <libavutil/opt.h>
#include <libavutil/mathematics.h>
#include <libavutil/rational.h>
*/


#ifdef AV_PIX_FMT_FLAG_PSEUDOPAL
#define CFF_PIX_FMT_FLAG_PSEUDOPAL	AV_PIX_FMT_FLAG_PSEUDOPAL
#else
#define CFF_PIX_FMT_FLAG_PSEUDOPAL	0
#endif

#ifdef HAVE_FFMPEG_ITERATE
#define CffMuxerIterate(ofmt, opaque)	av_muxer_iterate(opaque)
#else
#define CffMuxerIterate(ofmt, opaque)	av_oformat_next(ofmt)
#endif

#ifdef HAVE_FFMPEG_ITERATE
#define CffDemuxerIterate(ifmt, opaque)	av_demuxer_iterate(opaque)
#else
#define CffDemuxerIterate(ifmt, opaque)	av_iformat_next(ifmt)
#endif

#ifdef HAVE_FFMPEG_ITERATE
#define CffCodecIterate(codec, opaque)	av_codec_iterate(opaque)
#else
#define CffCodecIterate(codec, opaque)	av_codec_next(codec)
#endif

#ifdef HAVE_FFMPEG_ITERATE
#define CffBsfIterate(opaque)		av_bsf_iterate(opaque)
#else
#define CffBsfIterate(opaque)		av_bsf_next(opaque)
#endif

#ifdef HAVE_FFMPEG_ITERATE
#define CffRegisterAll()		{;}
#else
#define CffRegisterAll()		av_register_all()
#endif


#define kCFFOpenFileGetFirstFrame	0
#define kCFFOpenFileSeekLastGop		5
#define kCFFOpenFileGetLastFrame	6

#define kCFFUndo			0
#define kCFFRedo			1

/* Prototypes */
static void      cff_ffmpeg_video_undo_clear(cff_undo_t **list);
static int       cff_ffmpeg_video_cut_add_dts(cff_video_t *video, cff_dts_t *src);
static int       cff_ffmpeg_video_cut_remove_dts(cff_video_t *video, cff_dts_t *src);
static AVStream *cff_ffmpeg_get_stream_from_index(cff_video_t *video, int index);

/* Some error strings */
static const char g_alloc_codec_context_errstr[] =
			gettext_nop("Could not alloc codec context!");
static const char g_alloc_frame_errstr[] =
			gettext_nop("Could not alloc a frame!");
static const char g_alloc_packet_errstr[] = 
			gettext_nop("Could not alloc a packet!");
static const char g_alloc_stream_errstr[] =
			gettext_nop("Could not alloc stream!");
static const char g_clone_frame_errstr[] =
			gettext_nop("Could not clone the frame!");
static const char g_find_oformat_errstr[] =
			gettext_nop("Could not find an output format!");
static const char g_open_codec_errstr[] =
			gettext_nop("Could not open codec!");
static const char g_open_ocontext_errstr[] =
			gettext_nop("Could not open output context!");
static const char g_open_ofile_errstr[] =
			gettext_nop("Could not open ouput file!");
static const char g_video_codec_errstr[] =
			gettext_nop("No video codec found!");
static const char g_avseek_frame_errstr[] =
			gettext_nop("'av_seek_frame()' failed!");
static const char g_header_write_errstr[] =
			gettext_nop("Header write failed!");
static const char g_trailer_write_errstr[] =
			gettext_nop("Trailer write failed!");
static const char g_missing_filename_errstr[] =
			gettext_nop("Missing the file name!");
static const char g_copy_codec_parameters_errstr[] =
			gettext_nop("Could not copy codec parameters!");

/* Global */
cff_video_t gVideo;

/* 'opaque' is a pointer for the iteration state */
struct CFFCodecNext
{
	const AVCodec	*codec;
	void		*opaque;
};


/*
	For 'cff_ffmpeg_log_callback' and
	'cff_file_info_callback' (av_dump_format)
*/

#define kCffLogBufferSize	1024

typedef struct
{
	char		*str;
	size_t		max;
	size_t		len;
	unsigned long	prev;

	void			*data;
	func_log_callback_t	*func;

} cff_logbuffer_t;

static cff_logbuffer_t *gLogBufferPtr = NULL;

/* Semaphore */

#ifdef HAVE_SEMAPHORE
static sem_t gLogSemaphore;
#define CFF_SEMAPHORE_INIT()	sem_init(&gLogSemaphore, 0, 1)
#define CFF_SEMAPHORE_LOCK()	sem_wait(&gLogSemaphore)
#define CFF_SEMAPHORE_UNLOCK()	sem_post(&gLogSemaphore)
#else
#define CFF_SEMAPHORE_INIT()
#define CFF_SEMAPHORE_LOCK()
#define CFF_SEMAPHORE_UNLOCK()
#endif

/* Loglevel */

typedef struct
{
	const char	*name;
	int		value;

} LoglevelTab;

static const LoglevelTab loglevel_tab[] =
{
	{ "quiet"  , AV_LOG_QUIET   },
	{ "panic"  , AV_LOG_PANIC   },
	{ "fatal"  , AV_LOG_FATAL   },
	{ "error"  , AV_LOG_ERROR   },
	{ "warning", AV_LOG_WARNING },
	{ "info"   , AV_LOG_INFO    },
	{ "verbose", AV_LOG_VERBOSE },
	{ "debug"  , AV_LOG_DEBUG   },
	{ NULL }

}; /* static const LoglevelTab loglevel_tab[] = */

static int gLoglevel;


/* DVD-Subtitles */

#define kCFFDvbSubHeaderSize		6
#define kCFFDvbSubDisplayDefinition	0x14
#define kCFFDvbSubPageComposition	0x10
#define kCFFDvbSubRegionComposition	0x11
#define kCFFDvbSubEnd			0x80


/* Softcut */

typedef struct
{
	AVPacket	*packet;
	cff_pktqueue_t	*pktqueue;
	unsigned int	qidx;
	int		decoder_in;
	int		encode_max;
	int		q0_end;
	int		q0_decoded;
	int		q0_encoded;
	int		q3_decoded;
	int		q3_encoded;
	int		done;
	int64_t		dts_last;

} cff_encdata_t;



/* Codecs */

enum
{
	kLIBSTART,
	kLIBAVUTIL = kLIBSTART,
	kLIBAVCODEC,
	kLIBAVFORMAT,
	kLIBSWSCALE,
	kLIBEND
};

typedef struct
{
	const char	*name;
	unsigned int	version_int;
	unsigned int	(*versionFunc)(void);

} AVLibTab;

static const AVLibTab avlib_tab[] =
{
   	{ "avutil",
		LIBAVUTIL_VERSION_INT,
		avutil_version
	},

   	{ "avcodec",
		LIBAVCODEC_VERSION_INT,
		avcodec_version
	},

   	{ "avformat",
		LIBAVFORMAT_VERSION_INT,
		avformat_version
	},

   	{ "swscale",
		LIBSWSCALE_VERSION_INT,
		swscale_version
	}

}; /* static const AVLibTab avlib_tab[] */


static void *cff_ffmpeg_memdup(const void *src, int sz)
{
	void *ptr;

	if (sz <= 0)
	{
		return NULL;
	}

	ptr = malloc(sz);

	if (ptr == NULL)
	{
		return NULL;
	}

	return memcpy(ptr, src, sz);

} /* static void *cff_ffmpeg_memdup(const void*, int) */

static char *cff_ffmpeg_strdup(const char *str, int slen)
{
	int  sz;
	char *ptr;

	if (slen < 0)
	{
		slen = strlen(str);

		/* slen plus '\0' */
		sz = slen + 1;
	}

	else
	{
		sz = slen;

		if (str[slen] != '\n')
		{
			/* slen plus '\0' */
			sz++;
		}
	}

	ptr = cff_ffmpeg_memdup(str, sz);

	if (ptr != NULL)
	{
		ptr[slen] = '\0';
	}

	return ptr;

} /* static char *cff_ffmpeg_strdup(const char*, int) */

static char *cff_ffmpeg_string_truncate(char *str, size_t slen)
{
	if (str != NULL)
	{
		char *ptr;

		/* Include terminating '\0' */
		ptr = realloc(str, slen + 1);

		if (ptr != NULL)
		{
			/* Terminate string */
			ptr[slen] = '\0';

			return ptr;
		}

	} /* if (str != NULL) */

	return str;

} /* static char *cff_ffmpeg_string_truncate(char*, size_t) */

static char *cff_ffmpeg_printf_concat(char *str, size_t *slen,
					size_t *max, const char *fmt, ...)
{
	int	n;
	size_t  bufsz, sz;
	va_list args;

	va_start(args, fmt);
	n = vsnprintf(NULL, 0, fmt, args);
	va_end(args);

	if (n <= 0)
	{
		n = 1024;
	}

	bufsz = *slen + n + 1;
	sz    = *max;

	if (sz == 0)
	{
		sz = bufsz;
	}

	while (bufsz > sz)
	{
		sz <<= 1;
	}

	if (str == NULL || sz > *max)
	{
		str = realloc(str, sz);
	}

	if (str != NULL)
	{
		*max = sz;

		sz -= *slen;

		va_start(args, fmt);
		n  = vsnprintf(&str[*slen], sz, fmt, args);
		va_end(args);

		if (n > 0 && n < sz)
		{
			*slen += n;
		}
	}

	return str;

} /* static char *cff_ffmpeg_printf_concat(char *str, size_t*, size_t*,
							const char*, ...) */


/* FFmpeg => cmdutil.h */

char *cff_ffmpeg_get_libs_info(void)
{
	int		lib, slen;
	const AVLibTab	*avlib;
	char		buf[4096];

	avlib = avlib_tab;
	slen  = 0;
	lib   = kLIBSTART;

	do
	{
		int		n;
		size_t		sz;
		unsigned int	version;

		version = avlib->versionFunc();

		sz = sizeof(buf) - slen;
		n  = snprintf(&buf[slen], sz,
			"lib%-11s %2d.%3d.%2d / %2d.%3d.%2d\n",
			avlib->name,
			avlib->version_int >> 16,		/* major */
			(avlib->version_int >> 8) & 0xff,	/* minor */
			avlib->version_int & 0xff,		/* micro */
			version >> 16,				/* major */
			(version >> 8) & 0xff,			/* minor */
			version & 0xff);			/* micro */

		if (n > 0 && n < sz)
		{
			slen += n;
		}

		avlib++;
	}
	while (++lib < kLIBEND);

	/* remove last '\n' */
	return cff_ffmpeg_strdup(buf, slen - 1);

} /* char *cff_ffmpeg_get_libs_info(void) */


/* Formats */

struct FormatInfo
{
	const char	*name;
	const char	*long_name;
	unsigned int	encode;
	unsigned int	decode;
};

struct FormatInfo *cff_ffmpeg_get_format_entry(struct FormatInfo **buf,
				size_t *num, size_t *max, const char *name)
{
	int			n;
	size_t			sz;
	struct FormatInfo	*ptr;

	ptr = *buf;
	sz  = *max;

	if (*num >= sz)
	{
		sz  = *num + 128;
		ptr = realloc(ptr, sizeof(struct FormatInfo) * sz);

		if (ptr == NULL)
		{
			return NULL;
		}

		*buf = ptr;
		*max = sz;

	} /* if (*num >= sz) */

	n = *num;

	while (--n >= 0)
	{
		int i = strcmp(ptr->name, name);

		if (i == 0)
		{
			return ptr;
		}

		if (i > 0)
		{
			memmove(&ptr[1], ptr,
				(n + 1) * sizeof(struct FormatInfo));
			break;
		}

		ptr++;

	} /* while (--n >= 0) */

	(*num)++;

	return ptr;

} /* struct FormatInfo *cff_ffmpeg_get_format_entry(struct FormatInfo**,
					size_t*, size_t*, const char*) */

char *cff_ffmpeg_get_formats_info(void)
{
	size_t			n, num, max, bufsz, slen;
	const AVInputFormat	*ifmt;
	const AVOutputFormat	*ofmt;
	char			*str;
#ifdef HAVE_FFMPEG_ITERATE
	void			*opaque;
#endif

	struct FormatInfo *ptr, *buf;

	max  = 0;
	num  = 0;
 	n    = 0;
	buf  = NULL;

#ifdef HAVE_FFMPEG_ITERATE
	opaque = NULL;
#else
	ofmt = NULL;
#endif

	while ((ofmt = CffMuxerIterate(ofmt, &opaque)) != NULL)
	{
		ptr = cff_ffmpeg_get_format_entry(&buf, &num, &max, ofmt->name);

		if (ptr == NULL)
		{
			break;
		}

		if (num > n)
		{
			n = num;

			ptr->name	= ofmt->name;
			ptr->long_name	= ofmt->long_name;
			ptr->encode	= 'E';
			ptr->decode	= '.';
		}

	} /* while ((ofmt = CffMuxerIterate(ofmt, &onext)) != NULL) */


#ifdef HAVE_FFMPEG_ITERATE
	opaque = NULL;
#else
	ifmt = NULL;
#endif

	while ((ifmt = CffDemuxerIterate(ifmt, &opaque)) != NULL)
	{
		ptr = cff_ffmpeg_get_format_entry(&buf, &num, &max, ifmt->name);

		if (ptr == NULL)
		{
			break;
		}

		ptr->decode = 'D';

		if (num > n)
		{
			n = num;

			ptr->name	= ifmt->name;
			ptr->long_name	= ifmt->long_name;
			ptr->encode	= '.';
		}

	} /* while ((ifmt = CffDemuxerIterate(ofmt, &inext)) != NULL) */


	ptr = buf;

	if (ptr == NULL)
	{
		return NULL;
	}

	bufsz = 4096;
	slen  = 0;
	str   = cff_ffmpeg_printf_concat(NULL, &slen, &bufsz,
			" D. = Demuxing supported\n"
			" .E = Muxing supported\n"
			" --\n");
	if (str == NULL)
	{
		return str;
	}

	do
	{
		char *s;

		s = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
			" %c%c %-20s %s\n",
			(char) ptr->decode,
			(char) ptr->encode,
			ptr->name,
			ptr->long_name != NULL ? ptr->long_name : "");

		if (s == NULL)
		{
			break;
		}

		str = s;
		ptr++;
	}
	while (--num != 0);

	free(buf);

	/* remove last '\n' */
	return cff_ffmpeg_string_truncate(str, slen - 1);

} /* char *cff_ffmpeg_get_formats_info(void) */



/* Codecs */

static char cff_ffmpeg_get_media_type_char(enum AVMediaType type)
{
	switch (type)
	{
		case AVMEDIA_TYPE_VIDEO:	return 'V';
		case AVMEDIA_TYPE_AUDIO:	return 'A';
		case AVMEDIA_TYPE_DATA:		return 'D';
		case AVMEDIA_TYPE_SUBTITLE:	return 'S';
		case AVMEDIA_TYPE_ATTACHMENT:	return 'T';
		default:			break;
	}

	return '?';

} /* static char cff_ffmpeg_get_media_type_char(enum AVMediaType) */

static int cff_ffmpeg_compare_codec_desc(const void *a, const void *b)
{
	const AVCodecDescriptor * const *da = a;
	const AVCodecDescriptor * const *db = b;

	return ((*da)->type != (*db)->type) ?
			(*da)->type - (*db)->type :
				strcmp((*da)->name, (*db)->name);

} /* static int cff_ffmpeg_compare_codec_desc(const void*, const void*) */

static const AVCodecDescriptor **cff_ffmpeg_get_codecs_sorted(void)
{
	unsigned int		nb_codecs, i;
	const AVCodecDescriptor	*desc, **codecs;

	nb_codecs	= 0;
	desc		= NULL;

	while ((desc = avcodec_descriptor_next(desc)) != NULL)
	{
		nb_codecs++;
	}

	if (nb_codecs == 0)
	{
		return NULL;
	}

	codecs = av_calloc(nb_codecs + 1, sizeof(*codecs));

	if (codecs == NULL)
	{
		return NULL;
	}

	codecs[nb_codecs] = NULL;

	i	= 0;
	desc	= NULL;

	while ((desc = avcodec_descriptor_next(desc)) != NULL)
	{
		codecs[i++] = desc;
	}

	qsort(codecs, nb_codecs, sizeof(*codecs), cff_ffmpeg_compare_codec_desc);

	return codecs;

} /* static const AVCodecDescriptor **cff_ffmpeg_get_codecs_sorted(void) */

static int cff_ffmpeg_next_codec_for_id(enum AVCodecID id,
				struct CFFCodecNext *prev, int is_encoder)
{
	while ((prev->codec = CffCodecIterate(prev->codec,
						&(prev->opaque))) != NULL)
	{
		if ((prev->codec)->id == id && ((is_encoder == 0) ?
			av_codec_is_decoder(prev->codec) :
					av_codec_is_encoder(prev->codec)))
		{
			return 0;
		}
	}

	return -1;

} /* static int cff_ffmpeg_next_codec_for_id(enum AVCodecID, struct CFFCodecNext*, int) */

static char *cff_ffmpeg_print_codecs_for_id(char *str, size_t *slen, size_t *bufsz, enum AVCodecID id, int is_encoder)
{
	struct CFFCodecNext codec;

	str = cff_ffmpeg_printf_concat(str, slen, bufsz,
			" (%s: ", is_encoder ? "encoders" : "decoders");
	if (str == NULL)
	{
		return str;
	}

	codec.codec  = NULL;
	codec.opaque = NULL;

	while (cff_ffmpeg_next_codec_for_id(id, &codec, is_encoder) == 0)
	{
		str = cff_ffmpeg_printf_concat(str, slen, bufsz,
						"%s, ", codec.codec->name);
		if (str == NULL)
		{
			return str;
		}
	}

	if (str[*slen - 1] == ' ')
	{
		(*slen)--;
	}

	if (str[*slen - 1] == ',')
	{
		(*slen)--;
	}

	return cff_ffmpeg_printf_concat(str, slen, bufsz, "%c", ')');

} /* static char *cff_ffmpeg_print_codecs_for_id(char*, size_t*, size_t*, enum AVCodecID, int) */

char *cff_ffmpeg_get_codecs_info(void)
{
	size_t			bufsz, slen;
	const AVCodecDescriptor *desc, **codecs, **codecs_orig;
	char			*str;

	bufsz = 4096;
	slen  = 0;
	str   = cff_ffmpeg_printf_concat(NULL, &slen, &bufsz,
			" D..... = Decoding supported\n"
			" .E.... = Encoding supported\n"
			" ..V... = Video codec\n"
			" ..A... = Audio codec\n"
			" ..S... = Subtitle codec\n"
			" ...I.. = Intra frame-only codec\n"
			" ....L. = Lossy compression\n"
			" .....S = Lossless compression\n"
			" -------\n");

	if (str == NULL)
	{
		return str;
	}

	codecs = cff_ffmpeg_get_codecs_sorted();

	if (codecs == NULL)
	{
		return NULL;
	}

	codecs_orig = codecs;

	while ((desc = *(codecs++)) != NULL)
	{
		char			*ptr, decoder, encoder;
		struct CFFCodecNext	codec;

		decoder = encoder = '.';

		if (avcodec_find_decoder(desc->id))
		{
			decoder = 'D';
		}

		if (avcodec_find_encoder(desc->id))
		{
			encoder = 'E';
		}

		if (decoder == '.' && encoder == '.')
		{
			continue;
		}

		ptr = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
			" %c%c%c%c%c%c %-20s %s",
			(char) decoder,
			(char) encoder,
			cff_ffmpeg_get_media_type_char(desc->type),
			(desc->props & AV_CODEC_PROP_INTRA_ONLY)? 'I' : '.',
			(desc->props & AV_CODEC_PROP_LOSSY)	? 'L' : '.',
			(desc->props & AV_CODEC_PROP_LOSSLESS)	? 'S' : '.',
			desc->name,
			desc->long_name ? desc->long_name : "");

		if (ptr == NULL)
		{
			break;
		}

		str = ptr;

		/*
			Print decoders/encoders when there's more than
			one or their names are different from codec name
		*/

		codec.codec  = NULL;
		codec.opaque = NULL;

		while (cff_ffmpeg_next_codec_for_id(desc->id, &codec, 0) == 0)
		{
			if (strcmp(codec.codec->name, desc->name) != 0)
			{
				ptr = cff_ffmpeg_print_codecs_for_id(str, &slen,
								&bufsz, desc->id, 0);
				if (ptr != NULL)
				{
					str = ptr;
				}

				break;
			}
		}


		codec.codec  = NULL;
		codec.opaque = NULL;

		while (cff_ffmpeg_next_codec_for_id(desc->id, &codec, 1) == 0)
		{
			if (strcmp(codec.codec->name, desc->name) != 0)
			{
				ptr = cff_ffmpeg_print_codecs_for_id(str, &slen,
								&bufsz, desc->id, 1);
				if (ptr != NULL)
				{
					str = ptr;
				}

				break;
			}
		}

		ptr = cff_ffmpeg_printf_concat(str, &slen, &bufsz, "%c", '\n');

		if (ptr == NULL)
		{
			break;
		}

		str = ptr;

	} /* while ((desc = *(codecs++)) != NULL) */

	av_free(codecs_orig);

	return cff_ffmpeg_string_truncate(str, slen - 1);

} /* char *cff_ffmpeg_get_codecs_info(void) */

static char *cff_ffmpeg_get_bitstream_codecs(char *str, size_t *slen, size_t *bufsz, const enum AVCodecID *id_ptr)
{
	const char	*end;
	enum AVCodecID	codec_id;

	end = "";

	while ((codec_id = *(id_ptr++)) != AV_CODEC_ID_NONE)
	{
		char		*ptr;
		const AVCodec	*codec;

		codec = avcodec_find_decoder(codec_id);

		if (codec != NULL)
		{
			if (end[0] == '\0')
			{
				end = "(";
			}

			else if (end[0] == '(')
			{
				end = ", ";
			}

			ptr = cff_ffmpeg_printf_concat(str, slen, bufsz,
						"%s%s", end, codec->name);
			if (ptr == NULL)
			{
				break;
			}

			str = ptr;

		} /* if (codec != NULL) */

	} /* while ((codec_id = *(codec_ptr++)) != AV_CODEC_ID_NONE) */

	if (end[0] == '(')
	{
		end = ")";
	}

	else if (end[0] == ',')
	{
		end = ")";
	}

	return cff_ffmpeg_printf_concat(str, slen, bufsz, "%s\n", end);

} /* static char *cff_ffmpeg_get_bitstream_codecs(char*, size_t*, size_t*, const enum AVCodecID*) */

char *cff_ffmpeg_get_bitstream_filters_info(void)
{
	size_t			bufsz, slen;
	char			*str;
	void			*opaque;
	const AVBitStreamFilter	*bsf;

	opaque = NULL;
	str    = NULL;
	bufsz  = 4096;
	slen   = 0;

	while ((bsf = CffBsfIterate(&opaque)) != NULL)
	{
		char *ptr;

		ptr = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
							"%s ", bsf->name);
		if (ptr == NULL)
		{
			break;
		}

		str = ptr;

		if (bsf->codec_ids == NULL)
		{
			ptr = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
							"%s\n", "(all codecs)");
			if (ptr == NULL)
			{
				break;
			}

			str = ptr;
		}

		else
		{
			ptr = cff_ffmpeg_get_bitstream_codecs(str, &slen, &bufsz, bsf->codec_ids);

			if (ptr == NULL)
			{
				break;
			}

			str = ptr;

		} /* if (id_ptr == NULL) .. else */

	} /* while ((bsf = CffBsfIterate(&opaque)) != NULL) */

	/* remove last '\n' */
	return cff_ffmpeg_string_truncate(str, slen - 1);

} /* char *cff_ffmpeg_get_bitstream_filters_info(void) */

cff_fmtlist_t *cff_ffmpeg_get_format_names(cff_video_t *video)
{
	int			i, n, nb_codecs;
	unsigned int		num;
	const AVOutputFormat	*ofmt;
	cff_fmtlist_t		buf[512];
	const AVCodecDescriptor	*codec_descriptors[128], *codecs_found[128];
#ifdef HAVE_FFMPEG_ITERATE
	void			*opaque;
#endif

	if (video == NULL || video->inFormatCtx == NULL ||
					video->nb_streams == 0)
	{
		return NULL;
	}

	i = 0;
	nb_codecs = 0;

	do
	{
		AVStream *is = cff_ffmpeg_get_stream_from_index(video, i);

		if (is == NULL)
		{
			continue;
		}

		codec_descriptors[nb_codecs] =
			avcodec_descriptor_get((is->codecpar)->codec_id);

		if (codec_descriptors[nb_codecs] != NULL)
		{
			const AVCodecDescriptor	**d;

			d = codec_descriptors;
			n = nb_codecs;

			while (--n >= 0)
			{
				if (*(d++) == codec_descriptors[nb_codecs])
				{
					break;
				}
			}

			if (n < 0)
			{
				nb_codecs++;

				if (nb_codecs >= (sizeof(codec_descriptors) /
						sizeof(AVCodecDescriptor*)) - 1)
				{
					break;
				}
			}

		} /* if (codec_descriptors[nb_codecs] != NULL) */
	}
	while (++i < video->nb_streams);

	if (nb_codecs == 0)
	{
		return NULL;
	}

	codec_descriptors[nb_codecs] = NULL;


	num = 0;

#ifdef HAVE_FFMPEG_ITERATE
	opaque = NULL;
#else
	ofmt = NULL;
#endif

	while ((ofmt = CffMuxerIterate(ofmt, &opaque)) != NULL)
	{
		const AVCodecDescriptor	*desc, **desc_ptr, **found;
		cff_fmtlist_t		*ptr;

		desc_ptr = codec_descriptors;
		found    = codecs_found;

		while ((desc = *(desc_ptr++)) != NULL)
		{
			if (av_codec_get_tag(ofmt->codec_tag, desc->id) != 0)
			{
				*(found++) = desc;
			}

			else if (ofmt->video_codec == desc->id ||
					ofmt->audio_codec == desc->id ||
						ofmt->subtitle_codec == desc->id)
			{
				*(found++) = desc;
			}

			else if (strcmp(ofmt->name, "mpegts") == 0 &&
				(ofmt->name == (video->inFormatCtx->iformat)->name ||
				strcmp(ofmt->name, (video->inFormatCtx->iformat)->name) == 0))
			{
				/* 'mpegts' does not show all codecs (e.g. 'hevc') */
				*(found++) = desc;
			}

		} /* while ((desc = *(desc_ptr++)) != NULL) */

		if (found == codecs_found)
		{
			continue;
		}

		*found	= NULL;
		ptr	= buf;
		n	= num;
		i	= -1;

		while (--n >= 0)
		{
			i = strcmp(ptr->name, ofmt->name);

			if (i == 0)
			{
				break;
			}

			if (i > 0)
			{
				memmove(&ptr[1], ptr,
					(n + 1) * sizeof(cff_fmtlist_t));
				break;
			}

			ptr++;

		} /* while (--n >= 0) */

		if (i != 0)
		{
			const char		*s1, *s2;
			size_t			slen;
			const AVCodecDescriptor	*d;
			char			sbuf[128];

			s1 = s2 = "";

			found	= codecs_found;
			sbuf[0]	= '(';
			slen	= sizeof(char);

			while ((d = *(found++)) != NULL)
			{
				const char end[] = ")";

				if (*found == NULL)
				{
					s2 = end;
				}

				n = snprintf(&sbuf[slen], sizeof(buf) - slen,
							"%s%s%s", s1, d->name, s2);
				if (n > 0)
				{
					slen += n;

					if (slen >= sizeof(sbuf))
					{
						slen = sizeof(sbuf) -
								sizeof(char);
						break;
					}

					if (s2 != end)
					{
						s1 = ", ";
					}
				}

			} /* while ((d = *(found++)) != NULL) */


			ptr->name   = ofmt->name;
			ptr->codecs = NULL;

			if (slen > sizeof(char))
			{
				ptr->codecs = cff_ffmpeg_memdup(sbuf,
							slen + sizeof(char));
			}

			num++;

			if (num >= (sizeof(buf) / sizeof(cff_fmtlist_t)) - 1)
			{
				break;
			}

		} /* if (i != 0) */

	} /* while ((ofmt = CffMuxerIterate(ofmt, &opaque)) != NULL) */

	if (num == 0)
	{
		return NULL;
	}

	buf[num].name   = NULL;
	buf[num].codecs = NULL;

	return cff_ffmpeg_memdup(buf, sizeof(cff_fmtlist_t) * (num + 1));

} /* cff_fmtlist_t *cff_ffmpeg_get_format_names(cff_video_t*) */

cff_bsflist_t *cff_ffmpeg_get_bitstream_filter_names(cff_video_t *video, int media_type)
{
	int			i, nb_codecs, has_media;
	unsigned int		num;
	void			*opaque;
	const AVBitStreamFilter	*bsf;
	const AVCodec		*codec;
	cff_bsflist_t		*ptr, buf[512];
	enum AVCodecID		codec_ids[128];

	if (video == NULL || video->inFormatCtx == NULL ||
					video->nb_streams == 0)
	{
		return NULL;
	}

	i = 0;
	nb_codecs = 0;
	has_media = 0;

	do
	{
		AVStream		*is;
		AVCodecParameters	*codecpar;
		
		is = cff_ffmpeg_get_stream_from_index(video, i);

		if (is == NULL)
		{
			continue;
		}

		codecpar = is->codecpar;

		if (codecpar->codec_type == AVMEDIA_TYPE_VIDEO)
		{
			if (media_type != kCFFMediaVideo)
			{
				continue;
			}
		}

		else if (codecpar->codec_type == AVMEDIA_TYPE_AUDIO)
		{
			if (media_type != kCFFMediaAudio)
			{
				continue;
			}
		}

		else
		{
			if (media_type != kCFFMediaOther)
			{
				continue;
			}
		}

		has_media = 1;
		codec = avcodec_find_decoder(codecpar->codec_id);

		if (codec != NULL)
		{
			codec_ids[nb_codecs] = codecpar->codec_id;
			nb_codecs++;

			if (nb_codecs >= (sizeof(codec_ids) /
						sizeof(enum AVCodecID)) - 1)
			{
				break;
			}

		} /* if (codec != NULL) */
	}
	while (++i < video->nb_streams);

	if (has_media == 0)
	{
		return NULL;
	}

	codec_ids[nb_codecs] = AV_CODEC_ID_NONE;


	num	= 0;
	ptr	= buf;
	opaque	= NULL;

	while ((bsf = CffBsfIterate(&opaque)) != NULL)
	{
		const char		*name;
		const enum AVCodecID	*id_ptr;

		name		= NULL;
		ptr->codecs	= NULL;
		id_ptr		= bsf->codec_ids;

		if (id_ptr == NULL)
		{
			/* All codecs */
			name = bsf->name;
		}

		else
		{
			enum AVCodecID codec_id;

			while ((codec_id = *(id_ptr++)) != AV_CODEC_ID_NONE)
			{
				enum AVCodecID		cid;
				const enum AVCodecID	*cid_ptr;

				cid_ptr = codec_ids;

				while ((cid = *(cid_ptr++)) != AV_CODEC_ID_NONE)
				{
					if (cid == codec_id)
					{
						break;
					}
				}

				if (cid != AV_CODEC_ID_NONE)
				{
					break;
				}

			} /* while ((codec_id = *(codec_ptr++)) != AV_CODEC_ID_NONE) */

			if (codec_id != AV_CODEC_ID_NONE)
			{
				char	*str;
				size_t	bufsz, slen;

				slen  = 0;
				bufsz = 1024;
				str   = cff_ffmpeg_get_bitstream_codecs(NULL, &slen, &bufsz, bsf->codec_ids);

				/* remove last '\n' */
				ptr->codecs = cff_ffmpeg_string_truncate(str, slen - 1);

				name = bsf->name;
			}

		} /* if (id_ptr == NULL) .. else */

		if (name != NULL)
		{
			ptr->name = name;

			ptr++;
			num++;

			if (num >= (sizeof(buf) / sizeof(cff_bsflist_t)) - 1)
			{
				break;
			}
		}

	} /* while ((bsf = CffBsfIterate(&opaque)) != NULL) */

	if (num == 0)
	{
		return NULL;
	}

	ptr->name   = NULL;
	ptr->codecs = NULL;

	return cff_ffmpeg_memdup(buf, sizeof(cff_bsflist_t) * (num + 1));

} /* cff_bsflist_t *cff_ffmpeg_get_bitstream_filter_names(cff_video_t*, int) */

int cff_ffmpeg_get_image_format_names(char ***long_name, char ***format_name)
{
	int			num;
	const AVOutputFormat	*ofmt;
	const char		*namebuf[512], *formatbuf[512];
#ifdef HAVE_FFMPEG_ITERATE
	void			*opaque;
#endif

	num = 0;

#ifdef HAVE_FFMPEG_ITERATE
	opaque = NULL;
#else
	ofmt = NULL;
#endif

	while ((ofmt = CffMuxerIterate(ofmt, &opaque)) != NULL)
	{
		enum AVCodecID codec_id;

		if (ofmt->mime_type == NULL ||
			strncmp(ofmt->mime_type, "image/", 6) != 0)
		{
			static const enum AVCodecID image_codecs[] =
			{
				AV_CODEC_ID_MJPEG,
				AV_CODEC_ID_LJPEG,
				AV_CODEC_ID_PNG,
				AV_CODEC_ID_BMP,
				AV_CODEC_ID_TIFF,
				AV_CODEC_ID_GIF,
				AV_CODEC_ID_NONE
			};

			enum AVCodecID		c;
			const enum AVCodecID	*p;

			c = ofmt->video_codec;
			p = image_codecs;

			while ((codec_id = *(p++)) != AV_CODEC_ID_NONE)
			{
				if (codec_id == c && (ofmt->mime_type == NULL ||
					strncmp(ofmt->mime_type, "video/", 6) == 0))
				{
					break;
				}
			}
		}

		else
		{
			codec_id = ofmt->video_codec;
		}

		if (codec_id != AV_CODEC_ID_NONE)
		{
			const AVCodec *codec =
					avcodec_find_encoder(codec_id);

			if (codec != NULL)
			{
				namebuf[num] = ofmt->long_name;

				if (namebuf[num] == NULL)
				{
					namebuf[num] = ofmt->name;
				}

				formatbuf[num] = ofmt->name;

				num++;

				if (num >= 510)
				{
					break;
				}

			} /* if (codec != NULL) */

		} /* if (codec_id != AV_CODEC_ID_NONE) */

	} /* while ((ofmt = CffMuxerIterate(ofmt, &next)) != NULL) */

	if (num == 0)
	{
		if (long_name != NULL)
		{
			*long_name = NULL;
		}

		if (format_name != NULL)
		{
			*format_name = NULL;
		}

		return 0;
	}

	namebuf[num]   = NULL;
	formatbuf[num] = NULL;
	num++;

	if (long_name != NULL)
	{
		*long_name = cff_ffmpeg_memdup(namebuf, sizeof(char*) * num);
	}

	if (format_name != NULL)
	{
		*format_name = cff_ffmpeg_memdup(formatbuf, sizeof(char*) * num);
	}

	return num - 1;

} /* int cff_ffmpeg_get_image_format_names(char***, char***) */

int cff_ffmpeg_get_encoder_names(cff_video_t *video, char ***encoder_names)
{
	int			index, num;
	AVStream		*is;
	enum AVCodecID		codec_id;
	struct CFFCodecNext	codec;
	const char		*namebuf[512];

	if (cff_ffmpeg_video_cut_list_distance(video) == 0)
	{
		goto _fail;
	}

	index = video->vframe.stream_index;

	if (index >= (video->inFormatCtx)->nb_streams)
	{
		goto _fail;
	}

	is = (video->inFormatCtx)->streams[index];

	if (is == NULL)
	{
		goto _fail;
	}

	codec_id = (is->codecpar)->codec_id;

	num		= 0;
	codec.codec 	= NULL;
	codec.opaque	= NULL;

	while (cff_ffmpeg_next_codec_for_id(codec_id, &codec, 1) == 0)
	{
		const enum AVPixelFormat *pixptr = codec.codec->pix_fmts;

		if (pixptr != NULL)
		{
			enum AVPixelFormat pixfmt;

			while ((pixfmt = *(pixptr++)) != -1)
			{
				if (pixfmt == (is->codecpar)->format)
				{
					namebuf[num] = codec.codec->name;
					num++;
					break;
				}
			}

			if (num >= 510)
			{
				break;
			}

		} /* if (pixptr != NULL) */

	} /* while (cff_ffmpeg_next_codec_for_id(codec_id, &codec, 1) == 0) */

	if (num == 0)
	{
_fail:
		if (encoder_names != NULL)
		{
			*encoder_names = NULL;
		}

		return 0;

	} /* if (num == 0) */

	namebuf[num] = NULL;

	if (encoder_names != NULL)
	{
		/* num + 1 => terminating NULL */
		*encoder_names = cff_ffmpeg_memdup(namebuf,
						sizeof(char*) * (num + 1));
	}

	return num;

} /* int cff_ffmpeg_get_encoder_names(cff_video_t*, char***) */

char* cff_ffmpeg_get_outformat_extension(const char *format_name, int *slen)
{
	unsigned int		n;
	char			*str;
	const AVOutputFormat	*format;

	if (format_name == NULL)
	{
		return NULL;
	}

	format = av_guess_format(format_name, NULL, NULL);

	if (format == NULL)
	{
		return NULL;
	}

	if (format->extensions == NULL)
	{
		return NULL;
	}

	if ((n = strcspn(format->extensions, ",")) == 0)
	{
		return NULL;
	}

	/* Plus '.' and '\0' */
	str = malloc(n + 2);

	if (str != NULL)
	{
		if (slen != NULL)
		{
			*slen = n + 1;
		}

		memcpy(&str[1], format->extensions, n);
		str[0]     = '.';
		str[n + 1] = '\0';
	}

	return str;

} /* char *cff_ffmpeg_get_outformat_extension(const char*, int*) */

static void cff_ffmpeg_dts2time(uint64_t dts, AVStream *is, unsigned int buf[4])
{
	unsigned int hour, min, sec;

	dts = av_rescale_q(dts, is->time_base, AV_TIME_BASE_Q);

	sec  = dts / AV_TIME_BASE;
	min  = sec / 60;
	hour = min / 60;

	buf[0] = hour;
	buf[1] = min % 60;
	buf[2] = sec % 60;
	buf[3] = (dts % AV_TIME_BASE) / 1000;

	if (buf[3] < 999 && (dts % AV_TIME_BASE) >= AV_TIME_BASE / 2)
	{
		buf[3]++;
	}

} /* static void cff_ffmpeg_dts2time(uint64_t, unsigned int[4]) */

static int64_t cff_ffmpeg_sec2dts(int sec, cff_video_t *video)
{
	int64_t		dts, dts_mod;
	AVStream	*is;

	is = (video->inFormatCtx->streams[video->vframe.stream_index]);

	dts =  av_rescale_q((int64_t) sec * video->vframe.frame_duration,
			(AVRational) {1, video->vframe.frame_duration}, is->time_base);

	dts_mod = dts % video->vframe.frame_duration;

	if (dts_mod == 0)
	{
		return dts;
	}

	if (sec > 0)
	{
		/* round up */
		return dts + video->vframe.frame_duration - dts_mod;
	}

	/* round down */
	return dts - video->vframe.frame_duration - dts_mod;

} /* static int64_t cff_ffmpeg_sec2dts(int, cff_video_t*) */

static const char *cff_ffmpeg_get_pictype_str(const AVFrame *frame)
{
	static const char *picture_type[] =
		{ "?", "I", "P", "B", "S", "SI", "SP", "BI" };

	static char pictype_buf[5];

	int key_frame, pict_type;

	key_frame = 0;
	pict_type = 0;

	if (frame != NULL)
	{
		if (frame->key_frame)
		{
			strcpy(pictype_buf, "K/");
			key_frame = 2;
		}

		if (frame->pict_type <= AV_PICTURE_TYPE_BI)
		{
			pict_type = frame->pict_type;
		}
	}

	strcpy(&pictype_buf[key_frame], picture_type[pict_type]);

	return pictype_buf;

} /* static const char *cff_ffmpeg_get_pictype_str(const AVFrame*) */

int cff_ffmpeg_get_current_loglevel_index(void)
{
	if (gLogBufferPtr != NULL)
	{
		int			loglevel;
		const LoglevelTab	*ptr;

		loglevel = av_log_get_level();

		ptr = loglevel_tab;

		do
		{
			if (loglevel == ptr->value)
			{
				return ptr - loglevel_tab;
			}

			ptr++;
		}
		while (ptr->name != NULL);

	} /* if (gLogBufferPtr != NULL) */

	return -1;

} /* int cff_ffmpeg_get_current_loglevel_index(void) */

static const LoglevelTab *cff_ffmpeg_get_loglevel_ptr(const char *logstr)
{
	if (logstr != NULL)
	{
		const LoglevelTab *p = loglevel_tab;

		do
		{
			if (strcmp(logstr, p->name) == 0)
			{
				return p;
			}

			p++;
		}
		while (p->name != NULL);

	} /* if (logstr != NULL) */

	return NULL;

} /* static const LoglevelTab *cff_ffmpeg_get_loglevel_ptr(const char*) */

const char *cff_ffmpeg_get_loglevel_name(int logindex)
{
	if (logindex < 0 ||
		logindex >= sizeof(loglevel_tab) / sizeof(LoglevelTab))
	{
		return NULL;
	}

	return loglevel_tab[logindex].name;

} /* const char *cff_ffmpeg_get_loglevel_name(int) */

static void cff_ffmpeg_log_callback(void *ptr, int level, const char *fmt, va_list vl)
{
	/* FFmpeg => libavutil/log.c */

	static int skip_prefix = 0;

	int	n, len, sz;
	AVClass	*avc;
	char	*str, buf[kCffLogBufferSize];

	level &= 0x00FF;

	if (level > av_log_get_level())
	{
		return;
	}

	len = 0;
	avc = (ptr == NULL) ? NULL : *(AVClass **) ptr;

	if (avc != NULL && skip_prefix == 0)
	{
		static const char null_str[] = "NULL";
		static const char avcfmt[]   = "[%s @ %p] ";

		const char *item_name;

		if (avc->parent_log_context_offset > 0 &&
					avc->category != AV_CLASS_CATEGORY_NA)
		{
			AVClass **parent = *(AVClass ***)
				(((uint8_t *) ptr) + avc->parent_log_context_offset);

			if (parent != NULL && *parent != NULL)
			{
				item_name = (*(*parent)->item_name)(parent);

				if (item_name == NULL)
				{
					item_name = null_str;
				}

				sz = snprintf(buf, sizeof(buf),
						avcfmt, item_name, parent);
				if (sz > 0)
				{
					if (sz >= sizeof(buf))
					{
						sz = sizeof(buf) - sizeof(char);
					}

					len = sz;
				}
			}

		} /* if (avc->parent_log_context_offset > 0 &&
					avc->category != AV_CLASS_CATEGORY_NA) */

		item_name = (*avc->item_name)(ptr);

		if (item_name == NULL)
		{
			item_name = null_str;
		}

		sz = sizeof(buf) - len;
		n  = snprintf(&buf[len], sz, avcfmt, item_name, ptr);

		if (n > 0)
		{
			if (n >= sz)
			{
				n = sz - sizeof(char);
			}

			len += n;
		}

	} /* if (avc != NULL && skip_prefix == 0) */

	sz = sizeof(buf) - len;
	n  = vsnprintf(&buf[len], sz, fmt, vl);

	if (n > 0)
	{
		if (n >= sz)
		{
			n = sz - sizeof(char);
		}

		len += n;
	}

	if (len > 0)
	{
		cff_logbuffer_t *logbuf;

		CFF_SEMAPHORE_LOCK();
		
		logbuf = gLogBufferPtr;

		if (logbuf != NULL)
		{
			int last_prefix = skip_prefix;

			if (buf[len - 1] == '\n' || buf[len - 1] == '\r')
			{
				skip_prefix = 0;
			}

			else
			{
				skip_prefix = 1;
			}

			if (logbuf->func != NULL &&
				logbuf->str != NULL &&
					len == logbuf->len &&
						(last_prefix | skip_prefix) == 0 &&
							strcmp(buf, logbuf->str) == 0)
			{
				static const char s_str[] = "s";

				const char *s = s_str;

				logbuf->prev++;

				if (logbuf->prev == 1)
				{
					/* One: time - else: times */
					s++;
				}

				len = snprintf(buf, sizeof(buf),
					"\rLast message repeated %lu time%s\n",
					logbuf->prev, s);

				str = (len > 0) ? buf : NULL;
			}

			else
			{
				logbuf->prev = 0;

				if (last_prefix == 0 && logbuf->func != NULL)
				{
					/* New line */
					logbuf->len = 0;
				}

				str = cff_ffmpeg_printf_concat(logbuf->str,
					&(logbuf->len),
					&(logbuf->max),
					"%s",
					buf);

				if (str != NULL)
				{
					logbuf->str = str;
					len = logbuf->len;
				}

			} /* if (logbuf->func != NULL &&
					logbuf->str != NULL &&
						len == logbuf->len && ... */

			if (str != NULL && skip_prefix == 0)
			{
				/* logbuf->func == NULL => cff_ffmpeg_file_info_io */

				if (logbuf->func != NULL &&
					(*logbuf->func)(str, len, logbuf->data) != 0)
				{
					/* Restore previous level */
					cff_ffmpeg_set_loglevel(NULL, 0, NULL, NULL);
					logbuf->len  = 0;
					logbuf->prev = 0;
				}
			}

		} /* if (logbuf != NULL) */

		CFF_SEMAPHORE_UNLOCK();

	} /* if (len > 0) */

} /* static void cff_ffmpeg_log_callback(void*, int, const char*, va_list) */

int cff_ffmpeg_set_loglevel(const char *logname, int lock, func_log_callback_t func, void *user_data)
{
	cff_logbuffer_t		*logbuf;
	const LoglevelTab	*ptr;

#ifdef HAVE_SEMAPHORE
	if (lock)
	{
		CFF_SEMAPHORE_LOCK();
	}
#endif
	if (func == NULL ||
		(ptr = cff_ffmpeg_get_loglevel_ptr(logname)) == NULL ||
						ptr->value == AV_LOG_QUIET)
	{
		if ((logbuf = gLogBufferPtr) != NULL)
		{
			if (logbuf->str != NULL)
			{
				free(logbuf->str);
			}

			free(logbuf);
			gLogBufferPtr = NULL;
		}

		av_log_set_level(gLoglevel);
		av_log_set_callback(av_log_default_callback);

#ifdef HAVE_SEMAPHORE
		if (lock)
		{
			CFF_SEMAPHORE_UNLOCK();
		}
#endif
		return -1;
	}

	if ((logbuf = gLogBufferPtr) == NULL)
	{
		logbuf = malloc(sizeof(cff_logbuffer_t));

		if (logbuf == NULL)
		{
#ifdef HAVE_SEMAPHORE
			if (lock)
			{
				CFF_SEMAPHORE_UNLOCK();
			}
#endif
			return -1;
		}

		gLogBufferPtr = memset(logbuf, '\0', sizeof(cff_logbuffer_t));
		logbuf->max   = 512;
	}

	logbuf->data = user_data;
	logbuf->func = func;

	av_log_set_level(ptr->value);
	av_log_set_callback(cff_ffmpeg_log_callback);

#ifdef HAVE_SEMAPHORE
	if (lock)
	{
		CFF_SEMAPHORE_UNLOCK();
	}
#endif
	return 0;

} /* int cff_ffmpeg_set_loglevel(const char*, int, func_log_callback_t, void*) */

int cff_ffmpeg_init(const char *logname)
{
	int loglevel;

	if (logname == NULL)
	{
		loglevel = AV_LOG_QUIET;
	}

	else
	{
		const LoglevelTab *ptr;

		ptr = cff_ffmpeg_get_loglevel_ptr(logname);

		if (ptr == NULL)
		{
			return -1;
		}

		loglevel = ptr->value;
	}

	gLoglevel = loglevel;

	memset(&gVideo, '\0', sizeof(cff_video_t));
	CffRegisterAll();
	av_log_set_level(loglevel);
	gVideo.vframe.pict_type = cff_ffmpeg_get_pictype_str(NULL);

	CFF_SEMAPHORE_INIT();

	return 0;

} /* int cff_ffmpeg_init(const char *loglevel) */

static enum AVCodecID cff_ffmpeg_is_video_stream(cff_video_t *video, int index)
{
	enum AVCodecID	codec_id;
	AVStream	*is;

	codec_id = AV_CODEC_ID_NONE;

	is = (video->inFormatCtx)->streams[index];

	if ((is->codecpar)->codec_type == AVMEDIA_TYPE_VIDEO)
	{
		video->vframe.stream_index = index;
		codec_id = (is->codecpar)->codec_id;
	}

	return codec_id;

} /* static enum AVCodecID cff_ffmpeg_is_video_stream(cff_video_t *video, int index) */

static enum AVCodecID cff_ffmpeg_get_program_video_stream(cff_video_t *video, int index)
{
	if (index < (video->inFormatCtx)->nb_programs &&
			(video->inFormatCtx)->programs != NULL)
	{
		unsigned int		nb_idx, *stream_index;
		const AVProgram		*prg;

		prg = (video->inFormatCtx)->programs[index];

		if (prg == NULL)
		{
			return AV_CODEC_ID_NONE;
		}

		stream_index = prg->stream_index;

		if (stream_index == NULL)
		{
			return AV_CODEC_ID_NONE;
		}

		nb_idx = prg->nb_stream_indexes;

		while (--nb_idx >= 0)
		{
			enum AVCodecID codec_id =
				cff_ffmpeg_is_video_stream(video,
						stream_index[nb_idx]);

			if (codec_id != AV_CODEC_ID_NONE)
			{
				video->nb_streams = prg->nb_stream_indexes;
				return codec_id;
			}
		}

	} /* if (index < (video->inFormatCtx)->nb_programs &&
				(video->inFormatCtx)->programs != NULL) */

	return AV_CODEC_ID_NONE;

} /* static enum AVCodecID cff_ffmpeg_get_program_video_stream(cff_video_t*, int) */

static int cff_ffmpeg_read_packet(cff_video_t *video, AVPacket *packet, const char **error)
{
	int retv;

	retv = av_read_frame(video->inFormatCtx, packet);

	if (retv < 0)
	{
		AVIOContext *pb = (video->inFormatCtx)->pb;

		if (avio_feof(pb) != 0 ||
			avio_tell(pb) >= video->vframe.io_seek_end)
		{
			*error = NULL;
		}

		else
		{
			*error = gettext_nop("'av_read_frame()' failed!");
		}

	} /* if (retv < 0) */

	return retv;

} /* static int cff_ffmpeg_read_packet(cff_video_t*, AVPacket*, const char**) */

static const char *cff_ffmpeg_get_frame_dts(cff_video_t *video)
{
	/*
		Get first or last frame dts,
		min and max gop_size,
		and max frame duration
	*/

	int		frame_count, retv;
	unsigned int	i;
	const char	*str;
	AVFrame		*frame;
	AVPacket	*pktptr;
	int64_t		duration, lastgop, dts_next, dts_drain[20];

	str         = NULL;
	i           = 0;
	frame_count = 0;
	duration    = 0;
	dts_next    = AV_NOPTS_VALUE;
	lastgop     = AV_NOPTS_VALUE;
	pktptr      = video->vframe.tmppkt;
	frame       = video->vframe.avframe;

	if (video->vframe.frame_width > kCFFOpenFileGetFirstFrame)
	{
		/* Key-frame frome previous loop */
		frame_count++;
	}

	do
	{
		if (pktptr != NULL)
		{
			retv = cff_ffmpeg_read_packet(video, pktptr, &str);

			if (retv < 0)
			{
				/* EOF or Error */

				if (str == NULL)
				{
					/* EOF */
					pktptr = NULL;
				}

				else
				{
					if (video->vframe.frame_width !=
							kCFFOpenFileGetFirstFrame)
					{
						str = NULL;
					}

					break;
				}
			}

			else if (pktptr->stream_index != video->vframe.stream_index)
			{
				/* Another stream */
				continue;
			}

			else
			{
				if (video->vframe.frame_width == kCFFOpenFileGetFirstFrame)
				{
					if (video->vframe.bsf_mp4toannexb == NULL &&
						((video->vframe.ic)->codec_id == AV_CODEC_ID_H264 ||
							(video->vframe.ic)->codec_id == AV_CODEC_ID_H265))
					{
						/* libavformat->rawenc.c */

						if (pktptr->data != NULL && pktptr->size >=5 &&
								AV_RB32(pktptr->data) != 0x0000001 &&
									AV_RB24(pktptr->data) != 0x000001)
						{
							if ((video->vframe.ic)->codec_id == AV_CODEC_ID_H264)
							{
								video->vframe.bsf_mp4toannexb =
											"h264_mp4toannexb";
							}

							else /* AV_CODEC_ID_H265 */
							{
								video->vframe.bsf_mp4toannexb =
											"hevc_mp4toannexb";
							}
							
						} /* if (pktptr->data != NULL && pktptr->size >=5 && .. */

					} /* if (video->copy.softcut.bsf_mp4toannexb == NULL && .. */
				}

				else if (frame_count == 1 &&
					video->vframe.frame_width ==
							kCFFOpenFileGetLastFrame)
				{
					/* Wait for a key-frame */

					if ( ! (pktptr->flags & AV_PKT_FLAG_KEY))
					{
						continue;
					}

					/* Found */
					frame_count++;
				}
			}

		} /* if (pktptr != NULL) */

		if (pktptr != NULL || retv != 0)
		{
			retv = avcodec_send_packet(video->vframe.ic, pktptr);

			if (retv != 0)
			{
				/* Failed */
				if (retv == AVERROR(EAGAIN))
				{
					retv = 0;
					continue;
				}

				if ((video->vframe.frame_duration == 0 ||
						video->vframe.frame_width ==
							kCFFOpenFileGetLastFrame) &&
					pktptr != NULL &&
					! (pktptr->flags & AV_PKT_FLAG_KEY))
				{
					/* First or last frame(s) is no keyframe */
					retv = 0;
				}

				continue;
			}

		} /* if (pktptr != NULL || retv != 0) */

		retv = avcodec_receive_frame(video->vframe.ic, frame);

		if (retv != 0)
		{
			/* Failed */
			if (retv == AVERROR(EAGAIN))
			{
				retv = 0;
			}

			continue;
		}

		if (frame->key_frame != 0 && frame->pkt_dts != AV_NOPTS_VALUE)
		{
			/* Start of the last gop */
			lastgop = frame->pkt_dts;
		}

		if (video->vframe.frame_height == 0)
		{
			int64_t frame_duration;

			frame_duration = av_rescale_q_rnd(video->vframe.ic->ticks_per_frame,
				(video->vframe.ic)->time_base,
				(video->inFormatCtx)->streams[video->vframe.stream_index]->time_base,
				AV_ROUND_DOWN);

			if (frame_duration < frame->pkt_duration)
			{
				/* Maybe wrong - disable */
				frame_duration = 0;
			}

			if (frame_duration > 0)
			{
				video->vframe.frame_duration = frame_duration;
				video->vframe.frame_height   = 1;
			}

			else
			{
				frame_duration = frame->pkt_duration;

				if (frame_duration == 0 && dts_next != AV_NOPTS_VALUE)
				{
					frame_duration = frame->pkt_dts - dts_next;
				}

				if (frame_duration > video->vframe.frame_duration)
				{
					/* Duration (may differ from frame to frame */
					video->vframe.frame_duration = frame_duration;
				}
			}

		} /* if (video->vframe.frame_height == 0) */

		if (video->vframe.frame_width < kCFFOpenFileGetLastFrame)
		{
			if (frame->pkt_dts != AV_NOPTS_VALUE)
			{
				dts_next = frame->pkt_dts + frame->pkt_duration;

				if (video->vframe.dts_start == AV_NOPTS_VALUE)
				{
					/* Time of the first frame */
					video->vframe.dts_start = frame->pkt_dts - duration;
				}
			}

			if (video->vframe.pktpos_start == -1)
			{
				/* Position of the first frame */
				video->vframe.pktpos_start = frame->pkt_pos;
			}

			duration += frame->pkt_duration;

			if (frame->key_frame != 0)
			{
				if (frame_count > 0)
				{
					if (video->vframe.min_gop_size == 0 ||
							video->vframe.min_gop_size > frame_count)
					{
						/* Min gop size */
						video->vframe.min_gop_size = frame_count;
					}

					if (frame_count > video->vframe.max_gop_size)
					{
						/* Max gop size */
						video->vframe.max_gop_size = frame_count;
					}

					if (pktptr == NULL)
					{
						goto _last_frame;
					}

					/* Break */
					retv = -1;
					continue;
				}

				else
				{
					frame_count++;
				}
			}

			else if (frame_count > 0)
			{
				frame_count++;
			}

			if (pktptr == NULL)
			{
				goto _last_frame;
			}
		}

		else
		{
_last_frame:
			if (frame->pkt_dts != AV_NOPTS_VALUE)
			{
				/* Last frame */
				video->vframe.dts_stop = frame->pkt_dts;
				dts_next = frame->pkt_dts + frame->pkt_duration;
			}

			else if (pktptr == NULL && dts_next != AV_NOPTS_VALUE && i < 20)
			{
				/* Last frame */
				video->vframe.dts_stop = dts_next;
				dts_drain[i] = dts_next;
				i++;
				dts_next += frame->pkt_duration;
			}

		} /* if (video->vframe.frame_width < kCFFOpenFileGetLastFrame) .. else */
	}
	while (av_packet_unref(video->vframe.tmppkt), retv == 0);

	if (i > 0)
	{
		/* Drain mode */
		video->vframe.drain = malloc(sizeof(cff_drain_t) +
						((i - 1) * sizeof(int64_t)));
		if (video->vframe.drain != NULL)
		{
			(video->vframe.drain)->num = i;
			(video->vframe.drain)->idx = 0;
			memcpy((video->vframe.drain)->pkt_dts, dts_drain,
							i * sizeof(int64_t));
		}

		video->vframe.dts_lastgop = lastgop;
	}

	else if (video->vframe.frame_width == kCFFOpenFileGetLastFrame)
	{
		/* No drain mode */
		video->vframe.dts_lastgop = lastgop;

	} /* if (i > 0) .. else */

	return str;

} /* static const char *cff_ffmpeg_get_frame_dts(cff_video_t*) */

int cff_ffmpeg_file_open_cb(cff_video_t *video, double *progress, const char **error)
{
	const char	*str;
	int64_t		dts_seek;

	if (video->vframe.frame_width == kCFFOpenFileSeekLastGop)
	{
		dts_seek = video->vframe.dts_stop;

		for (;;)
		{
			dts_seek -= video->vframe.max_gop_size *
						video->vframe.frame_duration;

			if (av_seek_frame(video->inFormatCtx,
				video->vframe.stream_index, dts_seek, 0) == 0)
			{
				avcodec_flush_buffers(video->vframe.ic);
				str = NULL;
				break;
			}

			if (dts_seek <= video->vframe.dts_start)
			{
				str = gettext_nop("Cannot seek to last frame!");
				break;
			}

		} /* for (;;) */
	}

	else
	{
		/* Get frame infos */

		int	n;
		int64_t	dts_stop;

		dts_stop = video->vframe.dts_stop;
		n = 1;

		for (;;)
		{
			str = cff_ffmpeg_get_frame_dts(video);

			if (video->vframe.frame_width < kCFFOpenFileGetLastFrame)
			{
				/* Not seeking the last frame */
				break;
			}

			if (video->vframe.dts_lastgop != AV_NOPTS_VALUE)
			{
				/* Last gop found */
				break;
			}

			if (--n < 0)
			{
				/* Stop seeking */
				break;
			}

			/* Last gop not found, try once again */

			dts_seek = dts_stop - (2 * video->vframe.max_gop_size *
							video->vframe.frame_duration);

			if (av_seek_frame(video->inFormatCtx,
				video->vframe.stream_index, dts_seek, 0) < 0)
			{
				break;
			}

			if (video->vframe.drain != NULL)
			{
				free(video->vframe.drain);
			}

			avcodec_flush_buffers(video->vframe.ic);

		} /* for (;;) */

		if (str == NULL)
		{
			static const char cannot_get_frame_duration_str[] =
				gettext_nop("Cannot get the frame duration!");

			if (video->vframe.frame_duration == 0)
			{
				str = cannot_get_frame_duration_str;
			}

			else if (video->vframe.frame_width == kCFFOpenFileSeekLastGop - 1)
			{
				if (video->vframe.min_gop_size == 0 ||
						video->vframe.max_gop_size == 0)
				{
					str = gettext_nop("Cannot get the gop size!");
				}
			}

			else if (video->vframe.frame_width == kCFFOpenFileGetLastFrame)
			{
				if (video->vframe.frame_duration == 0)
				{
					str = cannot_get_frame_duration_str;
				}
			}
		}

		else if (video->vframe.frame_width == kCFFOpenFileGetLastFrame)
		{
			if (dts_stop != video->vframe.dts_stop)
			{
				/* Not EOF, but 'dts_stop' has changed */
				str = NULL;
			}

		} /* if (str == NULL) .. else */

	} /* if (video->vframe.frame_width == kCFFOpenFileSeekLastGop) .. else */

	if (error != NULL)
	{
		*error = str;
	}

	if (str == NULL)
	{
		if (video->vframe.frame_width < kCFFOpenFileGetLastFrame)
		{
			video->vframe.frame_width++;

			if (progress != NULL)
			{
				*progress = (double) video->vframe.frame_width /
							(double) kCFFOpenFileGetLastFrame;
			}

			return 0;
		}
	}

	else
	{
		/* Failure */
		avcodec_free_context(&(video->vframe.ic));

	} /* if (str == NULL) .. else */

	return 1;

} /* int cff_ffmpeg_file_open_cb(cff_video_t*, double*, const char**) */

static const char *cff_ffmpeg_find_codec(cff_video_t *video,
				enum AVCodecID codec_id,
				func_cff_callback_t func, void *user_data)
{
	int		retv;
	const char	*str;
	const AVCodec	*codec;
	AVCodecContext	*ic;
	AVStream	*is;
	AVDictionary	*opts;

	codec = avcodec_find_decoder(codec_id);

	if (codec == NULL)
	{
		return g_video_codec_errstr;
	}

	ic = avcodec_alloc_context3(NULL);
	video->vframe.ic = ic;

	if (ic == NULL)
	{
		return g_alloc_codec_context_errstr;
	}

	ic->flags = AV_CODEC_CAP_DR1 | AV_CODEC_CAP_DELAY;

	is = (video->inFormatCtx)->streams[video->vframe.stream_index];

	avcodec_parameters_to_context(ic, is->codecpar);
	ic->pkt_timebase = is->time_base;

	opts = NULL;

	av_dict_set(&opts, "threads", "auto", 0);
	av_dict_set_int(&opts, "refcounted_frames", 1, 0);
	av_dict_set(&opts, "flags", "+low_delay", 0);

	retv = avcodec_open2(ic, codec, &opts);
	av_dict_free(&opts);

	if (retv < 0)
	{
		avcodec_free_context(&(video->vframe.ic));
		return g_open_codec_errstr;
	}

	video->vframe.tmppkt = av_packet_alloc();

	if (video->vframe.tmppkt == NULL)
	{
		avcodec_free_context(&(video->vframe.ic));
		return g_alloc_packet_errstr;
	}

	video->vframe.avframe = av_frame_alloc();

	if (video->vframe.avframe == NULL)
	{
		avcodec_free_context(&(video->vframe.ic));
		return g_alloc_frame_errstr;
	}

	/* ffmpeg.c 4.2.3 line 2388 */

	if (ic->codec_id == AV_CODEC_ID_H264 &&
			ic->has_b_frames > is->codecpar->video_delay)
	{
		(is->codecpar)->video_delay = ic->has_b_frames;
	}

	/* Gather some informations */

	if (is->start_time == AV_NOPTS_VALUE || is->duration == AV_NOPTS_VALUE)
	{
		video->vframe.stream_duration =
			av_rescale_q((video->inFormatCtx)->duration,
					AV_TIME_BASE_Q, is->time_base);
		video->vframe.dts_start =
			av_rescale_q((video->inFormatCtx)->start_time,
					AV_TIME_BASE_Q, is->time_base);
	}

	else
	{
		video->vframe.stream_duration  = is->duration;
		video->vframe.dts_start = is->start_time;
	}

	video->vframe.dts_stop = video->vframe.dts_start + video->vframe.stream_duration;

	is->discard = AVDISCARD_DEFAULT;

	/* Get the start position */

	av_seek_frame(video->inFormatCtx, video->vframe.stream_index, -1, AVSEEK_FLAG_ANY);
	video->vframe.io_seek_start = avio_tell((video->inFormatCtx)->pb);
	video->fsize = avio_size((video->inFormatCtx)->pb);

	/* Get first and last frame, gop_size and frame duration */

	video->vframe.frame_width  = kCFFOpenFileGetFirstFrame;
	video->vframe.dts_start    = AV_NOPTS_VALUE;
	video->vframe.pktpos_start = -1;
	video->vframe.dts_lastgop  = AV_NOPTS_VALUE;

	if (func == NULL)
	{
		while ((retv = cff_ffmpeg_file_open_cb(video, NULL, &str)) == 0) {;}
	}

	else
	{
		retv = (*func)(user_data);
		str  = NULL;
	}

	av_frame_free(&video->vframe.avframe);
	video->vframe.frame_width  = 0;
	video->vframe.frame_height = 0;

	if (retv == 0 || video->vframe.ic == NULL)
	{
		/* Error, or user canceled */
		if (video->vframe.ic != NULL)
		{
			avcodec_free_context(&(video->vframe.ic));
		}

		av_packet_free(&video->vframe.tmppkt);

		return str;
	}

	video->vframe.io_seek_end = avio_tell((video->inFormatCtx)->pb);

	video->vframe.min_gop_size *= video->vframe.frame_duration;
	video->vframe.max_gop_size *= video->vframe.frame_duration;

	video->vframe.current.frame	= video->vframe.dts_stop;
	video->vframe.current.packet	= video->vframe.dts_stop;
	video->vframe.stream_duration	= video->vframe.dts_stop - video->vframe.dts_start;
	video->vframe.frame_total	= video->vframe.stream_duration / video->vframe.frame_duration;

	/* Duration (h:m:s) */
	cff_ffmpeg_dts2time(video->vframe.stream_duration, is, video->vframe.time_total);

	return NULL;

} /* static const char *cff_ffmpeg_find_codec(cff_video_t*, enum AVCodecID, func_cff_callback_t, void*) */

const char *cff_ffmpeg_file_open(cff_video_t *video, const char *fname,
					func_cff_callback_t func, void *user_data)
{
	int		i, retv;
	enum AVCodecID	codec_id;
	AVFormatContext	*inFormatCtx;
	AVDictionary	*opts;

	if (fname == NULL || fname[0] == '\0')
	{
		return g_missing_filename_errstr;
	}

	cff_ffmpeg_file_close(video);

	video->fname = cff_ffmpeg_strdup((char *) fname, -1);

	if (video->fname == NULL)
	{
		return gettext_nop("Cannot copy the filename!");
	}

	inFormatCtx = avformat_alloc_context();
	video->inFormatCtx = inFormatCtx;

	if (inFormatCtx == NULL)
	{
		return gettext_nop("Cannot allocate input context!");
	}

	opts = NULL;
	av_dict_set(&opts, "fflags", "genpts", 0);
	av_dict_set(&opts, "avoid_negative_ts", "make_zero", 0);

	/* Open video file */
	retv = avformat_open_input(&inFormatCtx, video->fname, NULL, &opts);

	av_dict_free(&opts);

	if (retv != 0)
	{
		av_dict_free(&opts);
		avformat_free_context(inFormatCtx);
		video->inFormatCtx = NULL;
	 	return gettext_nop("Could not open the input file!");
	}

	/* Retrieve stream information */

	if (avformat_find_stream_info(video->inFormatCtx, NULL) != 0)
	{
		avformat_close_input(&inFormatCtx);
	 	return gettext_nop("Could not find stream information!");
	}


	video->nb_streams = inFormatCtx->nb_streams;

	if (video->nb_streams == 0)
	{
		return gettext_nop("No video stream found!");
	}

	codec_id = AV_CODEC_ID_NONE;
	i = inFormatCtx->nb_programs;

	if (i > 0)
	{
		unsigned int	*ptr;
		AVProgram	**program;

		program = inFormatCtx->programs;

		if (program != NULL &&
			(ptr = av_malloc(i * sizeof(int))) != NULL)
		{
			int index;

			video->program.index = ptr;
			index = 0;

			do
			{
				codec_id = cff_ffmpeg_get_program_video_stream(video, index);

				if (codec_id != AV_CODEC_ID_NONE)
				{
					*(ptr++) = index;
					video->program.num++;
				}
			}
			while (++index < i);

			if (video->program.num == 0)
			{
				/* No video stream found */
				av_freep(&(video->program.index));
			}

			else
			{
				/* Open first program in the list */
				i = video->program.index[0];
				codec_id = cff_ffmpeg_get_program_video_stream(video, i);
			}

		} /* if (program != NULL &&
				(ptr = av_malloc(i * sizeof(int))) != NULL) */
	} /* if (i > 0) */

	if (codec_id == AV_CODEC_ID_NONE)
	{
		i = 0;

		do
		{
			codec_id = cff_ffmpeg_is_video_stream(video, i);
		}
		while (codec_id == AV_CODEC_ID_NONE &&
					++i < inFormatCtx->nb_streams);

		if (codec_id == AV_CODEC_ID_NONE)
		{
			return g_video_codec_errstr;
		}

	} /* if (codec_id == AV_CODEC_ID_NONE) */

	i = 0;

	do
	{
		AVStream *is = (video->inFormatCtx)->streams[i];

		is->discard = AVDISCARD_ALL;
	}
	while (++i < inFormatCtx->nb_streams);

	return cff_ffmpeg_find_codec(video, codec_id, func, user_data);

} /* const char *cff_ffmpeg_file_open(cff_video_t*, const char*, func_cff_callback_t, void*) */

const char *cff_ffmpeg_program_open(cff_video_t *video, unsigned int index,
				func_cff_callback_t func, void *user_data)
{
	if (index < video->program.num && video->program.index != NULL)
	{
		int		i, prev_index;
		enum AVCodecID	codec_id;

		prev_index = video->vframe.stream_index;
		video->program.selected = index;

		if (video->vframe.ic)
		{
			avcodec_free_context(&(video->vframe.ic));
		}

		if (video->vframe.drain)
		{
			free(video->vframe.drain);
		}

		av_freep(&(video->vframe.rgb24.data));
		av_frame_free(&(video->vframe.avframe));

		cff_ffmpeg_video_cut_list_clear(video);

		cff_ffmpeg_video_undo_clear(video->vframe.undo);
		cff_ffmpeg_video_undo_clear(video->vframe.redo);

		memset(&(video->vframe), '\0', sizeof(cff_vframe_t));

		video->vframe.pict_type = cff_ffmpeg_get_pictype_str(NULL);

		avio_flush((video->inFormatCtx)->pb);
		avformat_flush(video->inFormatCtx);

		i = video->program.index[index];
		codec_id = cff_ffmpeg_get_program_video_stream(video, i);

		if (codec_id != AV_CODEC_ID_NONE)
		{
			const char	*str;
			AVStream	*is;

			is = (video->inFormatCtx)->streams[prev_index];
			is->discard = AVDISCARD_ALL;

			str = cff_ffmpeg_find_codec(video,
						codec_id, func, user_data);
			if (str != NULL)
			{
				/* failed */

				is = (video->inFormatCtx)->streams[prev_index];
				is->discard = AVDISCARD_ALL;
			}

			return str;

		} /* if (codec_id != AV_CODEC_ID_NONE) */

	} /* if (index < video->program.num && video->program.index != NULL) */

	return g_video_codec_errstr;

} /* const char *cff_ffmpeg_program_open(cff_video_t*, unsigned int, func_cff_callback_t, void*) */

const char *cff_ffmpeg_get_program_name(cff_video_t *video, unsigned int index, int *prgnum)
{
	if (index < video->program.num && video->program.index != NULL)
	{
		unsigned int	i;
		const AVProgram	*prg;

		i   = video->program.index[index];
		prg = (video->inFormatCtx)->programs[i];

		if (prg != NULL)
		{
			AVDictionaryEntry *entry;

			if (prgnum != NULL)
			{
				*prgnum = prg->program_num;
			}

			entry = av_dict_get(prg->metadata,
						"service_name", NULL, 0);
			if (entry != NULL)
			{
				return entry->value;
			}

		} /* if (prg != NULL) */

	} /* if (index < video->program.num && video->program.index != NULL) */

	if (prgnum != NULL)
	{
		*prgnum = -1;
	}

	return NULL;

} /* const char *cff_ffmpeg_get_program_number(cff_video_t*, unsigned int, int*) */

void cff_ffmpeg_file_close(cff_video_t *video)
{
	if (video->fname)
	{
		/* malloc'ed */
		free(video->fname);
	}

	if (video->vframe.ic)
	{
		avcodec_free_context(&(video->vframe.ic));
	}

	if (video->vframe.drain)
	{
		/* malloc'ed */
		free(video->vframe.drain);
	}

	if (video->inFormatCtx)
	{
		avformat_close_input(&(video->inFormatCtx));
	}

	if (video->cut.list)
	{
		/* av_malloc'ed */
		av_free(video->cut.list);
	}

	if (video->program.index)
	{
		/* av_malloc'ed */
		av_free(video->program.index);
	}

	cff_ffmpeg_video_undo_clear(video->vframe.undo);
	cff_ffmpeg_video_undo_clear(video->vframe.redo);

	/* av_malloc'ed */
	av_freep(&(video->vframe.rgb24.data));
	av_packet_free(&(video->vframe.tmppkt));
	av_frame_free(&(video->vframe.avframe));

	memset(video, '\0', sizeof(cff_video_t));

	video->vframe.pict_type = cff_ffmpeg_get_pictype_str(NULL);

} /* void cff_ffmpeg_file_close(cff_video_t*) */

static char *cff_ffmpeg_file_info_io(AVFormatContext *ctx,
					const char *fname, int is_output)
{
	if (ctx != NULL)
	{
		int		loglevel;
		cff_logbuffer_t	info, *old;

		old = gLogBufferPtr;
		gLogBufferPtr = memset(&info, '\0',
					sizeof(cff_logbuffer_t));

		av_log_set_callback(cff_ffmpeg_log_callback);
		loglevel = av_log_get_level();
		av_log_set_level(AV_LOG_DEBUG);

		info.max = 2048;
		av_dump_format(ctx, 0, fname, is_output);

		av_log_set_level(loglevel);

		if (old == NULL)
		{
			av_log_set_callback(av_log_default_callback);
		}

		gLogBufferPtr = old;

		if (info.str != NULL)
		{
			if (info.str[0] != '\0')
			{
				/* Remove last '\n' */
				return cff_ffmpeg_string_truncate(info.str, info.len - 1);
			}

			free(info.str);
		}

	} /* if (ctx != NULL) */

	return NULL;

} /* static char *cff_ffmpeg_file_info_io(AVFormatContext*, const char*, int) */

char *cff_ffmpeg_file_info(cff_video_t *video)
{
	return cff_ffmpeg_file_info_io(video->inFormatCtx, video->fname, 0);

} /* char *cff_ffmpeg_file_info(cff_video_t*) */

static void cff_ffmpeg_scale_dimension(cff_video_t *video, int *width, int *height)
{
	int		h, w, scale;
	double		aspect;
	AVRational	sar;

	/* from 'ffplay.c' */

	sar = av_guess_sample_aspect_ratio(video->inFormatCtx,
		(video->inFormatCtx)->streams[video->vframe.stream_index],
		video->vframe.avframe);

	if (sar.den == 0)
	{
		aspect = 1.0;
	}

	else
	{
		aspect = av_q2d(sar);

		if (aspect <= 0.0)
		{
			aspect = 1.0;
		}
	}

	aspect *= (double) video->vframe.avframe->width /
				(double) video->vframe.avframe->height;

	scale	= 0;
	w	= *width;
	h	= *height;

	if (w != video->vframe.avframe->width)
	{
		/* Scale the width */
		scale = 0x01;
	}

	if (h != video->vframe.avframe->height)
	{
		/* Scale the height (or width and height) */
		scale |= 0x02;
	}

	if (scale == 0 && sar.num != sar.den)
	{
		/* Scale only the width */
		scale = 0x01;
	}

	if (scale & 0x01)
	{
		h = lrint((double) w / aspect) & ~1;
	}

	if (scale & 0x02)
	{
		w = lrint((double) *height * aspect) & ~1;
	}

	if (scale == 0x03)
	{
		if (w > *width)
		{
			w = *width;
			h = lrint((double) w / aspect) & ~1;
		}

		else if (h > *height)
		{
			h = *height;
			w = lrint((double) h * aspect) & ~1;
		}

	} /* if (scale == 0x03) */

	*width  = w;
	*height = h;

} /* static void cff_ffmpeg_scale_dimension(cff_video_t*, int*, int*) */

static int cff_ffmpeg_scale_frame(AVFrame *frame, int width, int height,
	enum AVPixelFormat pixfmt, int align, uint8_t *data[4], int linesize[4])
{
	int			retv;
	struct SwsContext	*sws_context;

	/*
		'av_image_alloc' may return with an error without
		freeing the allocated memory (on wrong align) !
	*/

	data[0] = NULL;

	if (av_image_alloc(data, linesize, width, height, pixfmt, align) < 0)
	{
		return -1;
	}

	sws_context = sws_getContext(frame->width, frame->height,
					frame->format,
					width, height,
					pixfmt,
					SWS_BICUBIC, NULL, NULL, NULL);
	if (sws_context == NULL)
	{
		return -1;
	}

	retv = sws_scale(sws_context,
			(const uint8_t* const*) frame->data, frame->linesize,
			0, frame->height,
			data, linesize);

	sws_freeContext(sws_context);

	return retv;

} /* static int cff_ffmpeg_scale_frame(AVFrame*, int, int,
				enum AVPixelFormat, int, uin8t_t[4]*, int[4]) */

int cff_ffmpeg_frame2rgb24(cff_video_t *video, int width, int height)
{
	int		h, w, retv;
	AVFrame		*frame;

	struct
	{
		uint8_t	*data[4];
		int	linesize[4];

	} rgb;

	frame = video->vframe.avframe;

	if (frame == NULL)
	{
		return -1;
	}

	w = width;
	h = height;
	cff_ffmpeg_scale_dimension(video, &w, &h);

	retv = cff_ffmpeg_scale_frame(frame,
				w, h, AV_PIX_FMT_RGB24,
					1, rgb.data, rgb.linesize);
	if (retv > 0)
	{
		av_freep(&(video->vframe.rgb24));

		video->vframe.rgb24.data   = rgb.data[0];
		video->vframe.rgb24.width  = w;
		video->vframe.rgb24.height = h;
	}

	else
	{
		av_freep(&(rgb.data[0]));
	}

	return retv;

} /* int cff_ffmpeg_frame2rgb24(cff_video_t*, int, int) */

static int cff_ffmpeg_seek(cff_video_t *video, int64_t dts)
{
	int	any, avio, flags;
	int64_t	dts_seek;

	/*
		Keep this order (else 'avi' with 'dts_start' don't work):
		 1) flags = 0
		 2) flags = AVSEEK_FLAG_FRAME
		 3) flags = AVSEEK_FLAG_BYTE
		 4) flags = AVSEEK_FLAG_ANY
	*/

	flags	= 0;
	any	= 0;
	avio	= 0;

	if (dts <= video->vframe.dts_start + video->vframe.min_gop_size)
	{
		/* Sometimes neccessary for proper decoding */
		dts = video->vframe.dts_start;
	}

	dts_seek = dts;

	for (;;)
	{
		int retv;

		if (dts == video->vframe.dts_start)
		{
			/* Seek first frame */

			if (flags == 0)
			{
				if (avio == 0)
				{
					avio = 1;

					/* Rewind - for some files necessary */
					avio_seek((video->inFormatCtx)->pb,
						video->vframe.io_seek_start, SEEK_SET);

					/* Get first frame - works with 'mp4' and 'matroska' */
					dts_seek = video->vframe.dts_start - video->vframe.max_gop_size;
				}

				else
				{
					/* Get first frame - works with 'avi' and 'matroska' */
					flags = AVSEEK_FLAG_FRAME; 
					dts_seek = 0;
				}
			}

			else if (flags == AVSEEK_FLAG_FRAME)
			{
				/* Get first frame - works with 'mpeg-ts' and 'mpeg-2ps' */
				flags = AVSEEK_FLAG_BYTE;
				dts_seek = video->vframe.io_seek_start;
			}

			else if (flags == AVSEEK_FLAG_BYTE)
			{
				/* Get first frame - last chance */
				flags = AVSEEK_FLAG_ANY;
				dts_seek = dts;
			}

			else
			{
				/* Failed */
				return -1;
			}

		} /* if (dts == video->vframe.dts_start) */

		retv = av_seek_frame(video->inFormatCtx,
				video->vframe.stream_index,
				dts_seek, flags);

		if (retv >= 0)
		{
			int64_t pos;

			if (dts > video->vframe.dts_start)
			{
				break;
			}

			pos = avio_tell((video->inFormatCtx)->pb);

			if (pos <= video->vframe.io_seek_start)
			{
				break;
			}
		}

		else if (flags == AVSEEK_FLAG_ANY || dts == video->vframe.dts_start)
		{
			if (flags == AVSEEK_FLAG_ANY)
			{
				if (any)
				{
					return -1;
				}

				/* Getting first frame failed - try second frame */
				dts += video->vframe.frame_duration;
				dts_seek = dts;
				any = 1;
			}
		}

		else
		{
			flags = AVSEEK_FLAG_ANY;
			any = 1;
		}

	} /* for (;;) */

	return 0;

} /* static int cff_ffmpeg_seek(cff_video_t*, int64_t) */

static int cff_ffmpeg_get_packet_with_dts(cff_video_t *video, int64_t dts, AVPacket *packet)
{
	/* Get the packet with the given timestamp */

	int	n;
	int64_t	dts_sub;
	
	n = 5;
	dts_sub = video->vframe.frame_duration;

	do
	{
		unsigned int	index;
		int64_t		dts_last, dts_seek;

		/*
			Move a few frames back, else the frame
			we are looking for might be skipped (???)
		*/

		if (dts == AV_NOPTS_VALUE)
		{
			/* May happen if seeking the first frame */
			dts      = video->vframe.dts_start;
			dts_seek = dts;
			n = 0;
		}

		else
		{
			dts_sub <<= 2;
			dts_seek = dts - dts_sub;
		}

		if (dts_seek <= video->vframe.dts_start)
		{
			if (cff_ffmpeg_seek(video, dts_seek) != 0)
			{
				return -1;
			}

			n = 0;
		}

		else if (av_seek_frame(video->inFormatCtx,
					video->vframe.stream_index,
					dts_seek,
					AVSEEK_FLAG_ANY) < 0)
		{
			continue;
		}

		dts_last = dts - video->vframe.frame_duration;

		do
		{
			if (av_read_frame(video->inFormatCtx, packet) < 0)
			{
				break;
			}

			index = packet->stream_index;

			if (index == video->vframe.stream_index)
			{
				if (dts == packet->dts ||
					(dts == video->vframe.dts_start &&
						packet->pos == video->vframe.pktpos_start))
				{
					return 0;
				}

				dts_last = packet->dts;
			}

			av_packet_unref(packet);
		}
		while (index != video->vframe.stream_index || dts > dts_last);
	}
	while (--n > 0);

	return -1;

} /* static int cff_ffmpeg_get_packet_with_dts(cff_video_t*, int64_t, AVPacket*) */

static int cff_ffmpeg_get_frame(cff_video_t *video, int64_t dts, int width, int height, enum cff_FrameFlags frame_flags)
{
	int		retv, rewind, num_last, dts_error, flush_decoder, iframe_num;
	unsigned int	drain_mode;
	int64_t		diff, pkt_dts, pkt_last, pkt_next, iframe_last, dts_seek, gop_size;
	AVFrame		*frame;
	AVPacket	*packet;
	cff_packet_t	*pkt_ptr;

	struct
	{
		AVFrame	*frame;
		int64_t	dts;
	} last;


	if (dts < video->vframe.dts_start)
	{
		dts = video->vframe.dts_start;
	}

	else if (dts > video->vframe.dts_stop)
	{
		dts = video->vframe.dts_stop;
	}

	diff = dts - video->vframe.current.frame;

	if (diff == 0)
	{
		/* Nothing to do */
		return -1;
	}


	pkt_next = dts;

	if (diff > 0)
	{
		if (diff == 1)
		{
			/* Next frame */
			pkt_next--;
			pkt_next += video->vframe.frame_duration;

			if (pkt_next > video->vframe.dts_stop)
			{
				pkt_next = video->vframe.dts_stop;
			}
		}

		else if (diff < video->vframe.frame_duration)
		{
			/* Clicked at same position */
			return -1;
		}
	}

	else
	{
		for (;;)
		{
			if (diff == -1)
			{
				/* Previous frame */
				pkt_next++;
				pkt_next -= video->vframe.frame_duration;

				if (pkt_next < video->vframe.dts_start)
				{
					pkt_next = video->vframe.dts_start;
				}

				break;
			}

			if (diff > -video->vframe.frame_duration)
			{
				/* Clicked at same position */
				return -1;
			}

			if (diff > -video->vframe.frame_duration * 2)
			{
				/* Clicked near same position */
				diff = -1;
			}

			else
			{
				break;
			}

		} /* for (;;) */

	} /* if (diff > 0) .. else */


	if (pkt_next == video->vframe.dts_stop &&
		video->vframe.current.frame > video->vframe.dts_stop)
	{
		/* May happen if frame duration differs */
		return -1;
	}

	gop_size      = video->vframe.max_gop_size;
	pkt_last      = AV_NOPTS_VALUE;
	iframe_last   = AV_NOPTS_VALUE;
	iframe_num    = 0;
	dts_seek      = dts;
	rewind        = 0;
	num_last      = 0;
	dts_error     = 0;
	drain_mode    = 0;
	flush_decoder = 0;
	last.frame    = NULL;

	/*
		Do not seek if going to the next frame
		and all picture types or P-Frames should
		be displayed
	*/

	if ((video->vframe.current.frame != video->vframe.dts_start &&
		(frame_flags != 0 && (frame_flags & kPFrameFlag) == 0)) ||
			diff != 1 || video->vframe.packet[0].pos == -1)
	{
		if (pkt_next < video->vframe.dts_stop - gop_size)
		{
			gop_size = video->vframe.min_gop_size;
		}

		if (pkt_next > video->vframe.dts_start)
		{
			if (video->vframe.dts_lastgop != AV_NOPTS_VALUE &&
					pkt_next >= video->vframe.dts_lastgop)
			{
				dts_seek = video->vframe.dts_lastgop - gop_size;
			}

			else
			{
				dts_seek -= gop_size;
			}
		}

		rewind = 1;

	} /* if ((video->vframe.current.frame != video->vframe.dts_start &&
		(frame_flags != 0 && (frame_flags & kPFrameFlag) == 0)) ||
			diff != 1 || video->vframe.packet[0].pos == -1) */

	frame = av_frame_alloc();

	if (frame == NULL)
	{
		return -1;
	}

	packet = video->vframe.tmppkt;


	/* Start */

	do
	{
		int		i, send_err;
		AVRational	sar;
		AVFrame		*tmp_frame;

		if (rewind > 0)
		{
			if (num_last > 20 ||
				dts_seek < video->vframe.dts_start - gop_size)
			{
				/* num_last > 3: see 'avcodec_receive_frame' below */

				if (flush_decoder != 0 && num_last <= 3)
				{
					num_last++;
				}

				else
				{
					retv = -1;
					break;
				}
			}

			retv = cff_ffmpeg_seek(video, dts_seek);

			if (retv != 0)
			{
				break;
			}

			dts_seek -= gop_size;

			drain_mode = 0;

			if (video->vframe.drain != NULL)
			{
				(video->vframe.drain)->idx = 0;
			}

			avcodec_flush_buffers(video->vframe.ic);

			flush_decoder	= 0;
			rewind		= -1;

		} /* if (rewind > 0) */

		if (drain_mode || (retv = av_read_frame(video->inFormatCtx, packet)) != 0)
		{
			if ((video->vframe.dts_lastgop == AV_NOPTS_VALUE) ?
				pkt_next + video->vframe.max_gop_size >= video->vframe.dts_stop :
				pkt_next >= video->vframe.dts_lastgop)
			{
				/* Reading last packed failed */

				int max;

				retv = 0;
				max  = 3;

				if (video->vframe.drain != NULL &&
					video->vframe.dts_lastgop == AV_NOPTS_VALUE)
				{
					max = 5;
				}

				if (drain_mode == 0)
				{
					if (num_last >= max)
					{
						if (last.frame != NULL)
						{
							goto _got_picture;
						}

						/* Finish */
						retv = -1;
						continue;
					}

					if (num_last == 0 && last.frame == NULL)
					{
						if (diff == 1 && video->vframe.drain != NULL &&
							(video->vframe.drain)->num >= (video->vframe.drain)->idx)
						{
							drain_mode = (video->vframe.drain)->idx;
						}

						else if (video->vframe.drain == NULL ||
								dts < (video->vframe.drain)->pkt_dts[0])
						{
							/* Nothing to flush */
							num_last++;
							rewind = 1;
							continue;
						}
					}

					if (drain_mode == 0 && avcodec_send_packet(video->vframe.ic, NULL) == 0)
					{
						drain_mode = 1;
					}

				} /* if (drain_mode == 0) */

				if (drain_mode)
				{
					if (avcodec_receive_frame(video->vframe.ic, frame) == 0)
					{
						goto _got_frame;
					}

					if (drain_mode == 1 && num_last < max)
					{
						/* Try again */
						num_last++;
						rewind = 1;
						continue;
					}

					drain_mode = 0;

				} /* if (drain_mode) */

				if (video->vframe.drain != NULL)
				{
					(video->vframe.drain)->idx = 0;
				}

				if (last.frame != NULL)
				{
					if (last.dts != video->vframe.current.frame)
					{
						goto _got_picture;
					}

					if (diff > 0)
					{
						/* Finish */
						retv = -1;
						continue;
					}
				}

				if (num_last > 1)
				{
					/* Try once again, stop if not found */
					dts -= video->vframe.frame_duration;
				}

				/* Clear last packet */
				pkt_last = AV_NOPTS_VALUE;

				/* Store the last frame */
				pkt_next = video->vframe.dts_stop;

				/* Try again */
				num_last++;
				rewind = 1;

			} /* if ((video->vframe.dts_lastgop == AV_NOPTS_VALUE) ?
					pkt_next + video->vframe.max_gop_size >= video->vframe.dts_stop :
					pkt_next >= video->vframe.dts_lastgop) */

			/* Finish */
			continue;

		} /* if (drain_mode || (retv = av_read_frame(video->inFormatCtx, packet)) != 0) */

		if (packet->stream_index != video->vframe.stream_index)
		{
			continue;
		}

		if (packet->dts != AV_NOPTS_VALUE)
		{
			int same_packet = 0;

			if (pkt_last == AV_NOPTS_VALUE)
			{
				pkt_last = packet->dts;
			}

			else if (pkt_last == packet->dts)
			{
				if (frame_flags == 0 &&
					pkt_next != packet->dts &&
						dts_seek > video->vframe.dts_start)
				{
					same_packet = 1;
				}
			}

			if (rewind < 0)
			{
				rewind = 1;

				if (dts_seek > video->vframe.dts_start)
				{
					int64_t distance = packet->dts + video->vframe.max_gop_size * 8;

					if (gop_size > video->vframe.max_gop_size * 8 &&
							(dts_seek > distance || dts_seek < distance))
					{
						dts_error = 1;
					}

					if (packet->dts > dts)
					{
						/*
							The first packet dts after seeking should
							be less or equal than the requested dts!
						*/

						if (dts > distance)
						{
							if (diff > 0)
							{
								/*
									The packet dts should be
									less than the requested dts!
								*/

								if (dts < video->vframe.dts_stop)
								{
									dts_seek += gop_size * 2;
								}

								else if (num_last > 1)
								{
									dts_seek += gop_size * 2;
								}
							}

							else if (pkt_next > video->vframe.dts_start)
							{
								/*
									diff < 0
									The requested dts should be
									less than the packet dts!
								*/

								if (dts_seek > (video->vframe.dts_start << 1))
								{
									dts_seek >>= 1;
									continue;
								}

								if (dts_seek > video->vframe.dts_start + gop_size)
								{
									dts_seek -= gop_size * 2;
								}

							} /* if (diff > 0) .. else */

						} /* if (dts > distance) */

						if (same_packet)
						{
							if (num_last > video->vframe.max_gop_size / video->vframe.frame_duration)
							{
								/* Finish */
								retv = -1;
								continue;
							}

							num_last++;

							if (gop_size < video->vframe.max_gop_size)
							{
								gop_size = video->vframe.max_gop_size;
							}

							else if (dts_seek > video->vframe.dts_start + gop_size)
							{
								if (num_last & 1)
								{
									gop_size <<= num_last;
								}

								else
								{
									gop_size <<= 1;
								}
							}

						} /* if (same_packet) */

						continue;

					} /* if (packet->dts > dts) */


					/*
						The first packet after seeking should
						be a Key-frame for proper decoding!
					*/

			 		if ( ! (packet->flags & AV_PKT_FLAG_KEY))
					{
						rewind = -1;
						continue;
					}

					if (iframe_last == packet->dts)
					{
						/* Same Key-frame, try again */

						if (num_last > 3 && gop_size < video->vframe.max_gop_size)
						{
							gop_size  = video->vframe.max_gop_size;
							dts_seek -= gop_size;
						}

						iframe_num++;
						num_last++;
						continue;
					}

					iframe_last = packet->dts;


					/*
						The first packet after seeking should have at least
						a distance of seven frames for proper decoding!
						Seldom needed!
					*/

					if (packet->dts + (video->vframe.frame_duration * 7) >= pkt_next)
					{
						num_last++;
						continue;
					}

				} /* if (dts_seek > video->vframe.dts_start) */

				if (same_packet)
				{
					num_last++;
					continue;
				}

				rewind = 0;

			} /* if (rewind < 0) */

		} /* if (packet->dts != AV_NOPTS_VALUE) */


		/*
			Store the packet dts, may differ from the
			displayed frame
			If only Key-frame and/or I-frame is selected,
			store only the packet dts of Key-frames
		*/

		if (frame_flags == 0 || (frame_flags & kPFrameFlag) ||
					(packet->flags & AV_PKT_FLAG_KEY))
		{
			if (num_last < 2 &&
				packet->dts == AV_NOPTS_VALUE &&
					pkt_next > video->vframe.dts_start)
			{
				if (dts_seek > video->vframe.dts_start + video->vframe.max_gop_size)
				{
					/*
						The packet dts is invalid. If this
						is the first packet and the minimum
						gop size is selected, change it to
						the maximum gop size and try again!
					*/

					if (gop_size < video->vframe.max_gop_size)
					{
						gop_size = video->vframe.max_gop_size;
						dts_seek = dts - gop_size;
					}

					rewind = 1;
				}

				num_last++;
				continue;

			} /* if (num_last < 2 && .. */

			if (frame_flags == kPFrameFlag &&
				(packet->flags & AV_PKT_FLAG_KEY) &&
							packet->dts > dts)
			{
				num_last++;

				if (num_last > 7)
				{
					/*
						No P-frame found!
						If not searching the first
						frame after 'Open' break,
						else display the first frame
					*/

					if (video->vframe.avframe != NULL)
					{
						retv = -1;
						continue;
					}

					frame_flags = 0;
					num_last = 0;
					pkt_last = AV_NOPTS_VALUE;
					dts_seek = dts;
					rewind = 1;
					continue;

				} /* if (num_last > 7) */

			} /* if (frame_flags == kPFrameFlag &&
				(packet.flags & AV_PKT_FLAG_KEY) &&
							packet->dts > dts) */


			if (packet->dts != AV_NOPTS_VALUE ||
				packet->pos == video->vframe.pktpos_start ||
						pkt_next == video->vframe.dts_start)
			{
				if (video->vframe.packet[0].dts != packet->dts ||
						video->vframe.packet[0].pos != packet->pos)
				{
					memmove(&video->vframe.packet[1], video->vframe.packet,
						(kCFFPacketArrayEntries - 1) * sizeof(cff_packet_t));
					video->vframe.packet[0].dts = packet->dts;
					video->vframe.packet[0].pos = packet->pos;
				}
			}

		} /* if (frame_flags == 0 || (frame_flags & kPFrameFlag) ||
						(packet.flags & AV_PKT_FLAG_KEY)) */


		/* Decode */

		send_err = avcodec_send_packet(video->vframe.ic, packet);

		if (send_err != 0)
		{
			if (send_err != AVERROR(EAGAIN) &&
					num_last < 3 + iframe_num && packet->dts >= pkt_next)
			{
				num_last++;
				rewind = 1;
			}

			else if (flush_decoder != 0)
			{
				avcodec_flush_buffers(video->vframe.ic);
			}

			flush_decoder = 0;
			continue;
		}

		if (avcodec_receive_frame(video->vframe.ic, frame) != 0)
		{
			if ( ! (packet->flags & AV_PKT_FLAG_DISCARD) &&
				num_last < 3 + iframe_num && packet->dts != AV_NOPTS_VALUE &&
					((diff == -1) ?
					packet->dts >= pkt_next :
					packet->dts >= dts))
			{
				/*
					Rewind if next frame is
					requested and receiving it failed
				*/

				if (diff == 1)
				{
					if (dts <= video->vframe.dts_start + gop_size)
					{
						dts_seek = video->vframe.dts_start;
					}

					num_last++;
					rewind = 1;
				}

				else if (diff < 0 || frame_flags == 0)
				{
					if (dts_seek < video->vframe.dts_start)
					{
						dts_seek = video->vframe.dts_start;
					}

					num_last++;
					rewind = 1;
				}
			}

			else
			{
				/* Flush decoder on next 'send_err' */
				flush_decoder = 1;

			} /* if ( ! (packet->flags & AV_PKT_FLAG_DISCARD) &&
					num_last < 3 + iframe_num && packet->dts != AV_NOPTS_VALUE &&
						((diff == -1) ?
						packet->dts >= pkt_next :
						packet->dts >= dts)) .. else */
			continue;

		} /* if (avcodec_receive_frame(video->vframe.ic, frame) != 0) */

_got_frame:
		pkt_dts = frame->pkt_dts;

		if (pkt_dts == AV_NOPTS_VALUE)
		{
			if (frame->pkt_pos == video->vframe.pktpos_start)
			{
				pkt_dts = video->vframe.dts_start;
			}

			else if (drain_mode && video->vframe.drain != NULL &&
					(video->vframe.drain)->num >= drain_mode)
			{
				pkt_dts = (video->vframe.drain)->pkt_dts[drain_mode - 1];
				drain_mode++;
				(video->vframe.drain)->idx = drain_mode;
			}
		}

		if (pkt_dts == AV_NOPTS_VALUE ||
			pkt_dts < video->vframe.dts_start)
		{
			/* Discard invalid packets */
			continue;
		}

		if (frame_flags != 0)
		{
			enum cff_FrameFlags tmp_flags = 0;

			if (frame->key_frame != 0)
			{
				tmp_flags |= kKeyFrameFlag;
			}

			if (frame->pict_type == AV_PICTURE_TYPE_I)
			{
				tmp_flags |= kIFrameFlag;
			}

			else if (frame->pict_type == AV_PICTURE_TYPE_P)
			{
				tmp_flags |= kPFrameFlag;
			}

			if ((frame_flags & tmp_flags) == 0)
			{
				/* Not the wanted picture type */
				continue;
			}

			if (dts > pkt_dts)
			{
				if (dts != video->vframe.dts_start)
				{
					if (last.frame == NULL)
					{
						last.frame = av_frame_clone(frame);
					}

					else
					{
						av_frame_unref(last.frame);
						av_frame_ref(last.frame, frame);
					}

					last.dts = pkt_dts;
				}

				if (diff > 0 || dts_error == 0)
				{
					continue;
				}

			} /* if (dts > pkt_dts) */

			if (diff < 0 && dts != video->vframe.dts_start)
			{
				if (pkt_dts > dts)
				{
					if (last.frame != NULL)
					{
						goto _got_picture;
					}

					rewind = 1;
					continue;
				}

				if (pkt_dts >= video->vframe.current.frame)
				{
					rewind = 1;
					continue;
				}

			} /* if (diff < 0 && dts != video->vframe.dts_start) */
		}

		else /* frame_flags == 0 */
		{
			if (dts > pkt_dts)
			{
				if (diff < 0 || pkt_next == video->vframe.dts_stop)
				{
					if (last.frame == NULL)
					{
						last.frame = av_frame_clone(frame);
						last.dts = pkt_dts;
					}

					else if (pkt_dts > last.dts)
					{
						av_frame_unref(last.frame);
						av_frame_ref(last.frame, frame);
						last.dts = pkt_dts;
					}

					if (diff < 0)
					{
						continue;
					}
				}

				if (diff > 0 || dts_error == 0)
				{
					continue;
				}

			} /* if (dts > pkt_dts) */

			if (diff < 0 && pkt_dts >= video->vframe.current.frame)
			{
				if (last.frame != NULL)
				{
					goto _got_picture;
				}

				if (diff == -1 || num_last < 3)
				{
					num_last++;
					rewind = 1;
					continue;
				}
			}

		} /* if (frame_flags != 0) .. else */

		/* Picture found */

		if (last.frame != NULL)
		{
			av_frame_free(&last.frame);
		}

		/* Success */
_got_picture:
		if (last.frame != NULL)
		{
			av_frame_free(&frame);
			frame = last.frame;
			pkt_dts = last.dts;
		}

		if (diff < 0)
		{
			if (pkt_dts >= video->vframe.current.frame)
			{
				/* Not found */
				retv = -1;
				continue;
			}
		}

		else /* diff > 0 */
		{
			if (pkt_dts <= video->vframe.current.frame)
			{
				/* Not found */
				retv = -1;
				continue;
			}
		}

		/* Get the packet dts belonging to this frame */

		pkt_ptr = video->vframe.packet;
		i = kCFFPacketArrayEntries;

		do
		{
			if (pkt_ptr->pos == -1)
			{
				/* Packet not found */
				i = 0;
				break;
			}

			if (pkt_ptr->pos == frame->pkt_pos)
			{
				/* Packet found */
				video->vframe.current.packet = pkt_ptr->dts;
				video->vframe.current.pktpos = pkt_ptr->pos;
				break;
			}
		}
		while (++pkt_ptr, --i > 0);

		if (last.frame != NULL)
		{
			last.frame = NULL;
			video->vframe.packet[0].pos = -1;
		}

		if (i == 0)
		{
			/* No packet found */

			if (num_last < 5 &&
				pkt_next > video->vframe.dts_start &&
					pkt_next < video->vframe.dts_stop)
			{
				/* Try again */
				num_last++;
				rewind = 1;
				continue;
			}

			/* Packet and frame dts are equal (may incorrect) */
			video->vframe.current.packet = pkt_dts;
			video->vframe.current.pktpos = -1;

		} /* if (i == 0) */

		video->vframe.current.frame = pkt_dts;


		/* Take this frame */

		tmp_frame = video->vframe.avframe;
		video->vframe.avframe = frame;

		retv = cff_ffmpeg_frame2rgb24(video, width, height);

		if (retv <= 0)
		{
			video->vframe.avframe = tmp_frame;
			retv = -1;
			continue;
		}

		av_frame_free(&tmp_frame);

		video->vframe.pict_type =
				cff_ffmpeg_get_pictype_str(frame);

		video->vframe.frame_width  = frame->width;
		video->vframe.frame_height = frame->height;

		sar = av_guess_sample_aspect_ratio(video->inFormatCtx,
			(video->inFormatCtx)->streams[video->vframe.stream_index],
			frame);

		if (sar.num == 0)
		{
			double aspect;

			aspect = (double) video->vframe.avframe->width /
					(double) video->vframe.avframe->height;
			sar = av_d2q(aspect, 1024 * 1024);
			video->vframe.frame_aspect[0] = sar.num;
			video->vframe.frame_aspect[1] = sar.den;
		}

		else if (sar.den == 0)
		{
			video->vframe.frame_aspect[0] = 1;
			video->vframe.frame_aspect[1] = 1;
		}

		else
		{
			av_reduce((int *) &video->vframe.frame_aspect[0],
				(int *) &video->vframe.frame_aspect[1],
				frame->width * (int64_t) sar.num,
				frame->height * (int64_t) sar.den,
				1024 * 1024);
		}

		dts = pkt_dts - video->vframe.dts_start;

		if (dts < 0)
		{
			dts = -dts;
		}

		cff_ffmpeg_dts2time(dts,
			video->inFormatCtx->streams[video->vframe.stream_index],
						video->vframe.time_current);

		video->vframe.frame_current = dts / video->vframe.frame_duration;
	}
	while (av_packet_unref(packet), retv == 0);

	if (last.frame != NULL)
	{
		/* Avoid double freeing */

		if (retv == 0 || last.frame != frame)
		{
			av_frame_free(&last.frame);
		}
	}

	if (retv < 0)
	{
		av_frame_free(&frame);
		video->vframe.packet[0].pos = -1;

		retv = -1;

	} /* if (retv < 0) */

	return retv;

} /* static int cff_ffmpeg_get_frame(cff_video_t*, int64_t, int, int, enum cff_FrameFlags) */

int cff_ffmpeg_get_frame_next(cff_video_t *video, enum cff_VideoFrame next, int width, int height, enum cff_FrameFlags frame_flags)
{
	int		i;
	int64_t		dts, diff, t;
	cff_dts_t	*ptr;

	switch (next)
	{
		case kVideoFramePrev05Sec:
			dts = cff_ffmpeg_sec2dts(-5, video);
			break;
		case kVideoFramePrevHalfSec:
			dts = cff_ffmpeg_sec2dts(-1, video) / 2;
			break;
		case kVideoFramePrevSingle:
			dts = -1;
			break;
		case kVideoFrameNextSingle:
			dts = 1;
			break;
		case kVideoFrameNextHalfSec:
			dts = cff_ffmpeg_sec2dts(1, video) / 2;
			break;
		case kVideoFrameNext05Sec:
			dts = cff_ffmpeg_sec2dts(5, video);
			break;
		case kVideoFramePrev10Min:
			dts = cff_ffmpeg_sec2dts(-(10 * 60), video);
			break;
		case kVideoFramePrev05Min:
			dts = cff_ffmpeg_sec2dts(-(5 * 60), video);
			break;
		case kVideoFramePrev01Min:
			dts = cff_ffmpeg_sec2dts(-(1 * 60), video);
			break;
		case kVideoFrameNext01Min:
			dts = cff_ffmpeg_sec2dts(1 * 60, video);
			break;
		case kVideoFrameNext05Min:
			dts = cff_ffmpeg_sec2dts(5 * 60, video);
			break;
		case kVideoFrameNext10Min:
			dts = cff_ffmpeg_sec2dts(10 * 60, video);
			break;
		case kVideoFrameNextCut:
		case kVideoFrameEnd:
			dts = video->vframe.dts_stop;
			break;
		default: /* kVideoPrevCut and kVideoFrameStart */
			dts = video->vframe.dts_start;
			break;

	}; /* switch (next) */

	if (next == kVideoFramePrevCut)
	{
		i = video->cut.num;

		if (i > 0 && video->cut.list != NULL)
		{
			ptr = &video->cut.list[i];

			do
			{
				ptr--;
				t = ptr->frame;

				if (video->vframe.current.frame > t)
				{
					frame_flags = 0;
					dts = t;
					break;
				}
			}
			while (--i > 0);

		} /* if (ptr != NULL) */
	}

	else if (next == kVideoFrameNextCut)
	{
		i = video->cut.num;

		if (i > 0 && video->cut.list != NULL)
		{
			ptr = video->cut.list;

			do
			{
				t = ptr->frame;
				ptr++;

				if (t > video->vframe.current.frame)
				{
					frame_flags = 0;
					dts = t;
					break;
				}
			}
			while (--i > 0);

		} /* if (ptr != NULL) */
	}

	else if (next != kVideoFrameStart && next != kVideoFrameEnd)
	{
		dts += video->vframe.current.frame;

	} /* if (next == kVideoFrameCutPrev) ..
			else if (next == kVideoFrameCutNext) .. else if .. ) */


	if (dts < video->vframe.dts_start)
	{
		dts = video->vframe.dts_start;
	}

	else if (dts > video->vframe.dts_stop)
	{
		dts = video->vframe.dts_stop;
	}

	diff = dts - video->vframe.current.frame;

	if (diff == 0)
	{
		return -1;
	}

	if (diff > 0)
	{
		if (diff < video->vframe.frame_duration)
		{
			dts = video->vframe.current.frame + 1;
		}
	}

	else
	{
		if (diff > -video->vframe.frame_duration)
		{
			dts = video->vframe.current.frame - 1;
		}

	} /* if (diff > 0) .. else */

	return cff_ffmpeg_get_frame(video, dts, width, height, frame_flags);

} /* int cff_ffmpeg_get_frame_next(cff_video_t*, enum cff_VideoFrame, int, int, enum cff_FrameFlags) */

int cff_ffmpeg_get_frame_cutpoint(cff_video_t *video, unsigned int next, int width, int height, enum cff_FrameFlags frame_flags)
{
	int64_t dts, diff;

	if (next >= video->cut.num || video->cut.list == NULL)
	{
		return -1;
	}

	dts  = video->cut.list[next].frame;
	diff = dts - video->vframe.current.frame;

	if (diff == 0)
	{
		return -1;
	}

	if (diff > 0)
	{
		if (diff < video->vframe.frame_duration)
		{
			dts = video->vframe.current.frame + 1;
		}
	}

	else
	{
		if (diff > -video->vframe.frame_duration)
		{
			dts = video->vframe.current.frame - 1;
		}

	} /* if (diff > 0) .. else */

	return cff_ffmpeg_get_frame(video, dts, width, height, 0);

} /* int cff_ffmpeg_get_frame_cutpoint(cff_video_t*, unsigned int, int, int, enum cff_FrameFlags) */

int cff_ffmpeg_get_frame_slider_click(cff_video_t *video, double next, int width, int height, enum cff_FrameFlags frame_flags)
{
	int64_t dts;
 
	dts  = next * (double) video->vframe.stream_duration;
	dts += video->vframe.dts_start;

	return cff_ffmpeg_get_frame(video, dts, width, height, frame_flags);

} /* int cff_ffmpeg_get_frame_slider_click(cff_video_t*, double, int, int, enum cff_FrameFlags) */

int cff_ffmpeg_get_frame_slider_move(cff_video_t *video, unsigned int frame_nr, int width, int height, enum cff_FrameFlags frame_flags)
{
	int64_t dts;
 
	dts  = frame_nr * video->vframe.frame_duration;
	dts += video->vframe.dts_start;

	return cff_ffmpeg_get_frame(video, dts, width, height, frame_flags);

} /* int cff_ffmpeg_get_frame_slider_move(cff_video_t*, unsigned int, int, int, enum cff_FrameFlags) */

static void cff_ffmpeg_video_undo_clear(cff_undo_t **list)
{
	int		n;
	cff_undo_t	*ptr;

	n = kCFFUndoArrayEntries;

	do
	{
		ptr = *list;

		if (ptr == NULL)
		{
			break;
		}

		free(ptr);
		*(list++) = NULL;
	}
	while (--n > 0);

} /* static void cff_ffmpeg_video_undo_clear(cff_undo_t**) */

static int cff_ffmpeg_video_undo_redo(cff_video_t *video, int action)
{
	int		retv;
	unsigned int	n;
	cff_undo_t	*cut, **src, **dst;

	int (*func)(cff_video_t*, cff_dts_t*);

	if (action == kCFFUndo)
	{
		src = video->vframe.undo;
		dst = video->vframe.redo;
	}

	else
	{
		src = video->vframe.redo;
		dst = video->vframe.undo;
	}

	cut = src[0];

	if (cut->action == kCutpointAdd)
	{
		cut->action = kCutpointRemove;
		func = cff_ffmpeg_video_cut_remove_dts;
	}

	else
	{
		cut->action = kCutpointAdd;
		func = cff_ffmpeg_video_cut_add_dts;
	}

	n = 0;

	do
	{
		/* Add or remove cutpoints */
		retv = (*func)(video, &(cut->dts[n]));

		if (retv < 0)
		{
			return retv;
		}
	}
	while (++n < cut->num);

	/* Move one entry down */
	memmove(&src[0], &src[1],
		sizeof(cff_undo_t*) * (kCFFUndoArrayEntries - 1));

	/* Clear last entry */
	src[kCFFUndoArrayEntries - 1] = NULL;

	if (dst[kCFFUndoArrayEntries - 1] != NULL)
	{
		/* Free last entry (this shouldn't happen!) */
		free(dst[kCFFUndoArrayEntries - 1]);
	}

	/* Move one entry up */
	memmove(&dst[1], &dst[0],
		sizeof(cff_undo_t*) * (kCFFUndoArrayEntries - 1));

	/* Copy new entry */
	dst[0] = cut;

	/* Clear 'dirty' if list is empty */
	video->dirty = (video->cut.num > 0) ? 1 : 0;

	return retv;

} /* static int cff_ffmpeg_video_undo_redo(cff_video_t*, int) */

int cff_ffmpeg_video_undo(cff_video_t *video)
{
	if (video->vframe.undo[0] == NULL)
	{
		return -1;
	}

	return cff_ffmpeg_video_undo_redo(video, kCFFUndo);

} /* int cff_ffmpeg_video_undo(cff_video_t*) */

int cff_ffmpeg_video_redo(cff_video_t *video)
{
	if (video->vframe.redo[0] == NULL)
	{
		return -1;
	}

	return cff_ffmpeg_video_undo_redo(video, kCFFRedo);

} /* int cff_ffmpeg_video_redo(cff_video_t*) */

static void cff_ffmpeg_video_undo_add(cff_video_t *video, cff_undo_t *cut)
{
	if (cut != NULL)
	{
		cff_undo_t **undo = video->vframe.undo;

		if (undo[kCFFUndoArrayEntries - 1] != NULL)
		{
			/* Free last entry */
			free(undo[kCFFUndoArrayEntries - 1]);
		}

		/* Move one entry up */
		memmove(&undo[1], &undo[0],
				sizeof(cff_undo_t*) * (kCFFUndoArrayEntries - 1));

		undo[0] = cut;

		cff_ffmpeg_video_undo_clear(video->vframe.redo);

		/* Clear 'dirty' if list is empty */
		video->dirty = (video->cut.num > 0) ? 1 : 0;

	} /* if (cut != NULL) */

} /* static void cff_ffmpeg_video_undo_add(cff_video_t*, cff_undo_t*) */

static void cff_ffmpeg_video_undo_add_current(cff_video_t *video, enum cff_UndoAction action)
{
	cff_undo_t undo;

	undo.action	= action;
	undo.num	= 1;
	undo.dts[0]	= video->vframe.current;

	cff_ffmpeg_video_undo_add(video,
			cff_ffmpeg_memdup(&undo, sizeof(cff_undo_t)));

} /* static void cff_ffmpeg_video_undo_add_current(cff_video_t*, enum cff_UndoAction) */

unsigned int cff_ffmpeg_get_cut_frame_number(const cff_video_t *video, unsigned int index)
{
	if (index < video->cut.num && video->cut.list != NULL)
	{
		return (video->cut.list[index].frame -
					video->vframe.dts_start) /
						video->vframe.frame_duration;
	}

	return (unsigned int) -1;

} /* int cff_ffmpeg_get_cut_frame_number(cff_video_t*, unsigned int) */

int cff_ffmpeg_video_cut_list_distance(const cff_video_t *video)
{
	int64_t dts, dts_next;

	if (video == NULL ||
		video->inFormatCtx == NULL ||
			video->nb_streams == 0 ||
				video->cut.list == NULL ||
						video->cut.num == 0)
	{
		return 0;
	}

	dts = video->vframe.dts_start;
	dts_next = video->cut.list[0].packet;

	if (video->vframe.max_gop_size > (dts_next - dts))
	{
		if (video->cut.num <= 2)
		{
			return 0;
		}

		dts = video->cut.list[1].packet;
		dts_next = video->cut.list[2].packet;

		if (video->vframe.max_gop_size > (dts_next - dts))
		{
			return 0;
		}

	} /* if (video->vframe.max_gop_size > (dts_next - dts)) */

	return 1;

} /* int cff_ffmpeg_video_cut_list_distance(const cff_video_t*) */

static int cff_ffmpeg_video_cut_find_dts(const cff_video_t *video, const cff_dts_t *src, int all)
{
	/*
		Return the index if
		  all
		    frame, packet and position
		  else
		    either frame, packet or position
		do exist in the list.	
	*/

	int		i, match;
	cff_dts_t	*ptr;

	i   = video->cut.num;
	ptr = video->cut.list;

	if (video->vframe.ic == NULL || i <= 0 || ptr == NULL)
	{
		return -1;
	}

	match = (all == 0) ? 1 : 3;

	i--;
	ptr += i;

	do
	{
		int found = 0;

		if (src->frame == ptr->frame)
		{
			found++;
		}

		if (src->packet == ptr->packet)
		{
			found++;
		}

		if (src->pktpos == ptr->pktpos)
		{
			found++;
		}

		if (found >= match)
		{
			break;
		}

		ptr--;
	}
	while (--i >= 0);

	return i;

} /* static int cff_ffmpeg_video_cut_find_dts(const cff_video_t*, const cff_dts_t*, int) */

int cff_ffmpeg_video_cut_find_current(const cff_video_t *video)
{
	return cff_ffmpeg_video_cut_find_dts(video, &(video->vframe.current), 1);

} /* int cff_ffmpeg_video_cut_find_current(const cff_video_t*) */

static int cff_ffmpeg_video_cut_add_dts(cff_video_t *video, cff_dts_t *src)
{
	int		i;
	unsigned int	n;
	cff_dts_t	*dst;

	if (video->vframe.ic == NULL ||
		src->frame < video->vframe.dts_start ||
			src->frame > video->vframe.dts_stop)
	{
		/* Invalid */
		return -1;
	}

	if (src->pktpos == -1 &&
		src->packet != AV_NOPTS_VALUE &&
			src->packet > video->vframe.dts_start)
	{
		/*
			With no B-frames the DTS should be
			equal, else the frame DTS should be
			greater (or equal) than the packet DTS!
		*/

		if ((video->vframe.ic)->has_b_frames == 0)
		{
			if (src->frame != src->packet)
			{
				return -1;
			}
		}

		else if (src->frame < src->packet)
		{
			return -1;
		}

	} /* if (src->pktpos == -1 &&
			src->packet != AV_NOPTS_VALUE &&
				src->packet > video->vframe.dts_start) */

	i = cff_ffmpeg_video_cut_find_dts(video, src, 0);

	if (i >= 0)
	{
		/* Entry already exists */
		return -1;
	}

	i   = video->cut.num;
	n   = video->cut.max;
	dst = video->cut.list;

	if (i >= n)
	{
		n  += 10;
		dst = av_realloc(dst, n * sizeof(cff_dts_t));

		if (dst == NULL)
		{
			/* Out of memory */
			return -1;
		}

		video->cut.max  = n;
		video->cut.list = dst;
	}

	i   = video->cut.num;
	dst = video->cut.list;

	while (--i >= 0)
	{
		if (src->frame < dst->frame)
		{
			/* Insert - entries one position up (entries == i + 1) */
			memmove(&dst[1], &dst[0], (i + 1) * sizeof(cff_dts_t));
			break;
		}

		dst++;
	}

	*dst = *src;
	video->cut.num++;

	/* Index of the new entry */
	return dst - video->cut.list;

} /* static int cff_ffmpeg_video_cut_add_dts(cff_video_t*, cff_dts_t*) */

int cff_ffmpeg_video_cut_add(cff_video_t *video)
{
	int retv;

	if (video->vframe.ic == NULL)
	{
		return -1;
	}

	retv = cff_ffmpeg_video_cut_add_dts(video, &(video->vframe.current));

	if (retv >= 0)
	{
		cff_ffmpeg_video_undo_add_current(video, kCutpointAdd);
	}

	return retv;

} /* int cff_ffmpeg_video_cut_add(cff_video_t*) */

static int cff_ffmpeg_video_cut_remove_dts(cff_video_t *video, cff_dts_t *src)
{
	int i;

	i = cff_ffmpeg_video_cut_find_dts(video, src, 1);

	if (i >= 0)
	{
		int		n;
		cff_dts_t	*ptr;

		ptr = &video->cut.list[i];
		n   = video->cut.num - i - 1;

		if (n > 0)
		{
			/* Remove - upper entries one position down */
			memmove(&ptr[0], &ptr[1], n * sizeof(cff_dts_t));
		}

		i = video->cut.num - 1;
		video->cut.num = i;

		/* Clear 'dirty' if list is empty */
		video->dirty = (i > 0) ? 1 : 0;

	} /* if (i >= 0) */

	return i;

} /* static int cff_ffmpeg_video_cut_remove_dts(cff_video_t*, cff_dts_t*) */

int cff_ffmpeg_video_cut_remove(cff_video_t *video)
{
	int retv;

	if (video->vframe.ic == NULL)
	{
		return -1;
	}

	retv = cff_ffmpeg_video_cut_remove_dts(video, &(video->vframe.current));

	if (retv >= 0)
	{
		cff_ffmpeg_video_undo_add_current(video, kCutpointRemove);
	}

	return retv;

} /* int cff_ffmpeg_video_cut_remove(cff_video_t*) */

void cff_ffmpeg_video_cut_list_clear(cff_video_t *video)
{
	if (video->cut.list != NULL)
	{
		if (video->cut.num > 0)
		{
			cff_undo_t *undo;

			undo = malloc(sizeof(cff_undo_t) +
				((video->cut.num - 1) * sizeof(cff_dts_t)));

			if (undo != NULL)
			{
				unsigned int i;

				undo->action	= kCutpointRemove;
				undo->num	= video->cut.num;

				i = 0;

				do
				{
					undo->dts[i] = video->cut.list[i];
				}
				while (++i < video->cut.num);

				cff_ffmpeg_video_undo_add(video, undo);

			} /* if (undo != NULL) */

		} /* if (video->cut.num > 0) */

		av_freep(&(video->cut.list));

	} /* if (video->cut.list != NULL) */

	video->cut.num = 0;
	video->cut.max = 0;
	video->dirty   = 0;

} /* void cff_ffmpeg_video_cut_list_clear(cff_video_t*) */

static int64_t cff_ffmpeg_video_rescale_dts(cff_video_t *video, int64_t dts, int write)
{
	if (dts != AV_NOPTS_VALUE)
	{
		AVStream   *is;
		AVRational q, *p1, *p2;

		q  = AV_TIME_BASE_Q;
		is = (video->inFormatCtx)->streams[video->vframe.stream_index];

		if (write)
		{
			p1 = &(is->time_base);
			p2 = &q;
		}

		else
		{
			p1 = &q;
			p2 = &(is->time_base);
		}

		dts = av_rescale_q(dts, *p1, *p2);

	} /* if (dts != AV_NOPTS_VALUE) */

	return dts;

} /* static int64_t cff_ffmpeg_video_rescale_dts(cff_video_t*, int64_t, int) */

int cff_ffmpeg_video_cut_list_save(cff_video_t *video, const char *fname)
{
	int		i;
	FILE		*fd;
	cff_dts_t	*dts;

	i   = video->cut.num;
	dts = video->cut.list;

	if (video->vframe.ic == NULL || i <= 0 || dts == NULL)
	{
		return -1;
	}

#ifdef BUILD_MINGW
	fd = av_fopen_utf8(fname, "w");
#else
	fd = fopen(fname, "w");
#endif

	if (fd == NULL)
	{
		return -1;
	}

	video->dirty = 0;

	fputs("#\n# Cut List (generated from 'CutterFF')\n#\n\n", fd);

	do
	{
		int64_t frame, packet;

		frame  = cff_ffmpeg_video_rescale_dts(video, dts->frame, 1);
		packet = cff_ffmpeg_video_rescale_dts(video, dts->packet, 1);

		fprintf(fd, "%"PRId64" %"PRId64" %"PRId64"\n",
					frame, packet, dts->pktpos);
		dts++;
	}
	while (--i > 0);

	fclose(fd);

	return 0;

} /* int cff_ffmpeg_video_cut_list_save(cff_video_t*, const char*) */

static char *cff_ffmpeg_strtoll(char *line, int64_t *ptr)
{
	char	*end;
	size_t	i;
	int64_t	val;

	line += strspn(line, " \t");

	if ((i = strcspn(line, " \t#\n\r")) == 0)
	{
		/* Empty line or no second or third entry */
		return (char *) -1;
	}

	line[i]	= '\0';

	val = (int64_t) strtoll(line, &end, 10);

	if (end[0] != '\0')
	{
		return NULL;
	}

	*ptr = val;

	return &end[1];

} /* static char *cff_ffmpeg_strtoll(char*, int64_t*) */

static char *cff_ffmpeg_video_cut_get_dts(cff_video_t *video, char *line, int64_t *ptr)
{
	int64_t dts;

	line = cff_ffmpeg_strtoll(line, &dts);

	if (line != NULL && line != (char *) - 1)
	{
		*ptr = cff_ffmpeg_video_rescale_dts(video, dts, 0);
	}

	return line;

} /* static char *cff_ffmpeg_video_cut_get_dts(cff_video_t*, char*, int64_t*) */

int cff_ffmpeg_video_cut_list_load(cff_video_t *video, const char *fname)
{
	int		retv, num;
	unsigned int	max;
	char		*line, buf[2048];
	FILE		*fd;
	AVPacket	*packet;
	cff_undo_t	*undo, *ptr;

	if (video->vframe.ic == NULL)
	{
		return -1;
	}

#ifdef BUILD_MINGW
	fd = av_fopen_utf8(fname, "r");
#else
	fd = fopen(fname, "r");
#endif

	if (fd == NULL)
	{
		return -1;
	}

	max  = 0;
	undo = NULL;

	num  = video->cut.num;
	retv = 0;

	packet = video->vframe.tmppkt;

	while ((line = fgets(buf, sizeof(buf), fd)) != NULL)
	{
		int		i;
		int64_t		pktpos;
		cff_dts_t	list;

		line = cff_ffmpeg_video_cut_get_dts(video, line, &(list.frame));

		if (line == (char *) - 1)
		{
			/* Empty line */
			continue;
		}

		if (line == NULL)
		{
			/* DTS not found */
			retv++;
			continue;
		}

		if (cff_ffmpeg_get_packet_with_dts(video, list.frame, packet) < 0)
		{
			int i = -1;

			if (video->vframe.drain != NULL)
			{
				i = (video->vframe.drain)->num;

				while (--i >= 0)
				{
					if ((video->vframe.drain)->pkt_dts[i] == list.frame)
					{
						break;
					}
				}
			}

			if (i < 0)
			{
				/* DTS is not existing */
				retv++;
				continue;
			}
		}

		else
		{
			av_packet_unref(packet);
		}

		line = cff_ffmpeg_video_cut_get_dts(video, line, &(list.packet));

		if (line == NULL || line == (char *) - 1)
		{
			/* DTS not found or no second entry */
			retv++;
			continue;
		}

		if (cff_ffmpeg_get_packet_with_dts(video, list.packet, packet) < 0)
		{
			/* DTS is not existing */
			retv++;
			continue;
		}

		pktpos = packet->pos;
		av_packet_unref(packet);

		line = cff_ffmpeg_strtoll(line, &(list.pktpos));

		if (line == NULL || line == (char *) - 1)
		{
			/* Position invalid or no third entry */
			list.pktpos = -1;
		}

		else if (pktpos != list.pktpos)
		{
			/* Position do not match */
			retv++;
			continue;
		}

		i = cff_ffmpeg_video_cut_add_dts(video, &list);

		if (i < 0)
		{
			/* Frame already stored */
			retv++;
			continue;
		}

		if (list.pktpos == -1)
		{
			/* Update position */
			list.pktpos = pktpos;
			video->cut.list[i].pktpos = pktpos;
		}

		if (undo == NULL || undo->num >= max)
		{
			max += 10;

			ptr = realloc(undo, sizeof(cff_undo_t) +
					((max - 1) * sizeof(cff_dts_t)));

			if (ptr == NULL)
			{
				continue;
			}

			if (undo == NULL)
			{
				ptr->action = kCutpointAdd;
				ptr->num    = 0;
			}

			undo = ptr;
		}

		if (undo != NULL)
		{
			undo->dts[undo->num] = list;
			undo->num++;
		}

	} /* while ((line = fgets(buf, sizeof(buf), fd)) != NULL) */

	fclose(fd);

	video->vframe.packet[0].pos = -1;

	if (undo == NULL)
	{
		if (video->cut.num > num)
		{
			video->dirty = 1;
		}
	}

	else
	{
		if (max > undo->num)
		{
			/* Free unused memory */

			ptr = realloc(undo, sizeof(cff_undo_t) +
				((undo->num - 1) * sizeof(cff_dts_t)));

			if (ptr != NULL)
			{
				undo = ptr;
			}
		}

		cff_ffmpeg_video_undo_add(video, undo);

	} /* if (undo == NULL) .. else */

	if (num == 0)
	{
		video->dirty = 0;
	}

	return retv;

} /* int cff_ffmpeg_video_cut_list_load(cff_video_t*, const char*) */

static int cff_ffmpeg_stream_info_to_buf(AVStream *s, char *buf, int buf_sz)
{
	static const char* const fmt_string[] =
	{
		"%d] %s: %s%s%s",
		"%d] %s: %s%s (%s)",
		"%d] %s: (%s) %s%s",
		"%d] %s: (%s) %s (%s)"
	};

	int			n;
	const char		*codec_type, *language, *codec_name, *profile;
	AVCodecParameters	*codecpar;
	AVDictionaryEntry	*dict;

	dict = av_dict_get(s->metadata, "language", NULL, 0);

	if (dict == NULL)
	{
		language = "";
		n = 0;
	}

	else
	{
		language = dict->value;
		n = 2;
	}

	codecpar = s->codecpar;

	codec_type = av_get_media_type_string(codecpar->codec_type);
	codec_name = avcodec_get_name(codecpar->codec_id);
	profile    = avcodec_profile_name(codecpar->codec_id, codecpar->profile);

	if (codec_type == NULL)
	{
		codec_type = "unknown";
	}

	if (profile == NULL)
	{
		profile	= "";
	}

	else
	{
		n++;
	}

	return snprintf(buf, buf_sz, fmt_string[n],
		s->index, codec_type, language, codec_name, profile);

} /* static int cff_ffmpeg_stream_info_to_buf(AVStream*, char*, int) */

static AVStream *cff_ffmpeg_get_stream_from_index(cff_video_t *video, int index)
{
	int i;

	if (video->program.index != NULL)
	{
		const AVProgram	*prg;

		i   = video->program.index[video->program.selected];
		prg = (video->inFormatCtx)->programs[i];

		if (index >= prg->nb_stream_indexes)
		{
			return NULL;
		}

		i = prg->stream_index[index];
	}

	else
	{
		if (index >= (video->inFormatCtx)->nb_streams)
		{
			return NULL;
		}

		i = index;
	}

	return (video->inFormatCtx)->streams[i];

} /* static AVStream *cff_ffmpeg_get_stream_from_index(cff_video_t*, int) */

int cff_ffmpeg_get_stream_info(cff_video_t *video, int index, int *stream_idx, char *buf, int buf_sz)
{
	AVStream *is = cff_ffmpeg_get_stream_from_index(video, index);

	if (is == NULL)
	{
		return -1;
	}

	if (stream_idx != NULL)
	{
		*stream_idx = is->index;
	}

	return cff_ffmpeg_stream_info_to_buf(is, buf, buf_sz);

} /* int cff_ffmpeg_get_stream_info(cff_video_t*, int, int*, char*, int) */

static void cff_ffmpeg_copy_free_softcut_packets(cff_pktqueue_t *pktqueue, int all)
{
	int q = 4;

	do
	{
		int      n;
		AVPacket **pkthdl;

		n			= pktqueue->num;
		pktqueue->num		= 0;
		pktqueue->key_frames	= 0;

		pkthdl = pktqueue->packets;

		if (pkthdl != NULL)
		{
			while (--n >= 0)
			{
				av_packet_free(pkthdl);
				pkthdl++;
			}

			if (all)
			{
				av_free(pktqueue->packets);
			}

		} /* if (pkthdl != NULL) */
	}
	while (pktqueue++, --q != 0);

} /* static void cff_ffmpeg_copy_free_softcut_packets(cff_pktqueue_t*, int) */

static int cff_ffmpeg_copy_timebase_reduce(enum AVCodecID codec_id, AVRational *timebase)
{
	int64_t max;

	if (codec_id == AV_CODEC_ID_H264)
	{
		/* Reduce frame rate within spec limits */
		max = 65535;
	}

	else
	{
		max = INT_MAX;
	}

	return av_reduce(&(timebase->num), &(timebase->den),
				timebase->num, timebase->den, max);

} /* static int cff_ffmpeg_copy_timebase_reduce(enum AVCodecID, AVRational*) */

static void cff_ffmpeg_copy_free(cff_video_t *video)
{
	if (video->copy.outFormatCtx != NULL)
	{
		if ((video->copy.outFormatCtx)->pb != NULL)
		{
			/* Close the output file */
			avio_closep(&(video->copy.outFormatCtx->pb));
		}

		avformat_free_context(video->copy.outFormatCtx);

		if (video->copy.ast != NULL)
		{
			av_free(video->copy.ast);
		}

		if (video->copy.streamData != NULL)
		{
			int i;
			struct StreamData *st;

			st = video->copy.streamData;
			i  = 0;

			do
			{
				if (st->oStream == NULL)
				{
					/* No more output streams */
					break;
				}

				if ((st->iStream)->index == video->vframe.stream_index)
				{
					(st->iStream)->discard = AVDISCARD_DEFAULT;
				}

				else
				{
					(st->iStream)->discard = AVDISCARD_ALL;
				}

				av_bsf_free(&(st->bsf_ctx));
			}
			while (st++, ++i < video->copy.nb_streams);

			av_free(video->copy.streamData);

		} /* if (video->copy.streamData != NULL) */

		if (video->copy.softcut.oc != NULL)
		{
			avcodec_free_context(&(video->copy.softcut.oc));
		}

		if (video->copy.softcut.prev_frame != NULL)
		{
			av_frame_free(&(video->copy.softcut.prev_frame));

		}

		cff_ffmpeg_copy_free_softcut_packets(video->copy.softcut.pktqueue, 1);

	} /* if (video->copy.outFormatCtx != NULL) */

	memset(&(video->copy), '\0', sizeof(cff_copy_t));

	/* Back to last position */
	video->vframe.packet[0].pos = -1;

} /* static void cff_ffmpeg_copy_free(cff_video_t*) */

static const char *cff_ffmpeg_encoder_new(cff_video_t *video)
{
	int		n, retv;
	int64_t		bitrate;
	AVCodecContext	*oc;
	AVDictionary	*opts;
	AVRational	timebase;
	const char	*fieldorder_str;
	char		buf[1024];

	oc = avcodec_alloc_context3(video->copy.softcut.codec);

	if (oc == NULL)
	{
		cff_ffmpeg_copy_free(video);
		return gettext_nop("Could not alloc encoder context!");
	}

	video->copy.softcut.oc = oc;

	if (avcodec_parameters_to_context(oc,
			(video->copy.softcut.vst->iStream)->codecpar) < 0)
	{
		cff_ffmpeg_copy_free(video);
		return g_copy_codec_parameters_errstr;
	}


	n = 0;

	do
	{
		if (n == 0)
		{
			timebase = (video->vframe.ic)->time_base;

			if ((video->vframe.ic)->ticks_per_frame != 0)
			{
				timebase.num *= (video->vframe.ic)->ticks_per_frame;
			}
		}

		else if (n == 1)
		{
			timebase =
				av_add_q(av_stream_get_codec_timebase((video->copy.softcut.vst)->iStream),
									(AVRational) {0, 1});

			if ((video->vframe.ic)->ticks_per_frame != 0)
			{
				timebase.num *= (video->vframe.ic)->ticks_per_frame;
			}
		}

		else if (n == 2)
		{
			timebase = (video->copy.softcut.vst->iStream)->time_base;
		}

		else /* (n == 3) */
		{
			timebase.num = 1;
			timebase.den = (video->vframe.ic)->ticks_per_frame;
		}
	}
	while ((timebase.num <= 0 || timebase.den <= 0) && ++n < 4);

	cff_ffmpeg_copy_timebase_reduce(oc->codec_id, &timebase);


	n = 0;

	do
	{
		if (n == 0)
		{
			bitrate = ((video->copy.softcut.vst->iStream)->codecpar)->bit_rate;
		}

		else if (n == 1)
		{
			bitrate = (video->vframe.ic)->bit_rate;
		}

		else /* (n == 2) */
		{
			bitrate = (video->inFormatCtx)->bit_rate;
		}
	}
	while (bitrate == 0 && ++n < 3);


	opts = NULL;

	av_dict_set(&opts, "threads", "auto", 0);
	av_dict_set_int(&opts, "level", (video->vframe.ic)->level, 0);
	av_dict_set_int(&opts, "trellis", (video->vframe.ic)->trellis, 0);

	snprintf(buf, sizeof(buf), "%"PRId64, bitrate);
	av_dict_set(&opts, "b", buf, 0);

	/* Let the encoder choose */
	av_dict_set_int(&opts, "maxrate", 0, 0);
	av_dict_set_int(&opts, "minrate", 0, 0);
	av_dict_set_int(&opts, "bufsize", 0, 0);

	av_dict_set_int(&opts, "qmin", 2, 0);
	av_dict_set_int(&opts, "qmax", 27, 0);
	av_dict_set_int(&opts, "qdiff", 25, 0);

	av_dict_set_int(&opts, "refs", (video->vframe.ic)->refs, 0);
	av_dict_set_int(&opts, "slice_count", (video->vframe.ic)->slices, 0);
	av_dict_set_int(&opts, "slice_flags", (video->vframe.ic)->slice_flags, 0);


	/*
		If the video contains different aspect ratios, then
		display a frame in the window which has the ratio your
		video should have!
	*/

	if (video->vframe.avframe != NULL)
	{
		AVRational sar;

		sar = av_guess_sample_aspect_ratio(video->inFormatCtx,
					(video->copy.softcut.vst)->iStream,
					video->vframe.avframe);

		if (sar.den != 0 && sar.num != 0)
		{
			snprintf(buf, sizeof(buf), "%i/%i", sar.num, sar.den);
			av_dict_set(&opts, "sar", buf, 0);
		}

	} /* if (video->vframe.avframe != NULL) */


	/* AVFieldOrder 'avcodec.h' */

	switch (oc->field_order)
	{
		case AV_FIELD_TT:
			fieldorder_str = "tt";
			break;
		case AV_FIELD_BB:
			fieldorder_str = "bb";
			break;
		case AV_FIELD_TB:
			fieldorder_str = "tb";
			break;
		case AV_FIELD_BT:
			fieldorder_str = "bt";
			break;
		default:
			fieldorder_str = "progressive";
			break;

	} /* switch (oc->field_order) */

	av_dict_set(&opts, "field_order", fieldorder_str, 0);

	av_dict_set_int(&opts, "ticks_per_frame", (video->vframe.ic)->ticks_per_frame, 0);

	snprintf(buf, sizeof(buf), "%i/%i", timebase.num, timebase.den);
	av_dict_set(&opts, "time_base", buf, 0);

	/* Use closed gop and loop filter */
	av_dict_set(&opts, "flags", "+cgop+loop", 0);

	/* Skip bitstream encoding, ignore cropping information from sps */
	av_dict_set(&opts, "flags2", "+noout+ignorecrop", 0);

	if (video->copy.softcut.flags &
		(kCFFSoftcutFlagLibx264 | kCFFSoftcutFlagLibx265))
	{
		av_dict_set_int(&opts, "forced-idr", 1, 0);
	}

	if (oc->field_order != AV_FIELD_UNKNOWN &&
			oc->field_order != AV_FIELD_PROGRESSIVE)
	{
		/* Use interlace DCT and interlace motion estimation */
		av_dict_set(&opts, "flags", "+ildct+ilme", 0);
	}

	if (oc->codec_id == AV_CODEC_ID_H264)
	{
		/* Use four motion vector by macroblock */
		av_dict_set(&opts, "flags", "+mv4", 0);

		if (video->copy.softcut.flags & kCFFSoftcutFlagLibx264)
		{
			av_dict_set(&opts, "x264-params",
					"crf=25:deblock=1,1:ref=1:b-adapt=0",
					0);
		}
	}

	else if (oc->codec_id == AV_CODEC_ID_H265)
	{
		if (video->copy.softcut.flags & kCFFSoftcutFlagLibx265)
		{
			av_dict_set(&opts, "x265-params",
					"signhide=0:scaling-list=default:"
					"max-merge=1:crf=25:deblock=1,1:"
					"ref=1:b-adapt=0",
					0);
		}

	} /* if (oc->codec_id == AV_CODEC_ID_H264) .. else */


	retv = avcodec_open2(oc, NULL, &opts);

	if (opts != NULL)
	{
		AVDictionaryEntry *entry = NULL;

		while ((entry = av_dict_get(opts, "",
					entry, AV_DICT_IGNORE_SUFFIX)) != NULL)
		{
			av_log(NULL, AV_LOG_WARNING,
					"cutterff @ rejected encoder opts] "
					"Key: \"%s\" -- Value: \"%s\"\n",
					entry->key, entry->value);
		}

		av_dict_free(&opts);

	} /* if (opts != NULL) */

	if (retv < 0)
	{
		cff_ffmpeg_copy_free(video);
		return gettext_nop("Could not open the encoder!");
	}

	return NULL;

} /* static const char *cff_ffmpeg_encoder_new(cff_video_t*) */

static const char *cff_ffmpeg_copy_write_frame(cff_video_t *video, struct StreamData *st, AVPacket *packet)
{
	uint64_t sz;

	if (st->flags & kCFFStreamFlagTsRescale)
	{
		av_packet_rescale_ts(packet,
				(st->iStream)->time_base,
					(st->oStream)->time_base);
	}

	sz = packet->size;

	if (av_interleaved_write_frame(video->copy.outFormatCtx, packet) < 0)
	{
		av_packet_unref(packet);
		return gettext_nop("'av_interleaved_write_frame()' failed!");
	}

	st->nbBytesWritten += sz;
	st->nbPackets++;

	return NULL;

} /* static const char *cff_ffmpeg_copy_write_frame(cff_video_t*, struct StreamData*, AVPacket*) */

static const char *cff_ffmpeg_copy_write_frame_softcut(cff_video_t *video, cff_encdata_t *enc)
{
	AVPacket *packet, *pptr;

	if (enc->qidx >= (enc->pktqueue)->num)
	{
		if (enc->q0_end != 0)
		{
			/* Overflow - shouldn't happen */
			enc->done = 1;
			return NULL;
		}

		enc->q0_end    = 1;
		enc->qidx      = 0;
		enc->pktqueue += 3;

		/* Now count 'q3_encoded' */
		enc->q0_decoded = enc->q0_encoded;

	} /* if (enc->qidx >= (enc->pktqueue)->num) */

	pptr   = (enc->pktqueue)->packets[enc->qidx];
	packet = enc->packet;

	if (packet->dts >= packet->pts)
	{
		packet->pts = pptr->dts;
	}

	else
	{
		int64_t pts_diff = packet->dts - pptr->dts;

		packet->pts -= pts_diff;

	} /* if (packet->dts >= packet->pts) .. else */

	packet->dts          = pptr->dts;
	packet->duration     = pptr->duration;
	packet->stream_index = pptr->stream_index;

	if (enc->q0_decoded > enc->q0_encoded)
	{
		enc->q0_encoded++;
	}

	else
	{
		enc->q3_encoded++;
	}

	enc->qidx++;

	/* DTS for the last written frame */
	enc->dts_last = packet->dts;

	return cff_ffmpeg_copy_write_frame(video, video->copy.softcut.vst, packet);

} /* static const char *cff_ffmpeg_copy_write_frame_softcut(cff_video_t*, cff_encdata_t*) */

static void cff_ffmpeg_copy_softcut_coding_warn(int gap_num, const char *coded_str, int diff_enc, int key_frames)
{
	av_log(NULL, AV_LOG_WARNING, "cutterff @ gap #%i] %i frame%s %s!\n%s",
		gap_num,
		diff_enc,
		(diff_enc == 1) ? " wasn't" : "s weren't",
		coded_str,
		key_frames >= 2 ? "" :
			"\tMake shure there are at least 2 GOPs before"
				" and 2 GOPs after the gap!\n");

} /* static void cff_ffmpeg_copy_softcut_coding_warn(int, const char*, int, int) */

static void cff_ffmpeg_copy_softcut_pts_lastdec_warn(int64_t pts, int64_t last_pts, const char *str)
{

#ifdef BUILD_MINGW

	/*
		MingW print the warning "unknown conversion type character"
		'PRId64', below (Linux) do not show a warning (???)
	*/

	char buf [256];

	snprintf(buf, sizeof(buf), "cutterff @ encode] "
		"Warning: pts (%"PRId64") <= last pts (%"PRId64") - %s",
		pts, last_pts, str);
	av_log(NULL, AV_LOG_WARNING, "%s\n", buf);
#else
	av_log(NULL, AV_LOG_WARNING, "cutterff @ encode] "
		"Warning: pts (%"PRId64") <= last pts (%"PRId64") - %s\n",
		pts, last_pts, str);
#endif

} /* static void cff_ffmpeg_copy_softcut_decoded_pts_warn(int64_t, int64_t, const char*) */

static const char *cff_ffmpeg_copy_softcut_decode(AVCodecContext *ic, AVPacket *packet, AVFrame **frame)
{
	int retv;

	if (packet != NULL)
	{	
		retv = avcodec_send_packet(ic, packet);

		if (retv != 0 && retv != AVERROR(EAGAIN))
		{
			/* Failed */
			return gettext_nop("Could not write a packet to the decoder!");
		}
	}

	retv = avcodec_receive_frame(ic, *frame);

	if (retv != 0)
	{
		if (retv != AVERROR(EAGAIN) && retv != AVERROR_EOF)
		{
			/* Failed */
			return gettext_nop("Could not read a frame from the decoder!");
		}

		*frame = NULL;
	}

	return NULL;

} /* static const char *cff_ffmpeg_copy_softcut_decode(AVCodecContext*, AVPacket*, AVFrame**) */

static const char *cff_ffmpeg_copy_softcut_encode(cff_softcut_t *softcut, AVFrame *frame, AVPacket **packet)
{
	int retv;

	if (frame != NULL || softcut->prev_frame != NULL)
	{
		AVFrame *displ_frame, *prev_frame;

		prev_frame  = softcut->prev_frame;
		displ_frame = prev_frame;

		if (frame != NULL)
		{
			if (prev_frame == NULL)
			{
				*packet = NULL;
				softcut->prev_frame = av_frame_clone(frame);

				if (softcut->prev_frame == NULL)
				{
					/* Failed */
					return g_clone_frame_errstr;
				}

				return NULL;
			}

			if (prev_frame->pts > frame->pts)
			{
				int64_t    frame_pts, prev_pts;
				const char *str;

				frame_pts = frame->pts;
				prev_pts  = prev_frame->pts;

				if (frame->interlaced_frame)
				{
					displ_frame = frame;
					str = "swap frame";
				}

				else
				{
					int64_t tmp_pts;

					tmp_pts = displ_frame->pts;
					displ_frame->pts = frame->pts;
					frame->pts = tmp_pts;
					str = "swap values";
				}

				cff_ffmpeg_copy_softcut_pts_lastdec_warn(frame_pts,
									prev_pts, str);

			} /* if (prev_frame->pts > frame->pts) */

		} /* if (frame != NULL) */

		if (softcut->pts_lastdec != AV_NOPTS_VALUE && softcut->pts_lastdec >= displ_frame->pts)
		{
			cff_ffmpeg_copy_softcut_pts_lastdec_warn(displ_frame->pts,
					softcut->pts_lastdec, "assigning last pts + 1");

			softcut->pts_lastdec++;
			displ_frame->pts = softcut->pts_lastdec;
		}

		else
		{
			softcut->pts_lastdec = displ_frame->pts;
		}


		/*
			Got problems with 'libx264' and mp4 with level 'high',
			some frames will not be decoded. Setting the picture
			type to 'NONE' seems to solve this.
		*/

		displ_frame->pict_type = AV_PICTURE_TYPE_NONE;

		retv = avcodec_send_frame(softcut->oc, displ_frame);

		if (retv != 0 && retv != AVERROR(EAGAIN))
		{
			/* Failed */
			return gettext_nop("Could not write a frame to the encoder!");
		}

		if (frame == NULL)
		{
			/* Flush the encoder */
			av_frame_free(&(softcut->prev_frame));
			avcodec_send_frame(softcut->oc, NULL);
		}

		else if (displ_frame == prev_frame)
		{
			/* Ref frame, if previous frame was send to the encoder */

			av_frame_unref(prev_frame);

			if (av_frame_ref(prev_frame, frame) != 0)
			{
				/* Failed */
				return gettext_nop("Could not ref the frame!");
			}

		} /* if (frame == NULL) .. else */

	} /* if (frame != NULL || softcut->prev_frame != NULL) */

	retv = avcodec_receive_packet(softcut->oc, *packet);

	if (retv != 0)
	{
		if (retv != AVERROR(EAGAIN) && retv != AVERROR_EOF)
		{
			/* Failed */
			return gettext_nop("Could not read a packet from the encoder!");
		}

		*packet = NULL;
	}

	return NULL;

} /* static const char *cff_ffmpeg_copy_softcut_encode(cff_softcut_t*, AVFrame*, AVPacket**) */

static int cff_ffmpeg_copy_softcut_position_index(const AVFrame *frame, const AVPacket **pkthdl, int num)
{
	/*
		Search the frame position in the given
		packet queue (0 or 3). If found, return
		the index, else return -1;
	*/

	pkthdl += num;

	while (--num >= 0)
	{
		const AVPacket *pkt = *(--pkthdl);

		if (frame->pkt_pos == pkt->pos)
		{
			return num;
		}
	}

	return -1;

} /* static int cff_ffmpeg_copy_softcut_position_index(const AVFrame*, const AVPacket**, int max) */

static const char *cff_ffmpeg_copy_softcut(cff_video_t *video, AVPacket *packet, int *stop)
{
	unsigned int		num_pkts, queue_idx;
	const char		*str;
	cff_softcut_t		*softcut;
	cff_pktqueue_t		*pktqueue;
	AVPacket		*pkt, **pkthdl;
	struct StreamData	*st;

	if (stop != NULL)
	{
		*stop = 0;
	}

	softcut	= &(video->copy.softcut);
	st	= softcut->vst;

	if (st == NULL)
	{
		return NULL;
	}

	queue_idx = softcut->queue_idx;
	pktqueue  = &softcut->pktqueue[queue_idx];
	num_pkts  = pktqueue->num;

	if (queue_idx == 3 && (packet == NULL ||
		(pktqueue->key_frames > 1 && (packet->flags & AV_PKT_FLAG_KEY))))
	{
		/* Decode - Encode */

		int		cancel, got_start, cutidx, firstrun;
		unsigned int	offset_lastnum, mvpkt, key_frames, remaining_packets;
		int64_t		dts_offset;

		AVFrame			*frame, *fptr;
		AVPacket		*pptr;
		cff_encdata_t		enc;
		cff_softcut_pbar_t	pbar;
		const cff_pktqueue_t	*pq;

		if (softcut->oc == NULL)
		{
			str = cff_ffmpeg_encoder_new(video);

			if (str != NULL)
			{
				return str;
			}
		}

		frame = av_frame_alloc();

		if (frame == NULL)
		{
			return g_alloc_frame_errstr;
		}

		softcut->pts_lastdec = AV_NOPTS_VALUE;
		softcut->gap_num++;

		pktqueue = softcut->pktqueue;
		str      = NULL;

		/* Remove duration from queue 1 + 2 */
		st->duration -= softcut->duration_last;

		memset(&enc, '\0', sizeof(enc));
		enc.pktqueue = pktqueue;
		enc.packet   = av_packet_alloc();

		if (enc.packet == NULL)
		{
			av_frame_free(&frame);
			return g_alloc_packet_errstr;
		}

		cancel = 0;

		pbar.decoded   = 0;
		pbar.encoded   = 0;
		pbar.func_data = softcut->func_data;

		queue_idx  = 0;
		cutidx     = 0;
		got_start  = 0;
		dts_offset = 0;
		firstrun   = 0;

		/* Default: dont't copy remaining packets to queue 0 */
		mvpkt			= 0;
		remaining_packets	= 0;
		key_frames		= softcut->pktqueue[3].key_frames;

		if (pktqueue->num == 0)
		{
			queue_idx     = 2;
			pktqueue     += 2;
			enc.pktqueue += 3;
			enc.q0_end    = 1;

			if (st->nbPackets == 0)
			{
				firstrun = 1;

				pq = &(softcut->pktqueue[3]);

				if (st->flags & kCFFStreamFlagTsFromDuration)
				{
					/*
						Replace offset with dts of
						the first packet to write
					*/

					st->dts_offset = pq->packets[0]->dts;
				}

				else if ( ! (video->copy.flags & kCFFCopyFlagTsKeepStart))
				{
					/*
						The dts of the first packet
						written will be '0'
					*/

					dts_offset = pq->packets[0]->dts;
				}
			}

			else
			{
				dts_offset = softcut->pktdts_last;

			} /* if (st->nbPackets == 0) .. else */

		} /* if (pktqueue->num == 0) */


		avcodec_flush_buffers(softcut->oc);

		do
		{
			num_pkts = pktqueue->num;

			if (num_pkts == 0)
			{
				str = gettext_nop("One queue is empty!");
				goto _enc_done;
			}

			pkthdl = pktqueue->packets;

			if (queue_idx == 0)
			{
				if (pktqueue->key_frames > 1)
				{
					cutidx		= INT_MAX;
					enc.encode_max	= INT_MAX;
				}

				else
				{
					enc.encode_max = num_pkts;
				}

				dts_offset = pkthdl[num_pkts - 1]->dts;
				avcodec_flush_buffers(video->vframe.ic);
			}

			else if (queue_idx == 2)
			{
				int diff_enc;

				got_start = 0;

				if (firstrun == 0)
				{
					dts_offset = pkthdl[num_pkts - 1]->dts - dts_offset;
				}

				if ( ! (st->flags & kCFFStreamFlagTsFromDuration))
				{
					st->dts_offset += dts_offset;
				}

				if (softcut->pktqueue[0].packets == NULL)
				{
					pq = &(softcut->pktqueue[3]);
					video->copy.softcut.pktdts_last =
						pq->packets[pq->num - 1]->dts - dts_offset;
				}

				diff_enc = enc.decoder_in - enc.q0_decoded;

				if (diff_enc > 0)
				{
					cff_ffmpeg_copy_softcut_coding_warn(softcut->gap_num,
						"decoded", diff_enc, softcut->pktqueue[0].key_frames);
				}

				enc.decoder_in	= 0;
				enc.encode_max	= INT_MAX;
				cutidx		= INT_MAX;

				avcodec_flush_buffers(video->vframe.ic);

			} /* if (queue_idx == 0) .. else if (queue_idx == 2) */


			do
			{
				int q_pos;

				pkt = *(pkthdl++);

				if (enc.decoder_in >= enc.encode_max &&
					(queue_idx == 3 || enc.q0_decoded == enc.decoder_in))
				{
					break;
				}

				if (got_start == 0)
				{
					if ( ! (pkt->flags & AV_PKT_FLAG_KEY))
					{
						if (queue_idx == 0)
						{
							enc.qidx++;
							str = cff_ffmpeg_copy_write_frame(video, st, pkt);

							if (str != NULL)
							{
								goto _enc_done;
							}
						}

						continue;

					} /* if ( ! (pkt->flags & AV_PKT_FLAG_KEY)) */

					got_start = 1;

				} /* if (got_start == 0) */

				if (queue_idx == 0)
				{
					if ((pkt->flags & AV_PKT_FLAG_KEY) &&
						cutidx == INT_MAX && num_pkts < pktqueue->num)
					{
						cutidx		= pktqueue->num - num_pkts;
						enc.encode_max	= num_pkts;
					}

					if (enc.encode_max != INT_MAX)
					{
						enc.decoder_in++;
					}
				}

				else if (queue_idx == 3)
				{
					if ((pkt->flags & AV_PKT_FLAG_KEY) &&
						cutidx == INT_MAX && num_pkts < pktqueue->num)
					{
						cutidx		= pktqueue->num - num_pkts;
						enc.encode_max	= cutidx;
					}

					if (enc.encode_max > enc.decoder_in)
					{
						enc.decoder_in++;
					}

				} /* if (queue_idx == 0) .. else if (queue_idx == 3) */


				if (queue_idx >= 2 && firstrun == 0)
				{
					/* Substract offset before decoding */

					pkt->dts -= dts_offset;
					pkt->pts -= dts_offset;
				}

				fptr = frame;
				str  = cff_ffmpeg_copy_softcut_decode(video->vframe.ic, pkt, &fptr);

				if (str != NULL)
				{
					goto _enc_done;
				}

				if (queue_idx == 3 && firstrun)
				{
					/*
						Encode starting GOP, substract
						offset after decoding, avoids
						negative DTS/PTS at start of video.
					*/

					pkt->dts -= dts_offset;
					pkt->pts -= dts_offset;
				}

				if (fptr == NULL)
				{
					if (queue_idx != 0)
					{
						continue;
					}
				}

				else
				{
					pbar.decoded++;
				}


				if (queue_idx == 0 && cutidx == INT_MAX)
				{
					/* First packets needed only for decoding */

					enc.qidx++;
					str = cff_ffmpeg_copy_write_frame(video, st, pkt);

					if (str != NULL)
					{
						goto _enc_done;
					}

					q_pos = -1;
				}

				else if (queue_idx == 2 || fptr == NULL)
				{
					q_pos = -1;
				}

				else
				{
					AVPacket **phdl;

					if (queue_idx < 2)
					{
						phdl = &(softcut->pktqueue[0].packets[cutidx]);
					}

					else /* queue_idx == 3 */
					{
						phdl = &(softcut->pktqueue[3].packets[0]);
					}

					/* Compiler warning without typecast */
					q_pos = cff_ffmpeg_copy_softcut_position_index(frame,
							(const AVPacket **) phdl, enc.decoder_in);

				} /* if (queue_idx == 0 && cutidx == INT_MAX) .. else if .. else */

				if (q_pos < 0)
				{
					/* Do not encode this frame */

					if (softcut->func != NULL)
					{
						cancel = (*softcut->func)(&pbar);

						if (cancel != 0)
						{
							goto _enc_done;
						}
					}

					continue;

				} /* if (q_pos < 0) */

				if (queue_idx < 2)
				{
					enc.q0_decoded++;
				}

				else /* queue_idx == 3 */
				{
					enc.q3_decoded++;
				}

				pptr = enc.packet;
				str  = cff_ffmpeg_copy_softcut_encode(softcut, frame, &pptr);

				if (str != NULL)
				{
					goto _enc_done;
				}

				if (pptr != NULL)
				{
					pbar.encoded++;
				}

				if (softcut->func != NULL)
				{
					cancel = (*softcut->func)(&pbar);

					if (cancel != 0)
					{
						goto _enc_done;
					}
				}

				if (pptr == NULL)
				{
					continue;
				}

				str = cff_ffmpeg_copy_write_frame_softcut(video, &enc);

				if (str != NULL || enc.done != 0)
				{
					goto _enc_done;
				}
			}
			while (av_frame_unref(frame), av_packet_unref(enc.packet), --num_pkts != 0);
		}
		while (pktqueue++, ++queue_idx <= 3);


		pktqueue	= &softcut->pktqueue[3];
		offset_lastnum	= num_pkts + 1;


		/* Flush the decoder */

		if (avcodec_send_packet(video->vframe.ic, NULL) == 0)
		{
			fptr = frame;

			for (;;)
			{
				if (enc.q3_decoded >= enc.encode_max)
				{
					break;
				}

				str = cff_ffmpeg_copy_softcut_decode(video->vframe.ic,
										NULL, &fptr);
				if (str != NULL)
				{
					goto _enc_done;
				}

				if (fptr == NULL)
				{
					break;
				}

				pbar.decoded++;

				/* Compiler warning without typecast */

				if (cff_ffmpeg_copy_softcut_position_index(frame,
						(const AVPacket **) pktqueue->packets,
									enc.decoder_in) < 0)
				{
					/* Do not encode this frame */
					continue;
				}

				enc.q3_decoded++;

				pptr = enc.packet;
				str  = cff_ffmpeg_copy_softcut_encode(softcut, frame, &pptr);

				if (str != NULL)
				{
					goto _enc_done;
				}

				if (pptr != NULL)
				{
					pbar.encoded++;
				}

				if (softcut->func != NULL)
				{
					cancel = (*softcut->func)(&pbar);

					if (cancel != 0)
					{
						goto _enc_done;
					}
				}

				if (pptr == NULL)
				{
					continue;
				}

				str = cff_ffmpeg_copy_write_frame_softcut(video, &enc);

				if (str != NULL || enc.done != 0)
				{
					goto _enc_done;
				}

			} /* for (;;) */

		} /* if (avcodec_send_packet(video->vframe.ic, NULL) == 0) */


		/* Encode last frame, then flush the encoder */

		pptr = enc.packet;

		for (;;)
		{
			str = cff_ffmpeg_copy_softcut_encode(softcut, NULL, &pptr);

			if (str != NULL)
			{
				goto _enc_done;
			}

			if (pptr == NULL)
			{
				break;
			}

			pbar.encoded++;

			if (softcut->func != NULL)
			{
				cancel = (*softcut->func)(&pbar);

				if (cancel != 0)
				{
					goto _enc_done;
				}
			}

			str = cff_ffmpeg_copy_write_frame_softcut(video, &enc);

			if (str != NULL || enc.done != 0)
			{
				goto _enc_done;
			}

		} /* for(;;) */

		if (pktqueue->num > enc.decoder_in)
		{
			int diff_enc, out_max;

			diff_enc = enc.decoder_in - enc.q3_decoded;

			if (diff_enc > 0)
			{
				cff_ffmpeg_copy_softcut_coding_warn(softcut->gap_num,
								"decoded", diff_enc, key_frames);
			}

			else if ((diff_enc = enc.decoder_in - enc.q3_encoded) > 0)
			{
				cff_ffmpeg_copy_softcut_coding_warn(softcut->gap_num,
								"encoded", diff_enc, key_frames);
			}

			out_max = pktqueue->num - enc.decoder_in;

			if (video->copy.dts_next != INT64_MAX &&
				video->cut.num > video->copy.cutidx &&
					softcut->pktqueue[0].packets != NULL &&
						softcut->pktqueue[0].max >= out_max &&
							enc.q3_decoded == enc.encode_max)
			{
				/* Copy remaining packets to queue 0 */
				mvpkt = 1;
			}

			pkthdl   = pktqueue->packets;
			num_pkts = pktqueue->num;

			do
			{
				pptr = *(pkthdl++);

				if (offset_lastnum > num_pkts)
				{
					/*
						Main loop was finished before
						subtracting the offset
					*/

					pptr->dts -= dts_offset;
					pptr->pts -= dts_offset;
				}

				if (enc.dts_last >= pptr->dts)
				{
					if (pptr->flags & AV_PKT_FLAG_KEY)
					{
						/* One key frame less in queue 3 */
						key_frames--;
					}

					continue;
				}

				if (mvpkt != 0)
				{
					if (pptr->flags & AV_PKT_FLAG_KEY)
					{
						/*
							Move remaining packets to queue 0,
							if not the last cut point
						*/

						if (mvpkt == 1)
						{
							/* Packets to move from q3 to q0 */
							remaining_packets = num_pkts;

							/* Do move */
							mvpkt = 2;

							if (offset_lastnum == 1)
							{
								/* Offset for all packets okay */
								break;
							}
						}

						continue;

					} /* if (pptr->flags & AV_PKT_FLAG_KEY) */

					if (mvpkt > 1)
					{
						/* Continue subtracting the offset */
						continue;
					}

				} /* if (mvpkt != 0) */

				str = cff_ffmpeg_copy_write_frame(video, st, pptr);

				if (str != NULL)
				{
					break;
				}
			}
			while (--num_pkts > 0);

		} /* if (pktqueue->num > enc.decoder_in) */


_enc_done:
		av_frame_free(&frame);
		av_packet_free(&enc.packet);

		if (softcut->prev_frame != NULL)
		{
			av_frame_free(&softcut->prev_frame);
		}

		if (str != NULL)
		{
			/* Error */
			return str;
		}

		if (cancel != 0)
		{
			/* Halt */

			if (stop != NULL)
			{
				*stop = 1;
			}

			return NULL;
		}

		if (softcut->func != NULL)
		{
			/* Disable text in the progressbar */
			pbar.decoded = -1;
			pbar.encoded = -1;
			(*softcut->func)(&pbar);
		}

		if (mvpkt > 1)
		{
			if (remaining_packets > 0)
			{
				/* Do not unref packets moving to queue 0 */
				pktqueue->num -= remaining_packets;

				/* Remember number (index) of packets of q3 */
				num_pkts = pktqueue->num;
			}
		}

		else
		{
			/* No packets to move */
			remaining_packets = 0;
		}


		cff_ffmpeg_copy_free_softcut_packets(softcut->pktqueue, 0);
		avcodec_free_context(&(softcut->oc));


		if (remaining_packets > 0)
		{
			softcut->pktqueue[0].num = remaining_packets;
			softcut->pktqueue[0].key_frames = key_frames;

			memcpy(softcut->pktqueue[0].packets,
				&pktqueue->packets[num_pkts],
					remaining_packets * sizeof(AVPacket*));

		} /* if (remaining_packets > 0) */


		if (video->copy.dts_next == INT64_MAX ||
			video->copy.cutidx >= video->cut.num)
		{
			/* Nothing to decode-encode */
			softcut->vst = NULL;
		}

		softcut->queue_idx = 0;

		if (packet == NULL)
		{
			/* Ready */
			return NULL;
		}

		packet->dts -= dts_offset;
		packet->pts -= dts_offset;

		if (softcut->vst == NULL)
		{
			/* Nothing to decode-encode */
			return cff_ffmpeg_copy_write_frame(video, st, packet);
		}

		/* Queue 0 */
		queue_idx = 0;
		pktqueue  = softcut->pktqueue;
		num_pkts  = pktqueue->num;

	} /* if (queue_idx == 3 && (packet == NULL ||
		(pktqueue->key_frames > 1 && (packet->flags & AV_PKT_FLAG_KEY)))) */

	if (packet == NULL)
	{
		pktqueue = &video->copy.softcut.pktqueue[0];
		num_pkts = pktqueue->num;
		pkthdl   = pktqueue->packets;

		if (pkthdl != NULL && num_pkts > 0)
		{
			av_log(NULL, AV_LOG_WARNING, "cutterff @ flush] "
					"Warning: frames left in queue 0: %i\n",
					num_pkts);
			do
			{
				str = cff_ffmpeg_copy_write_frame(video,
							st, *(pkthdl++));
				if (str != NULL)
				{
					return str;
				}
			}
			while (--num_pkts != 0);

		} /* if (pkthdl != NULL && num_pkts > 0) */

		return NULL;

	} /* if (packet == NULL) */

	if (pktqueue->packets == NULL)
	{
		if (queue_idx == 0)
		{
			video->copy.softcut.pktdts_last = packet->dts;

			return cff_ffmpeg_copy_write_frame(video, st, packet);
		}

		return NULL;
	}

	if ((packet->flags & AV_PKT_FLAG_KEY) && queue_idx == 0)
	{
		int x = 1;

		pkthdl	= pktqueue->packets;
		pkt	= pkthdl[1];

		if (pkt != NULL && pkt->flags & AV_PKT_FLAG_KEY)
		{
			x++;
		}

		if (pktqueue->key_frames > x)
		{
			do
			{
				str = cff_ffmpeg_copy_write_frame(video,
							st, *(pkthdl++));
				if (str != NULL)
				{
					return str;
				}

				pkt = *pkthdl;

				if (pkt != NULL &&
					(pkt->flags & AV_PKT_FLAG_KEY))
				{
					/*
						Move remaining packets
						to the start of the queue
					*/

					num_pkts--;
					pktqueue->key_frames--;
					memmove(pktqueue->packets, pkthdl,
						num_pkts * sizeof(AVPacket*));
					break;
				}
			}
			while (--num_pkts != 0);

		} /* if (pktqueue->key_frames > x) */

	} /* if ((packet->flags & AV_PKT_FLAG_KEY) && queue_idx == 0 */

	if (num_pkts >= pktqueue->max)
	{
		unsigned int	max;
		AVPacket	**ptr;

		max = pktqueue->max * 2;

		ptr = av_realloc(pktqueue->packets, max * sizeof(AVPacket*));

		if (ptr == NULL)
		{
			return gettext_nop("'av_realloc()' failed!");
		}

		pktqueue->packets = ptr;
		pktqueue->max     = max;
		memset(&ptr[num_pkts], '\0', (max - num_pkts) * sizeof(AVPacket*));

	} /* if (num_pkts >= pktqueue->max) */

	if (packet->flags & AV_PKT_FLAG_KEY)
	{
		pktqueue->key_frames++;
	}

	pktqueue->packets[num_pkts] = av_packet_clone(packet);

	if (pktqueue->packets[num_pkts] == NULL)
	{
		return gettext_nop("Could not clone a packet");
	}

	pktqueue->num = num_pkts + 1;

	if (queue_idx == 3 && packet->dts == softcut->dts_stop)
	{
		return cff_ffmpeg_copy_softcut(video, NULL, stop);
	}

	return NULL;

} /* static const char *cff_ffmpeg_copy_softcut(cff_video_t*, AVPacket*, int*) */

static int cff_ffmpeg_copy_dvbsub_clear(const AVPacket *packet)
{
	/*
		Simple function to detect if a
		dvb-subtitle should be cleared
	*/

	uint16_t	page_id;
	const uint8_t	*ptr, *end;

	ptr = packet->data;

	if (ptr == NULL || packet->size < kCFFDvbSubHeaderSize)
	{
		return 0;
	}

	end	= &ptr[packet->size];
	page_id	= (ptr[2] << 8 ) | ptr[3];

	do
	{
		/*
			struct
			{
				int8_t	sync_byte;
				int8_t	segment_type;
				int16_t	page_id;
				int16_t	segment_length;

			} DVBSUB_header;
		*/

		uint16_t id;
		size_t   segment_length;

		/* Sync byte */

		if (ptr[0] != 0x0F)
		{
			break;
		}

		id = (ptr[2] << 8 ) | ptr[3];

		if (id != page_id)
		{
			break;
		}

		segment_length = (ptr[4] << 8 ) | ptr[5];

		if (ptr[1] == kCFFDvbSubDisplayDefinition)
		{
			if (segment_length != 5)
			{
				break;
			}
		}

		else if (ptr[1] == kCFFDvbSubPageComposition)
		{
			if (segment_length != 2)
			{
				break;
			}
		}

		else if (ptr[1] == kCFFDvbSubRegionComposition)
		{
			if (segment_length != 10)
			{
				break;
			}
		}

		else if (ptr[1] == kCFFDvbSubEnd)
		{
			if (segment_length != 0)
			{
				break;
			}

			/* Clear subtitle */
			return 1;
		}

		/* Next segment */
		ptr += segment_length + kCFFDvbSubHeaderSize;
	}
	while (end >= &ptr[kCFFDvbSubHeaderSize]);

	/* Display subtitle */
	return 0;

} /* static int cff_ffmpeg_copy_dvbsub_clear(const AVPacket*) */

static int cff_ffmpeg_copy_need_rescaling(AVStream *a, AVStream *b)
{
	/* Test whether rescaling is needed */

	if (a != b)
	{
		if (a->time_base.num != b->time_base.num)
		{
			return 1;
		}

		if (a->time_base.den != b->time_base.den)
		{
			return 1;
		}

	} /* if (a != b) */

	return 0;

} /* static int cff_ffmpeg_copy_need_rescaling(AVStream*, AVStream*) */

static int64_t cff_ffmpeg_convert_dts_video_to_stream(cff_video_t *video, struct StreamData *st, int64_t dts)
{
	/*
		Convert the timestamp of the video sequence
		to the timebase of another stream
	*/

	if (dts != AV_NOPTS_VALUE && dts != INT64_MAX)
	{
		AVStream *is = (video->inFormatCtx->streams[video->vframe.stream_index]);

		if (cff_ffmpeg_copy_need_rescaling(is, st->iStream))
		{
			dts = av_rescale_q(dts, is->time_base, (st->iStream)->time_base);
		}
	}

	return dts;

} /* static int64_t cff_ffmpeg_convert_dts_video_to_stream(cff_video_t*, struct StreamData*, int64_t) */

static int64_t cff_ffmpeg_copy_get_offset(cff_video_t *video, struct StreamData *st)
{
	/*
		Convert the start time of the video sequence
		to the time base of another stream
	*/

	int64_t dts = video->copy.gap_offset;

	if ( ! (video->copy.flags & kCFFCopyFlagTsKeepStart))
	{
		dts += video->copy.dts_start;
	}

	return cff_ffmpeg_convert_dts_video_to_stream(video, st, dts);

} /* static int64_t cff_ffmpeg_copy_get_offset(cff_video_t*, struct StreamData*) */

static int cff_ffmpeg_copy_stream_finish(const cff_video_t *video, struct StreamData *st)
{
	int flags = video->copy.flags | kCFFCopyFlagSeek;

	if (flags & kCFFCopyFlagStreamSync)
	{
		int n;

		st->flags &= ~(kCFFStreamFlagCopy);
		st->flags |=   kCFFStreamFlagCopyDone;

		st = video->copy.streamData;
		n  = video->copy.nb_streams;

		do
		{
			if (st->flags & kCFFStreamFlagCopy)
			{
				/* Stream is active */
				flags &= ~(kCFFCopyFlagSeek);
				break;
			}

			st++;
		}
		while (--n > 0);

	} /* if (flags & kCFFCopyFlagStreamSync) */

	return flags;

} /* static int cff_ffmpeg_copy_stream_finish(const cff_video_t*, struct StreamData*) */

static int64_t cff_ffmpeg_copy_get_next(cff_video_t *video, int incr)
{
	/* Get next start or stop frame */

	int64_t dts_next;

	if (video->copy.cutidx >= video->cut.num)
	{
		/* Stop or copy till end */
		dts_next = INT64_MAX;
	}

	else
	{
		/* Next start or stop */
		dts_next = video->cut.list[video->copy.cutidx].packet;

		if (incr)
		{
			video->copy.cutidx++;
		}

		if (dts_next >= video->vframe.dts_stop)
		{
			/* Stop or copy till end */
			dts_next = INT64_MAX;
		}
	}

	return dts_next;

} /* static int64_t cff_ffmpeg_copy_get_next(cff_video_t*, int) */

int cff_ffmpeg_copy_cb(cff_video_t *video, double *progress, const char **error)
{
	/* Read frames and copy them to the output file */

	int		i, end, is_stop, retv;
	const char	*str;
	AVPacket	*packet;

	packet  = video->vframe.tmppkt;
	str	= NULL;
	end	= 0;
	retv	= 0;
	i	= 100;

	/* Softcut DTS */
	is_stop  = 0;

	do
	{
		int			n, skip_encoding;
		int64_t			dts, pts, dts_start;
		struct StreamData	*st;

		if (video->copy.flags & kCFFCopyFlagSeek)
		{
			int64_t dts_next;

			if (video->copy.flags & kCFFCopyFlagEOF)
			{
				goto _stop;
			}

			/* Get next start frame */
			video->copy.flags &= ~(kCFFCopyFlagSeek);

			dts_next = cff_ffmpeg_copy_get_next(video, 1);

			if (dts_next == INT64_MAX)
			{
				goto _stop;
			}

			if (dts_next == AV_NOPTS_VALUE)
			{
				/* May happen if seeking the first frame */
				dts_next = video->vframe.dts_start;
			}

			dts_start = dts_next;

			/* Start time of sequence */
			video->copy.dts_beg = dts_start;

			if (video->copy.dts_start == AV_NOPTS_VALUE)
			{
				/* Start time of first video packet */
				video->copy.dts_start = dts_start;
			}

			if (video->copy.dts_end != AV_NOPTS_VALUE)
			{
				video->copy.gap_offset +=
					dts_start - video->copy.dts_end;		}

			skip_encoding = 0x00;

			if (video->copy.softcut.vst != NULL &&
				video->copy.softcut.queue_idx == 0 &&
						video->copy.cutidx == 1)
			{
				video->copy.softcut.dts_stop = INT64_MAX;
				skip_encoding = 0x01;
			}

			if ((video->copy.softcut.vst != NULL ||
				(video->copy.flags & kCFFCopyFlagStreamSync)) &&
						dts_next > video->vframe.dts_start)
			{
				int64_t	dts_found, dts_seek;

				if (video->copy.softcut.vst != NULL)
				{
					video->copy.softcut.dts_next = dts_next;
				}

				dts_seek  = dts_next;
				dts_found = AV_NOPTS_VALUE;

				n = 3;

				do
				{
					unsigned int index;

					dts_seek -= video->vframe.max_gop_size;

					retv = av_seek_frame(video->inFormatCtx,
							video->vframe.stream_index, dts_seek, 0);

					if (retv != 0)
					{
						retv = -1;
						break;
					}

					do
					{
						retv = av_read_frame(video->inFormatCtx, packet);

						if (retv < 0)
						{
							n = 0;
							break;
						}


						dts   = AV_NOPTS_VALUE;
						index = packet->stream_index;

						if (index == video->vframe.stream_index &&
									(packet->flags & AV_PKT_FLAG_KEY))
						{
							dts = packet->dts;

							if (dts != AV_NOPTS_VALUE && dts_next > dts)
							{
								if (video->vframe.max_gop_size >
											(dts_next - dts))
								{
									if ( ! (skip_encoding & 0x01))
									{
										/* We need a full gop, try again */
										dts = dts_next;
									}
								}

								else
								{
									dts_found = dts;
								}

							} /* if (dts != AV_NOPTS_VALUE && dts_next > dts) */

						} /* if (index == video->vframe.stream_index &&
									(packet->flags & AV_PKT_FLAG_KEY)) */

						av_packet_unref(packet);
					}
					while (dts == AV_NOPTS_VALUE || dts_next > dts);

					if (retv >= 0 && (skip_encoding & 0x01))
					{
						if (dts_next == dts)
						{
							/* First frame is a keyframe */
							skip_encoding = 0x02;

							if ( ! (video->copy.flags & kCFFCopyFlagStreamSync))
							{
								/* Start with first frame */
								break;
							}
						}

						skip_encoding &= ~(0x01);

					} /* if (retv >= 0 && (skip_encoding & 0x01)) */
				}
				while (dts_found == AV_NOPTS_VALUE && --n > 0);

				if (retv < 0)
				{
					retv = -1;
					str = g_avseek_frame_errstr;
					break;
				}

				if (dts_found == AV_NOPTS_VALUE)
				{
					if (video->copy.softcut.queue_idx == 1)
					{
						retv = -1;
						str = g_avseek_frame_errstr;
						break;
					}
				}

				else
				{
					if ( ! (skip_encoding & 0x02) &&
						video->copy.softcut.vst != NULL)
					{
						video->copy.softcut.queue_idx = 2;
					}

					if ( ! (skip_encoding & 0x02) ||
						(video->copy.flags & kCFFCopyFlagStreamSync))
					{
						dts_next = dts_found;
					}

				} /* if (dts_found == AV_NOPTS_VALUE) .. else */

			} /* if ((video->copy.softcut.vst != NULL ||
				(video->copy.flags & kCFFCopyFlagStreamSync)) &&
						dts_next > video->vframe.dts_start) */

			if (video->copy.softcut.vst != NULL &&
				video->copy.softcut.queue_idx == 0 &&
							! (skip_encoding & 0x02))
			{
				/*
					May encoded if both gops are enabled and
					the sequence has only one keyframe!
				*/

				av_log(NULL, AV_LOG_WARNING,
						"cutterff @ start] "
						"encoding start frame disabled\n");
			}

			retv = cff_ffmpeg_get_packet_with_dts(video,
							dts_next, packet);
			if (retv != 0)
			{
				str = g_avseek_frame_errstr;
				break;
			}

			/* Get next stop frame */

			dts_next = cff_ffmpeg_copy_get_next(video, 1);
			video->copy.dts_next = dts_next;

			if (video->copy.flags & kCFFCopyFlagStreamSync)
			{
				/* Break if missing a streams stop package */
				video->copy.dts_finish =
						cff_ffmpeg_copy_get_next(video, 0);
			}

			if (video->copy.softcut.queue_idx == 2)
			{
				video->copy.softcut.dts_stop = dts_next;
			}

			st = video->copy.streamData;
			n  = video->copy.nb_streams;

			do
			{
				unsigned int flags;

				/* Clear flags except these */
				flags = kCFFStreamFlagGotStart |
						kCFFStreamFlagTsFromDuration |
							kCFFStreamFlagTsRescale |
								kCFFStreamFlagBsfInit |
									kCFFStreamFlagDvbSub;

				if (st->flags & kCFFStreamFlagTsFromDuration)
				{
					/* Offset needed only at start */
					flags |= kCFFStreamFlagGotOffset;
				}

				st->flags &= flags;

				if (video->copy.flags & kCFFCopyFlagStreamSync)
				{
					st->dts_start =
						cff_ffmpeg_convert_dts_video_to_stream(video, st, dts_start);
					st->dts_stop =
						cff_ffmpeg_convert_dts_video_to_stream(video, st, dts_next);
				}

				st++;
			}
			while (--n > 0);

			st = video->copy.ast[packet->stream_index];

			if (st == NULL)
			{
				/* Video stream not selected */
				continue;
			}

			/*
				If start of the video stream, and encoding
				is not neccassary, or seeking backward failed,
				softcut.queue_idx has a value of '0', else '2'
			*/

			if (video->copy.softcut.vst == NULL)
			{
				if ((video->copy.flags & kCFFCopyFlagStreamSync) &&
					(dts_start == packet->dts ||
						packet->pos == video->vframe.pktpos_start))
				{
					st->flags |= kCFFStreamFlagCopy;
				}
			}

			else if ( ! (skip_encoding & 0x02))
			{
				st->flags |= kCFFStreamFlagCopy;
			}
		}

		else /* ! (video->copy.flags & kCFFCopyFlagSeek) */
		{
			if (video->copy.flags & kCFFCopyFlagEOF)
			{
				goto _stop;
			}

			retv = cff_ffmpeg_read_packet(video, packet, &str);

			if (retv < 0)
			{
_stop:
				video->copy.flags |= kCFFCopyFlagEOF;

				st = video->copy.streamData;
				n  = video->copy.nb_streams;

				do
				{
					if (st->bsf_ctx != NULL &&
						(st->flags & kCFFStreamFlagBsfInit) &&
							! (st->flags & kCFFStreamFlagBsfDone))
					{
						goto _bsf;
					}

					st++;
				}
				while (--n > 0);


				if (video->copy.softcut.vst != NULL)
				{
					str = cff_ffmpeg_copy_softcut(video, NULL, NULL);

					if (str != NULL)
					{
						retv = -1;
						break;
					}
				}

				retv = av_write_trailer(video->copy.outFormatCtx);

				if (retv < 0 && str == NULL)
				{
					str = g_trailer_write_errstr;
				}

				end = 1;
				break;

			} /* if (retv < 0) */

			if (packet->stream_index > (video->inFormatCtx)->nb_streams)
			{
				continue;
			}

			st = video->copy.ast[packet->stream_index];

			if (st == NULL)
			{
				/*
					If this packet belongs to the video
					stream, and 'stream_sync' is not
					selected, test the stop condition
					below. Else read the next package.
				*/

				if (video->copy.flags & kCFFCopyFlagStreamSync)
				{
					continue;
				}

				if (packet->stream_index != video->vframe.stream_index)
				{
					continue;
				}

			} /* if (st == NULL) */

		} /* if (video->copy.flags & kCFFCopyFlagSeek) .. else */


		if (st == video->copy.softcut.vst && st != NULL)
		{
			cff_softcut_t *softcut = &(video->copy.softcut);

			if (st->flags & kCFFStreamFlagCopyDone)
			{
				if (packet->dts > video->copy.dts_finish)
				{
					/* Next sequence reached (Stream-sync) */
					video->copy.flags |= kCFFCopyFlagSeek;
				}

				continue;
			}

			if (softcut->queue_idx == 0)
			{
				if ( ! (st->flags & kCFFStreamFlagCopy))
				{
					if (video->copy.dts_beg > packet->dts)
					{
						continue;
					}

					st->flags |= kCFFStreamFlagCopy;
				}

				/* One GOP before stop frame */

				if (packet->dts > video->copy.dts_next)
				{
					/* End time of sequence */
					video->copy.dts_end = packet->dts;

					/* Duration till queue 1 or 2 */
					softcut->duration_last = st->duration;
					softcut->queue_idx = 1;

					if (softcut->pktqueue[0].packets == NULL)
					{
						/* De-/encode only GOP after gap */
						video->copy.flags =
							cff_ffmpeg_copy_stream_finish(video, st);

						continue;
					}

				} /* if (packet->dts > video->copy.dts_next) */
			}

			else if (softcut->queue_idx == 1)
			{
				/* One GOP after queue 0 stop frame */

				if ((packet->flags & AV_PKT_FLAG_KEY) &&
						softcut->pktqueue[1].key_frames != 0)
				{
					video->copy.flags =
						cff_ffmpeg_copy_stream_finish(video, st);
				}
			}

			else if (softcut->queue_idx == 2)
			{
				/* One GOP before queue 3 start frame */

				if (packet->dts >= softcut->dts_next)
				{
					/* Duration of queue 1 + 2 */
					softcut->duration_last =
							st->duration - softcut->duration_last;

					softcut->queue_idx = 3;
				}
			}

		} /* if (st == video->copy.softcut.vst && st != NULL) */

		else if ( ! (video->copy.flags & kCFFCopyFlagStreamSync))
		{
			if (packet->stream_index == video->vframe.stream_index)
			{
				if (video->copy.dts_beg == AV_NOPTS_VALUE)
				{
					/* video->cut.list == NULL */

					dts = packet->dts;

					if (dts == AV_NOPTS_VALUE)
					{
						dts = video->vframe.dts_start;
					}

					/* Start time of first video packet */
					video->copy.dts_beg   = dts;
					video->copy.dts_start = dts;
				}

				if (packet->dts > video->copy.dts_next)
				{
					/* End time of sequence */
					video->copy.dts_end  = packet->dts;
					video->copy.flags   |= kCFFCopyFlagSeek;
					continue;
				}

				if (st == NULL)
				{
					continue;
				}
			}

			else if (video->copy.softcut.queue_idx == 1 ||
					video->copy.softcut.queue_idx == 2)
			{
				continue;
			}
		}

		else /* video->copy.flags & kCFFCopyFlagStreamSync */
		{
			if (st->flags & kCFFStreamFlagCopyDone)
			{
				if (packet->stream_index == video->vframe.stream_index &&
							packet->dts > video->copy.dts_finish)
				{
					/* Next sequence reached */
					video->copy.flags |= kCFFCopyFlagSeek;
				}

				/* Nothing to do */
				continue;
			}

			if (st->flags & kCFFStreamFlagCopy)
			{
				if (packet->dts > st->dts_stop)
				{
					if (packet->stream_index == video->vframe.stream_index)
					{
						/* End time of sequence */
						video->copy.dts_end = packet->dts;
					}

					video->copy.flags =
						cff_ffmpeg_copy_stream_finish(video, st);

					/*
						If a dvb subtitle is displayed and should be
						cleared, write this packet, else drop it.
					*/

					if ( ! (st->flags & kCFFStreamFlagDvbSub) ||
							cff_ffmpeg_copy_dvbsub_clear(packet) == 0)
					{
						/* No subtitle to clear */
						continue;
					}

					/* Clip timestamps of dvb-subtitle to stop */
					packet->pts = st->dts_stop;
					packet->dts = st->dts_stop;

				} /* if (packet->dts > st->dts_stop) */
			}

			else /* ! (st->flags & kCFFStreamFlagCopy) */
			{
				if (st->dts_start > packet->dts)
				{
					continue;
				}

				if (packet->dts > st->dts_stop)
				{
					/* Overlap */
					video->copy.flags =
						cff_ffmpeg_copy_stream_finish(video, st);

					continue;
				}

				st->flags |= kCFFStreamFlagCopy;

			} /* if (st->flags & kCFFStreamFlagCopy) .. else */

		} /* if (st == video->copy.softcut.vst && st != NULL) .. else if .. else */


		if ( ! (st->flags & kCFFStreamFlagGotOffset))
		{
			if ( ! (st->flags & kCFFStreamFlagGotStart))
			{
				if (video->copy.dts_beg == AV_NOPTS_VALUE &&
					packet->stream_index != video->vframe.stream_index)
				{
					/* No start time of sequence */
					continue;
				}

				dts = packet->dts;

				if (dts == AV_NOPTS_VALUE)
				{
					/* Zero if not kCFFCopyFlagStreamSync */
					dts = st->dts_start;
				}

				st->dts_offset = cff_ffmpeg_copy_get_offset(video, st);

				st->pkt_dts = dts - st->dts_offset;

				if (st->flags & kCFFStreamFlagTsFromDuration)
				{
					/* Timestamps starts from first packet */
					st->dts_offset = st->pkt_dts;
				}

				if ((st->oStream->codecpar)->codec_type ==
							AVMEDIA_TYPE_VIDEO &&
					(packet->flags & AV_PKT_FLAG_KEY) &&
					strcmp((video->copy.outFormatCtx->oformat)->name, "mp4") == 0)
				{
					/* Format 'MP4' */
					packet->pts = packet->dts;
				}

				/* Done getting start */
				st->flags |= kCFFStreamFlagGotStart;
			}

			else if (packet->dts == AV_NOPTS_VALUE ||
					(st->oStream->codecpar)->codec_id == AV_CODEC_ID_DVB_SUBTITLE)
			{
				st->dts_offset = cff_ffmpeg_copy_get_offset(video, st);
			}

			else
			{
				st->dts_offset = packet->dts - st->pkt_dts;

			} /* if ( ! (st->flags & kCFFStreamFlagGotStart)) .. else if .. else*/

			/* Done getting offset */
			st->flags |= kCFFStreamFlagGotOffset;

		} /* if ( ! (st->flags & kCFFStreamFlagGotOffset)) */


		packet->stream_index = (st->oStream)->index;


		/* PTS */

		pts = packet->pts;

		if (pts == AV_NOPTS_VALUE)
		{
			if (st->flags & kCFFStreamFlagTsFromDuration)
			{
				pts = st->duration + st->dts_offset;
			}
		}

		else
		{
			if (st->flags & kCFFStreamFlagTsFromDuration)
			{
				int64_t pts_diff;

				pts      = st->duration + st->dts_offset;
				pts_diff = packet->pts - packet->dts;

				if (pts_diff > 0)
				{
					pts += pts_diff;
				}
			}

			else
			{
				pts -= st->dts_offset;
			}

		} /* if (pts == AV_NOPTS_VALUE) .. else */

		packet->pts = pts;


		/* DTS */

		dts = packet->dts;

		if (st == video->copy.softcut.vst &&
				dts == video->copy.softcut.dts_stop)
		{
			is_stop = 1;

		}

		if (st->flags & kCFFStreamFlagTsFromDuration)
		{
			dts = st->duration + st->dts_offset;
		}

		else
		{
			if (dts == AV_NOPTS_VALUE)
			{
				dts = st->pkt_dts;
			}

			else
			{
				dts -= st->dts_offset;
			}

			if ((st->oStream->codecpar)->codec_id ==
							AV_CODEC_ID_DVB_SUBTITLE)
			{
				/*
					DVB subtitle need one packet to display them
					and another packet to clear them (ffmpeg.c)
				*/

				if (st->pkt_dts > dts)
				{
					dts = st->pkt_dts;
					packet->pts = dts;
				}

				if (cff_ffmpeg_copy_dvbsub_clear(packet))
				{
					if ( ! (st->flags & kCFFStreamFlagDvbSub))
					{
						/* No subtitle to clear */
						continue;
					}

					/* This is the stop frame */

					if (packet->duration == 0)
					{
						st->duration += dts - st->pkt_dts;
					}

					st->flags &= ~kCFFStreamFlagDvbSub;
				}

				else
				{
					st->flags |= kCFFStreamFlagDvbSub;
				}

			} /* if ((st->oStream->codecpar)->codec_id ==
							AV_CODEC_ID_DVB_SUBTITLE) */

		} /* if (st->flags & kCFFStreamFlagTsFromDuration) .. else */


		st->pkt_dts = dts + packet->duration;


		if ((st->oStream->codecpar)->codec_type == AVMEDIA_TYPE_AUDIO &&
			 ! (st->flags & kCFFStreamFlagTsFromDuration) &&
			 				packet->dts == AV_NOPTS_VALUE)
		{
			int duration;

			duration = av_get_audio_frame_duration2((st->iStream)->codecpar, packet->size);

			if (duration == 0)
			{
				duration = (st->iStream->codecpar)->frame_size;
			}

			dts = av_rescale_delta((st->iStream)->time_base,
					dts,
					(AVRational) {1,
						(st->iStream->codecpar)->sample_rate},
					duration,
					&(st->delta_last),
					(st->oStream)->time_base);

		} /* if ((st->oStream->codecpar)->codec_type == AVMEDIA_TYPE_AUDIO && ... */

		packet->dts = dts;


		/* BSF */

		if (st->bsf_ctx != NULL)
		{
_bsf:
			if ( ! (st->flags & kCFFStreamFlagBsfFlush))
			{
				AVPacket *pktptr = packet;

				if (video->copy.flags & kCFFCopyFlagEOF)
				{
					pktptr = NULL;
					st->flags |= kCFFStreamFlagBsfFlush;
				}

				if (av_bsf_send_packet(st->bsf_ctx, pktptr) < 0)
				{
					if (video->copy.flags & kCFFCopyFlagEOF)
					{
						st->flags |= kCFFStreamFlagBsfDone;
					}

					continue;
				}

			} /* if ( ! (st->flags & kCFFStreamFlagBsfFlush)) */

			if (av_bsf_receive_packet(st->bsf_ctx, packet) < 0)
			{
				if (video->copy.flags & kCFFCopyFlagEOF)
				{
					st->flags |= kCFFStreamFlagBsfDone;
				}

				continue;
			}

			if ( ! (st->flags & kCFFStreamFlagBsfInit))
			{
				if (avcodec_parameters_copy((st->oStream)->codecpar,
								(st->bsf_ctx)->par_out) < 0)
				{
					continue;
				}

				st->flags |= kCFFStreamFlagBsfInit;
			}

		} /* if (st->bsf_ctx != NULL) */

		st->duration += packet->duration;

		if (st == video->copy.softcut.vst)
		{
			int cancel;

			if (is_stop)
			{
				is_stop = 0;
				video->copy.softcut.dts_stop = packet->dts;
			}

			str = cff_ffmpeg_copy_softcut(video, packet, &cancel);

			if (cancel != 0)
			{
				i = 0;
			}
		}

		else
		{
			str = cff_ffmpeg_copy_write_frame(video, st, packet);
		}

		if (str != NULL)
		{
			i    = 0;
			retv = -1;
		}
	}
	while (av_packet_unref(packet), --i > 0);

	if (error != NULL)
	{
		*error = str;
	}

	if (retv == 0 && progress != NULL)
	{
		if (end)
		{
			*progress = 1.0;
		}

		else
		{
			*progress = (double)
				avio_tell((video->inFormatCtx)->pb) /
							(double) video->fsize;
		}
	}

	return end | retv;

} /* int cff_ffmpeg_copy_cb(cff_video_t*, double*, const char**) */

const char *cff_ffmpeg_copy_open(cff_video_t *video, const cff_copy_param_t *param)
{
	static const char malloc_err[] =
			gettext_nop("Could not allocate memory!");

	int			retv;
	unsigned int		i, duration[4];
	char			*str;
	const char		*error, *short_name, *mime_type;
	size_t			bufsz, slen, sz;
	int64_t			streamSize, totalSize;
	double			diffSize;
	AVFormatContext		*inFormatCtx, *outFormatCtx;
	AVProgram		*program;
	AVDictionary		*opts;
	struct StreamData	*st;
#ifdef HAVE_FFMPEG_AVOUTPUTFORMAT_NOCONST
	AVOutputFormat		*format;
#else
	/* Since 'ffmpeg-5' const */
	const AVOutputFormat	*format;
#endif

	if (param->info != NULL)
	{
		*param->info = NULL;
	}

	if (param->fname == NULL || param->fname[0] == '\0')
	{
		return g_missing_filename_errstr;
	}

	if (param->stream_idx == NULL || param->nb_streams == 0 ||
				param->nb_streams > video->nb_streams)
	{
		return gettext_nop("Wrong stream index!");
	}


	sz = sizeof(struct StreamData*) * (video->inFormatCtx)->nb_streams;
	video->copy.ast = av_mallocz(sz);

	if (video->copy.ast == NULL)
	{
		return malloc_err;
	}

	video->copy.nb_streams = param->nb_streams;

	sz = sizeof(struct StreamData) * param->nb_streams;
	video->copy.streamData = av_mallocz(sz);

	if (video->copy.streamData == NULL)
	{
		av_free(video->copy.ast);
		return malloc_err;
	}


	/* Open output file */

	inFormatCtx = video->inFormatCtx;
	short_name  = param->format_name;
	mime_type   = NULL;

	if (short_name == NULL)
	{
		short_name = param->fname;
		mime_type  = (inFormatCtx->iformat)->mime_type;
	}

	i = 1;

	do
	{
		if (i == 0 && strcmp((inFormatCtx->iformat)->name, "mpeg") == 0)
		{
			short_name = "vob";
		}

		format = av_guess_format(short_name, param->fname, mime_type);
	}
	while (format == NULL && param->format_name == NULL && --i == 0);

	retv = avformat_alloc_output_context2(&(video->copy.outFormatCtx),
						format, NULL, param->fname);

	if (retv != 0 || video->copy.outFormatCtx == NULL)
	{
		cff_ffmpeg_copy_free(video);

		return (format == NULL) ?
				g_find_oformat_errstr :
					g_open_ocontext_errstr;
	}

	outFormatCtx = video->copy.outFormatCtx;
	program = (AVProgram *) inFormatCtx->programs;

	if (program != NULL)
	{
		program = av_find_program_from_stream(inFormatCtx, NULL,
						video->vframe.stream_index);
		if (program != NULL)
		{
			AVProgram *dst = av_new_program(outFormatCtx, program->id);

			if (dst != NULL)
			{
				dst->stream_index = av_malloc(sizeof(unsigned int) * param->nb_streams);

				if (dst->stream_index == NULL)
				{
					av_free(dst);
					av_freep(&(outFormatCtx->programs));
					outFormatCtx->nb_programs = 0;
					program = NULL;
				}

				else
				{
					dst->nb_stream_indexes = param->nb_streams;
					dst->program_num = program->program_num;
					dst->pmt_pid = program->pmt_pid;
					dst->pcr_pid = program->pcr_pid;

					av_dict_copy(&(dst->metadata), program->metadata, 0);

					program = dst;

				} /* if (dst->stream_index == NULL) .. else */

			} /* if (dst != NULL) .. else */

		} /* if (program != NULL) */

	} /* if (program != NULL) */


	error	= NULL;
	st	= video->copy.streamData;
	i	= 0;

	do
	{
		int			index, n;
		enum AVCodecID		codec_id;
		enum AVMediaType	codec_type;
		const AVCodec		*codec;
		AVCodecParameters	*oCodecPar;
		const char		*bsf_mp4toannexb, *bsf_name;
		const cff_media_t	*media;

		index = param->stream_idx[i];

		if (index >= (inFormatCtx)->nb_streams)
		{
			error = gettext_nop("Stream index greater than number of streams!");
			break;
		}

		video->copy.ast[index] = st;
		st->iStream = inFormatCtx->streams[index];

		codec_id = (st->iStream->codecpar)->codec_id;
		codec = avcodec_find_decoder(codec_id);

		if (codec != NULL)
		{
			codec_id = codec->id;
		}

		st->oStream = avformat_new_stream(outFormatCtx, codec);

		if (st->oStream == NULL)
		{
			error = g_alloc_stream_errstr;
			break;
		}

		av_dict_copy(&((st->oStream)->metadata),
					(st->iStream)->metadata, 0);

		oCodecPar = (st->oStream)->codecpar;

		retv = avcodec_parameters_copy(oCodecPar, (st->iStream)->codecpar);

		if (retv < 0)
		{
			error = g_copy_codec_parameters_errstr;
			break;
		}

		if (av_codec_get_id((outFormatCtx->oformat)->codec_tag,
				(st->iStream->codecpar)->codec_tag) != codec_id)
		{
			unsigned int tag;

			if (av_codec_get_tag2((outFormatCtx->oformat)->codec_tag,
									codec_id, &tag) != 0)
			{
				oCodecPar->codec_tag = tag;
			}
		}

		/*
			Copy the frame rates before calling
			'avformat_transfer_internal_stream_timing_info'.
			See 'ffmpeg/libavformat/utils.c'
		*/

		if (oCodecPar->codec_type == AVMEDIA_TYPE_VIDEO)
		{
			(st->oStream)->avg_frame_rate =
						(st->iStream)->avg_frame_rate;
			(st->oStream)->r_frame_rate =
						(st->iStream)->r_frame_rate;
		}

		if (avformat_transfer_internal_stream_timing_info(outFormatCtx->oformat,
						st->oStream, st->iStream, AVFMT_TBCF_AUTO) < 0)
		{
			error = gettext_nop("Could not transfer internal stream timing info!");
			break;
		}

		n = 0;

		while (((st->oStream)->time_base.num <= 0 ||
				(st->oStream)->time_base.den <= 0) && n < 3)
		{
			if (n == 0)
			{
				(st->oStream)->time_base = (st->iStream)->time_base;
			}

			else if (n == 1)
			{
				(st->oStream)->time_base =
					av_add_q(av_stream_get_codec_timebase(st->oStream),
									(AVRational) {0, 1});
			}

			else /* (n == 2) */
			{
				(st->oStream)->time_base.num = 1;
				(st->oStream)->time_base.den =
							(video->vframe.ic)->ticks_per_frame;
			}

			n++;

		} /* while (((st->oStream)->time_base.num <= 0 ||
				(st->oStream)->time_base.den <= 0) && n < 3) */

		cff_ffmpeg_copy_timebase_reduce(codec_id, &((st->oStream)->time_base));

		(st->iStream)->discard = AVDISCARD_DEFAULT;

		if (program != NULL && program->stream_index != NULL)
		{
			program->stream_index[i] = (st->oStream)->index;
		}

		(st->oStream)->id = (st->iStream)->id;
		(st->oStream)->disposition = (st->iStream)->disposition;

		/* Clear duration and bitrate */
		(st->oStream)->duration = 0;
		oCodecPar->bit_rate = 0;

		if ((st->iStream)->side_data != NULL && (st->iStream)->nb_side_data != 0)
		{
			int nsd = 0;

			do
			{
				uint8_t			*dst;
				const AVPacketSideData	*src;

				src = &(st->iStream->side_data[nsd]);
				dst = av_stream_new_side_data(st->oStream,
								src->type, src->size);
				if (dst == NULL)
				{
					error = gettext_nop("Could not copy side data!");
					break;
				}

				memcpy(dst, src->data, src->size);
			}
			while (++nsd < (st->iStream)->nb_side_data);

			if (error != NULL)
			{
				break;
			}

		} /* if ((st->iStream)->side_data != NULL && (st->iStream)->nb_side_data != 0) */


		bsf_mp4toannexb = NULL;

		codec_type = oCodecPar->codec_type;

		if (codec_type == AVMEDIA_TYPE_VIDEO)
		{
			AVRational sar;

			sar = (st->iStream)->sample_aspect_ratio;
			oCodecPar->sample_aspect_ratio = sar;
			(st->oStream)->sample_aspect_ratio = sar;

			if ((st->iStream)->index == video->vframe.stream_index &&
				param->encoder_name != NULL &&
					video->cut.list != NULL && video->cut.num != 0)
			{
				unsigned int	max;
				cff_pktqueue_t	*pktqueue;

				video->copy.softcut.codec = avcodec_find_encoder_by_name(param->encoder_name);

				if (video->copy.softcut.codec == NULL)
				{
					error = gettext_nop("No encoder found!");
					break;
				}


				/* The video stream */
				video->copy.softcut.vst = st;

				if (strncmp(param->encoder_name, "libx264", 7) == 0)
				{
					video->copy.softcut.flags = kCFFSoftcutFlagLibx264;
				}

				else if (strcmp(param->encoder_name, "libx265") == 0)
				{
					video->copy.softcut.flags = kCFFSoftcutFlagLibx265;
				}

				error = cff_ffmpeg_encoder_new(video);

				if (error != NULL)
				{
					break;
				}

				if (param->softcut.data != NULL)
				{
					video->copy.softcut.func_data = param->softcut.data;
					video->copy.softcut.func = param->softcut.func;
				}


				/* Max packets in the queue */
				max = (video->vframe.max_gop_size * 3) /
						video->vframe.frame_duration;
				n = 0;

				if (param->softcut.both_gops == 0)
				{
					n = 2;
				}

				pktqueue = &video->copy.softcut.pktqueue[n];

				do
				{
					pktqueue->max = max;

					pktqueue->packets = av_mallocz(max * sizeof(AVPacket*));

					if (pktqueue->packets == NULL)
					{
						error = malloc_err;
						break;
					}

					pktqueue++;
				}
				while (++n < 4);

				if (error != NULL)
				{
					break;
				}

				video->copy.softcut.dts_stop = INT64_MAX;

				/* Bitstream filter (see 'cff_ffmpeg_get_frame_dts') */
				bsf_mp4toannexb = video->vframe.bsf_mp4toannexb;

			} /* if ((st->iStream)->index == video->vframe.stream_index &&
				param->encoder_name != NULL &&
					video->cut.list != NULL && video->cut.num != 0) */

			media = &(param->media[kCFFMediaVideo]);
		}

		else if (codec_type == AVMEDIA_TYPE_AUDIO)
		{
			if (oCodecPar->codec_id == AV_CODEC_ID_MP3)
			{
				if (oCodecPar->block_align == 1 ||
					oCodecPar->block_align == 576 ||
						oCodecPar->block_align == 1152)
				{
					oCodecPar->block_align = 0;
				}
			}

			else if (oCodecPar->codec_id == AV_CODEC_ID_AC3)
			{
				oCodecPar->block_align = 0;
			}

			media = &(param->media[kCFFMediaAudio]);
		}

		else
		{
			media = &(param->media[kCFFMediaOther]);
		}


		bsf_name = (media->bsf_name != NULL) ?
				media->bsf_name : bsf_mp4toannexb;

		if (bsf_name != NULL)
		{
			const AVBitStreamFilter *bsf;

			bsf = av_bsf_get_by_name(bsf_name);

			if (bsf != NULL)
			{
				enum AVCodecID		codec_bsf;
				const enum AVCodecID	*id_ptr;

				codec_bsf = codec_id;
				id_ptr = bsf->codec_ids;

				if (id_ptr != NULL)
				{
					while ((codec_bsf = *(id_ptr++)) != codec_id)
					{
						if (codec_bsf == AV_CODEC_ID_NONE)
						{
							break;
						}
					}
				}

				if (codec_bsf == codec_id && av_bsf_alloc(bsf, &(st->bsf_ctx)) == 0)
				{
					if (avcodec_parameters_copy((st->bsf_ctx)->par_in, oCodecPar) < 0)
					{
						error = gettext_nop("Error copying codec parameters to the bitstream filter!");
						break;
					}

					(st->bsf_ctx)->time_base_in = (st->oStream)->time_base;

					if (av_bsf_init(st->bsf_ctx) < 0)
					{
						error = gettext_nop("Error initializing the bitstream filter!");
						break;
					}

					if (avcodec_parameters_copy(oCodecPar, (st->bsf_ctx)->par_out) < 0)
					{
						error = gettext_nop("Error copying bitstream filter parameters to the codec!");
						break;
					}

					(st->oStream)->time_base = (st->bsf_ctx)->time_base_out;

					if (error == 0 && bsf_name == bsf_mp4toannexb)
					{
						av_log(NULL, AV_LOG_WARNING, "cutterff @ copy_open] Added bitstream filter '%s'!\n",
							bsf_mp4toannexb);
					}

				} /* if (codec_bsf == codec_id && av_bsf_alloc(bsf, &(st->bsf_ctx)) == 0) */

			} /* if (bsf != NULL) */

		} /* if (bsf_name != NULL) */


		if (media->ts_from_duration)
		{
			/* Compute timestamps from duration */
			st->flags |= kCFFStreamFlagTsFromDuration;
		}

		st++;
	}
	while (++i < param->nb_streams);

	if (error != NULL)
	{
		cff_ffmpeg_copy_free(video);
		return error;
	}

	retv = avio_open(&(outFormatCtx->pb), param->fname, AVIO_FLAG_WRITE);

	if (retv < 0)
	{
		cff_ffmpeg_copy_free(video);
		return g_open_ofile_errstr;
	}

	av_dict_copy(&(outFormatCtx->metadata), inFormatCtx->metadata, 0);

	opts = NULL;

	if (param->shortest)
	{
		av_dict_set(&opts, "fflags", "shortest", 0);
	}

	retv = avformat_write_header(outFormatCtx, &opts);

	av_dict_free(&opts);

	if (retv < 0)
	{
		cff_ffmpeg_copy_free(video);
		return g_header_write_errstr;
	}

	st = video->copy.streamData;
	i  = 0;

	do
	{
		if (cff_ffmpeg_copy_need_rescaling(st->iStream, st->oStream))
		{
			st->flags |= kCFFStreamFlagTsRescale;
		}

		st++;
	}
	while (++i < param->nb_streams);

	if (video->cut.list == NULL || video->cut.num == 0)
	{
		/* First frame */
		retv = cff_ffmpeg_seek(video, video->vframe.dts_start);

		if (retv != 0)
		{
			cff_ffmpeg_copy_free(video);
			return g_avseek_frame_errstr;
		}

		/* Copy all */
		video->copy.dts_next = INT64_MAX;
		video->copy.cutidx   = (unsigned int) -1;
	}

	else
	{
		/* Seek first packet */
		video->copy.flags |= kCFFCopyFlagSeek;

		if (param->stream_sync != 0 &&
			cff_ffmpeg_video_cut_list_distance(video) != 0)
		{
			/*
				Enable stream sync only if
				at least one cut point is set.
			*/

			video->copy.flags |= kCFFCopyFlagStreamSync;
		}

	} /* if (video->cut.list == NULL || video->cut.num == 0) .. else */

	if (param->ts_keep_start != 0 &&
		strcmp((outFormatCtx->oformat)->name, "avi") != 0)
	{
		/*
			Keep original start timestamps if selected
			'AVI' starts from zero
		*/

		video->copy.flags |= kCFFCopyFlagTsKeepStart;
	}

	/* Disable start time of video (wait for first packet) */
	video->copy.dts_start = AV_NOPTS_VALUE;
	video->copy.dts_beg   = AV_NOPTS_VALUE;
	video->copy.dts_end   = AV_NOPTS_VALUE;


	/*
	   ---------------
	    AND NOW START
	   ---------------
	*/

	if (param->func == NULL)
	{
		while ((retv = cff_ffmpeg_copy_cb(video, NULL, &error)) == 0) {;}
	}

	else
	{
		error = NULL;
		retv  = (*param->func)(param->user_data);
	}

	if (retv <= 0 || error != NULL)
	{
		/* Error, or user canceled */
		cff_ffmpeg_copy_free(video);
		return error;
	}

	if (param->info == NULL)
	{
		cff_ffmpeg_copy_free(video);
		return gettext_nop("File successfully copied!");
	}

	bufsz = 4096;
	slen  = 0;
	str   = cff_ffmpeg_printf_concat(NULL, &slen, &bufsz,
				"Output: %s\n\n", param->fname);
	streamSize = 0;

	st = video->copy.streamData;
	i  = 0;

	do
	{
		char buf[256];

		streamSize += st->nbBytesWritten;

		if (cff_ffmpeg_stream_info_to_buf(st->oStream,
							buf, sizeof(buf)) < 0)
		{
			continue;
		}

		/* 'st->duration' uses time base from input stream */
		cff_ffmpeg_dts2time(st->duration, st->iStream, duration);

		str = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
					"   Stream #%s"
					" => Packets:%u Size:%1.0fkB "
					"Duration: %02u:%02u:%02u.%03u\n",
					buf,
					st->nbPackets,
					(double) st->nbBytesWritten / 1024.0,
					duration[0],
					duration[1],
					duration[2],
					duration[3]);
	}
	while (st++, ++i < video->copy.nb_streams);

	totalSize = avio_size((video->copy.outFormatCtx)->pb);

	if (totalSize < 0)
	{
		totalSize = avio_tell((video->copy.outFormatCtx)->pb);

		if (totalSize < 0)
		{
			totalSize = 0;
		}
	}

	if (totalSize > streamSize)
	{
		diffSize = (double) (totalSize - streamSize);
	}

	else
	{
		diffSize = 0.0;
	}

	str = cff_ffmpeg_printf_concat(str, &slen, &bufsz,
		"\n"
		"Size:%1.0fkB Global headers:%1.0fkB "
		"Muxing overhead:%1.2f%%",
			(double) totalSize / 1024.0,
			diffSize / 1024.0,
			(totalSize == 0 || streamSize == 0) ?
			0.0 : 100.0 * (diffSize / (double) streamSize));

	cff_ffmpeg_copy_free(video);

	*(param->info) = cff_ffmpeg_string_truncate(str, slen);

	return NULL;

} /* const char *cff_ffmpeg_copy_open(cff_video_t*, const cff_copy_param_t*) */

const char *cff_ffmpeg_save_image(cff_video_t *video, const cff_image_param_t *param)
{
	int				i, align, retv;
	const char			*error, *mime_type;
	enum AVCodecID			codec_id;
	enum AVPixelFormat		pixfmt;
	const enum AVPixelFormat	*pix_ptr;
	const AVPixFmtDescriptor	*pix_desc;
	const AVCodec         		*codec;
	AVStream			*os;
	AVCodecContext			*oc;
	AVFormatContext			*outFormatCtx;
	AVFrame				*frame;
	AVDictionary			*opts;
	AVPacket			*packet;
#ifdef HAVE_FFMPEG_AVOUTPUTFORMAT_NOCONST
	AVOutputFormat			*format;
#else
	/* Since 'ffmpeg-5' const */
	const AVOutputFormat		*format;
#endif

	if (param->info != NULL)
	{
		*param->info = NULL;
	}

	if (param->fname == NULL || param->fname[0] == '\0')
	{
		return g_missing_filename_errstr;
	}

	i = 1;
	mime_type = NULL;

	do
	{
		if (i == 0)
		{
			const char *ext;

			if (param->format_name != NULL)
			{
				break;
			}

			ext = strrchr(param->fname, '.');

			if (ext == NULL ||
				(strcmp(&ext[1], "jpg") != 0 &&
					strcmp(&ext[1], "jpeg") != 0))
			{
				break;
			}

			mime_type = "image/jpeg";

		} /* if (i == 0) */

		format = av_guess_format(param->format_name, param->fname, mime_type);
	}
	while (format == NULL && --i == 0);

	oc   = NULL;
	retv = avformat_alloc_output_context2(&outFormatCtx,
						format, NULL, param->fname);

	if (retv != 0 || outFormatCtx == NULL)
	{
		error = (format == NULL) ?
				g_find_oformat_errstr :
					g_open_ocontext_errstr;
		goto _done;
	}

	codec_id = (format == NULL) ?
		(outFormatCtx->oformat)->video_codec : format->video_codec;
	codec = avcodec_find_encoder(codec_id);

	if (codec == NULL)
	{
		error = gettext_nop("No image codec found!");
		goto _done;
	}


	oc = avcodec_alloc_context3(NULL);

	if (oc == NULL)
	{
		error = g_alloc_codec_context_errstr;
		goto _done;
	}

	if (param->keep_dimension)
	{
		oc->width  = (video->vframe.ic)->width;
		oc->height = (video->vframe.ic)->height;
	}

	else
	{
		oc->width  = param->width;
		oc->height = param->height;
	}

	if (param->keep_ratio)
	{
		cff_ffmpeg_scale_dimension(video, &(oc->width), &(oc->height));
	}

	oc->time_base	= (AVRational) {1, 1};
	pixfmt		= AV_PIX_FMT_NONE;
	pix_ptr		= codec->pix_fmts;

	if (pix_ptr != NULL)
	{
		if (param->gray != 0)
		{
			while ((pixfmt = *(pix_ptr++)) != AV_PIX_FMT_NONE)
			{
				pix_desc = av_pix_fmt_desc_get(pixfmt);

				if (strncmp(pix_desc->name, "gray", 4) == 0)
				{
					break;
				}
			}

		} /* if (param->gray != 0) */

		if (pixfmt == AV_PIX_FMT_NONE)
		{
			/* 'avcodec_default_get_format' may return an invalid format */
			pixfmt = codec->pix_fmts[0];

			if (pixfmt == AV_PIX_FMT_NONE)
			{
				pixfmt = avcodec_default_get_format(oc, codec->pix_fmts);
			}

		} /* if (pixfmt == AV_PIX_FMT_NONE) */

	} /* if (pix_ptr != NULL) */

	oc->pix_fmt = pixfmt;

	opts = NULL;

	if (param->gray)
	{
		av_dict_set(&opts, "flags", "gray", 0);
	}

	/* Side Aspect Ratio */
	av_dict_set(&opts, "aspect", "1:1", 0);

	retv = avcodec_open2(oc, codec, &opts);

	av_dict_free(&opts);

	if (retv < 0)
	{
		error = g_open_codec_errstr;
		goto _done;
	}

	os = avformat_new_stream(outFormatCtx, codec);

	if (os == NULL)
	{
		error = g_alloc_stream_errstr;
		goto _done;
	}

	avcodec_parameters_from_context(os->codecpar, oc);

	retv = avio_open(&(outFormatCtx->pb),
				param->fname, AVIO_FLAG_WRITE);
	if (retv < 0)
	{
		error = g_open_ofile_errstr;
		goto _done;
	}

	retv = avformat_write_header(outFormatCtx, NULL);

	if (retv < 0)
	{
		error = g_header_write_errstr;
		goto _done;
	}


	if (pixfmt == (video->vframe.ic)->pix_fmt &&
		oc->width == (video->vframe.ic)->width &&
			oc->height == (video->vframe.ic)->height)
	{
		frame = av_frame_clone(video->vframe.avframe);

		if (frame == NULL)
		{
			error = g_clone_frame_errstr;
			goto _done;
		}
	}

	else
	{
		frame = av_frame_alloc();

		if (frame == NULL)
		{
			error = g_alloc_frame_errstr;
			goto _done;
		}

		align = 1;
		pix_desc = av_pix_fmt_desc_get(oc->pix_fmt);

		if (pix_desc != NULL &&
			(pix_desc->flags & (AV_PIX_FMT_FLAG_PAL |
					CFF_PIX_FMT_FLAG_PSEUDOPAL)))
		{
			/* Has a palette */
			align = 4;
		}

		retv = cff_ffmpeg_scale_frame(video->vframe.avframe,
					oc->width, oc->height, oc->pix_fmt,
						align, frame->data, frame->linesize);
		if (retv < 0)
		{
			av_frame_free(&frame);
			error = gettext_nop("Could not convert the image!");
			goto _done;
		}

		frame->width     = oc->width;
		frame->height    = oc->height;
		frame->format    = oc->pix_fmt;

	} /* if (pixfmt == (video->vframe.ic)->pix_fmt &&
			oc->width == (video->vframe.ic)->width &&
				oc->height == (video->vframe.ic)->height) .. else */

	frame->pts       = 0;
	frame->pkt_dts   = 0;
	frame->key_frame = 1;

	retv = avcodec_send_frame(oc, frame);
	av_frame_free(&frame);

	if (retv < 0)
	{
		error = gettext_nop("Could not write the frame to the codec!");
		goto _done;
	}

	packet	= video->vframe.tmppkt;
	retv	= avcodec_receive_packet(oc, packet);

	if (retv < 0)
	{
		if (retv == AVERROR(EAGAIN))
		{
			avcodec_send_frame(oc, NULL);
			retv = avcodec_receive_packet(oc, packet);
		}

		if (retv < 0)
		{
			error = gettext_nop("Could not read the packet from the codec!");
			goto _done;
		}

	} /* if (retv < 0) */

	retv = av_write_frame(outFormatCtx, packet);
	av_packet_unref(packet);

	if (retv < 0)
	{
		error = gettext_nop("'av_write_frame()' failed!");
		goto _done;
	}

	retv = av_write_trailer(outFormatCtx);

	if (retv < 0)
	{
		error = g_trailer_write_errstr;
		goto _done;
	}

	if (param->info == NULL)
	{
		error = gettext_nop("Image successfully saved!");
	}

	else
	{
		*(param->info) = cff_ffmpeg_file_info_io(outFormatCtx,
							param->fname, 1);
		error = NULL;
	}

_done:
	if (oc)
	{
		avcodec_free_context(&oc);
	}

	if (outFormatCtx != NULL)
	{
		if (outFormatCtx->pb != NULL)
		{
			/* Close the output file */
			avio_closep(&(outFormatCtx->pb));
		}

		avformat_free_context(outFormatCtx);

	} /* if (outFormatCtx != NULL) */

	return error;

} /* const char *cff_ffmpeg_save_image(cff_video_t*, const cff_image_param_t*) */

/* EOF */
