/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017 .. 2019,2021,2022 Harald Foerster	*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include <gdk/gdkkeysyms.h>
#include <glib/gstdio.h>

#include "cff_macros-gtk.h"
#include "cff_accel-gtk.h"
#include "cff_gprop.h"
#include "cff_nls.h"
#include "cff_charconv.h"
#include "cff_ffmpeg.h"

#ifdef HAVE_SYS_NPROCESSORS

#ifdef BUILD_MINGW
#include <windows.h>
#elif _MACOS
#include <sys/param.h>
#include <sys/sysctl.h>
#else
#include <unistd.h>
#endif

#endif /* HAVE_SYS_NPROCESSORS */

/*
#include <stdio.h>
#include <gtk/gtk.h>
#include <sys/types.h>
#include "cff_tools-gtk.h"
*/

#ifndef BUILD_MINGW
/* Application icon */
#include "../cutterff.xpm"
#endif

#if GTK_MAJOR_VERSION > 2
#ifdef HAVE_GTKOVERWRITECONFIRMATION
#undef HAVE_GTKOVERWRITECONFIRMATION
#endif
#ifdef HAVE_CAIRO
#undef HAVE_CAIRO
#endif
#define HAVE_GTKOVERWRITECONFIRMATION	1
#define HAVE_CAIRO			1
#endif /* GTK_MAJOR_VERSION > 2 */

#ifdef BUILD_MINGW
#define CFF_IS_DIR_SEPARATOR(chr)	((chr) == '/' || (chr) == '\\')
#else
#define CFF_IS_DIR_SEPARATOR(chr)	((chr) == '/')
#endif

#ifdef GTK_TEXT_VIEW_PRIORITY_VALIDATE
#define CFF_TEXT_VIEW_PRIORITY_VALIDATE	GTK_TEXT_VIEW_PRIORITY_VALIDATE
#else
#define CFF_TEXT_VIEW_PRIORITY_VALIDATE	Gui_PRIORITY_REDRAW + 5
#endif


typedef struct
{
	GThread			*thread;
	GThreadFunc		func;
	gpointer		data;
	gpointer		result;
	volatile gint		step;
	GuiMutex		mutex;

} cff_thread_t;

static cff_thread_t gThread = { NULL };

typedef struct
{
	GtkWidget	*some;
	GtkWidget	*all;
	cff_ListStore_t	list;

} cff_streamlist_t;

typedef struct
{
	GtkWidget	*ts_from_duration;
	GtkWidget	*bsf_option;
	cff_bsflist_t	*bsf_list;

} cff_mediaopts_t;

typedef struct
{
	GtkWidget		*window;
	GtkEntry		*entry;

	struct
	{
		GtkWidget	*option;
		gchar		*extension_orig;

		union
		{
			cff_fmtlist_t	*list;   /* Video */
			gchar		**names; /* Image */
		} u;

	} format;

	struct
	{
		GtkWidget	*widget;
		guint		bits;

	} okay;

	struct
	{
		gboolean	(*func)(GtkWidget*, GdkEvent*, gpointer);
		gpointer	data;

	} cancel;

	GtkWidget		**mbar;
	cff_video_t		*video;
	gchar			*filechooser_utf8str;
	gint			path_prop;

} cff_save_common_t;

typedef struct
{
	cff_save_common_t	common;
	GtkWidget		*up;
	GtkWidget		*down;
	GtkWidget		*shortest;
	GtkWidget		*stream_sync;
	GtkWidget		*ts_keep_start;

	struct
	{
		GtkWidget	*option;
		GtkWidget	*both_gops;
		gchar		**encoder_names;

	} softcut;

	cff_streamlist_t	avail;
	cff_streamlist_t	out;
	cff_mediaopts_t		media[kCFFMediaTotal];

} cff_save_video_t;

typedef struct
{
	gint			bsfname_idx;	/* Filter to use */
	gboolean		ts_from_duration; /* Compute timestamps from duration */

} cff_save_media_prev_t;

typedef struct
{
	gint			*stream_idx;	/* List of stream id's */
	gint			nb_streams;	/* Number of streams */
	gint			format_idx;	/* Muxer to use */
	gint			encoder_idx;	/* Softcut => decode / encode */
	gboolean		shortest;	/* Finish with shortest stream duration */
	gboolean		ts_keep_start;	/* Keep original start timestamps */
	gboolean		stream_sync;	/* For better synchron */
	gboolean		both_gops;	/* Encode both (end/start) GOPs */
	cff_save_media_prev_t	media[kCFFMediaTotal];

} cff_save_video_prev_options_t;

static cff_save_video_prev_options_t *gSaveVideoPrevOptionsPtr = NULL;


typedef struct
{
	GtkSpinButton		*widget;
	gint			prev_value;

} cff_spinbutton_t;

typedef struct
{
	cff_save_common_t	common;
	cff_spinbutton_t	width;
	cff_spinbutton_t	height;
	GtkWidget		*keep_dimension;
	GtkWidget		*keep_ratio;
	GtkWidget		*gray;

} cff_save_image_t;

typedef struct
{
	volatile GtkProgressBar	*widget;
	GuiMutex		*mutex;

	union
	{
		gdouble		progress;

		struct
		{
			gint		decoded;
			gint		encoded;
			gboolean	show_text;

		} softcut;
	} u;

} cff_progressbar_callback_t;

typedef struct
{
	int (*func)(cff_video_t*, double*, const char**);

	cff_video_t			*video;
	const gchar			*title;
	GtkWidget			**mbar;
	GuiMutex			*mutex;
	GtkProgressBar			*progress_bar;
	
	/*
		Lock and data for 'cff_gui_dialog_progress_set_fraction'
		and 'cff_gui_dialog_progress_set_text'
	*/

	cff_progressbar_callback_t	cb;
	
	const gchar			*error;
	gint				retv;
	volatile gboolean		is_locked;

} cff_progressbar_t;

typedef struct
{
	GtkWidget	*window;
	GtkWidget	*textview;

	struct
	{
		GtkAdjustment	*h_widget;
		gdouble		v_upper;
		gdouble		v_pagesize;

	} scrollbar;

	struct
	{
		gchar		*str;
		guint		len;
		guint		max;
		guint		lastchar;
		guint		cr_len;
		gboolean	overwrite;

	} utf8;

	guint		nThreads;
	guint		timeout_id;
	gint		busy_loop;
	GuiMutex	mutex;

} cff_logwindow_t;

static cff_logwindow_t gLogWindow = { NULL };

static const gchar gLogWinTextColor[] = "#00EE00";
static const gchar gLogWinBackgroundColor[] = "#000000";


/* Open, Program or Save */
#define kIsBusyFile		1
/* 'cff_gui_video_frame' */
#define kIsBusyVideoFrame	2

static gint gIsBusy = 0;


/* Ignore these events if busy */
#define kButtonPressEvent	0
#define kKeyPressEvent		1


/* Filter */

#define kFileFilterEntriesMax	10
#define kFileFilterCharsMax	15

typedef struct
{
	GtkFileChooser	*chooser;
	GtkWidget	*btn_add;
	GtkWidget	*btn_remove;
	GtkFileFilter	*buf[kFileFilterEntriesMax];
	guint		num;
	const gchar	*ext;
	gint		prop;

} cff_filefilter_t;

static const gchar cff_video_ext[] = "* avi mov mp4 mpg mpeg mkv ts webm";
static const gchar cff_image_ext[] = "* jpg";

typedef struct
{
	enum
	{
		kFrameSeekNext,
		kFrameSeekCutpoint,
		kFrameSeekSliderClick,
		kFrameSeekSliderMove

	} type;

	union
	{
		enum cff_VideoFrame	e_next;
		unsigned int		i_next;
		gdouble			d_next;

	} value;

} cff_frame_seek_t;



/*
	------------
	 PROTOTYPES
	------------
*/

/* File Menu */
static void cff_gui_file_open(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_file_save(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_file_image(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_file_close(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_file_info(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_file_save_rc(GtkWidget *w, gpointer data);
static void cff_gui_file_quit(GtkWidget *w, GtkWidget **mbar);

/* Edit Menu */
static void cff_gui_edit_undo(GtkWidget *w, cff_video_t *video);
static void cff_gui_edit_redo(GtkWidget *w, cff_video_t *video);
static void cff_gui_edit_load(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_edit_save(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_edit_clear(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_edit_add(GtkWidget *w, cff_video_t *video);
static void cff_gui_edit_remove(GtkWidget *w, cff_video_t *video);

/* Submenu for frame numbers */
static void cff_gui_build_frame_submenu(GtkWidget *submenu, gpointer data);

/* Submenu for programs */
static void cff_gui_build_program_submenu(GtkWidget *submenu, gpointer data);

/* Submenu for loglevels */
static void cff_gui_build_loglevel_submenu(GtkWidget *submenu, gpointer data);

/* Options Menu */
static void cff_gui_opt_toggle_bool(GtkWidget *w, gpointer data);

/* Help Menu */
static void cff_gui_help_ffmpeg(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_help_cutterff(GtkWidget *w, GtkWidget **mbar);
static void cff_gui_help_about(GtkWidget *w, GtkWidget **mbar);

/* Prototype needed for 'cff_gui_calculate_frame_rects' */
static void cff_gui_widget_queue_draw_area(GtkWidget *widget);

/* Called from 'cff_gaccel.c' and 'cff_gui-gtk.c' */
void cff_gui_adjust_menu(GtkWidget *w, gpointer data);

/* cff_main.c */
extern char *cff_main_save_rc(void);


/* Function for drawing rectangles */

#ifdef HAVE_CAIRO

#if GTK_MAJOR_VERSION == 2

#define CffGuiDrawRectangle(cr, event, color, fill, rect) \
		cff_gui_cairo_draw_rect_gtk2((cr), \
			(GdkEventExpose *) (event), (color), (fill), (rect))
#else

#define CffGuiDrawRectangle(cr, event, color, fill, rect) \
		cff_gui_cairo_draw_rectangle((cr), (color), (fill), (rect))
#endif

#else /* ! HAVE_CAIRO */

#define CffGuiDrawRectangle(cr, event, color, fill, rect) \
		cff_gui_gdk_draw_rectangle((GdkEventExpose *) (event), \
						(color), (fill), (rect))

#endif /* #ifdef HAVE_CAIRO .. #else */


/*
	--------
	 CURSOR
	--------
*/

typedef struct
{
	GdkWindow	*window;
	GdkCursor	*busy;

} cff_cursor_t;

static cff_cursor_t gCursor = { NULL };


/*
	-------------
	 VIDEO FRAME
	-------------
*/

typedef struct
{
	GtkWidget	*scale;
	GtkAdjustment	*adjust;
	gdouble		value;
	gdouble		x;
	gboolean	is_slider;
	volatile gint	button;

} cff_slider_t;

static cff_slider_t gSlider = { NULL };


typedef struct
{
	GtkWidget	*drawingArea;
	GtkLabel	*info;
	GtkLabel	*time;
	const guchar	**dataRGB24;
#ifdef HAVE_CAIRO
	cairo_surface_t	*surfaceRGB24;
#else
	gint		rowstrideRGB24;
#endif
	GdkRectangle	rectRGB24;
	GdkRectangle	rectLeftTop;
	GdkRectangle	rectRightBottom;

} cff_video_frame_t;

static cff_video_frame_t gCurrentVideoFrame;


static GtkWidget *gDrawingAreaBar;

enum
{
	kColorBlack,
	kColorDarkGrey,
	kColorGrey,
	kColorGreen,
	kColorRed,
	kColorTotal
};

#if GTK_MAJOR_VERSION == 2

#ifdef HAVE_CAIRO
static GdkColor	gColor[kColorTotal];
#else
static GdkGC *gContext[kColorTotal];
#endif

static const char cff_configure_event[]	= "configure-event";
static const char cff_expose_event[]	= "expose-event";

#else

static GdkRGBA gRGBA[kColorTotal];

static const char cff_configure_event[]	= "size-allocate";
static const char cff_expose_event[]	= "draw";

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */


/*
	----------------
	 BUTTON WIDGETS
	----------------
*/

enum
{
	kButtonVideoFramePrev4,
	kButtonVideoFramePrev3,
	kButtonVideoFramePrev2,
	kButtonVideoFramePrev1,
	kButtonVideoFrameNext1,
	kButtonVideoFrameNext2,
	kButtonVideoFrameNext3,
	kButtonVideoFrameNext4,
	kButtonVideoFrameTotal
};

static gchar *gButtonVideoFrameLabel[] =
{
	"|<",
	"<<<",
	"<<",
	"<",
	">",
	">>",
	">>>",
	">|"
};

static GtkWidget *gui_button_video_frame[kButtonVideoFrameTotal];

enum
{
	kButtonVideoCutAdd,
	kButtonVideoCutRemove,
	kButtonVideoCutTotal
};

static GtkWidget *gui_button_video_cut[kButtonVideoCutTotal];


/*
	--------------
	 MENU WIDGETS
	--------------
*/

enum
{
	kMenuBarFile,
	kMenuBarEdit,
	kMenuBarOption,
	kMenuBarHelp,
	kMenuBarTotal
};

/* Enbled or disabled */

enum
{
	kMenuFileSave,
	kMenuFileImage,
	kMenuFileClose,
	kMenuFileInfo,
	kMenuEditUndo,
	kMenuEditRedo,
	kMenuEditLoad,
	kMenuEditSave,
	kMenuEditClear,
	kMenuEditAdd,
	kMenuEditRemove,
	kMenuEditFrame,
	kMenuEditProgram,
	kMenuItemsTotal
};

enum
{
	kPopupFileSave,
	kPopupEditUndo,
	kPopupEditRedo,
	kPopupEditLoad,
	kPopupEditSave,
	kPopupEditClear,
	kPopupEditFrame,
	kPopupItemsTotal
};

static GtkWidget *cff_gui_menu_bar[kMenuBarTotal];
static GtkWidget *cff_gui_menu_widget[kMenuItemsTotal];
static GtkWidget *cff_gui_popup_widget[kPopupItemsTotal];

/* kPropBoolOptKeyFrame, kPropBoolOptIFrame and kPropBoolOptPFrame */
static GtkWidget *cff_gui_checkmenu_frame_widget[3];
static GtkWidget *cff_gui_popup_checkmenu_frame_widget[3];

static GtkWidget *cff_gui_popup_menu;


/*
	------------
	 MENU TYPES
	------------
*/

enum
{
	kMTypeEndOfList,
	kMTypeItem,
	kMTypeCheckItem,
	kMTypeSeparator,
	kMTypeSubmenuCheckItem
};


/*
	------------
	 MENU ITEMS
	------------
*/

static const gchar mn_file_open_str[]	= gettext_nop("Open ...");
static const gchar mn_file_save_str[]	= gettext_nop("Save ...");
static const gchar mn_edit_undo_str[]	= gettext_nop("Undo");
static const gchar mn_edit_redo_str[]	= gettext_nop("Redo");
static const gchar mn_edit_load_str[]	= gettext_nop("Load list ...");
static const gchar mn_edit_save_str[]	= gettext_nop("Save list ...");
static const gchar mn_edit_clear_str[]	= gettext_nop("Clear list ...");
static const gchar mn_edit_frame_str[]	= gettext_nop("Goto");
static const gchar mn_opt_keyframe_str[]= gettext_nop("Key-Frame");
static const gchar mn_opt_iframe_str[]	= gettext_nop("I-Frame");
static const gchar mn_opt_pframe_str[]	= gettext_nop("P-Frame");

typedef struct
{
	gint		type;		/* See above */
	const gchar	*label;		/* Name to display */
	gint		accel_id;	/* Accel key */
	GtkWidget	**storage;	/* cff_gui_menu_widget */
	GuiSignalFunc	callback;	/* Function pointer */
	gpointer	data;		/* Data for it */

} cff_guiMenuItem_t;

static const cff_guiMenuItem_t cff_gui_menuFile[] =
{
	{kMTypeItem, mn_file_open_str,
		kGAccelFileOpen, NULL,
		(GuiSignalFunc) cff_gui_file_open, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeItem, mn_file_save_str,
		kGAccelFileSave, &cff_gui_menu_widget[kMenuFileSave],
		(GuiSignalFunc) cff_gui_file_save, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeItem, gettext_nop("Save image ..."),
		kGAccelFileImage, &cff_gui_menu_widget[kMenuFileImage],
		(GuiSignalFunc) cff_gui_file_image, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Close"),
		kGAccelFileClose, &cff_gui_menu_widget[kMenuFileClose],
		(GuiSignalFunc) cff_gui_file_close, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeItem, gettext_nop("Info ..."),
		kGAccelFileInfo, &cff_gui_menu_widget[kMenuFileInfo],
		(GuiSignalFunc) cff_gui_file_info, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Save rc file"),
		kGAccelFileSaveRC, NULL,
		(GuiSignalFunc) cff_gui_file_save_rc, NULL},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Quit"),
		kGAccelFileQuit, NULL,
		(GuiSignalFunc) cff_gui_file_quit, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuFile[] */

static const cff_guiMenuItem_t cff_gui_menuEdit[] =
{
	{kMTypeItem, mn_edit_undo_str,
		kGAccelEditUndo, &cff_gui_menu_widget[kMenuEditUndo],
		(GuiSignalFunc) cff_gui_edit_undo, &gVideo},
	{kMTypeItem, mn_edit_redo_str,
		kGAccelEditRedo, &cff_gui_menu_widget[kMenuEditRedo],
		(GuiSignalFunc) cff_gui_edit_redo, &gVideo},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_edit_load_str,
		kGAccelEditLoad, &cff_gui_menu_widget[kMenuEditLoad],
		(GuiSignalFunc) cff_gui_edit_load, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeItem, mn_edit_save_str,
		kGAccelEditSave, &cff_gui_menu_widget[kMenuEditSave],
		(GuiSignalFunc) cff_gui_edit_save, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeItem, mn_edit_clear_str,
		kGAccelEditClear, &cff_gui_menu_widget[kMenuEditClear],
		(GuiSignalFunc) cff_gui_edit_clear, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Add"),
		kGAccelEditAdd, &cff_gui_menu_widget[kMenuEditAdd],
		(GuiSignalFunc) cff_gui_edit_add, &gVideo},
	{kMTypeItem, gettext_nop("Remove"),
		kGAccelEditRemove, &cff_gui_menu_widget[kMenuEditRemove],
		(GuiSignalFunc) cff_gui_edit_remove, &gVideo},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenuCheckItem, mn_edit_frame_str,
		kGAccelKeyNone, &cff_gui_menu_widget[kMenuEditFrame],
		(GuiSignalFunc) cff_gui_build_frame_submenu, &gVideo},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenuCheckItem, gettext_nop("Program"),
		kGAccelKeyNone, &cff_gui_menu_widget[kMenuEditProgram],
		(GuiSignalFunc) cff_gui_build_program_submenu, &gVideo},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuEdit[] */

static const cff_guiMenuItem_t cff_gui_menuOptions[] =
{
	{kMTypeItem, gettext_nop("Keyboard ..."),
		kGAccelOptAccel, NULL,
		(GuiSignalFunc) cff_gaccel_dialog_open, NULL},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeCheckItem, mn_opt_keyframe_str,
		kGAccelOptKeyFrame, &cff_gui_checkmenu_frame_widget[0],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptKeyFrame)},
	{kMTypeCheckItem, mn_opt_iframe_str,
		kGAccelOptIFrame, &cff_gui_checkmenu_frame_widget[1],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptIFrame)},
	{kMTypeCheckItem, mn_opt_pframe_str,
		kGAccelOptPFrame, &cff_gui_checkmenu_frame_widget[2],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptPFrame)},
	{kMTypeCheckItem, gettext_nop("Cursor"),
		kGAccelOptCursor, NULL,
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptCursor)},
	{kMTypeCheckItem, gettext_nop("Overwrite"),
		kGAccelOptOverwrite, NULL,
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptOverwrite)},
	{kMTypeCheckItem, gettext_nop("Delete copy on error"),
		kGAccelOptDelete, NULL,
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptDelete)},
	{kMTypeCheckItem, gettext_nop("Beep"),
		kGAccelOptBeep, NULL,
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptBeep)},
	{kMTypeCheckItem, gettext_nop("Save rc on exit"),
		kGAccelOptSave, NULL,
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptSaveConfig)},
	{kMTypeSubmenuCheckItem, gettext_nop("Log ..."),
		kGAccelKeyNone, NULL,
		(GuiSignalFunc) cff_gui_build_loglevel_submenu, NULL},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuOptions[] */

static const cff_guiMenuItem_t cff_gui_menuHelp[] =
{
	{kMTypeItem, gettext_nop("FFmpeg ..."),
		kGAccelHelpFFmpeg, NULL,
		(GuiSignalFunc) cff_gui_help_ffmpeg, &cff_gui_menu_bar[kMenuBarHelp]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("CutterFF ..."),
		kGAccelHelpCutterff, NULL,
		(GuiSignalFunc) cff_gui_help_cutterff, &cff_gui_menu_bar[kMenuBarHelp]},
	{kMTypeItem, gettext_nop("About ..."),
		kGAccelHelpAbout, NULL,
		(GuiSignalFunc) cff_gui_help_about, &cff_gui_menu_bar[kMenuBarHelp]},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuHelp[] */

static const cff_guiMenuItem_t cff_gui_menuPopup[] =
{
	{kMTypeItem, mn_edit_undo_str,
		kGAccelKeyNone,  &cff_gui_popup_widget[kPopupEditUndo],
		(GuiSignalFunc) cff_gui_edit_undo, &gVideo},
	{kMTypeItem, mn_edit_redo_str,
		kGAccelKeyNone,  &cff_gui_popup_widget[kPopupEditRedo],
		(GuiSignalFunc) cff_gui_edit_redo, &gVideo},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_file_open_str,
		kGAccelKeyNone, NULL,
		(GuiSignalFunc) cff_gui_file_open, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeItem, mn_file_save_str,
		kGAccelKeyNone, &cff_gui_popup_widget[kPopupFileSave],
		(GuiSignalFunc) cff_gui_file_save, &cff_gui_menu_bar[kMenuBarFile]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_edit_load_str,
		kGAccelKeyNone, &cff_gui_popup_widget[kPopupEditLoad],
		(GuiSignalFunc) cff_gui_edit_load, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeItem, mn_edit_save_str,
		kGAccelKeyNone, &cff_gui_popup_widget[kPopupEditSave],
		(GuiSignalFunc) cff_gui_edit_save, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeItem, mn_edit_clear_str,
		kGAccelKeyNone, &cff_gui_popup_widget[kPopupEditClear],
		(GuiSignalFunc) cff_gui_edit_clear, &cff_gui_menu_bar[kMenuBarEdit]},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenuCheckItem, mn_edit_frame_str,
		kGAccelKeyNone, &cff_gui_popup_widget[kPopupEditFrame],
		(GuiSignalFunc) cff_gui_build_frame_submenu, &gVideo},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeCheckItem, mn_opt_keyframe_str,
		kGAccelKeyNone, &cff_gui_popup_checkmenu_frame_widget[0],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptKeyFrame)},
	{kMTypeCheckItem, mn_opt_iframe_str,
		kGAccelKeyNone, &cff_gui_popup_checkmenu_frame_widget[1],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptIFrame)},
	{kMTypeCheckItem, mn_opt_pframe_str,
		kGAccelKeyNone, &cff_gui_popup_checkmenu_frame_widget[2],
		(GuiSignalFunc) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptPFrame)},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuPopup[] */


/* ------------------- */


/*
	------
	 MISC
	------
*/

static void cff_gui_slider_adjustment_configure(guint upper)
{
	/*
		Using the same upper and lower bounds for
		the GtkScale will hide the slider itself
	*/

	if (upper == 0)
	{
		upper = 1;
	}

	gtk_adjustment_configure(gSlider.adjust,
				0.0,		/* Value */
				0.0,		/* Lower */
				(gdouble) upper,
				0.0,		/* Step_increment */
				0.0,		/* Page_increment */
				0.0);		/* Page_size */

} /* static void cff_gui_slider_adjustment_configure(guint) */

static gboolean cff_gui_slider_adjustment_set_value(cff_slider_t *slider)
{
	gtk_adjustment_set_value(slider->adjust,
			(gdouble) gVideo.vframe.frame_current);

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_slider_adjustment_set_value(cff_slider_t*) */

static gboolean cff_gui_key_or_mouse_button_pressed(GtkWidget *widget,
					GdkEvent *event, gint *isBusy)
{
	/* Supress mouse or button event if not the main level */

	if (*isBusy != 0 || gtk_main_level() > 1)
	{
		return TRUE;
	}

	return FALSE;

} /* static gboolean cff_gui_key_or_mouse_button_pressed(GtkWidget*,
							GdkEvent*, gint*) */

static void cff_gui_signal_connect_key_or_mouse_button(GtkWidget *widget, gint type)
{
	const gchar *str;

	if (type == kButtonPressEvent)
	{
		str = "button-press-event";
	}

	else /* kKeyPressEvent */
	{
		str = "key-press-event";
	}

	GuiSignalConnect(widget, str,
		cff_gui_key_or_mouse_button_pressed, &gIsBusy);

} /* static void cff_gui_signal_connect_key_or_mouse_button(GtkWidget*, gint) */

static gboolean cff_gui_overwrite(const gchar *path, GtkWidget **mbar)
{
	gint overwrite;

	if (cff_gprop_get_bool(kPropBoolOptOverwrite, &overwrite) && overwrite)
	{
		return TRUE;
	}

	if (cff_tl_file_exists(path) == 0)
	{
		return TRUE;
	}

	if (cff_guitl_dialog_warning_2btn(gettext_nop("File already exists! Overwrite?"),
			gettext_nop("Cancel"),
			gettext_nop("Yes"),
			mbar) == GTK_RESPONSE_ACCEPT)
	{
		return TRUE;
	}

	return FALSE;

} /* static gboolean cff_gui_overwrite(const gchar*, GtkWidget **mbar) */

static gboolean cff_gui_file_discard_changes(GtkWidget **mbar)
{
	if (gVideo.dirty &&
		cff_guitl_dialog_warning_2btn(gettext_nop("Discard changes?"),
					gettext_nop("Cancel"),
					gettext_nop("Yes"),
					mbar) == GTK_RESPONSE_CANCEL)
	{
		return FALSE;
	}

	return TRUE;

} /* static gboolean cff_gui_file_discard_changes(GtkWidget**) */

static gpointer cff_gui_thread_remove_me(gpointer data)
{
	return Gui_SOURCE_REMOVE;

} /* static gpointer cff_gui_thread_remove_me(gpointer) */

static gpointer cff_gui_thread_wait(cff_thread_t *thread)
{
	for (;;)
	{
		GuiMutexLock(&(thread->mutex));

		GuiMutexUnlock(&(thread->mutex));
		thread->step = 1;

		if (thread->func == NULL)
		{
			break;
		}

		gLogWindow.nThreads++;
		thread->result = (*thread->func)(thread->data);
		gLogWindow.nThreads--;

		while (thread->step != 2) {;}

		/*
			Calling 'gtk_main_quit' directly may
			freeze the program, if the thread
			function returns before 'gtk_main' in the
			main thread has its job successfully done.
		*/

		g_idle_add_full(G_PRIORITY_HIGH,
			(GSourceFunc) cff_gui_thread_remove_me,
				NULL, (GDestroyNotify) gtk_main_quit);

	} /* for (;;) */

	return NULL;

} /* gpointer cff_gui_thread_wait(cff_thread_t*) */

static gpointer cff_gui_thread_invoke(GThreadFunc func, gpointer data)
{
	cff_thread_t *thread = &gThread;

	if (thread->thread == NULL)
	{
		return (*func)(data);
	}

	thread->func = func;
	thread->data = data;

	thread->step = 0;
	GuiMutexUnlock(&(thread->mutex));

	while (thread->step != 1) {;}

	GuiMutexLock(&(thread->mutex));
	thread->step = 2;

	gtk_main();

	return thread->result;

} /* static gpointer cff_gui_thread_invoke(GThreadFunc, gpointer) */

static gpointer cff_gui_thread_run(gpointer data)
{
	gpointer result;

	struct
	{
		GThreadFunc	func;
		gpointer	data;

	} *s = data;

	gLogWindow.nThreads++;
	result = (*s->func)(s->data);
	gLogWindow.nThreads--;

	gtk_main_quit();

	return result;

} /* static gpointer cff_gui_thread_run(gpointer) */

static gboolean cff_gui_thread_new(GThreadFunc func, gpointer data, gpointer *retv)
{
	int nThreads;

#ifdef HAVE_GNUMPROCESSORS
	nThreads = g_get_num_processors();
#else

#ifdef HAVE_SYS_NPROCESSORS

#ifdef BUILD_MINGW

	SYSTEM_INFO sysinfo;

	GetSystemInfo(&sysinfo);
	nThreads = sysinfo.dwNumberOfProcessors;

#elif _MACOS

	size_t		len;
	uint32_t	count;
	int		nm[2];

	len   = sizeof(uint32_t);
	nm[0] = CTL_HW;
	nm[1] = HW_AVAILCPU;

	sysctl(nm, 2, &count, &len, NULL, 0);

	if(count < 1)
	{
		nm[1] = HW_NCPU;
		sysctl(nm, 2, &count, &len, NULL, 0);
	}

	nThreads = count;

#else /* UNIX */

	nThreads = sysconf(_SC_NPROCESSORS_ONLN);
#endif

#else
	nThreads = 0;

#endif /* #ifdef HAVE_SYS_NPROCESSORS .. #else */
#endif /* #ifdef HAVE_GNUMPROCESSORS .. #else */

	if (nThreads > 1)
	{
		if (func == NULL)
		{
			cff_thread_t *thread = &gThread;

			GuiMutexLock(&(thread->mutex));

			thread->thread = GuiThreadNew(cff_gui_thread_wait, thread);

			if (thread->thread != NULL)
			{
				return TRUE;
			}

			GuiMutexUnlock(&(thread->mutex));
		}

		else
		{
			GThread *thread;

			struct
			{
				GThreadFunc	func;
				gpointer	data;
			} s;

			s.func = func;
			s.data = data;

			thread = GuiThreadNew(cff_gui_thread_run, &s);

			if (thread != NULL)
			{
				gpointer result;

				gtk_main();
				result = g_thread_join(thread);

				if (retv != NULL)
				{
					*retv = result;
				}

				return TRUE;
			}

		} /* if (func == NULL) .. else */

	} /* if (nThreads > 1) */

	return FALSE;

} /* static gboolean cff_gui_thread_new(GThreadFunc, gpointer, gpointer*) */


/*
	--------------
	 PROGRESS-BAR
	--------------
*/

static gboolean cff_gui_dialog_progress_halt(GtkWidget *w, GdkEvent *e, cff_progressbar_t *data)
{
	/* 'button-press-event' or 'delete-event' */

	/* Stop */
	data->progress_bar = NULL;

	return TRUE;

} /* static gboolean cff_gui_dialog_progress_halt(GtkWidget*, GdkEvent*, cff_progressbar_t*) */

static gboolean cff_gui_dialog_progress_set_text(cff_progressbar_callback_t *cb)
{
	if (cb != NULL)
	{
		if (gtk_main_level() > 1)
		{
			GtkProgressBar *pbar;

			if (cb->mutex != NULL)
			{
				GuiMutexLock(cb->mutex);
			}

			pbar = (GtkProgressBar *) cb->widget;

			if (pbar != NULL)
			{
				if (cb->u.softcut.decoded < 0 || cb->u.softcut.encoded < 0)
				{
						cb->u.softcut.show_text = FALSE;
						GuiProgressbarTextHide(pbar);
				}

				else
				{
					gchar str[256];

					if (cb->u.softcut.show_text == FALSE)
					{
						cb->u.softcut.show_text = TRUE;
#if GTK_MAJOR_VERSION > 2
						gtk_progress_bar_set_show_text(pbar, 1);
#endif
					}

					g_snprintf(str, sizeof(str), "d: %i / e: %i",
							cb->u.softcut.decoded, cb->u.softcut.encoded);
					gtk_progress_bar_set_text(pbar, str);
				}

			} /* if (pbar != NULL) */

			cb->widget = NULL;

			if (cb->mutex == NULL)
			{
				while (g_main_context_iteration(NULL, FALSE)) {;}
			}

			else
			{
				GuiMutexUnlock(cb->mutex);
			}
		}

		else
		{
			cb->widget = NULL;

		} /* if (gtk_main_level() > 1) .. else */

	} /* if (cb != NULL) */

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_set_text(cff_progressbar_callback_t*) */

static int cff_gui_dialog_progress_softcut_cb(cff_softcut_pbar_t *data)
{
	/*
		Called from 'cff_ffmpeg_copy_softcut' to
		display the number of decoded and encoded
		frames.
	*/

	cff_progressbar_t *pbar, **pbarhdl;

	if (data == NULL)
	{
		return 0;
	}

	pbarhdl = data->func_data;

	if (pbarhdl == NULL)
	{
		return 0;
	}

	pbar = *pbarhdl;

	if (pbar == NULL)
	{
		return 0;
	}

	if (pbar->progress_bar == NULL)
	{
		/* Stop */
		return -1;
	}

	if (pbar->cb.mutex != NULL)
	{
		GuiMutexLock(pbar->cb.mutex);
	}

	if (pbar->cb.widget == NULL)
	{
		pbar->cb.widget = pbar->progress_bar;

		pbar->cb.u.softcut.decoded = data->decoded;
		pbar->cb.u.softcut.encoded = data->encoded;

		g_main_context_invoke_full(NULL, Gui_PRIORITY_REDRAW,
			(GSourceFunc) cff_gui_dialog_progress_set_text, &(pbar->cb), NULL);
	}

	if (pbar->cb.mutex != NULL)
	{
		GuiMutexUnlock(pbar->cb.mutex);
	}

	return 0;

} /* static int cff_gui_dialog_progress_softcut_cb(cff_softcut_pbar_t*) */

static gboolean cff_gui_dialog_progress_set_fraction(cff_progressbar_callback_t *cb)
{
	if (cb != NULL)
	{
		if (gtk_main_level() > 1)
		{
			GtkProgressBar *pbar;

			if (cb->mutex != NULL)
			{
				GuiMutexLock(cb->mutex);
			}

			pbar = (GtkProgressBar *) cb->widget;

			if (pbar != NULL)
			{
				if (cb->u.softcut.show_text)
				{
					cb->u.softcut.show_text = FALSE;
					GuiProgressbarTextHide(pbar);
				}

				gtk_progress_bar_set_fraction(pbar, cb->u.progress);
			}

			cb->widget = NULL;

			if (cb->mutex != NULL)
			{
				GuiMutexUnlock(cb->mutex);
			}
		}

		else
		{
			cb->widget = NULL;

		} /* if (gtk_main_level() > 1) .. else */

	} /* if (cb != NULL) */

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_set_fraction(cff_progressbar_callback_t*) */

static gboolean cff_gui_dialog_progress_callback(cff_progressbar_t *pbar)
{
	double progress;

	if (pbar->cb.mutex != NULL)
	{
		GuiMutexInit(pbar->cb.mutex);
	}

	while ((pbar->retv = (*pbar->func)(pbar->video,
					&progress, &(pbar->error))) == 0)
	{
		if (pbar->error != NULL || pbar->progress_bar == NULL)
		{
			break;
		}

		if (pbar->cb.mutex != NULL)
		{
			GuiMutexLock(pbar->cb.mutex);
		}

		if (pbar->cb.widget == NULL)
		{
			pbar->cb.widget     = pbar->progress_bar;
			pbar->cb.u.progress = (gdouble) progress;

			g_main_context_invoke_full(NULL, G_PRIORITY_HIGH_IDLE,
				(GSourceFunc) cff_gui_dialog_progress_set_fraction, &(pbar->cb), NULL);
		}

		if (pbar->cb.mutex == NULL)
		{
			return Gui_SOURCE_CONTINUE;
		}

		GuiMutexUnlock(pbar->cb.mutex);
	}

	/* Wait */

	while (pbar->cb.widget != NULL)
	{
		g_main_context_wakeup(NULL);
		g_main_context_iteration(NULL, FALSE);
	}

	if (pbar->cb.mutex != NULL)
	{
		GuiMutexFree(pbar->cb.mutex);
	}

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_callback(cff_progressbar_t*) */

static gboolean cff_gui_dialog_progress_bar(cff_progressbar_t *data)
{
	GtkWidget *widget, *window, *frame, *button, *hbox, *vbox;
	GuiMutex  mutex;

	if (data->mutex)
	{
		/* 'cff_gui_open_callback' */
		GuiMutexLock(data->mutex);
		data->is_locked = TRUE;
	}

	vbox = GuiVBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);

	window = cff_guitl_window_new_gettext(data->title, vbox, 8);

	cff_guitl_dialog_set_modal(window, gWindata.topwin, data->mbar);
	gtk_window_set_resizable(GTK_WINDOW(window), FALSE);

	hbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	gtk_container_add(GTK_CONTAINER(vbox), frame);

	widget = gtk_progress_bar_new();
	data->progress_bar = GTK_PROGRESS_BAR(widget);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 4);

	button = gtk_button_new_with_label(gettext_nop("Halt"));
	GuiHBoxPackStart(hbox, button, FALSE, FALSE, 1);

	GuiWidgetCanDefault(button);

	GuiSignalConnect(window, "delete-event",
			cff_gui_dialog_progress_halt, data);
	GuiSignalConnect(button, "button-press-event",
			cff_gui_dialog_progress_halt, data);

	gtk_widget_show_all(window);

	/* Callback */
	data->cb.widget  = NULL;
	data->cb.mutex   = &mutex;

	/* Halt-button */
	data->retv = 0;

	if (cff_gui_thread_new((GThreadFunc)
		cff_gui_dialog_progress_callback, data, NULL) == FALSE)
	{
		data->cb.mutex = NULL;

		g_idle_add_full(Gui_PRIORITY_REDRAW,
			(GSourceFunc) cff_gui_dialog_progress_callback,
					data, (GDestroyNotify) gtk_main_quit);
		gtk_main();
	}

	gtk_widget_destroy(window);

	if (data->retv == 0 && data->error == NULL)
	{
		data->error = gettext_nop("User canceled");
	}

	if (data->mutex)
	{
		GuiMutexUnlock(data->mutex);
	}

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_bar(cff_progressbar_t*) */


/*
	----------------
	 FILE-SELECTION
	----------------
*/

static gchar *cff_gui_convert_filename(gchar *fname, gboolean delete)
{
	gchar *str = fname;

	if (str != NULL)
	{
		str = cff_charconv_utf8_to_filename(fname);

		if (str != fname && delete == TRUE)
		{
			g_free(fname);
		}
	}

	return str;

} /* static gchar *cff_gui_convert_filename(gchar*, gboolean) */

static gchar *cff_gui_get_gprop_path(gint prop)
{
	gchar *path, *sep;

	path = cff_gprop_get_str(prop);

	if (path == NULL)
	{
		return NULL;
	}

#ifdef BUILD_MINGW
	cff_file_path_unix_to_dos(path);
#endif
	path = cff_gui_convert_filename(path, TRUE);

	if (path == NULL)
	{
		return NULL;
	}

	sep = &path[strlen(path)];

	do
	{
		GStatBuf dirstat;

		sep[0] = '\0';

		if ( g_stat(path, &dirstat) == 0 &&
				S_ISDIR(dirstat.st_mode) )
		{
			return path;
		}
	}
	while ((sep = strrchr(path, G_DIR_SEPARATOR)) != NULL);

	g_free(path);

	return NULL;

} /* static gchar *cff_gui_get_gprop_path(gint) */

static gchar *cff_gui_get_gprop_fname(const gchar *fname, gint prop)
{
	gchar *path, *str;

	path = cff_gui_get_gprop_path(prop);

	if (path == NULL)
	{
		str = NULL;

		if (fname != NULL)
		{
			str = g_strdup(fname);
#ifdef BUILD_MINGW
			cff_file_path_unix_to_dos(str);
#endif
			str = cff_gui_convert_filename(str, TRUE);
		}
	}

	else
	{
		gchar *fn;

		fname	= cff_tl_get_filename(fname);
		fn	= cff_gui_convert_filename((gchar *) fname, FALSE);

		str = g_strconcat(path, G_DIR_SEPARATOR_S, fn, NULL);
		CFF_CHARCONV_UTF8_FREE(fn, fname);

		g_free(path);

	} /* if (path == NULL) .. elese */

	return str;

} /* static gchar *cff_gui_get_gprop_fname(const gchar*, gint) */

static void cff_gui_set_gprop_path(const gchar *fname, gint prop)
{
	gchar *path, *sep;

	if (fname == NULL || fname[0] == '\0')
	{
		return;
	}

	path = g_strdup(fname);

#ifdef BUILD_MINGW
	cff_file_path_dos_to_unix(path);
#endif

	if ((sep = strrchr(path, '/')) != NULL)
	{
		GStatBuf dirstat;

		*sep = '\0';

		if ( g_stat(path, &dirstat) == 0 &&
				S_ISDIR(dirstat.st_mode) )
		{
			gchar *str;

			str = cff_charconv_utf8_from_filename(path);

			if (str != NULL)
			{
				cff_gprop_set_str(prop, str);
				CFF_CHARCONV_UTF8_FREE(str, path);
			}
		}

	} /* if ((sep = strrchr(path, '/')) != NULL) */

	g_free(path);

} /* static void cff_gui_set_gprop_path(const gchar*, gint) */

static gchar *cff_gui_get_dirname(const gchar *fname)
{
	gchar		*path;
	const gchar	*fn;
	GStatBuf	dirstat;

	fn = cff_tl_get_filename(fname);

	if (fn == fname)
	{
		return NULL;
	}

	path = g_strndup(fname, fn - fname);

#ifdef BUILD_MINGW
	cff_file_path_unix_to_dos(path);
#endif
	path = cff_gui_convert_filename(path, TRUE);

	if ( g_stat(path, &dirstat) != 0 ||
			! S_ISDIR(dirstat.st_mode) )
	{
		g_free(path);
		path = NULL;
	}

	return path;

} /* static gchar *cff_gui_get_dirname(const gchar*) */

static gchar *cff_gui_file_filter_validate_string(const gchar *text, gint length, guint skip)
{
	gboolean	changed;
	gint		i, n;
	gchar		*str, *s;
	const gchar	*txt;

	n = length;

	if (n < 0)
	{
		n = strlen(text);
	}

	changed	= FALSE;

	while (n > 0 && *(text) == ' ')
	{
		n--;
		text++;
		changed = TRUE;
	}

	if (n == 0)
	{
		return NULL;
	}

	txt = text;
	i   = n;

	while (--i >= 0)
	{
		/* Skip space if string contains multiple extensions */

		guint chr = *(txt++);

		if (chr != skip)
		{
			if (g_ascii_isalnum(chr) == FALSE)
			{
				/* Not alphanumeric */
				return NULL;
			}

			if (g_ascii_isupper(chr))
			{
				/* Convert to lower */
				changed = TRUE;
			}
		}

	} /* while (--i >= 0) */

	if (changed == FALSE)
	{
		return (gchar *) text;
	}

	str = g_malloc(n + 1);
	s   = str;

	while (--n >= 0)
	{
		*(s++) = g_ascii_tolower(*(text++));
	}

	*s = '\0';

	return str;

} /* static gchar *cff_gui_file_filter_validate_string(const gchar*, gint, guint) */

static gchar *cff_gui_file_filter_extension_next(gchar *str, gchar **nextptr)
{
	guint chr;
	gchar *end, *next;

	end = strchr(str, ' ');

	if (end == NULL)
	{
		end = strchr(str, '\0');
	}

	chr  = *end;
	*end = '\0';
	next = end;

	while (chr == ' ')
	{
		chr = *(++next);
	}

	*nextptr = next;

	return end;

} /* static gchar *cff_gui_file_filter_extension_next(gchar *beg, gchar**) */

static gboolean cff_gui_file_filter_extension_compare(const gchar *str, const gchar *pattern)
{
	if (str != NULL)
	{
		const gchar *match;

		while ((match = strstr(str, pattern)) != NULL)
		{
			/* (match == pattern) => found itself */

			if (match != pattern)
			{
				guint  chr;
				size_t len;

				len = strlen(pattern);
				chr = match[len];

				if (chr == ' ' || chr == '\0')
				{
					/* Last characters are equal */

					chr = ' ';

					if (match > str)
					{
						chr = *(--match);
						match++;
					}

					if (chr == ' ')
					{
						/* First characters are equal */
						return TRUE;
					}
				}

			} /* if (match != pattern) */

			str = &match[1];

		} /* while ((match = strstr(str, pattern)) != NULL) */

	} /* if (str != NULL) */

	return FALSE;

} /* static gboolean cff_gui_file_filter_extension_compare(const gchar*, const gchar*) */

static gchar *cff_gui_file_filter_extension_remove(gchar *beg, gchar *end, gchar *start)
{
	if (*end == '\0')
	{
		/* Last extension in the string */

		while (beg > start)
		{
			/* Not the first entry */

			if (*(--beg) != ' ')
			{
				beg++;
				break;
			}
		}

		/* Remove last extension */
		*beg = '\0';
	}

	else
	{
		/* Move following extension(s) down */
		beg = memmove(beg, end, strlen(end) + 1);
	}

	return beg;

} /* static gchar *cff_gui_file_filter_extension_remove(gchar*, gchar*, gchar*) */

static void cff_gui_file_filter_validate_prop(gint prop)
{
	gchar *ext2 = cff_gprop_get_str(prop);

	if (ext2 != NULL)
	{
		gboolean changed;
		gchar    *str;

		changed	= FALSE;
		str	= cff_gui_file_filter_validate_string(ext2, -1, ' ');

		if (str != ext2)
		{
			/* 'ext2' was changed or is invalid */
			g_free(ext2);
			ext2	= str;
			changed	= TRUE;
		}

		if (ext2 != NULL)
		{
			gint		n;
			const gchar	*ext;

			if (prop == kPropStrFileVideoFilter)
			{
				ext = cff_video_ext;
			}

			else /* kPropStrFileImageFilter */
			{
				ext = cff_image_ext;
			}

			n = 0;

			do
			{
				gchar *end, *next;

				end = cff_gui_file_filter_extension_next(str, &next);

				if ((next - end) > 1)
				{
					/* More than one space */

					if (*next != '\0')
					{
						/* No following extension(s) */
						next--;
					}

					/* Move following spaces down */
					cff_gui_file_filter_extension_remove(end, next, ext2);
					changed = TRUE;
				}

				else if ((end - str) >= kFileFilterCharsMax ||
					cff_gui_file_filter_extension_compare(ext, str) ||
						cff_gui_file_filter_extension_compare(ext2, str))
				{
					/* Move following extensions down */
					str = cff_gui_file_filter_extension_remove(str, next, ext2);
					changed = TRUE;
				}

				else
				{
					if (++n >= kFileFilterEntriesMax)
					{
						/* String was terminated (end) */
						changed = TRUE;
						break;
					}

					if (*next != '\0')
					{
						/* Not the last extension */
						*end = ' ';
					}

					str = next;
				}
			}
			while (*str != '\0');

		} /* if (ext2 != NULL) */

		if (changed)
		{
			cff_gprop_set_str(prop, ext2);
		}

		g_free(ext2);

	} /* if (ext2 != NULL) */

} /* static void cff_gui_file_filter_validate_prop(gint) */

static const gchar *cff_gui_file_filter_add(GtkFileChooser *chooser, const gchar *ext, GtkFileFilter **filter)
{
	static const gchar all_files[] = gettext_nop("All files");

	guint chr;
	gchar *str, *name;
	gchar filter_glob[256], filter_name[256];

	do
	{
		chr = *(ext++);
	}
	while (chr == ' ');

	if (chr == '\0')
	{
		ext--;
		return ext;
	}

	str = filter_glob;

	if (chr == '*')
	{
		*(str++) = chr;
		chr = *(ext++);

		name = cff_charconv_utf8_gettext(all_files);
		strcpy(filter_name, name);
		CFF_CHARCONV_UTF8_FREE(name, all_files);
	}

	else
	{
		name = filter_name;

		*(str++) = '*';
		*(str++) = '.';

		do
		{
			/* Extension in lower and uppercase */
			*(str++) = '[';
			*(str++) = chr;

			/* Name in uppercase */
			chr = g_ascii_toupper(chr);
			*(str++)  = chr;
			*(name++) = chr;

			*(str++) = ']';

			chr = *(ext++);

			if (chr == '\0')
			{
				/* Done */
				ext--;
				break;
			}
		}
		while (chr != ' ');

		*name = '\0';

	} /* if (chr == '*') .. else */

	*str = '\0';

	*filter = gtk_file_filter_new();
	gtk_file_filter_set_name(*filter, filter_name);
	gtk_file_filter_add_pattern(*filter, filter_glob);
	gtk_file_chooser_add_filter(chooser, *filter);

	return ext;

} /* static const gchar *cff_gui_file_filter_add(GtkFileChooser*, const gchar*, GtkFileFilter**) */

static gint cff_gui_file_filter_init(GtkFileChooser *chooser, gint prop, GtkFileFilter *filterbuf[512])
{
	static const gchar cutterff_ext[] = "cff";

	gint		n, select;
	guint		num;
	gchar		*ext2;
	const gchar	*ext;

	if (prop == kPropStrEditPath)
	{
		ext = cutterff_ext;
		select = -1;
	}

	else if (prop == kPropStrFileImagePath)
	{
		ext	= cff_image_ext;
		select	= 1;
		prop	= kPropStrFileImageFilter;
	}

	else
	{
		if (cff_gprop_get_int(kPropIntFileFilterIndex, &select) == FALSE)
		{
			select = -1;
		}

		ext  = cff_video_ext;
		prop = kPropStrFileVideoFilter;
	}

	n   = -1;

	do
	{
		ext = cff_gui_file_filter_add(chooser, ext, &filterbuf[0]);

		if (++n == select)
		{
			gtk_file_chooser_set_filter(chooser, filterbuf[0]);
		}
	}
	while (*ext != '\0');


	num = 0;

	if (prop != kPropStrEditPath &&
			(ext2 = cff_gprop_get_str(prop)) != NULL)
	{
		ext = ext2;

		do
		{
			ext = cff_gui_file_filter_add(chooser, ext, &filterbuf[num]);

			if (++n == select)
			{
				gtk_file_chooser_set_filter(chooser, filterbuf[num]);
			}
		}
		while (++num < kFileFilterEntriesMax && *ext != '\0');

		g_free(ext2);

	} /* if (prop != kPropStrEditPath &&
				(ext2 = cff_gprop_get_str(prop)) != NULL) */

	return num;

} /* static gint cff_gui_file_filter_init(GtkFileChooser*, gint, GtkFileFilter*[512]) */

static void cff_gui_file_selection_add_filter_entry_cb(GtkEditable *editable, const gchar *text, gint length, gint *pos, gint *sigid)
{
	gchar *str;

	GuiSignalHandlerBlock(editable, *sigid);

	str = cff_gui_file_filter_validate_string(text, length, '\0');

	if (str != NULL)
	{
		gtk_editable_insert_text(editable, str, length, pos);

		if (str != text)
		{
			g_free(str);
		}
	}

	GuiSignalEmitStopByName(editable, "insert-text");
	GuiSignalHandlerUnblock(editable, *sigid);

} /* static void cff_gui_file_selection_add_filter_entry_cb(GtkEditable *editable, const gchar*, gint, gint*, gint*) */

static void cff_gui_file_selection_add_filter_cb(GtkButton *button, cff_filefilter_t *filter)
{
	guint		sigid;
	gchar		*ext, *ext2;
	const gchar	*space;
	GtkWidget	*border, *content, *entry, *frame, *label, *vbox, *widget;
	GtkDialog	*dialog;

	if (filter->num >= kFileFilterEntriesMax)
	{
		return;
	}

	widget = cff_guitl_dialog_new_gettext_2btn(gettext_nop("Add filter"),
						gettext_nop("Cancel"),
						gettext_nop("OK"),
						GTK_WIDGET(filter->chooser));
	dialog = GTK_DIALOG(widget);
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);

	vbox = GuiVBoxNew(FALSE, 2);
	border = cff_guitl_frame_new_add_child_gettext(NULL, vbox);
	gtk_container_set_border_width(GTK_CONTAINER(border), 8);

	entry = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(entry), kFileFilterCharsMax);

#if GTK_MAJOR_VERSION == 2
	widget = GuiHBoxNew(FALSE, 2);
	GuiHBoxPackStart(widget, entry, TRUE,  TRUE, 8);
	GuiVBoxPackStart(vbox, widget, FALSE, FALSE, 4);
#else
	gtk_widget_set_hexpand(entry, TRUE);
	GuiMiscSetPadding(entry, 8);
	GuiVBoxPackStart(vbox, entry, FALSE, FALSE, 4);
#endif

	sigid = GuiSignalConnect(GTK_EDITABLE(entry), "insert-text",
			cff_gui_file_selection_add_filter_entry_cb, &sigid);

	ext2  = cff_gprop_get_str(filter->prop);
	space = NULL;

	if (ext2 != NULL)
	{
		space = " ";
	}

	/* Skip "* " */
	ext = g_strconcat(&filter->ext[2], space, ext2, NULL);

	label = gtk_label_new(ext);
	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
	GuiMiscSetPadding(label, 8);

	g_free(ext);

	frame = cff_guitl_frame_new_add_child_gettext(gettext_nop("Filters"), label);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 8);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 2);

	gtk_widget_show_all(border);

	content	= gtk_dialog_get_content_area(dialog);
	gtk_container_add(GTK_CONTAINER(content), border);

	while (gtk_dialog_run(dialog) == GTK_RESPONSE_ACCEPT)
	{
		const gchar *str = gtk_entry_get_text(GTK_ENTRY(entry));

		if (str != NULL && str[0] != '\0')
		{
			gint		slen;
			gchar		*s;
			GtkFileFilter	**f;

			if (cff_gui_file_filter_extension_compare(filter->ext, str) ||
					cff_gui_file_filter_extension_compare(ext2, str))
			{
				/* Filter already exists */
				cff_guitl_beep();
				continue;
			}

			f = &filter->buf[filter->num];

			cff_gui_file_filter_add(filter->chooser, str, f);
			gtk_file_chooser_set_filter(filter->chooser, *f);
			filter->num++;

			slen = (ext2 == NULL) ? 0 : strlen(ext2);
			/* Plus '\0' or ' ' */
			ext2 = g_realloc(ext2, strlen(str) + slen + 1);
			s    = &ext2[slen];

			if (slen > 0)
			{
				*(s++) = ' ';
			}

			strcpy(s, str);
			cff_gprop_set_str(filter->prop, ext2);

			gtk_widget_set_sensitive(filter->btn_remove, TRUE);

			if (filter->num >= kFileFilterEntriesMax)
			{
				gtk_widget_set_sensitive(filter->btn_add, FALSE);
			}

			break;

		} /* if (str != NULL && str[0] != '\0') */

	} /* while (gtk_dialog_run(dialog) == GTK_RESPONSE_ACCEPT) */

	g_free(ext2);
	gtk_widget_destroy(GTK_WIDGET(dialog));

} /* static void cff_gui_file_selection_add_filter_cb(GtkButton*, cff_filefilter_t*) */

static void cff_gui_file_selection_remove_filter_cb(GtkButton *button, cff_filefilter_t *filter)
{
	guint			num;
	gchar			*ext2, *str;
	GtkWidget		*content, *frame, *swin, *widget;
	GtkDialog		*dialog;
	GtkListStore		*store;
	GtkTreeView		*view;
	GtkTreeViewColumn	*column;
	GtkTreeSelection	*selection;
	GtkCellRenderer		*renderer;
	GtkTreeIter		iter;

	if (filter->num == 0)
	{
		return;
	}

	ext2 = cff_gprop_get_str(filter->prop);

	if (ext2 == NULL)
	{
		return;
	}

	store = gtk_list_store_new(4, G_TYPE_STRING, G_TYPE_POINTER,
						G_TYPE_POINTER, G_TYPE_INT);
	view = (GtkTreeView *)
		gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
	gtk_tree_view_set_headers_visible(view, FALSE);

	g_object_unref(G_OBJECT(store));

	selection = gtk_tree_view_get_selection(view);
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);

	column   = gtk_tree_view_column_new();
	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_set_attributes(column, renderer, "text", 0, NULL);

	gtk_tree_view_append_column(view, column);


	num = 0;
	str = ext2;

	do
	{
		gchar *end, *next;

		end  = cff_gui_file_filter_extension_next(str, &next);

		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter,
				0, str, 1, str, 2, next, 3, num, -1);

		if (end != next)
		{
			*end = ' ';
		}

		str = next;
	}
	while (*str != '\0' && ++num < filter->num);


	frame = gtk_frame_new(NULL);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 8);

	swin = cff_guitl_scrolled_window_new(frame, -1, 150, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(swin), GTK_WIDGET(view));

	widget = cff_guitl_dialog_new_gettext_2btn(gettext_nop("Remove filter"),
						gettext_nop("Cancel"),
						gettext_nop("OK"),
						GTK_WIDGET(filter->chooser));
	dialog = GTK_DIALOG(widget);
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);

	gtk_widget_show_all(frame);

	content	= gtk_dialog_get_content_area(dialog);
	gtk_container_add(GTK_CONTAINER(content), frame);

	while (gtk_dialog_run(dialog) == GTK_RESPONSE_ACCEPT)
	{
		guint		n;
		gchar		*s, *e;
		GtkTreeModel	*model;

		if (gtk_tree_selection_get_selected(selection,
						&model, &iter) == FALSE)
		{
			/* Nothing selected */
			cff_guitl_beep();
			continue;
		}

		gtk_tree_model_get(model, &iter, 1, &s, 2, &e, 3, &n, -1);

		cff_gui_file_filter_extension_remove(s, e, ext2);
		gtk_file_chooser_remove_filter(filter->chooser, filter->buf[n]);

		num = filter->num - (n + 1);

		if (num != 0)
		{
			/* Move following entries one position down */
			memmove(&filter->buf[n], &filter->buf[n + 1],
						num * sizeof(GtkFileFilter*));
		}

		filter->num--;

		if (filter->num == 0)
		{
			gtk_widget_set_sensitive(filter->btn_remove, FALSE);
		}

		if (filter->num < kFileFilterEntriesMax)
		{
			gtk_widget_set_sensitive(filter->btn_add, TRUE);
		}

		cff_gprop_set_str(filter->prop, ext2);

		break;

	} /* while (gtk_dialog_run(dialog) == GTK_RESPONSE_ACCEPT) */

	g_free(ext2);
	gtk_widget_destroy(GTK_WIDGET(dialog));

} /* static void cff_gui_file_selection_remove_filter_cb(GtkButton*, cff_filefilter_t*) */

static gchar *cff_gui_file_selection_new_gettext(const gchar *title, const gchar *fname, GtkWidget **mbar, gint prop)
{
	gboolean		path_prop;
	gchar			*str, *btn_ok, *btn_cancel, *accept, *cancel;
	GtkWidget		*widget;
	GtkFileChooserAction	action;
#ifdef HAVE_GTKOVERWRITECONFIRMATION
	int			overwrite;
#endif
#if GTK_MAJOR_VERSION > 2
	gint			width, height;
	GtkWindow		*window;
#endif
	cff_filefilter_t	filter_s;

	if (mbar)
	{
		GuiWidgetSetState(*mbar, Gui_STATE_PRELIGHT);
	}

	if (fname == NULL)
	{
		if (prop == kPropStrFileOpenPath)
		{
			btn_ok = gettext_nop("Open");
		}

		else
		{
			btn_ok = gettext_nop("Load");
		}

		action = GTK_FILE_CHOOSER_ACTION_OPEN;
	}

	else
	{
		if (prop == kPropStrFileSavePath ||
				prop == kPropStrFileImagePath)
		{
			btn_ok = gettext_nop("OK");
		}

		else
		{
			btn_ok = gettext_nop("Save");
		}

		action = GTK_FILE_CHOOSER_ACTION_SAVE;

	} /* if (fname == NULL) .. else */

	btn_cancel = gettext_nop("Cancel");

	accept	= cff_charconv_utf8_gettext(btn_ok);
	cancel	= cff_charconv_utf8_gettext(btn_cancel);
	str	= cff_charconv_utf8_gettext(title);

	widget = gtk_file_chooser_dialog_new(str,
			GTK_WINDOW(gWindata.topwin),
			action,
			cancel, GTK_RESPONSE_CANCEL,
			accept, GTK_RESPONSE_ACCEPT,
			NULL);

	CFF_CHARCONV_UTF8_FREE(str, title);
	CFF_CHARCONV_UTF8_FREE(cancel, btn_cancel);
	CFF_CHARCONV_UTF8_FREE(accept, btn_ok);

	filter_s.chooser = GTK_FILE_CHOOSER(widget);

	if (prop != kPropStrEditPath)
	{
		/* Extra widget */

		GtkWidget *frame, *hbox;
#if GTK_MAJOR_VERSION == 2
		GtkWidget *bbox;
#endif

		if (prop == kPropStrFileImagePath)
		{
			filter_s.ext  = cff_image_ext;
			filter_s.prop = kPropStrFileImageFilter;
		}

		else
		{
			filter_s.ext  = cff_video_ext;
			filter_s.prop = kPropStrFileVideoFilter;
		}

		hbox = GuiHButtonBoxNew();
		gtk_button_box_set_layout(GTK_BUTTON_BOX(hbox), GTK_BUTTONBOX_START);
		frame = cff_guitl_frame_new_add_child_gettext(gettext_nop("Filter"), hbox);
		gtk_container_set_border_width(GTK_CONTAINER(frame), 4);

#if GTK_MAJOR_VERSION == 2
		bbox = GuiHBoxNew(FALSE, 1);
		GuiHBoxPackStart(bbox, frame, FALSE, FALSE, 1);
		frame = bbox;
#endif
		filter_s.btn_add = cff_guitl_button_box_add_button(hbox,
				gettext_nop("Add ..."),
				(GCallback) cff_gui_file_selection_add_filter_cb,
				&filter_s);
		filter_s.btn_remove = cff_guitl_button_box_add_button(hbox,
				gettext_nop("Remove ..."),
				(GCallback) cff_gui_file_selection_remove_filter_cb,
				&filter_s);

		gtk_widget_show_all(frame);
		gtk_file_chooser_set_extra_widget(filter_s.chooser, frame);

	} /* if (prop != kPropStrEditPath) */

	if ((prop == kPropStrFileSavePath ||
		prop == kPropStrFileImagePath) &&
			(fname == NULL || fname[0] == '\0'))
	{
		fname = "noname";
	}

	if (fname != NULL)
	{
		const gchar *fn;

		fn  = cff_tl_get_filename(fname);
		str = cff_gui_convert_filename((gchar *) fn, FALSE);

		if (str != NULL)
		{
			gtk_file_chooser_set_current_name(filter_s.chooser, str);
			CFF_CHARCONV_UTF8_FREE(str, fn);
		}

	} /* if (fname != NULL) */

	if (prop == kPropStrFileSavePath || prop == kPropStrFileImagePath)
	{
		/* Get path from filename (may changed in the entry) */
		str = cff_gui_get_dirname(fname);
		path_prop = FALSE;
	}

	else
	{
		path_prop = TRUE;
	}

	if (path_prop || str == NULL)
	{
		/* Get path stored in gprop */
		str = cff_gui_get_gprop_path(prop);

		if (str == NULL && path_prop)
		{
			/* Get path from filename */
			str = cff_gui_get_dirname(fname);
		}
	}

	if (str != NULL)
	{
		gtk_file_chooser_set_current_folder(filter_s.chooser, str);
		g_free(str);
	}

#ifdef HAVE_GTKOVERWRITECONFIRMATION
	if (action == GTK_FILE_CHOOSER_ACTION_SAVE &&
			(cff_gprop_get_bool(kPropBoolOptOverwrite,
				&overwrite) == FALSE || overwrite == FALSE))
	{
		gtk_file_chooser_set_do_overwrite_confirmation(filter_s.chooser, TRUE);
	}
#endif

	filter_s.num = cff_gui_file_filter_init(filter_s.chooser,
						prop, filter_s.buf);
	if (prop != kPropStrEditPath)
	{
		if (filter_s.num == 0)
		{
			gtk_widget_set_sensitive(filter_s.btn_remove, FALSE);
		}

		else if (filter_s.num >= kFileFilterEntriesMax)
		{
			gtk_widget_set_sensitive(filter_s.btn_add, FALSE);
		}

	} /* if (prop != kPropStrEditPath) */

#if GTK_MAJOR_VERSION > 2

	/*
		Running a 32-bit version on a 64-bit platform
		might display a huge filechooser, with the
		titlebar and butttons out of the screen.
	*/

	gtk_widget_show_all(widget);

	window = GTK_WINDOW(widget);
	gtk_window_get_size(window, &width, &height);

	if (cff_guitl_window_test_size(window, width, height) == FALSE)
	{
		gtk_window_unmaximize(window);
		gtk_window_move(window, 16, 16);
	}

#endif /* #if GTK_MAJOR_VERSION > 2 */

	str = NULL;

	if (gtk_dialog_run(GTK_DIALOG(widget)) == GTK_RESPONSE_ACCEPT)
	{
		if (prop == kPropStrFileOpenPath || prop == kPropStrFileSavePath)
		{
			gint		index;
			GtkFileFilter	*filter;
			GSList		*list;

			filter = gtk_file_chooser_get_filter(filter_s.chooser);
			list   = gtk_file_chooser_list_filters(filter_s.chooser);
			index  = g_slist_index(list, filter);
			g_slist_free(list);

			cff_gprop_set_int(kPropIntFileFilterIndex, index);

		} /* if (prop == kPropStrFileOpenPath || prop == kPropStrFileSavePath) */

		str = gtk_file_chooser_get_filename(filter_s.chooser);

		/* The save dialogs store the path itself */

		if (prop == kPropStrFileOpenPath || prop == kPropStrEditPath)
		{
			cff_gui_set_gprop_path(str, prop);
		}

	} /* if (gtk_dialog_run(GTK_DIALOG(widget)) == GTK_RESPONSE_ACCEPT) */

	gtk_widget_destroy(widget);

	if (mbar)
	{
		GuiWidgetSetState(*mbar, Gui_STATE_NORMAL);
	}

	return str;

} /* static gchar *cff_gui_file_selection_new_gettext(const gchar*, const gchar*, GtkWidget**, gint) */


/*
	-------------
	 SAVE COMMON
	-------------
*/

static void cff_gui_save_okay_button_set_sensitive(cff_save_common_t *vscommon, guint bits)
{
	if (bits != vscommon->okay.bits)
	{
		gboolean active;

		vscommon->okay.bits = bits;

		active = TRUE;

		if (vscommon->okay.bits != 0)
		{
			/* Invalid entry, or no stream in the output list */
			active = FALSE;
		}

		gtk_widget_set_sensitive(vscommon->okay.widget, active);

	} /* if (bits != vscommon->okay.bits) */

} /* static void cff_gui_save_okay_button_set_sensitive(cff_save_common_t*, guint) */

static void cff_gui_save_entry_changed(cff_save_common_t *vscommon)
{
	/* Disable the okay-button if the entry is empty or invalid */

	const gchar	*str;
	guint		bits;

	str  = gtk_entry_get_text(vscommon->entry);
	bits = vscommon->okay.bits;

	if (str == NULL)
	{
		/* Empty */
		bits |= 0x01;
	}

	else
	{
		guint    chr, last_chr;
		gboolean first, ext;

		first    = TRUE;
		ext      = FALSE;
		last_chr = '\0';

		while ((chr = *(str++)) != '\0')
		{
			if (isspace(chr))
			{
				if (first)
				{
					break;
				}

				if (last_chr == '.')
				{
					ext = TRUE;
				}
			}

			last_chr = chr;
			first    = CFF_IS_DIR_SEPARATOR(chr);

			if (first)
			{
				ext = FALSE;
			}

		} /* while ((chr = *(str++)) != '\0') */

		if (chr != '\0' || first == TRUE ||
					ext == TRUE || isspace(last_chr))
		{
			/* Empty, or contains leading spaces */
			bits |= 0x01;
		}

		else
		{
			/* Entry is okay */
			bits &= ~0x01;
		}

	} /* if (str == NULL .. else */

	cff_gui_save_okay_button_set_sensitive(vscommon, bits);

} /* static void cff_gui_save_entry_changed(cff_save_common_t*) */

static void cff_gui_save_entry_delete_callback(GtkEntryBuffer *buffer,
		guint position, guint n_chars, cff_save_common_t *vscommon)
{
	cff_gui_save_entry_changed(vscommon);

} /* static void cff_gui_save_entry_delete_callback(GtkEntryBuffer*,
					guint, guint, cff_save_common_t*) */

static void cff_gui_save_entry_insert_callback(GtkEntryBuffer *buffer,
		guint position, gchar *chars, guint n_chars, cff_save_common_t *vscommon)
{
	cff_gui_save_entry_changed(vscommon);

} /* static void cff_gui_save_entry_insert_callback(GtkEntryBuffer *buffer,
					guint, gchar*, guint, cff_save_common_t*) */

static void cff_gui_save_selection_callback(GtkWidget *w, cff_save_common_t *vscommon)
{
	gchar *ename, *fname, *str;

	ename	= (gchar *) gtk_entry_get_text(vscommon->entry);
	str	= ename;

#ifndef BUILD_MINGW

	if (ename && ename[0] != '\0')
	{
		str = cff_charconv_utf8_to_filename(ename);

		if (str == NULL)
		{
			str = ename;
		}
	}
#endif

	fname = cff_gui_file_selection_new_gettext(gettext_nop("Select ..."),
						str, NULL, vscommon->path_prop);

#ifndef BUILD_MINGW
	CFF_CHARCONV_UTF8_FREE(str, ename);
#endif

	if (fname != NULL)
	{
		g_free(vscommon->filechooser_utf8str);

#ifdef BUILD_MINGW
		/* Already utf8 */
		vscommon->filechooser_utf8str = fname; /* Do not free 'fname' */
		gtk_entry_set_text(vscommon->entry, fname);
#else
		str = cff_charconv_utf8_from_filename(fname);
		vscommon->filechooser_utf8str = str; /* Do not free 'str' */
		gtk_entry_set_text(vscommon->entry, str);

		if (fname != str)
		{
			g_free(fname);
		}
#endif

	} /* if (fname != NULL) */

} /* static void cff_gui_save_selection_callback(GtkWidget*, cff_save_common_t*) */

static void cff_gui_save_optionmenu_callback(GuiOptionMenu *option, cff_save_common_t *vscommon)
{
	gint		i, pos, slen;
	gchar		*extension;
	const gchar	*str, *end;
	GtkEditable	*editable;

	i = GuiOptionMenuGetActive(option);

	extension = NULL;

	if (i > 0)
	{
		const gchar *format_name = NULL;

		if (vscommon->path_prop == kPropStrFileSavePath)
		{
			if (vscommon->format.u.list != NULL)
			{
				/* Array without "Auto" */
				format_name =
					vscommon->format.u.list[i - 1].name;
			}
		}

		else /* kPropStrFileImagePath */
		{
			if (vscommon->format.u.names != NULL)
			{
				/* Array without "Auto" */
				format_name = vscommon->format.u.names[i - 1];
			}
		}

		if (format_name != NULL)
		{
			extension = cff_ffmpeg_get_outformat_extension(format_name, &slen);
		}

	} /* if (i > 0) */

	if (extension == NULL)
	{
		if ((extension = vscommon->format.extension_orig) == NULL)
		{
			return;
		}

		slen = -1;
	}

	str = gtk_entry_get_text(vscommon->entry);
	end = strrchr(str, '.');

	editable = GTK_EDITABLE(vscommon->entry);

	if (end != NULL)
	{
		if (strcmp(end, extension) == 0)
		{
			if (extension != vscommon->format.extension_orig)
			{
				/* malloc'ed */
				free(extension);
			}

			return;
		}

		pos = end - str;
		gtk_editable_delete_text(editable, pos, -1);
	}

	/* Append new extension */
	pos = -1;
	gtk_editable_insert_text(editable, extension, slen, &pos);
	gtk_editable_set_position(editable, -1);

	g_free(vscommon->filechooser_utf8str);
	vscommon->filechooser_utf8str = NULL;

	if (extension != vscommon->format.extension_orig)
	{
		/* malloc'ed */
		free(extension);
	}

} /* static void cff_gui_save_optionmenu_callback(GuiOptionMenu*, cff_save_common_t*) */

static void cff_gui_save_softcut_option_changed_cb(GuiOptionMenu *option, GtkWidget *both_gops)
{
	/*If softcut is "No", disable 'both gops' */

	gboolean sensitive = FALSE;

	if (GuiOptionMenuGetActive(option) > 0)
	{
		sensitive = TRUE;
	}

	gtk_widget_set_sensitive(both_gops, sensitive);

} /* static void cff_gui_save_softcut_option_changed_cb(GuiOptionMenu*, GtkWidget*) */

static gboolean cff_gui_save_common_cancel(GtkWidget *window, GdkEvent *e, cff_save_common_t *vscommon)
{
	GdkEventKey *kevent = (GdkEventKey *) e;

	if (kevent->keyval == Gui_KEY_Escape &&
		(kevent->state & (GDK_CONTROL_MASK | GDK_SHIFT_MASK)) == 0)
	{
		(*vscommon->cancel.func)(NULL, NULL, vscommon->cancel.data);
		gtk_widget_destroy(window);
		return TRUE;
	}

	return FALSE;

} /* static gboolean cff_gui_save_common_cancel(GtkWidget*, GdkEvent*, cff_save_common_t*) */

static GtkWidget *cff_gui_save_build_common(cff_save_common_t *vscommon, const gchar *title, gint prop)
{
	gchar		*fname, *str, *end;
	const gchar	*dot, *ext;
	GtkWidget	*entry, *hbox, *vbox, *frame;
	GtkEntryBuffer	*buffer;
	gchar		num[128];

	vscommon->path_prop = prop;

	vbox = GuiVBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);

	vscommon->window = cff_guitl_window_new_gettext(title, vbox, 8);
	cff_guitl_dialog_set_modal(vscommon->window, gWindata.mainwin, vscommon->mbar);

	frame = gtk_frame_new(NULL);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	gtk_container_add(GTK_CONTAINER(vbox), frame);

	vbox = GuiVBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
	gtk_container_add(GTK_CONTAINER(frame), vbox);

	hbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 4);

	entry = gtk_entry_new();
	vscommon->entry = GTK_ENTRY(entry);
	GuiHBoxPackStart(hbox, entry, TRUE, TRUE, 1);

#if GTK_MAJOR_VERSION > 2
	gtk_widget_set_hexpand(entry, TRUE);
#endif

	vscommon->filechooser_utf8str = NULL;
	cff_guitl_hbox_add_button_gettext(hbox, "..",
		(GCallback) cff_gui_save_selection_callback, vscommon);

	fname = cff_gui_get_gprop_fname((vscommon->video)->fname,
							vscommon->path_prop);
	end = strrchr(fname, '.');
	dot = end;
	ext = NULL;

	if (end != NULL)
	{
		*end = '\0';
		dot  = ".";
	}

	if (vscommon->path_prop == kPropStrFileSavePath)
	{
		if (end != NULL)
		{
			ext  = &end[1];
		}

		num[0] = '\0';
	}

	else
	{
		/* kPropStrFileImagePath */
		dot  = ".";
		ext  = "jpg";
		snprintf(num, sizeof(num), "_%u",
				(vscommon->video)->vframe.frame_current);
	}

	str = g_strconcat(fname, num, "_cff", dot, ext, NULL);
	g_free(fname);

	vscommon->format.extension_orig = NULL;


#ifdef BUILD_MINGW

	/* Already utf8 */
	gtk_entry_set_text(vscommon->entry, str);

	if ((ext = strrchr(fname, '.')) != NULL)
	{
		vscommon->format.extension_orig = g_strdup(ext);
	}
#else
	fname = cff_charconv_utf8_from_filename(str);

	if (fname != NULL)
	{
		gtk_entry_set_text(vscommon->entry, fname);

		if ((ext = strrchr(fname, '.')) != NULL)
		{
			vscommon->format.extension_orig = g_strdup(ext);
		}

		CFF_CHARCONV_UTF8_FREE(fname, str);
	}

#endif /* #ifdef BUILD_MINGW .. #else */


	g_free(str);

	buffer = gtk_entry_get_buffer(vscommon->entry);

	GuiSignalConnect(buffer, "deleted-text",
				cff_gui_save_entry_delete_callback, vscommon);
	GuiSignalConnect(buffer, "inserted-text",
				cff_gui_save_entry_insert_callback, vscommon);

	GuiSignalConnect(vscommon->window, "delete-event",
				vscommon->cancel.func, vscommon->cancel.data);
	GuiSignalConnect(vscommon->window, "key-press-event",
				cff_gui_save_common_cancel, vscommon);
	return vbox;

} /* static GtkWidget *cff_gui_save_build_common(cff_save_common_t *vscommon, const gchar*, gint) */

static gchar *cff_gui_save_get_fname(cff_save_common_t *vscommon, gboolean *fname_free)
{
	gboolean	ask_overwrite;
	gchar		*fname;
	const gchar	*utf8;

	if ((vscommon->video)->fname == NULL)
	{
		return NULL;
	}

	utf8 = gtk_entry_get_text(vscommon->entry);

	if (utf8 == NULL || utf8[0] == '\0')
	{
		cff_guitl_beep();
		return NULL;
	}

#ifdef BUILD_MINGW
	fname = (gchar *) utf8;
#else
	fname = cff_charconv_utf8_to_filename(utf8);

	if (fname == NULL)
	{
		fname = (gchar *) utf8;
	}
#endif

	if (strcmp(fname, (vscommon->video)->fname) == 0)
	{
		static const gchar text[] =
			gettext_nop("The filename or the folder must differ!");

#ifndef BUILD_MINGW
		CFF_CHARCONV_UTF8_FREE(fname, utf8);
#endif
		cff_guitl_beep();
		cff_guitl_dialog_error(text, NULL);
		return NULL;
	}

	ask_overwrite = FALSE;

#ifdef HAVE_GTKOVERWRITECONFIRMATION
	if (vscommon->filechooser_utf8str == NULL ||
			strcmp(utf8, vscommon->filechooser_utf8str) != 0)
#endif
	{
		ask_overwrite = TRUE;
	}

	if (ask_overwrite && cff_gui_overwrite(fname, NULL) == FALSE)
	{

#ifndef BUILD_MINGW
		CFF_CHARCONV_UTF8_FREE(fname, utf8);
#endif
		return NULL;
	}

	*fname_free = (fname == utf8) ? FALSE : TRUE;

	cff_gui_set_gprop_path(utf8, vscommon->path_prop);

	return fname;

} /* static gchar *cff_gui_save_get_fname(cff_save_common_t*, gboolean*) */


/*
	------------
	 SAVE VIDEO
	------------
*/

static void cff_gui_save_video_prev_settings_free(void)
{
	cff_save_video_prev_options_t *save_prev = gSaveVideoPrevOptionsPtr;

	if (save_prev != NULL)
	{
		g_free(save_prev->stream_idx);
		g_free(save_prev);

		gSaveVideoPrevOptionsPtr = NULL;
	}

} /* static void cff_gui_save_video_prev_settings_free(void) */

static int cff_gui_save_copy_callback(gpointer data)
{
	cff_progressbar_t user_data;

	struct
	{
		const gchar		*str;
		cff_progressbar_t	*pbar;

	} *s = data;

	if (s != NULL)
	{
		s->pbar = &user_data;
	}

	user_data.func  = cff_ffmpeg_copy_cb;
	user_data.video = &gVideo;
	user_data.title = gettext_nop("Copy Video File");
	user_data.mbar  = NULL;
	user_data.mutex = NULL;

	cff_gui_dialog_progress_bar(&user_data);

	if (s != NULL)
	{
		s->str = user_data.error;
	}

	return user_data.retv;

} /* static int cff_gui_save_copy_callback(gpointer) */

static void cff_gui_save_start(GtkWidget *w, cff_save_video_t *vsave)
{
	gboolean	retv, fname_free;
	gint		i, n, rows, nb_streams, *stream_idx;
	gint		encoder_idx, format_idx, bsfname_idx[kCFFMediaTotal];
	gchar		*fname;
	gchar		*info;
	const gchar	*result;
	GtkWidget	**mbar;
	GtkTreeModel	*model;
	GtkTreeIter	iter;
	cff_copy_param_t param;

	struct
	{
		const gchar		*error;
		cff_progressbar_t	*pbar;

	} s;


	model = GTK_TREE_MODEL(vsave->out.list.store);
	rows  = gtk_tree_model_iter_n_children(model, NULL);

	if (rows <= 0)
	{
		cff_guitl_beep();
		return;
	}

	fname = cff_gui_save_get_fname(&(vsave->common), &fname_free);

	if (fname == NULL)
	{
		return;
	}

	g_free(vsave->common.format.extension_orig);
	g_free(vsave->common.filechooser_utf8str);


	/* Save video ... */

	stream_idx = g_malloc(rows * sizeof(gint));

	retv = gtk_tree_model_get_iter_first(model, &iter);

	nb_streams = 0;

	while (retv == TRUE)
	{
		gtk_tree_model_get(model, &iter,
				1, &stream_idx[nb_streams], -1);
		nb_streams++;
		retv = gtk_tree_model_iter_next(model, &iter);
	}

	memset(&param, '\0', sizeof(cff_copy_param_t));

	format_idx = 0;

	if (vsave->common.format.u.list != NULL)
	{
		format_idx = GuiOptionMenuGetActive(vsave->common.format.option);

		if (format_idx > 0)
		{
			/* Array without "Auto" */
			param.format_name =
				vsave->common.format.u.list[format_idx - 1].name;
		}

		/* malloc'ed */
		free(vsave->common.format.u.list);
	}

	n = kCFFMediaVideo;

	do
	{
		cff_mediaopts_t *mediaopts = &(vsave->media[n]);

		bsfname_idx[n] = 0;

		if (mediaopts->bsf_list != NULL)
		{
			i = GuiOptionMenuGetActive(mediaopts->bsf_option);
			bsfname_idx[n] = i;

			if (i > 0)
			{
				/* Array without "None" */
				param.media[n].bsf_name = (mediaopts->bsf_list[i - 1]).name;
			}

			/* malloc'ed */
			free(mediaopts->bsf_list);
		}

		if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mediaopts->ts_from_duration)) == TRUE)
		{
			param.media[n].ts_from_duration = 1;
		}
	}
	while (++n < kCFFMediaTotal);

	info		= NULL;
	param.fname	= fname;
	param.info	= &info;
	param.func	= cff_gui_save_copy_callback;
	param.user_data	= &s;

	param.stream_idx = stream_idx;
	param.nb_streams = nb_streams;

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vsave->shortest)) == TRUE)
	{
		param.shortest = 1;
	}

	if (gtk_widget_get_sensitive(vsave->stream_sync) == TRUE &&
		gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vsave->stream_sync)) == TRUE)
	{
		param.stream_sync = 1;
	}

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vsave->ts_keep_start)) == TRUE)
	{
		param.ts_keep_start = 1;
	}

	encoder_idx = 0;

	if (vsave->softcut.encoder_names != NULL)
	{
		if (gtk_widget_get_sensitive(vsave->softcut.option) == TRUE &&
			(encoder_idx = GuiOptionMenuGetActive(vsave->softcut.option)) > 0)
		{
			/* Array without "No" */
			param.encoder_name = vsave->softcut.encoder_names[encoder_idx - 1];
			/* Callback */
			param.softcut.func = cff_gui_dialog_progress_softcut_cb;
			param.softcut.data = &(s.pbar);

			if (gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(vsave->softcut.both_gops)) == TRUE)
			{
				param.softcut.both_gops = 1;
			}
		}

		/* malloc'ed */
		free(vsave->softcut.encoder_names);

	} /* if (vsave->softcut.encoder_names != NULL) */


	gIsBusy |= kIsBusyFile;
	result   = cff_ffmpeg_copy_open(vsave->common.video, &param);
	gIsBusy &= ~kIsBusyFile;


	if (result == NULL)
	{
		result = s.error;
	}

	if (result == NULL)
	{	
		cff_save_video_prev_options_t *save_prev = gSaveVideoPrevOptionsPtr;

		if (save_prev != NULL)
		{
			g_free(save_prev->stream_idx);
			save_prev->stream_idx = param.stream_idx;
			save_prev->nb_streams = param.nb_streams;

			save_prev->format_idx  = format_idx;
			save_prev->encoder_idx = encoder_idx;

			save_prev->shortest      = param.shortest;
			save_prev->stream_sync   = param.stream_sync;
			save_prev->ts_keep_start = param.ts_keep_start;
			save_prev->both_gops     = param.softcut.both_gops;

			n = kCFFMediaVideo;

			do
			{
				save_prev->media[n].bsfname_idx =
						bsfname_idx[n];
				save_prev->media[n].ts_from_duration =
						param.media[n].ts_from_duration;
			}
			while (++n < kCFFMediaTotal);

		} /* if (save_prev != NULL) */
	}

	else
	{
		gint delete;

		if (cff_gprop_get_bool(kPropBoolOptDelete, &delete) == FALSE || delete)
		{
			/* Error: remove the output file */
			g_unlink(fname);
		}

		g_free(stream_idx);

	} /* if (result == NULL) .. else */

#ifndef BUILD_MINGW
	if (fname_free == TRUE)
	{
		g_free(fname);
	}
#endif
	gtk_widget_destroy(vsave->common.window);
	mbar = vsave->common.mbar;
	g_free(vsave);

	if (result != NULL)
	{
		cff_guitl_dialog_error(result, mbar);
	}

	else
	{
		cff_guitl_dialog_info((info == NULL) ?
			gettext_nop("No info available!") : info, mbar);

		if (info != NULL)
		{
			/* malloc'ed */
			free(info);
		}

	} /* if (result != NULL) .. else */

} /* static void cff_gui_save_start(GtkWidget*, cff_save_video_t*) */

static gboolean cff_gui_save_delevnt(GtkWidget *w, GdkEvent *e, gpointer data)
{
	gint			n;
	cff_save_video_t	*vsave;

	vsave = data;

	if (vsave->common.format.u.list != NULL)
	{
		/* malloc'ed */
		free(vsave->common.format.u.list);
	}

	n = kCFFMediaVideo;

	do
	{
		cff_mediaopts_t *mediaopts = &(vsave->media[n]);

		if (mediaopts->bsf_list != NULL)
		{
			/* malloc'ed */
			free(mediaopts->bsf_list);
		}
	}
	while (++n < kCFFMediaTotal);

	if (vsave->softcut.encoder_names != NULL)
	{
		/* malloc'ed */
		free(vsave->softcut.encoder_names);
	}

	g_free(vsave->common.format.extension_orig);
	g_free(vsave->common.filechooser_utf8str);
	g_free(vsave);

	return FALSE;

} /* static gboolean cff_gui_save_delevnt(GtkWidget*, GdkEvent*, gpointer) */

static void cff_gui_save_destroy(GtkWidget *button, cff_save_video_t *vsave)
{
	GtkWidget *window = vsave->common.window;

	cff_gui_save_delevnt(NULL, NULL, vsave);
	gtk_widget_destroy(window);

} /* static void cff_gui_save_destroy(GtkWidget*, cff_save_video_t*) */

static void cff_gui_save_list_changed(GtkTreeSelection *selection, cff_save_video_t *vsave)
{
	gint			rows, num;
	guint			bits;
	gboolean		button, up, down, first_row, last_row, stream_sync, softcut;
	GtkTreeModel		*model;
	cff_streamlist_t	*stream;

	if (selection == vsave->avail.list.selection)
	{
		stream = &(vsave->avail);
	}

	else
	{
		stream = &(vsave->out);
	}

	num = gtk_tree_selection_count_selected_rows(selection);

	button = FALSE;

	if (num > 0)
	{
		button = TRUE;
	}

	gtk_widget_set_sensitive(stream->some, button);

	rows = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(stream->list.store), NULL);

	button = FALSE;

	if (rows > 0)
	{
		button = TRUE;
	}

	gtk_widget_set_sensitive(stream->all, button);

	if (selection == vsave->avail.list.selection)
	{
		return;
	}


	bits = vsave->common.okay.bits;

	if (button == FALSE)
	{
		bits |= 0x02;
	}

	else
	{
		bits &= ~0x02;
	}

	cff_gui_save_okay_button_set_sensitive(&(vsave->common), bits);

	first_row = FALSE;
	last_row  = FALSE;

	if (num > 0 && rows > 1)
	{
		GList *list, *ptr;

		up   = TRUE;
		down = TRUE;

		list = gtk_tree_selection_get_selected_rows(selection, &model);
		ptr  = list;

		do
		{
			/*
				This is an array of integers, each representing
				a node in a tree. This value should not be freed.
			*/

			gint *i = gtk_tree_path_get_indices(ptr->data);

			if (*i == 0)
			{
				first_row = TRUE;
			}

			else if (*i == rows - 1)
			{
				last_row = TRUE;
			}

			gtk_tree_path_free(ptr->data);
		}
		while ((ptr = ptr->next) != NULL);

		g_list_free(list);
	}

	else
	{
		up   = FALSE;
		down = FALSE;

	} /* if (num > 0 && rows > 1) .. else */

	gtk_widget_set_sensitive(vsave->up,
			(first_row == FALSE) ? up : FALSE);
	gtk_widget_set_sensitive(vsave->down,
			(last_row == FALSE) ? down : FALSE);


	stream_sync = FALSE;
	softcut     = FALSE;

	if (rows > 0)
	{
		if (cff_ffmpeg_video_cut_list_distance(&gVideo) != 0)
		{
			stream_sync = TRUE;
		}

		if (vsave->softcut.encoder_names != NULL)
		{
			softcut = TRUE;
		}

		if (stream_sync || softcut)
		{
			gint        retv, video, other;
			GtkTreeIter iter;

			video = 0;
			other = 0;

			model = GTK_TREE_MODEL(vsave->out.list.store);
			retv  = gtk_tree_model_get_iter_first(model, &iter);

			while (retv == TRUE)
			{
				gint media, stream_index;

				gtk_tree_model_get(model, &iter,
					1, &stream_index, 2, &media, -1);

				if (media == kCFFMediaVideo &&
					stream_index == gVideo.vframe.stream_index)
				{
					video++;
				}

				else
				{
					other++;
				}

				retv = gtk_tree_model_iter_next(model, &iter);

			} /* while (retv == TRUE) */

			if (video == 0)
			{
				softcut     = FALSE;
				stream_sync = FALSE;
			}

			else if (other == 0)
			{
				stream_sync = FALSE;
			}

		} /* if (stream_sync || softcut) */

	} /* if (rows > 0) */

	gtk_widget_set_sensitive(vsave->stream_sync, stream_sync);
	gtk_widget_set_sensitive(vsave->softcut.option, softcut);
	
	if (softcut && GuiOptionMenuGetActive(vsave->softcut.option) == 0)
	{
		softcut = FALSE;
	}

	gtk_widget_set_sensitive(vsave->softcut.both_gops, softcut);

} /* static void cff_gui_save_list_changed(GtkTreeSelection*, cff_save_video_t*) */

static GList *cff_gui_save_list_get_all(GtkTreeModel *model)
{
	gboolean	retv;
	GList		*list;
	GtkTreeIter	iter;

	list = NULL;
	retv = gtk_tree_model_get_iter_first(model, &iter);

	while (retv == TRUE)
	{
		list = g_list_append(list, g_memdup(&iter, sizeof(GtkTreeIter)));
		retv = gtk_tree_model_iter_next(model, &iter);
	}

	return list;

} /* static GList *cff_gui_save_list_get_all(GtkTreeModel*) */

static GList *cff_gui_save_list_get_selected(GtkTreeSelection *selection)
{
	GtkTreeModel	*model;
	GList		*list, *ptr;

	list = gtk_tree_selection_get_selected_rows(selection, &model);

	if ((ptr = list) == NULL)
	{
		return list;
	}

	do
	{
		GtkTreeIter *iter;

		iter = g_malloc(sizeof(GtkTreeIter));
		gtk_tree_model_get_iter(model, iter, ptr->data);
		gtk_tree_path_free(ptr->data);

		ptr->data = iter;
	}
	while ((ptr = ptr->next) != NULL);

	return list;

} /* static GList *cff_gui_save_list_get_selected(GtkTreeSelection*) */

static void cff_gui_save_list_updown(GtkWidget *w, cff_save_video_t *vsave)
{
	gint		add, i, rows, stop, *order;
	GtkTreeModel	*model;
	GtkTreePath	*path;
	GList		*list, *selected;

	add = 1;

	if (w == vsave->up)
	{
		add = -1;
	}

	list = gtk_tree_selection_get_selected_rows(vsave->out.list.selection, &model);

	if (list == NULL)
	{
		return;
	}

	rows = gtk_tree_model_iter_n_children(model, NULL);

	if (rows <= 0)
	{
		return;
	}

	order = malloc(rows * sizeof(gint));

	i = 0;

	do
	{
		order[i] = i;
	}
	while (++i < rows);

	if (add == -1)
	{
		/* Move selected rows up */

		selected = list;
		i = 0;
		stop = rows;
	}

	else
	{
		/* Move selected rows down */

		selected = g_list_last(list);
		i = rows - 1;
		stop = -1;
	}

	path = gtk_tree_path_new_from_indices(i, -1);

	do
	{
		if (gtk_tree_path_compare(path, selected->data) == 0)
		{
			/* Exchange selected row with next (previous) row */

			gint prev = order[i + add];

			order[i + add] = i;
			order[i] = prev;

			gtk_tree_path_free(selected->data);

			if (add == -1)
			{
				selected = selected->next;
			}

			else
			{
				selected = selected->prev;
			}

			if (selected == NULL)
			{
				break;
			}

		} /* if (gtk_tree_path_compare(path, selected->data) == 0) */

		if (add == -1)
		{
			gtk_tree_path_next(path);
		}

		else
		{
			gtk_tree_path_prev(path);
		}
	}
	while ((i -= add) != stop);

	gtk_tree_path_free(path);
	g_list_free(list);

	gtk_list_store_reorder(vsave->out.list.store, order);
	g_free(order);

	/* Set buttons (up, down) sensitive or insensitive */
	cff_gui_save_list_changed(vsave->out.list.selection, vsave);

} /* static void cff_gui_save_list_updown(GtkWidget*, cff_save_video_t*) */

static void cff_gui_save_list_leftright(GtkWidget *w, cff_save_video_t *vsave)
{
	GList		*list, *ptr;
	cff_ListStore_t	*src, *dst;

	if (w == vsave->avail.some || w == vsave->avail.all)
	{
		src = &(vsave->avail.list);
		dst = &(vsave->out.list);
	}

	else
	{
		src = &(vsave->out.list);
		dst = &(vsave->avail.list);
	}

	if (w == vsave->avail.all || w == vsave->out.all)
	{
		list = cff_gui_save_list_get_all(GTK_TREE_MODEL(src->store));
	}

	else
	{
		list = cff_gui_save_list_get_selected(src->selection);
	}

	if ((ptr = list) == NULL)
	{
		return;
	}

	do
	{
		gint		media, stream_idx;
		gchar		*str;
		GtkTreeIter	*iter;

		iter = ptr->data;

		gtk_tree_model_get(GTK_TREE_MODEL(src->store), iter,
				0, &str, 1, &stream_idx, 2, &media, -1);

		gtk_list_store_remove(src->store, iter);

		gtk_list_store_append(dst->store, iter);
		gtk_list_store_set(dst->store, iter,
				0, str, 1, stream_idx, 2, media, -1);

		g_free(str);
		g_free(iter);
	}
	while ((ptr = ptr->next) != NULL);

	g_list_free(list);

	/*
		Set buttons (up, down, right, right-all,
		left, left-all and okay) sensitive or insensitive
	*/

	cff_gui_save_list_changed(src->selection, vsave);
	cff_gui_save_list_changed(dst->selection, vsave);

} /* static void cff_gui_save_list_leftright(GtkWidget*, cff_save_video_t*) */

static void cff_gui_save_list_selection_init(cff_ListStore_t *list, cff_save_video_t *vsave)
{
	GtkTreeIter iter;

	if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(list->store), &iter))
	{
		gtk_tree_selection_select_iter(list->selection, &iter);
	}

	GuiSignalConnect(list->selection, "changed",
					cff_gui_save_list_changed, vsave);

	/* Set buttons (up, down, left, left-all and okay) insensitive */
	cff_gui_save_list_changed(list->selection, vsave);

} /* static void cff_gui_save_list_selection_init(cff_ListStore_t*, cff_save_video_t*) */

static gboolean cff_gui_video_list_focus(GtkWidget *list, GdkEvent *e, gpointer data)
{
	if (gtk_widget_is_focus(list) == FALSE)
	{
		gtk_widget_grab_focus(list);
	}

	return FALSE;

} /* static gboolean cff_gui_video_list_focus(GtkWidget*, GdkEvent*, gpointer) */

static void cff_gui_build_list(const gchar *title, cff_ListStore_t *list, GtkWidget *hbox)
{
	gchar			*str;
	GtkWidget		*frame, *swin;
	GtkTreeViewColumn	*column;
	GtkCellRenderer		*renderer;

	/* Visible stream descripton, stream-index, media-type (kCFFMedia...) */
	list->store = gtk_list_store_new(3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

	list->view = (GtkTreeView *)
		gtk_tree_view_new_with_model(GTK_TREE_MODEL(list->store));

	GuiSignalConnect(GTK_WIDGET(list->view), "button-press-event",
						cff_gui_video_list_focus, NULL);

	gtk_widget_set_can_focus(GTK_WIDGET(list->view), TRUE);

	list->selection = gtk_tree_view_get_selection(list->view);
	gtk_tree_selection_set_mode(list->selection, GTK_SELECTION_MULTIPLE);

	column = gtk_tree_view_column_new();

	str = cff_charconv_utf8_gettext(title);
	gtk_tree_view_column_set_title(column, str);
	CFF_CHARCONV_UTF8_FREE(str, title);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_set_attributes(column, renderer, "text", 0, NULL);

	gtk_tree_view_append_column(list->view, column);
	g_object_set(list->view,
			"headers-visible", TRUE,
			"headers-clickable", FALSE,
			NULL);

	g_object_unref(G_OBJECT(list->store));

	frame = gtk_frame_new(NULL);
	GuiHBoxPackStart(hbox, frame, TRUE, TRUE, 4);
	
	swin = cff_guitl_scrolled_window_new(frame, -1, 150, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(swin), GTK_WIDGET(list->view));

} /* static void cff_gui_build_list(const gchar*, cff_ListStore_t*, GtkWidget*) */

static GtkWidget *cff_gui_build_option_menu(GtkWidget *parent, GtkWidget *vbox,
		const gchar *title, const gchar *entry, const gchar **name_list)
{
	gchar		*str;
	const gchar	*name;
	GtkWidget	*option, *frame, *hbox;

#if GTK_MAJOR_VERSION == 2
	gint		i;
	GtkWidget	*menu;
#endif

	option = GuiOptionMenuTextNew();
	gtk_container_set_border_width(GTK_CONTAINER(option), 4);

	if (vbox != NULL)
	{
		GuiVBoxPackStart(vbox, option, TRUE, FALSE, 2);
		frame = cff_guitl_frame_new_add_child_gettext(title, vbox);
		GuiHBoxPackStart(parent, frame, TRUE, FALSE, 2);
	}

	else
	{
		hbox = GuiHBoxNew(FALSE, 2);
		GuiVBoxPackStart(parent, hbox, TRUE, FALSE, 2);
		frame = cff_guitl_frame_new_add_child_gettext(title, option);
		GuiHBoxPackStart(hbox, frame, TRUE, FALSE, 2);
	}


#if GTK_MAJOR_VERSION == 2
	menu = gtk_menu_new();
	gtk_widget_realize(menu);
	gtk_option_menu_set_menu(GTK_OPTION_MENU(option), menu);
	i = 0;
#endif

	if (name_list == NULL || name_list[0] == NULL)
	{
		name_list = NULL;

		if (entry == NULL)
		{
			entry = gettext_nop("Empty");
		}
	}

	if (entry != NULL)
	{
		str = cff_charconv_utf8_gettext(entry);
		GuiOptionMenuTextAppend(option, menu, 0, str);
		CFF_CHARCONV_UTF8_FREE(str, entry);
	}

	if (name_list == NULL)
	{
		gtk_widget_set_sensitive(option, FALSE);
	}

	else
	{
		while ((name = *(name_list++)) != NULL)
		{
#if GTK_MAJOR_VERSION == 2
			i++;
#endif
			GuiOptionMenuTextAppend(option, menu, i, name);
		}

	} /* if (name_list == NULL) .. else */

	GuiOptionMenuSetActive(option, 0);

	return option;

} /* static GtkWidget *cff_gui_build_option_menu(GtkWidget*, GtkWidget*,
					const gchar*, const gchar*, const gchar**) */

static GtkWidget *cff_gui_notebook_append_page(GtkWidget *notebook, const gchar *title)
{
	GtkWidget *label, *vbox;

	vbox = GuiVBoxNew(FALSE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 2);

	label = cff_guitl_label_new_gettext(title);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, label);

	return vbox;

} /* static GtkWidget *cff_gui_notebook_append_page(GtkWidget*, const gchar*) */

static void cff_gui_build_media_option(GtkWidget *notebook, cff_save_video_t *vsave, gint media)
{
	gchar		*str;
	gchar		**ptr;
	const gchar	*title;
	GtkWidget	*vbox;
	cff_mediaopts_t *mediaopts;
	cff_bsflist_t	*bsf_list, *bsf_info;
	gchar		*buf[512];

	if (media == kCFFMediaVideo)
	{
		title = gettext_nop("Video");
	}

	else if (media == kCFFMediaAudio)
	{
		title = gettext_nop("Audio");
	}

	else
	{
		title = gettext_nop("Other");
	}

	vbox = cff_gui_notebook_append_page(notebook, title);

	mediaopts = &(vsave->media[media]);

	/* Check button */

	mediaopts->ts_from_duration =
		cff_guitl_vbox_add_check_button_gettext(vbox,
			gettext_nop("Create timestamps from packet duration"));

	/* Option menu */

	bsf_list = cff_ffmpeg_get_bitstream_filter_names(vsave->common.video, media);
	mediaopts->bsf_list = bsf_list;

	ptr = buf;

	if ((bsf_info = bsf_list) != NULL)
	{
		while (bsf_info->name != NULL)
		{
			if (ptr - buf < (sizeof(buf) / sizeof(gchar*)) - 1)
			{
				*(ptr++) = g_strconcat(bsf_info->name, " ",
							bsf_info->codecs, NULL);
			}

			if (bsf_info->codecs != NULL)
			{
				/* malloc'ed */
				free(bsf_info->codecs);
				bsf_info->codecs = NULL;
			}

			bsf_info++;
		}

	} /* if ((bsf_info = bsf_list) != NULL) */

	*ptr = NULL;

	mediaopts->bsf_option = cff_gui_build_option_menu(vbox, NULL,
					gettext_nop("Bitstream Filter"),
					gettext_nop("None"),
					(const gchar **) buf);
	/* Free strings in array */

	ptr = buf;

	while ((str = *(ptr++)) != NULL)
	{
		g_free(str);
	}

} /* static void cff_gui_build_media_option(GtkWidget*, cff_save_video_t*, gint) */


/*
	------------
	 SAVE IMAGE
	------------
*/

static void cff_gui_image_keep_dimension_toggled(GtkToggleButton *button, cff_save_image_t *vimage)
{
	gint			n;
	gboolean		active;
	cff_spinbutton_t	*spinbutton;

	active = ! gtk_toggle_button_get_active(button);

	spinbutton = &(vimage->width);

	n = 2;

	do
	{
		gint value;

		value = gtk_spin_button_get_value_as_int(spinbutton->widget);
		gtk_spin_button_set_value(spinbutton->widget,
						(gdouble) spinbutton->prev_value);
		spinbutton->prev_value = value;

		gtk_widget_set_sensitive(GTK_WIDGET(spinbutton->widget), active);
		spinbutton = &(vimage->height);
	}
	while (--n > 0);

} /* static void cff_gui_image_keep_dimension_toggled(GtkToggleButton*, cff_save_image_t*) */

static GtkSpinButton *cff_gui_image_spinbutton(GtkWidget *hbox, gint value, const gchar *title)
{
	GtkAdjustment	*adjustment;
	GtkWidget	*widget, *bbox, *frame;
	GtkSpinButton	*spinbutton;

	bbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(bbox), 4);

	frame = cff_guitl_frame_new_add_child_gettext(title, bbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 4);
	GuiHBoxPackStart(hbox, frame, TRUE, FALSE, 4);

	adjustment = (GtkAdjustment *)
			gtk_adjustment_new((gdouble) value,
						1.0,
						99999.0,
						1.0,
						10.0,
						0.0);

	widget = gtk_spin_button_new(adjustment, 1.0, 0);
	GuiHBoxPackStart(bbox, widget, FALSE, FALSE, 4);

	spinbutton = GTK_SPIN_BUTTON(widget);
	gtk_spin_button_set_numeric(spinbutton, TRUE);
	gtk_spin_button_set_update_policy(spinbutton, GTK_UPDATE_IF_VALID);

	return spinbutton;

} /* static GtkSpinButton *cff_gui_image_spinbutton(GtkWidget*, gint, const gchar*) */

static void cff_gui_save_image(GtkWidget *w, cff_save_image_t *vimage)
{
	gboolean	fname_free;
	gchar		*fname, *info;
	const gchar	*error;
	GtkWidget	**mbar;
	cff_image_param_t param;

	fname = cff_gui_save_get_fname(&(vimage->common), &fname_free);

	if (fname == NULL)
	{
		return;
	}

	g_free(vimage->common.format.extension_orig);
	g_free(vimage->common.filechooser_utf8str);


	/* Save image ... */

	memset(&param, '\0', sizeof(cff_image_param_t));

	info		= NULL;
	param.fname	= fname;
	param.info	= &info;

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vimage->keep_dimension)) == TRUE)
	{
		param.keep_dimension = 1;
	}

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vimage->keep_ratio)) == TRUE)
	{
		param.keep_ratio = 1;
	}

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(vimage->gray)) == TRUE)
	{
		param.gray = 1;
	}

	if (param.keep_dimension == FALSE)
	{
		param.width  = gtk_spin_button_get_value_as_int(vimage->width.widget);
		param.height = gtk_spin_button_get_value_as_int(vimage->height.widget);
	}

	if (vimage->common.format.u.names != NULL)
	{
		gint i = GuiOptionMenuGetActive(vimage->common.format.option);

		if (i > 0)
		{
			/* Array without "Auto" */
			param.format_name = vimage->common.format.u.names[i - 1];
		}

		/* malloc'ed */
		free(vimage->common.format.u.names);
	}


	gIsBusy |= kIsBusyFile;
	error    = cff_ffmpeg_save_image(vimage->common.video, &param);
	gIsBusy &= ~kIsBusyFile;

	if (error != NULL)
	{
		gint delete;

		if (cff_gprop_get_bool(kPropBoolOptDelete, &delete) == FALSE || delete)
		{
			/* Error: remove the image */
			g_unlink(fname);
		}
	}

#ifndef BUILD_MINGW
	if (fname_free == TRUE)
	{
		g_free(fname);
	}
#endif
	gtk_widget_destroy(vimage->common.window);
	mbar = vimage->common.mbar;
	g_free(vimage);

	if (error != NULL)
	{
		cff_guitl_dialog_error(error, mbar);
	}

	else
	{
		cff_guitl_dialog_info((info == NULL) ?
			gettext_nop("No info available!") : info, mbar);

		if (info != NULL)
		{
			/* malloc'ed */
			free(info);
		}

	} /* if (error != NULL) .. else */

} /* static void cff_gui_save_image(GtkWidget*, cff_save_image_t*) */

static gboolean cff_gui_image_delevnt(GtkWidget *w, GdkEvent *e, gpointer data)
{
	cff_save_image_t *vimage = data;

	if (vimage->common.format.u.names != NULL)
	{
		/* malloc'ed */
		free(vimage->common.format.u.names);
	}

	g_free(vimage->common.format.extension_orig);
	g_free(vimage->common.filechooser_utf8str);
	g_free(vimage);

	return FALSE;

} /* static gboolean cff_gui_image_delevnt(GtkWidget*, GdkEvent*, gpointer) */

static void cff_gui_image_destroy(GtkWidget *w, cff_save_image_t *vimage)
{
	GtkWidget *window = vimage->common.window;

	cff_gui_image_delevnt(NULL, NULL, vimage);
	gtk_widget_destroy(window);

} /* static void cff_gui_image_destroy(GtkWidget*, cff_save_image_t*) */


/*
	--------------
	 INFO WINDOWS
	--------------
*/

static void cff_gui_scrolled_info_window(const char *title, GtkWidget *ibox, GtkWidget **mbar, int prop)
{
	GtkWidget *button, *hbox, *vbox, *frame, *swin, *window;

	vbox = GuiVBoxNew(FALSE, 4);
	window = cff_guitl_window_new_gettext(title, vbox, 8);
	cff_guitl_dialog_set_modal(window, gWindata.mainwin, mbar);

	if (prop >= 0)
	{
		cff_guitl_window_setup(window, prop);
	}

	frame = gtk_frame_new(NULL);
	GuiVBoxPackStart(vbox, frame, TRUE, TRUE, 4);

	swin = cff_guitl_scrolled_window_new(frame, 320, 180, GTK_POLICY_AUTOMATIC);
	gtk_container_set_border_width(GTK_CONTAINER(ibox), 8);

#if GTK_MAJOR_VERSION == 2
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(swin), ibox);
#else
	/* Above is deprecated since version 3.8 */
	gtk_container_add(GTK_CONTAINER(swin), ibox);
#endif

	hbox = cff_guitl_hbutton_box_new(vbox, 8, GTK_BUTTONBOX_END);
	button = cff_guitl_button_box_add_button(hbox,
				gettext_nop("OK"), NULL, NULL);
	GuiSignalConnectSwapped(button, "clicked", gtk_widget_destroy, window);
	gtk_widget_grab_default(button);

	gtk_widget_show_all(window);

} /* static void cff_gui_scrolled_info_window(const char*, GtkWidget*, GtkWidget**, int) */

static void cff_gui_ffmpeg_info_label(const char *title, char *str, GtkWidget *parent)
{
	GtkWidget *frame, *label;

	if (str == NULL || str[0] == '\0')
	{
		label = cff_guitl_label_new_gettext(gettext_nop("No info available!"));
	}

	else
	{
		gchar			*text;
		PangoFontDescription	*desc;
#if GTK_MAJOR_VERSION > 2
		PangoAttribute		*attr;
		PangoAttrList		*attrlist;
#endif
		text = cff_charconv_utf8_from_locale(str, NULL);

		label = gtk_label_new(text);
		CFF_CHARCONV_UTF8_FREE(text, str);

		desc = pango_font_description_from_string("monospace");

#if GTK_MAJOR_VERSION == 2
		gtk_widget_modify_font(label, desc);
#else
		attr = pango_attr_font_desc_new(desc);
		attrlist = pango_attr_list_new();
		pango_attr_list_insert(attrlist, attr);
		gtk_label_set_attributes(GTK_LABEL(label), attrlist);
		pango_attr_list_unref(attrlist);
#endif
		pango_font_description_free(desc);

	} /* if (str == NULL || str[0] == '\0') .. else */

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	GuiMiscSetAlignment(label, Gui_ALIGN_START, Gui_ALIGN_CENTER);
	GuiMiscSetPadding(label, 15);

	frame = cff_guitl_frame_new_add_child_gettext(title, label);
	GuiVBoxPackStart(parent, frame, FALSE, FALSE, 5);

} /* static void cff_gui_ffmpeg_info_label(const char*, char*, GtkWidget*) */


/*
	-----
	 CSS
	-----
*/

#if GTK_MAJOR_VERSION > 2

static gboolean cff_gui_css_load_from_data(GtkWidget *widget, const gchar *str, guint priority)
{
	GtkCssProvider	*provider;
	GtkStyleContext	*context;
	GError		*error;
 
 	/* This memory is owned by widget and must not be freed */
	context  = gtk_widget_get_style_context(widget);

	provider = gtk_css_provider_new();

	gtk_style_context_add_provider(context,
				GTK_STYLE_PROVIDER(provider),
				priority);
	error = NULL;
	gtk_css_provider_load_from_data(GTK_CSS_PROVIDER(provider),
				str,
				-1,
				&error);

	g_object_unref(provider);

	if (error != NULL)
	{
		g_error_free(error);
		return FALSE;
	}

	return TRUE;

} /* static gboolean cff_gui_css_load_from_data(GtkWidget*, const gchar*, guint) */

#endif /* GTK_MAJOR_VERSION > 2 */


/*
	------------
	 LOG-WINDOW
	------------
*/

static void cff_gui_logwindow_color_parse(gint prop, gpointer color)
{
	gchar *colorstr, *defcolor;

#if GTK_MAJOR_VERSION == 2 && defined(HAVE_SSCANF)
	gchar buf[16];
#endif

	if (prop == kPropStrLogWinText)
	{
		defcolor = (gchar *) gLogWinTextColor;
	}

	else
	{
		defcolor = (gchar *) gLogWinBackgroundColor;
	}

	colorstr = cff_gprop_get_str(prop);

	if (colorstr == NULL)
	{
		colorstr = defcolor;
	}

#if GTK_MAJOR_VERSION == 2 && defined(HAVE_SSCANF)

	if (colorstr[0] == 'r')
	{
		/* If 'cutterffrc' was saved with GTK-3 */

		gint n, r, g, b;

		n = sscanf(colorstr, "rgb(%i,%i,%i)", &r, &g, &b);

		if (n == 3)
		{
			n = g_snprintf(buf, sizeof(buf),
					"#%02X%02X%02X", r, g, b);
			if (n == 7)
			{
				/* colorstr != defcolor */
				g_free(colorstr);
				colorstr = buf;
			}
		}

	} /* if (colorstr[0] == 'r') */

#endif /* GTK_MAJOR_VERSION == 2 && defined(HAVE_SSCANF) */

	while (GuiColorParse(colorstr, color) == FALSE)
	{
		if (colorstr == defcolor)
		{
			return;
		}

		g_free(colorstr);
		colorstr = defcolor;
	}

	if (colorstr != defcolor)
	{

#if GTK_MAJOR_VERSION == 2 && defined(HAVE_SSCANF)
		if (colorstr != buf)
#endif
		g_free(colorstr);
	}

	return;

} /* static void cff_gui_logwindow_color_parse(gint, gpointer) */

static void cff_gui_logwindow_color_set(cff_logwindow_t *log)
{

#if GTK_MAJOR_VERSION == 2

	GdkColor color;

	cff_gui_logwindow_color_parse(kPropStrLogWinText, &color);
	gtk_widget_modify_text(log->textview, GTK_STATE_NORMAL, &color);

	cff_gui_logwindow_color_parse(kPropStrLogWinBackground, &color);
	gtk_widget_modify_base(log->textview, GTK_STATE_NORMAL, &color);

	gtk_widget_modify_bg(log->window, GTK_STATE_NORMAL, &color);
#else
	guint i;
	gchar *background, *text, buf_wind[256], buf_view[256];

	text = cff_gprop_get_str(kPropStrLogWinText);

	if (text == NULL)
	{
		text = (gchar *) gLogWinTextColor;
	}

	background = cff_gprop_get_str(kPropStrLogWinBackground);

	if (background == NULL)
	{
		background = (gchar *) gLogWinBackgroundColor;
	}

	i = 2;

	do
	{
		g_snprintf(buf_view, sizeof(buf_view),
				"GtkTextView text,"
				"textview text {"
				"  color: %s;"
				"  background-color: %s;"
				"  background-image: none;"
				"}",
				text,
				background);

		g_snprintf(buf_wind, sizeof(buf_wind),
				"GtkWindow,"
				"window {"
				"  background-color: %s;"
				"  background-image: none;"
				"}",
				background);

		if (text != gLogWinTextColor)
		{
			g_free(text);
			text = (gchar *) gLogWinTextColor;
		}

		if (background != gLogWinBackgroundColor)
		{
			g_free(background);
			background = (gchar *) gLogWinBackgroundColor;
		}

		if (cff_gui_css_load_from_data(log->textview, buf_view,
				GTK_STYLE_PROVIDER_PRIORITY_USER) == TRUE)
		{
			cff_gui_css_load_from_data(log->window, buf_wind,
					GTK_STYLE_PROVIDER_PRIORITY_USER);
			break;
		}
	}
	while (--i > 0);

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */

} /* static void cff_gui_logwindow_color_set(cff_logwindow_t*) */

static GtkWidget *cff_gui_logwindow_colorbutton_new(gint prop, const gchar *title, gpointer color)
{
	gchar		*str;
	GtkWidget	*button, *label, *hbox, *vbox;
	GuiSignalFunc	func;

	vbox = GuiVBoxNew(FALSE, 5);
	hbox = GuiHBoxNew(FALSE, 5);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 3);

	cff_gui_logwindow_color_parse(prop, color);
	button = GuiColorButtonNewWithColor(color);
	GuiHBoxPackStart(hbox, button, TRUE, FALSE, 5);

	str = cff_charconv_utf8_gettext(title);
	gtk_color_button_set_title(GTK_COLOR_BUTTON(button), str);

#if GTK_MAJOR_VERSION == 2
	func = (GuiSignalFunc) gtk_color_button_get_color;
#else
	func = (GuiSignalFunc) gtk_color_chooser_get_rgba;
#endif

	GuiSignalConnect(button, "color-set", func, color);

	label = gtk_label_new(str);
	GuiVBoxPackStart(vbox, label, FALSE, FALSE, 3);

	CFF_CHARCONV_UTF8_FREE(str, title);

	return vbox;

} /* static GtkWidget *cff_gui_logwindow_colorbutton_new(gint, const gchar*, gpointer) */

static void cff_gui_logwindow_color(GtkWidget *w, cff_logwindow_t *log)
{
	static const gchar btn_cancel[] = gettext_nop("Cancel");
	static const gchar btn_okay[]   = gettext_nop("OK");

	gint		retv;
	gchar		*accept, *cancel, *text, *background;
	GtkWidget	*content, *frame, *hbox, *widget;
	GtkDialog	*dialog;

#if GTK_MAJOR_VERSION == 2
	GdkColor	fg, bg;
#else
	GdkRGBA		fg, bg;
#endif

	cancel = cff_charconv_utf8_gettext(btn_cancel);
	accept = cff_charconv_utf8_gettext(btn_okay);

	widget = gtk_dialog_new_with_buttons("CutterFF: Color",
		GTK_WINDOW(gWindata.topwin),
		GTK_DIALOG_MODAL,
		cancel, GTK_RESPONSE_CANCEL,
		accept, GTK_RESPONSE_ACCEPT,
		NULL);

	CFF_CHARCONV_UTF8_FREE(accept, btn_okay);
	CFF_CHARCONV_UTF8_FREE(cancel, btn_cancel);

	gtk_window_set_resizable(GTK_WINDOW(widget), FALSE);

	dialog = GTK_DIALOG(widget);

	hbox = GuiHBoxNew(TRUE, 5);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 5);

	widget = cff_gui_logwindow_colorbutton_new(kPropStrLogWinText,
						gettext_nop("Text"), &fg);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	widget = cff_gui_logwindow_colorbutton_new(kPropStrLogWinBackground,
						gettext_nop("Background"), &bg);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	gtk_widget_show_all(frame);

	content	= gtk_dialog_get_content_area(dialog);
	gtk_container_add(GTK_CONTAINER(content), frame);

	retv = gtk_dialog_run(dialog);

	gtk_widget_destroy(GTK_WIDGET(dialog));

	if (retv != GTK_RESPONSE_ACCEPT)
	{
		return;
	}

	text = GuiColorToString(&fg);
	cff_gprop_set_str(kPropStrLogWinText, text);
	g_free(text);

	background = GuiColorToString(&bg);
	cff_gprop_set_str(kPropStrLogWinBackground, background);
	g_free(background);

	cff_gui_logwindow_color_set(log);

} /* static void cff_gui_logwindow_color(GtkWidget*, cff_logwindow_t*) */

static void cff_gui_logwindow_popup(GtkTextView *textview, GtkMenu *menu, cff_logwindow_t *log)
{
	/* Append to the GtkTextView popup */

	GtkWidget *mi;

	mi = GuiMenuSeparatorNew();
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
	gtk_widget_show(mi);

	mi = cff_guitl_menu_item_new_gettext(gettext_nop("Color ..."));
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
	gtk_widget_show(mi);

	GuiSignalConnect(mi, "activate", cff_gui_logwindow_color, log);

} /* static void cff_gui_logwindow_popup(GtkTextView*, GtkMenu*, cff_logwindow_t*) */


#if defined(BUILD_MINGW) && GTK_MAJOR_VERSION == 2

static gboolean cff_gui_logwindow_copy_clipboard_cb(gpointer data)
{
	struct
	{
		gchar	*str;
		gint	len;
		gint	lines;

	} *s = data;

	if (s == NULL)
	{
		return FALSE;
	}

	if (s->str == NULL)
	{
		g_free(s);
		return FALSE;
	}

	if (s->lines >= 0 && s->len > 0 && s->str[0] != '\0')
	{
		if (OpenClipboard(NULL))
		{
			HGLOBAL	clipbuffer;

			EmptyClipboard();
			clipbuffer = GlobalAlloc(GMEM_MOVEABLE, s->len + s->lines + 1);

			if (clipbuffer != NULL)
			{
				char *dst = (char *) GlobalLock(clipbuffer);

				if (dst != NULL)
				{
					/* UNIX to DOS : LF to CR-LF */

					unsigned int	chr;
					char		*src;

					src = s->str;

					do
					{
						chr = *(src++);

						if (chr == 0x0A)
						{
							*(dst++) = 0x0D;
						}

						*(dst++) = chr;
					}
					while (chr != '\0');

					GlobalUnlock(clipbuffer);
					SetClipboardData(CF_TEXT, clipbuffer);
				}

			} /* if (clipbuffer != NULL) */

			CloseClipboard();

		} /* if (OpenClipboard(NULL)) */

	} /* if (s->lines >= 0 && s->len > 0 && s->str[0] != '\0') */

	g_free(s->str);
	g_free(s);

	return FALSE;

} /* static gboolean cff_gui_logwindow_copy_clipboard_cb(gpointer) */

static void cff_gui_logwindow_copy_clipboard(GtkTextView *textview, gpointer data)
{
	/* Copying to the global clipwork do not work with GTK-2 and Windows. */

	struct
	{
		gchar	*str;
		gint	len;
		gint	lines;

	} s;

	GtkTextBuffer	*buffer;
	GtkTextIter	start, end;

	buffer = gtk_text_view_get_buffer(textview);

	if (gtk_text_buffer_get_selection_bounds(buffer, &start, &end) == FALSE)
	{
		return;
	}

	s.str = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

	if (s.str == NULL)
	{
		return;
	}

	s.len = strlen(s.str);

	if (s.len == 0)
	{
		return;
	}

	s.lines  = gtk_text_iter_get_line(&end);
	s.lines -= gtk_text_iter_get_line(&start);

	if (s.lines < 0)
	{
		return;
	}

	/* Wait a litle bit, else copying does not work */

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) cff_gui_logwindow_copy_clipboard_cb,
					g_memdup(&s, sizeof(s)), NULL);

} /* static void cff_gui_logwindow_copy_clipboard(GtkTextView*, gpointer) */

#endif /* #if defined(BUILD_MINGW) && GTK_MAJOR_VERSION == 2 */


static void cff_gui_logwindow_adjustment(GtkAdjustment *vadj, cff_logwindow_t *log)
{
	/* Move the sliders to bottom left */

	if (log != NULL && log->window != NULL)
	{
		gdouble upper, page_size;

		upper     = GuiAdjustmentGetUpper(vadj);
		page_size = GuiAdjustmentGetPageSize(vadj);

		if (upper != log->scrollbar.v_upper ||
				page_size != log->scrollbar.v_pagesize)
		{
			gdouble new_value;

			log->scrollbar.v_upper    = upper;
			log->scrollbar.v_pagesize = page_size;

			new_value = upper - page_size;

			if (new_value > 0.0 &&
				new_value != GuiAdjustmentGetValue(vadj))
			{
				gtk_adjustment_set_value(vadj, new_value);
			}

			if (GuiAdjustmentGetValue(log->scrollbar.h_widget) != 0.0)
			{
				gtk_adjustment_set_value(log->scrollbar.h_widget, 0.0);
			}

		} /* if (upper != log->scrollbar.v_upper ||
					page_size != log->scrollbar.v_pagesize) */

	} /* if (log != NULL && log->window != NULL) */

} /* static void cff_gui_logwindow_adjustment(GtkAdjustment*, cff_logwindow_t*) */

static gboolean cff_gui_logwindow_mainthread(cff_logwindow_t *log)
{
	gint		lines;
	guint		lock;
	GtkTextBuffer	*buffer;
	GtkTextIter	start, end;

	if (log == NULL || log->window == NULL)
	{
		return Gui_SOURCE_REMOVE;
	}

	if (log->utf8.len == 0)
	{
		return Gui_SOURCE_CONTINUE;
	}

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(log->textview));
	lines  = gtk_text_buffer_get_line_count(buffer);

	lock = log->nThreads;

	if (lock)
	{
		GuiMutexLock(&(log->mutex));
	}

	if (log->utf8.overwrite)
	{
		log->utf8.overwrite = FALSE;

		/* Overwrite last line */
		gtk_text_buffer_get_iter_at_line(buffer, &start, lines - 1);
		gtk_text_buffer_get_end_iter(buffer, &end);
		gtk_text_buffer_delete(buffer, &start, &end);
	}

	else if (lines > 1000)
	{
		/* Delete some lines */
		gtk_text_buffer_get_start_iter(buffer, &start);
		gtk_text_buffer_get_iter_at_line(buffer, &end, lines - 250);
		gtk_text_buffer_delete(buffer, &start, &end);
	}

	gtk_text_buffer_get_end_iter(buffer, &end);
	gtk_text_buffer_insert(buffer, &end, log->utf8.str, log->utf8.len);

	if (log->utf8.max > 4096)
	{
		log->utf8.max = 4096;
		log->utf8.str = g_realloc(log->utf8.str, log->utf8.max);
	}

	log->utf8.len = 0;

	if (lock)
	{
		GuiMutexUnlock(&(log->mutex));
	}

	return Gui_SOURCE_CONTINUE;

} /* static gboolean cff_gui_logwindow_mainthread(cff_logwindow_t*) */

static int cff_gui_logwindow_cb(const char *str, int len, void *data)
{
	gint		slen;
	guint		lock, max, firstchar;
	gchar		*utf8, *ptr;
	cff_logwindow_t	*log;

	if (str == NULL || len == 0)
	{
		return 0;
	}

	log = data;

	if (log == NULL || log->window == NULL || log->timeout_id == 0)
	{
		/* Restore previous level */
		return -1;
	}

	firstchar = str[0];

	if (firstchar == '\r')
	{
		str++;
		len--;
	}
		
	slen = len;
	utf8 = cff_charconv_utf8_from_locale(str, &slen);
	lock = log->nThreads;

	if (lock)
	{
		if (GuiMutexTryLock(&(log->mutex)) == FALSE)
		{
			if (log->window == NULL)
			{
				CFF_CHARCONV_UTF8_FREE(utf8, str);
				return -1;
			}

			GuiMutexLock(&(log->mutex));
		}
	}

	len = log->utf8.len + slen;
	max = log->utf8.max;

	/* 2 => '\n' + '\0' */

	while (len + 2 > max)
	{
		max += 4096;
	}

	if (max > log->utf8.max)
	{
		log->utf8.max = max;
		log->utf8.str = g_realloc(log->utf8.str, max);
	}

	ptr = &log->utf8.str[log->utf8.len];

	if (firstchar == '\r')
	{
		if (log->utf8.cr_len != 0)
		{
			if (log->utf8.len == 0)
			{
				/* Repeat - delete last line */
				log->utf8.overwrite = TRUE;
			}

			else if (log->utf8.len >= log->utf8.cr_len)
			{
				/* Overwrite last characters */
				ptr -= log->utf8.cr_len;
				len -= log->utf8.cr_len;
			}

			/* Repeat - avoid double newline */
			log->utf8.lastchar = '\r';

		} /* if (log->utf8.cr_len != 0) */

		log->utf8.cr_len = slen;

		if (utf8[slen - 1] == '\n')
		{
			log->utf8.cr_len--;
		}
	}

	else
	{
		log->utf8.cr_len = 0;

	} /* if (firstchar == '\r') .. else */


	/*
		If the newline character from the
		last line was removed, append it now.
	*/

	if (log->utf8.lastchar == '\n')
	{
		*(ptr++) = '\n';
		len++;
	}

	ptr = strcpy(ptr, utf8);
	CFF_CHARCONV_UTF8_FREE(utf8, str);

	/* If the last character is newline, remove it. */

	log->utf8.lastchar = ptr[slen - 1];

	if (log->utf8.lastchar == '\n')
	{
		len--;
	}

	log->utf8.len = len;

	if (lock)
	{
		GuiMutexUnlock(&(log->mutex));
	}

	return 0;

} /* static int cff_gui_logwindow_cb(const char*, int, void*) */

static gboolean cff_gui_logwindow_free(GtkWidget *widget, GdkEvent *e, gpointer data)
{
	const gchar	*logname;
	cff_logwindow_t *log;

	if (widget == GINT_TO_POINTER(-1))
	{
		log = &gLogWindow;
	}

	else
	{
		log = data;
	}

	/*
		If the log-window is currently busy, and
		'log->nThreads' is not zero, ffmpeg runs
		in another thread.
		If 'widget' is NULL, this function was called
		from 'cff_gui_opt_logwindow_menu_selected'
	*/

	if (log->nThreads == 0)
	{
		if (widget != NULL)
		{
			/* Destroy the window */
			log->window = NULL;
		}

		if (log->timeout_id != 0)
		{
			g_source_remove(log->timeout_id);
		}
	}

	else /* widget != NULL => delete-event */
	{

#ifndef HAVE_SEMAPHORE
		/* Keep the window open */
		return TRUE;
#else
		/* Close the window */
		log->window = NULL;

		if (log->timeout_id != 0)
		{
			if (log->busy_loop != 0)
			{
				/* Try again */
				return Gui_SOURCE_CONTINUE;
			}

			g_idle_add_full(G_PRIORITY_LOW,
				(GSourceFunc) cff_gui_logwindow_free,
						GINT_TO_POINTER(-1), NULL);
			log->busy_loop = 10;

			/* Destroy the window */
			return FALSE;

		} /* if (log->timeout_id != 0) */

		if (--log->busy_loop > 0 &&
				cff_ffmpeg_get_current_loglevel_index() > 0)
		{
			/* While previous level was not restored */
			return Gui_SOURCE_CONTINUE;
		}

#endif /* #ifndef HAVE_SEMAPHORE .. #else */

	} /* if (log->nThreads == 0) .. else */


	g_free(log->utf8.str);
	log->utf8.str  = NULL;
	log->busy_loop = 0;

	/* If set to 'quiet', prepend newline in the next output */
	log->utf8.lastchar = '\n';

	if (cff_ffmpeg_get_current_loglevel_index() > 0)
	{
		/* Restore previous level */
		cff_ffmpeg_set_loglevel(NULL, log->nThreads, NULL, NULL);
	}

	logname = cff_ffmpeg_get_loglevel_name(0);
	cff_gprop_set_str(kPropStrOptLoglevel, logname);

	return FALSE;

} /* static gboolean cff_gui_logwindow_free(GtkWidget*, GdkEvent*, gpointer) */

static void cff_gui_logwindow_set_title(GtkWidget *window, const gchar *logname)
{
	static const gchar logwin_title[] = "Cutterff: FFmpeg-Log (%s)";

	if (logname != NULL)
	{
		gchar buf[256];

		g_snprintf(buf, sizeof(buf), logwin_title, logname);
		gtk_window_set_title(GTK_WINDOW(window), buf);
	}

} /* static void cff_gui_logwindow_set_title(GtkWidget*, const gchar*) */

static void cff_gui_logwindow_source_destroyed(cff_logwindow_t *log)
{
	log->timeout_id = 0;

} /* static void cff_gui_logwindow_source_destroyed(cff_logwindow_t*) */

static void cff_gui_logwindow_add_timeout(cff_logwindow_t *log)
{
	if (log->timeout_id == 0)
	{
		log->utf8.str		= NULL;
		log->utf8.len		= 0;
		log->utf8.max		= 0;
		log->utf8.cr_len	= 0;
		log->utf8.overwrite	= FALSE;
		log->busy_loop		= 0;

		/* 5ms */
		log->timeout_id =
			g_timeout_add_full(CFF_TEXT_VIEW_PRIORITY_VALIDATE, 5,
				(GSourceFunc) cff_gui_logwindow_mainthread,
					log, (GDestroyNotify)
						cff_gui_logwindow_source_destroyed);
	} /* if (log->timeout_id == 0) */

} /* static void cff_gui_logwindow_add_timeout(cff_logwindow_t*) */

static gboolean cff_gui_logwindow_open(const gchar *logname)
{
	GtkWidget		*widget;
	GtkScrolledWindow	*swin;
	GtkAdjustment		*vadj;
	cff_logwindow_t		*log;

#if GTK_MAJOR_VERSION == 2
	PangoFontDescription	*desc;
#endif
	log = &gLogWindow;

	if (cff_ffmpeg_set_loglevel(logname, 0, cff_gui_logwindow_cb, log) < 0)
	{
		return FALSE;
	}

	if (log->window != NULL)
	{
		cff_gui_logwindow_add_timeout(log);
		cff_gui_logwindow_set_title(log->window, logname);
		return TRUE;
	}

	log->textview = gtk_text_view_new();

#if GTK_MAJOR_VERSION == 2
	desc = pango_font_description_from_string("monospace");
	gtk_widget_modify_font(log->textview, desc);
	pango_font_description_free(desc);
#else
	cff_gui_css_load_from_data(log->textview,
				"GtkTextView,"
				"textview {"
				"  font-family: monospace;"
				"}",
				GTK_STYLE_PROVIDER_PRIORITY_USER);
#endif

	g_object_set(G_OBJECT(log->textview), "cursor-visible", FALSE,
						"editable", FALSE,
						"wrap-mode", GTK_WRAP_NONE,
						NULL);

	widget = cff_guitl_scrolled_window_new(NULL,
					320, 180, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(widget), log->textview);
	swin = GTK_SCROLLED_WINDOW(widget);

#if GTK_MAJOR_VERSION > 2
	gtk_scrolled_window_set_shadow_type(swin, GTK_SHADOW_NONE);
#endif
	log->scrollbar.v_upper  = 0.0;
	log->scrollbar.h_widget = gtk_scrolled_window_get_hadjustment(swin);

	vadj = gtk_scrolled_window_get_vadjustment(swin);
	GuiSignalConnect(vadj, "changed", cff_gui_logwindow_adjustment, log);

	log->window = cff_guitl_window_new(GTK_WIDGET(swin), 8);
	cff_gui_logwindow_set_title(log->window, logname);
	cff_guitl_window_setup(log->window, kPropIntLogWinWidth);

	GuiSignalConnect(log->window, "delete-event", cff_gui_logwindow_free, log);

	cff_gui_signal_connect_key_or_mouse_button(log->window, kButtonPressEvent);
	cff_gui_signal_connect_key_or_mouse_button(log->window, kKeyPressEvent);

	GuiSignalConnect(log->textview, "populate-popup",
					cff_gui_logwindow_popup, log);

#if defined(BUILD_MINGW) && GTK_MAJOR_VERSION == 2
	GuiSignalConnect(log->textview, "copy-clipboard",
					cff_gui_logwindow_copy_clipboard, NULL);
#endif

	cff_gui_signal_connect_key_or_mouse_button(log->textview, kButtonPressEvent);
	cff_gui_signal_connect_key_or_mouse_button(log->textview, kKeyPressEvent);

	log->utf8.lastchar = '\0';
	log->timeout_id    = 0;

	cff_gui_logwindow_add_timeout(log);

	cff_gui_logwindow_color_set(log);
	gtk_widget_show_all(log->window);

	return TRUE;

} /* static gboolean cff_gui_logwindow_open(const gchar*) */


/*
	-----------
	 RECTANGLE
	-----------
*/

#ifdef HAVE_CAIRO

static void cff_gui_cairo_draw_rectangle(cairo_t *cr, gint color,
					gint fill, const GdkRectangle *rect)
{
	/*
		The signal handler will get a 'cr' with a clip region
		already set to the widget's dirty region, i.e. to the
		area that needs repainting.
	*/

#if GTK_MAJOR_VERSION == 2
	gdk_cairo_set_source_color(cr, &gColor[color]);
#else
	gdk_cairo_set_source_rgba(cr, &gRGBA[color]);
#endif
	gdk_cairo_rectangle(cr, rect);

	if (fill)
	{
		cairo_fill(cr);
	}

	else
	{
		cairo_set_line_width(cr, 1.0);
		cairo_stroke(cr);
	}

} /* static void cff_gui_cairo_draw_rectangle(cairo_t*, gint, gint,
						const GdkRectangle*) */
#if GTK_MAJOR_VERSION == 2

static void cff_gui_cairo_draw_rect_gtk2(cairo_t *cr, const GdkEventExpose *e,
					gint color, gint fill, const GdkRectangle *rect)
{
	/* Calculates the intersection of two rectangles */

	GdkRectangle dst;

	if (gdk_rectangle_intersect(&(e->area), rect, &dst))
	{
		cff_gui_cairo_draw_rectangle(cr, color, fill,
					(fill == FALSE) ? rect : &dst);
	}

} /* static void cff_gui_cairo_draw_rect_gtk2(cairo_t*, const GdkEventExpose*, gint, gint, const GdkRectangle*) */

#endif /* GTK_MAJOR_VERSION == 2 */

#else /* ! HAVE_CAIRO */

static void cff_gui_gdk_draw_rectangle(GdkEventExpose *e,
			gint color, gint fill, const GdkRectangle *rect)
{
	/* Calculates the intersection of two rectangles */

	GdkRectangle dst;

	if (gdk_rectangle_intersect(&(e->area), rect, &dst))
	{
		const GdkRectangle *ptr = &dst;

		if (fill == FALSE)
		{
			ptr = rect;
		}	

		gdk_draw_rectangle(e->window,
				gContext[color],
				fill,
				ptr->x,
				ptr->y,
				ptr->width,
				ptr->height);
	}

} /*static void cff_gui_gdk_draw_rectangle(GdkEventExpose*, gint, gint, const GdkRectangle*) */

#endif /* #ifdef HAVE_CAIRO .. #else */


static void cff_gui_calculate_frame_rects(cff_video_frame_t *current_frame,
							const cff_video_t *video)
{
	gint x, y, w, h, width, height;

	width  = GuiWidgetGetAllocatedWidth(current_frame->drawingArea);
	height = GuiWidgetGetAllocatedHeight(current_frame->drawingArea);

	/* Current video frame */

	current_frame->rectRGB24.x = (width - video->vframe.rgb24.width) / 2;
	current_frame->rectRGB24.y = (height - video->vframe.rgb24.height) / 2;
	current_frame->rectRGB24.width  = video->vframe.rgb24.width;
	current_frame->rectRGB24.height = video->vframe.rgb24.height;

	/* Left or top bar */

	w = width;

	if (width > video->vframe.rgb24.width)
	{
		w = current_frame->rectRGB24.x;
	}

	h = height;

	if (height > video->vframe.rgb24.height)
	{
		h = current_frame->rectRGB24.y;
	}

	current_frame->rectLeftTop.x = 0;
	current_frame->rectLeftTop.y = 0;
	current_frame->rectLeftTop.width  = w;
	current_frame->rectLeftTop.height = h;

	/* Right or bottom bar */

	x = 0;
	y = 0;

	if (current_frame->rectRGB24.x != 0)
	{
		x = current_frame->rectRGB24.x + current_frame->rectRGB24.width;
	}

	if (current_frame->rectRGB24.y != 0)
	{
		y = current_frame->rectRGB24.y + current_frame->rectRGB24.height;
	}

	current_frame->rectRightBottom.x = x;
	current_frame->rectRightBottom.y = y;
	current_frame->rectRightBottom.width  = width - x;
	current_frame->rectRightBottom.height = height - y;

	current_frame->dataRGB24 = (const guchar **) &(video->vframe.rgb24.data);

#if ! defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2
	current_frame->rowstrideRGB24 = video->vframe.rgb24.width * 3;
#endif

	/* Clear the whole area */
	cff_gui_widget_queue_draw_area(current_frame->drawingArea);

} /* static void cff_gui_calculate_frame_rects(cff_video_frame_t*, const cff_video_t*) */


/*
	-------
	 CAIRO
	-------
*/

#ifdef HAVE_CAIRO

static gboolean cff_gui_cairo_surface_image_create(cff_video_frame_t *current_frame,
								const cff_video_t *video)
{
	cairo_surface_t *ptr;

	cff_gui_calculate_frame_rects(current_frame, video);

	if (current_frame->surfaceRGB24 != NULL)
	{
		cairo_surface_destroy(current_frame->surfaceRGB24);
	}

	ptr = cairo_image_surface_create(CAIRO_FORMAT_RGB24,
					video->vframe.rgb24.width,
					video->vframe.rgb24.height);
	if (ptr != NULL &&
		cairo_surface_status(ptr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy(ptr);
		ptr = NULL;
	}

	current_frame->surfaceRGB24 = ptr;

	return (ptr == NULL) ? FALSE : TRUE;

} /* static gboolean cff_gui_cairo_surface_image_create(cff_video_frame_t*,
								const cff_video_t*) */

static void cff_gui_cairo_surface_image_fill(cff_video_frame_t *current_frame)
{
	int			i;
	unsigned char		z, *dst;
	const unsigned char	*src;
	cairo_surface_t		*surface;

	surface = current_frame->surfaceRGB24;

	if (surface == NULL)
	{
		return;
	}

	dst = cairo_image_surface_get_data(surface);

	if (dst == NULL)
	{
		return;
	}

	if (current_frame->dataRGB24 == NULL)
	{
		return;
	}

	src = *(current_frame->dataRGB24);

	if (src == NULL)
	{
		return;
	}

#if G_BYTE_ORDER != G_BIG_ENDIAN
	src += 3;
#endif

	i = current_frame->rectRGB24.width * current_frame->rectRGB24.height;
	z = 0;

	do
	{

#if G_BYTE_ORDER == G_BIG_ENDIAN
		*(dst++) = z;
		*(dst++) = *(src++);
		*(dst++) = *(src++);
		*(dst++) = *(src++);
#else
		*(dst++) = *(--src);
		*(dst++) = *(--src);
		*(dst++) = *(--src);
		*(dst++) = z;
		src += 6;
#endif

	}
	while (--i > 0);

} /* static void cff_gui_cairo_surface_image_fill(cff_video_frame_t*) */

#endif /* HAVE_CAIRO */


/*
	-------
	 VIDEO
	-------
*/

static void cff_gui_video_button_set_sensitive(gint sensitive)
{
	/*
		Called from
		  'cff_gui_init'
		  'cff_gui_video_open'
		  'cff_gui_file_close'
	*/

	gint		i;
	GtkWidget	**button;

	button = gui_button_video_frame;
	i = 0;

	do
	{
		gtk_widget_set_sensitive(*(button++), sensitive);
	}
	while (++i < kButtonVideoFrameTotal);

	gtk_widget_set_sensitive(gSlider.scale, sensitive);
	gtk_widget_set_sensitive(gui_button_video_cut[kButtonVideoCutAdd], sensitive);
	gtk_widget_set_sensitive(gui_button_video_cut[kButtonVideoCutRemove], FALSE);

} /* static void cff_gui_video_button_set_sensitive(gint) */

static void cff_gui_video_button_cut_set_sensitive(gint sensitive)
{
	/* Either add or remove a cut point */

	gtk_widget_set_sensitive(gui_button_video_cut[kButtonVideoCutAdd], sensitive);

	sensitive = (sensitive == TRUE) ? FALSE : TRUE;
	gtk_widget_set_sensitive(gui_button_video_cut[kButtonVideoCutRemove], sensitive);

} /* static void cff_gui_video_button_cut_set_sensitive(gint) */

static void cff_gui_widget_queue_draw_area(GtkWidget *widget)
{
	GdkRectangle rect;

	rect.x = 0;
	rect.y = 0;
	rect.width  = GuiWidgetGetAllocatedWidth(widget);
	rect.height = GuiWidgetGetAllocatedHeight(widget);

	gdk_window_invalidate_rect(GuiWidgetGetWindow(widget), &rect, FALSE);

} /* static void cff_gui_widget_queue_draw_area(GtkWidget*) */

static void cff_gui_frame_info(cff_video_frame_t *current_frame, const cff_video_t *video)
{
	gchar buf[128];

	snprintf(buf, sizeof(buf), "[%ux%u] [%u:%u] %s-Frame : %u / %u",
			video->vframe.frame_width,
			video->vframe.frame_height,
			video->vframe.frame_aspect[0],
			video->vframe.frame_aspect[1],
			video->vframe.pict_type,
			video->vframe.frame_current,
			video->vframe.frame_total);

	gtk_label_set_text(current_frame->info, buf);

	snprintf(buf, sizeof(buf), "Time: %02u:%02u:%02u.%03u / %02u:%02u:%02u.%03u",
			video->vframe.time_current[0],
			video->vframe.time_current[1],
			video->vframe.time_current[2],
			video->vframe.time_current[3],
			video->vframe.time_total[0],
			video->vframe.time_total[1],
			video->vframe.time_total[2],
			video->vframe.time_total[3]);

	gtk_label_set_text(current_frame->time, buf);

} /* static void cff_gui_frame_info(cff_video_frame_t*, const cff_video_t*) */


#if GTK_MAJOR_VERSION == 2
static gboolean cff_gui_draw_frame(GtkWidget *widget, GdkEvent *e, cff_video_frame_t *current_frame)
#else
static gboolean cff_gui_draw_frame(GtkWidget *widget, cairo_t *cr, cff_video_frame_t *current_frame)
#endif
{

#if defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2
	/* Creates a Cairo context for drawing to 'drawable' */
	cairo_t *cr = gdk_cairo_create(widget->window);
#endif

	if (current_frame->dataRGB24 == NULL ||
				*(current_frame->dataRGB24) == NULL)
	{
		/* No frame available, clear the whole area */

		GdkRectangle rect;

		rect.x = 0;
		rect.y = 0;
		rect.width  = GuiWidgetGetAllocatedWidth(widget);
		rect.height = GuiWidgetGetAllocatedHeight(widget);

		CffGuiDrawRectangle(cr, e, kColorGrey, TRUE, &rect);
	}

	else
	{
		/*
			Draw rectangles left and right (or top
			and bottom), and the frame in the middle
		*/

		CffGuiDrawRectangle(cr, e, kColorDarkGrey, TRUE,
					&(current_frame->rectLeftTop));

		CffGuiDrawRectangle(cr, e, kColorDarkGrey, TRUE,
					&(current_frame->rectRightBottom));

#ifdef HAVE_CAIRO

		if (current_frame->surfaceRGB24 != NULL)
		{
			cairo_set_source_surface(cr,
				current_frame->surfaceRGB24,
					current_frame->rectRGB24.x,
						current_frame->rectRGB24.y);
			cairo_paint(cr);
		}
#else
		gdk_draw_rgb_image(widget->window,
			gContext[kColorBlack],
			current_frame->rectRGB24.x,
			current_frame->rectRGB24.y,
			current_frame->rectRGB24.width,
			current_frame->rectRGB24.height,
			GDK_RGB_DITHER_NONE,
			*(current_frame->dataRGB24),
			current_frame->rowstrideRGB24);

#endif /* #ifdef HAVE_CAIRO .. #else */

	} /* if (current_frame->dataRGB24 == NULL ||
				*(current_frame->dataRGB24) == NULL) .. else */

#if defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2
	cairo_destroy(cr);
#endif		

	return TRUE;

} /* static gboolean cff_gui_draw_frame(GtkWidget*, GdkEvent*, cff_video_frame_t*) */

static void cff_gui_mainwindow_cursor_new(GtkWidget *widget, GdkScreen *prev_screen, gpointer data)
{
	GdkDisplay	*display;
	cff_cursor_t	*cursor;

	cursor	= data;
	display = gtk_widget_get_display(widget);

	if (cursor->busy != NULL)
	{
		if (gdk_cursor_get_display(cursor->busy) == display)
		{
			return;
		}

		GuiCursorUnref(cursor->busy);
	}

	cursor->busy   = gdk_cursor_new_for_display(display, GDK_WATCH);
	cursor->window = GuiWidgetGetWindow(widget);

} /* static void cff_gui_mainwindow_cursor_new(GtkWidget*, GdkScreen*, gpointer) */


#if GTK_MAJOR_VERSION == 2
static gint cff_gui_draw_frame_configure_cb(GtkWidget *widget, GdkEvent *e, cff_video_t *video)
#else
static gint cff_gui_draw_frame_configure_cb(GtkWidget *widget, GdkRectangle *rect, cff_video_t *video)
#endif
{
	/* 'gCurrentVideoFrame.drawingArea' has changed its size */

	if (video->vframe.rgb24.data != NULL && video->vframe.avframe != NULL)
	{
		gint			width, height;
		cff_video_frame_t	*current_frame;

		width  = GuiWidgetGetAllocatedWidth(widget);
		height = GuiWidgetGetAllocatedHeight(widget);

		cff_ffmpeg_frame2rgb24(video, width, height);
		current_frame = &gCurrentVideoFrame;

#ifdef HAVE_CAIRO
		cff_gui_cairo_surface_image_create(current_frame, video);
		cff_gui_cairo_surface_image_fill(current_frame);
#else
		cff_gui_calculate_frame_rects(current_frame, video);
#endif
	}

	return TRUE;

} /* static gint cff_gui_draw_frame_configure_cb(GtkWidget*, GdkEvent*, cff_video_t*) */

static gpointer cff_gui_video_frame_thread(gpointer data)
{
	gint retv;

	struct
	{
		cff_video_t		*video;
		cff_frame_seek_t	*frame_seek;
		gint			width;
		gint			height;
		enum cff_FrameFlags	frame_flags;

	} *s = data;

	switch ((s->frame_seek)->type)
	{
		unsigned int next;

		case kFrameSeekNext:
			retv = cff_ffmpeg_get_frame_next(s->video,
				(s->frame_seek)->value.e_next,
					s->width, s->height, s->frame_flags);
			break;

		case kFrameSeekCutpoint:
			retv = cff_ffmpeg_get_frame_cutpoint(s->video,
				(s->frame_seek)->value.e_next,
					s->width, s->height, s->frame_flags);
			break;

		case kFrameSeekSliderClick:
			retv = cff_ffmpeg_get_frame_slider_click(s->video,
				(s->frame_seek)->value.d_next,
					s->width, s->height, s->frame_flags);
			break;

		case kFrameSeekSliderMove:
			next = (s->frame_seek)->value.i_next;
			retv = cff_ffmpeg_get_frame_slider_move(s->video,
				next, s->width, s->height, s->frame_flags);
			break;

		default:
			retv = -1;
			break;

	} /* switch ((s->frame_seek)->type) */

	return GINT_TO_POINTER(retv);

} /* static gpointer cff_gui_video_frame_thread(gpointer) */

static gint cff_gui_video_frame(cff_frame_seek_t *frame_seek)
{
	gint			cursor, retv;
	gpointer		result;
	cff_slider_t		*slider;
	cff_video_t		*video;
	cff_video_frame_t	*current_frame;

	struct
	{
		cff_video_t		*video;
		cff_frame_seek_t	*frame_seek;
		gint			width;
		gint			height;
		enum cff_FrameFlags	frame_flags;
	} s;

	if ((gIsBusy & kIsBusyVideoFrame) || gtk_main_level() > 1)
	{
		return -1;
	}

	video = &gVideo;

	if (video->vframe.ic == NULL)
	{
		return -1;
	}

	gIsBusy |= kIsBusyVideoFrame;

	s.video		= video;
	s.frame_seek	= frame_seek;

	current_frame = &gCurrentVideoFrame;

	s.width  = GuiWidgetGetAllocatedWidth(current_frame->drawingArea);
	s.height = GuiWidgetGetAllocatedHeight(current_frame->drawingArea);

	s.frame_flags = 0;

	slider = &gSlider;

	if (slider->is_slider == FALSE)
	{
		gint flag;

		if (cff_gprop_get_bool(kPropBoolOptKeyFrame, &flag) && flag != 0)
		{
			s.frame_flags |= kKeyFrameFlag;
		}

		if (cff_gprop_get_bool(kPropBoolOptIFrame, &flag) && flag != 0)
		{
			s.frame_flags |= kIFrameFlag;
		}

		if (cff_gprop_get_bool(kPropBoolOptPFrame, &flag) && flag != 0)
		{
			s.frame_flags |= kPFrameFlag;
		}

	} /* if (slider->is_slider == FALSE) */

	if (cff_gprop_get_bool(kPropBoolOptCursor, &cursor) == FALSE)
	{
		cursor = TRUE;
	}

	if (cursor == TRUE && gCursor.window != NULL)
	{
		gdk_window_set_cursor(gCursor.window, gCursor.busy);
	}

	if (cursor && gThread.thread == NULL)
	{
		/* Draw the cursor and/or the slider */
		while (g_main_context_iteration(NULL, FALSE)) {;}
	}

	result = cff_gui_thread_invoke((GThreadFunc)
				cff_gui_video_frame_thread, &s);

	retv = GPOINTER_TO_INT(result);

	if (cursor && gCursor.window != NULL)
	{
		gdk_window_set_cursor(gCursor.window, NULL);
	}

	if (retv == -1)
	{
		if (slider->is_slider == FALSE || slider->button == 0)
		{
			cff_guitl_beep();
		}
	}

	else
	{
		gboolean cut_add;

#ifdef HAVE_CAIRO

		if (current_frame->surfaceRGB24 == NULL ||
			current_frame->rectRGB24.width !=
					video->vframe.rgb24.width ||
				current_frame->rectRGB24.height !=
						video->vframe.rgb24.height)
		{
			/* Draw the whole area */
			if (cff_gui_cairo_surface_image_create(current_frame,
								video) == FALSE)
			{
				gIsBusy &= ~kIsBusyVideoFrame;
				return -1;
			}
		}
#else
		if (current_frame->rectRGB24.width !=
					video->vframe.rgb24.width ||
				current_frame->rectRGB24.height !=
						video->vframe.rgb24.height)
		{
			/* Draw the whole area */
			cff_gui_calculate_frame_rects(current_frame, video);
		}

#endif /* ifdef HAVE_CAIRO .. #else */

		else
		{
			/* Draw the video frame only */

			GdkWindow *wnd =
				GuiWidgetGetWindow(current_frame->drawingArea);

			gdk_window_invalidate_rect(wnd,
				&(current_frame->rectRGB24), FALSE);

		} /* if (current_frame->...) */


#ifdef HAVE_CAIRO
		cff_gui_cairo_surface_image_fill(current_frame);
#endif

		if (slider->is_slider == FALSE)
		{
			cff_gui_slider_adjustment_set_value(slider);
		}

		cff_gui_frame_info(current_frame, video);

		cut_add = TRUE;

		if (cff_ffmpeg_video_cut_find_current(video) >= 0)
		{
			cut_add = FALSE;
		}

		cff_gui_video_button_cut_set_sensitive(cut_add);

	} /* if (retv == -1) .. else */

	gIsBusy &= ~kIsBusyVideoFrame;

	return retv;

} /* static gint cff_gui_video_frame(cff_frame_seek_t*) */

static gboolean cff_gui_video_frame_button(GtkWidget *button, GdkEvent *e, gpointer data)
{
	enum cff_VideoFrame	next;
	GdkEventButton      	*bevent;
	cff_frame_seek_t	frame_seek;

	bevent = (GdkEventButton *) e;

	switch (GPOINTER_TO_INT(data))
	{
		case kButtonVideoFramePrev4:
			next = kVideoFramePrevCut;
			break;
		case kButtonVideoFramePrev3:
			next = kVideoFramePrev05Sec;
			break;
		case kButtonVideoFramePrev2:
			next = kVideoFramePrevHalfSec;
			break;
		case kButtonVideoFramePrev1:
			next = kVideoFramePrevSingle;
			break;
		case kButtonVideoFrameNext1:
			next = kVideoFrameNextSingle;
			break;
		case kButtonVideoFrameNext2:
			next = kVideoFrameNextHalfSec;
			break;
		case kButtonVideoFrameNext3:
			next = kVideoFrameNext05Sec;
			break;
		case kButtonVideoFrameNext4:
			next = kVideoFrameNextCut;
			break;
		default:
			return TRUE;

	} /* switch (GPOINTER_TO_INT(data)) */

	if ((bevent->state & GDK_CONTROL_MASK) != 0)
	{
		return FALSE;
	}

	if ((bevent->state & GDK_SHIFT_MASK) != 0 || bevent->button == 3)
	{
		/* Shift key or right mouse button */
		next += kVideoFrameShiftAdd;
	}

	frame_seek.type		= kFrameSeekNext;
	frame_seek.value.e_next	= next;
	cff_gui_video_frame(&frame_seek);

	return TRUE;

} /* static gboolean cff_gui_video_frame_button(GtkWidget*, GdkEvent*, gpointer) */

static gboolean cff_gui_video_frame_key(GtkWidget *window, GdkEvent *e, gpointer data)
{
	enum cff_VideoFrame	next;
	GdkEventKey		*kevent;
	cff_frame_seek_t	frame_seek;

	kevent = (GdkEventKey *) e;

	switch (kevent->keyval)
	{
		case Gui_KEY_Home:
		case Gui_KEY_KP_Home:
		case Gui_KEY_KP_7:
			next = kVideoFramePrevCut;
			break;
		case Gui_KEY_Page_Down:
		case Gui_KEY_KP_Page_Down:
		case Gui_KEY_KP_3:
			next = kVideoFramePrev05Sec;
			break;
		case Gui_KEY_Down:
		case Gui_KEY_KP_Down:
		case Gui_KEY_KP_2:
			next = kVideoFramePrevHalfSec;
			break;
		case Gui_KEY_Left:
		case Gui_KEY_KP_Left:
		case Gui_KEY_KP_4:
			next = kVideoFramePrevSingle;
			break;
		case Gui_KEY_Right:
		case Gui_KEY_KP_Right:
		case Gui_KEY_KP_6:
			next = kVideoFrameNextSingle;
			break;
		case Gui_KEY_Up:
		case Gui_KEY_KP_Up:
		case Gui_KEY_KP_8:
			next = kVideoFrameNextHalfSec;
			break;
		case Gui_KEY_Page_Up:
		case Gui_KEY_KP_Page_Up:
		case Gui_KEY_KP_9:
			next = kVideoFrameNext05Sec;
			break;
		case Gui_KEY_End:
		case Gui_KEY_KP_End:
		case Gui_KEY_KP_1:
			next = kVideoFrameNextCut;
			break;
		default:
			return FALSE;

	} /* switch (kevent->keyval) */

	if ((kevent->state & GDK_CONTROL_MASK) != 0)
	{
		return FALSE;
	}

	if ((kevent->state & GDK_SHIFT_MASK) != 0)
	{
		next += kVideoFrameShiftAdd;
	}

	frame_seek.type		= kFrameSeekNext;
	frame_seek.value.e_next	= next;
	cff_gui_video_frame(&frame_seek);

	return TRUE;

} /* static gboolean cff_gui_video_frame_key(GtkWidget*, GdkEvent*, gpointer) */

static gint cff_gui_video_frame_click(GtkWidget *widget, gdouble x)
{
	gint			width;
	gdouble			next;
	cff_frame_seek_t	frame_seek;

	width = GuiWidgetGetAllocatedWidth(widget);
	next  = x / (gdouble) width;

	frame_seek.type		= kFrameSeekSliderClick;
	frame_seek.value.d_next	= next;

	return cff_gui_video_frame(&frame_seek);

} /* static gint cff_gui_video_frame_click(GtkWidget*, gdouble) */

static gboolean cff_gui_video_frame_slider(GtkRange *range, GtkScrollType scroll, gdouble value, cff_slider_t *slider)
{
	unsigned int value_int;
	cff_video_t  *video;

	slider->value = value;

	if (slider->is_slider || gIsBusy)
	{
		return TRUE;
	}

	video     = &gVideo;
	value_int = (unsigned int) value;

	if (value_int > video->vframe.frame_total)
	{
		value_int = video->vframe.frame_total;
		value     = (gdouble) value_int;
	}

	else if (value < 0.0)
	{
		value_int = 0;
		value     = 0.0;
	}

	if (gtk_main_level() > 1 ||
			((scroll == GTK_SCROLL_JUMP) &&
				video->vframe.frame_current == value_int))
	{
		return TRUE;
	}

	if (scroll == GTK_SCROLL_JUMP)
	{
		cff_frame_seek_t frame_seek;

		frame_seek.type   = kFrameSeekSliderMove;
		slider->is_slider = TRUE;

		for (;;)
		{
			frame_seek.value.i_next	= value_int;
			cff_gui_video_frame(&frame_seek);

			if (slider->button)
			{
				break;
			}

			if (value == slider->value)
			{
				if (video->vframe.frame_current !=
					(unsigned int) GuiAdjustmentGetValue(slider->adjust))
				{
					/* Move slider to actual position */
					g_idle_add_full(Gui_PRIORITY_REDRAW,
						(GSourceFunc)
							cff_gui_slider_adjustment_set_value,
							slider, NULL);
				}

				slider->is_slider = FALSE;
				return TRUE;
			}

			value = slider->value;

		} /* for (;;) */

		slider->is_slider = FALSE;
		return FALSE;

	} /* if (scroll == GTK_SCROLL_JUMP) */

	if (slider->button == 1)
	{
		slider->button = 2;
		cff_gui_video_frame_click(slider->scale, slider->x);
	}

	return TRUE;

} /* static gboolean cff_gui_video_frame_slider(GtkRange*, GtkScrollType, gdouble, cff_slider_t*) */

static gboolean cff_gui_slider_button(GtkWidget *widget, GdkEvent *e, cff_slider_t *slider)
{
	GdkEventButton *bevent = (GdkEventButton *) e;

	if (bevent->type == GDK_BUTTON_RELEASE)
	{
		slider->button = 0;
	}

	else if (slider->button == 0)
	{
		slider->x	= bevent->x;
		slider->button	= 1;

	}

	return FALSE;

} /* static gboolean cff_gui_slider_button(GtkWidget*, GdkEvent*, cff_slider_t*) */


#if GTK_MAJOR_VERSION == 2
static gint cff_gui_draw_bar(GtkWidget *widget, GdkEvent *e, cff_video_t *video)
#else
static gint cff_gui_draw_bar(GtkWidget *widget, cairo_t *cr, cff_video_t *video)
#endif
{
	gint		x1, color, width;
	GdkRectangle	rect;

#if defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2
	/* Creates a Cairo context for drawing to 'drawable' */
	cairo_t *cr = gdk_cairo_create(widget->window);
#endif

	width	= GuiWidgetGetAllocatedWidth(widget);
	x1	= 0;
	color	= kColorGrey;

	rect.y      = 0;
	rect.height = GuiWidgetGetAllocatedHeight(widget);

	if (video->vframe.frame_total != 0 && video->vframe.stream_duration > 0)
	{
		gint		i;
		cff_dts_t	*dts;

		color	= kColorGreen;
		dts	= video->cut.list;
		i	= video->cut.num;

		if (i > 0 && dts != NULL)
		{
			double ratio =
				(double) width / (double) video->vframe.stream_duration;

			color = kColorRed;

			do
			{
				gint x2 = (double) (dts->frame - video->vframe.dts_start) * ratio;
				dts++;

				if (x2 > 0)
				{
					rect.x     = x1;
					rect.width = x2 - x1;

					CffGuiDrawRectangle(cr, e,
							color, TRUE, &rect);
					x1 = x2;

				}

				if (color == kColorGreen)
				{
					color = kColorRed;
				}

				else
				{
					color = kColorGreen;
				}

			}
			while (--i > 0);

		} /* if (i > 0 && ptr != NULL) */

	} /* if (video->vframe.frame_total != 0 && video->vframe.duration > 0) */


	/* Remaining rectangle */

	rect.x     = x1;
	rect.width = width - x1;

	CffGuiDrawRectangle(cr, e, color, TRUE, &rect);

	/* Border */

	rect.x     = 0;
	rect.width = width - 1;
	rect.height--;

	CffGuiDrawRectangle(cr, e, kColorBlack, FALSE, &rect);

#if defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2
	cairo_destroy(cr);
#endif

	return TRUE;

} /* static gint cff_gui_draw_bar(GtkWidget*, GdkEvent*, cff_video_t*) */

static gboolean cff_gui_draw_bar_button(GtkWidget *widget, GdkEvent *e, cff_video_t *video)
{
	GdkEventButton *bevent = (GdkEventButton *) e;

	if (video->vframe.frame_total != 0 &&
			bevent->type == GDK_BUTTON_PRESS &&
						bevent->button == 1)
	{
		cff_gui_video_frame_click(widget, bevent->x);
	}

	return TRUE;

} /* static gboolean cff_gui_draw_bar_button(GtkWidget*, GdkEvent*, cff_video_t*) */

static void cff_gui_open_result(cff_video_t *video, const gchar *str, guint mbar_item)
{
	gint			sensitive;
	cff_video_frame_t	*current_frame;

	current_frame = &gCurrentVideoFrame;

	if (str == NULL)
	{
		cff_frame_seek_t frame_seek;

#ifdef HAVE_CAIRO

		if (current_frame->surfaceRGB24 != NULL)
		{
			/* Create a new surface */
			cairo_surface_destroy(current_frame->surfaceRGB24);
			current_frame->surfaceRGB24 = NULL;
		}
#endif
		cff_gui_slider_adjustment_configure(video->vframe.frame_total);

		frame_seek.type		= kFrameSeekNext;
		frame_seek.value.e_next	= kVideoFrameStart;
		cff_gui_video_frame(&frame_seek);
		sensitive = TRUE;
	}

	else
	{
		cff_gui_slider_adjustment_configure(0);
		cff_gui_widget_queue_draw_area(current_frame->drawingArea);
		sensitive = FALSE;
	}

	cff_gui_video_button_set_sensitive(sensitive);

	cff_gui_frame_info(current_frame, video);
	cff_gui_widget_queue_draw_area(gDrawingAreaBar);

	if (str != NULL)
	{
		cff_guitl_beep();
		cff_guitl_dialog_error(str, &cff_gui_menu_bar[mbar_item]);
	}

} /* static void cff_gui_open_result(cff_video_t*, const gchar*, guint) */

static int cff_gui_open_callback(gpointer data)
{
	const gchar		**str;
	cff_progressbar_t	user_data;

	user_data.func  = cff_ffmpeg_file_open_cb;
	user_data.video = &gVideo;
	user_data.title = gettext_nop("Scanning ...");
	user_data.mbar  = &cff_gui_menu_bar[kMenuBarFile];

	if (gLogWindow.nThreads == 0)
	{
		user_data.mutex = NULL;
		cff_gui_dialog_progress_bar(&user_data);
	}

	else
	{
		GuiMutex mutex;

		user_data.mutex = &mutex;
		GuiMutexInit(&mutex);

		user_data.is_locked = FALSE;

		g_main_context_invoke_full(NULL, G_PRIORITY_HIGH,
			(GSourceFunc) cff_gui_dialog_progress_bar,
							&user_data, NULL);

		while (user_data.is_locked == FALSE)
		{
			g_usleep(250);
		}

		GuiMutexLock(&mutex);
		GuiMutexUnlock(&mutex);
		GuiMutexFree(&mutex);

	} /* if (gLogWindow.nThreads == 0) .. else */

	str = (const gchar **) data;

	if (str != NULL)
	{
		*str = user_data.error;
	}

	return user_data.retv;

} /* static int cff_gui_open_callback(gpointer) */

static gint cff_gui_program_open(gpointer data)
{
	const gchar	*error, *str;
	cff_video_t	*video;

	video = &gVideo;

	gIsBusy |= kIsBusyFile;

	str = cff_ffmpeg_program_open(video, GPOINTER_TO_INT(data),
					cff_gui_open_callback, &error);

	cff_gui_open_result(video, (str == NULL) ? error : str, kMenuBarEdit);

	gIsBusy &= ~kIsBusyFile;

	return Gui_SOURCE_REMOVE;

} /* static gint cff_gui_program_open(gpointer) */

static const gchar *cff_gui_video_open_cb(gpointer data)
{
	struct
	{
		cff_video_t	*video;
		gchar		*fname;
		const gchar	*error;

	} *s = data;

	return cff_ffmpeg_file_open(s->video, s->fname,
				cff_gui_open_callback, &(s->error));

} /* static const gchar *cff_gui_video_open_cb(gpointer) */

static gint cff_gui_video_open(gchar *fname)
{
	const gchar *str;

	struct
	{
		cff_video_t	*video;
		gchar		*fname;
		const gchar	*error;

	} s;

	s.fname  = fname;
	s.video  = &gVideo;
	gIsBusy |= kIsBusyFile;

	str = (const gchar *) cff_gui_thread_invoke((GThreadFunc)
						cff_gui_video_open_cb, &s);

	cff_gui_open_result(s.video, (str == NULL) ?
					s.error : str, kMenuBarFile);

	cff_guitl_window_set_title_filename(gWindata.mainwin,
				((s.video)->inFormatCtx == NULL) ?
							NULL : s.fname);
	g_free(s.fname);
	gIsBusy &= ~kIsBusyFile;

	return Gui_SOURCE_REMOVE;

} /* static gint cff_gui_video_open(gchar*) */


/*
	-------------
	 Menu 'File'
	-------------
*/

static void cff_gui_file_open(GtkWidget *w, GtkWidget **mbar)
{
	gchar *fname;

	if (cff_gui_file_discard_changes(mbar) == FALSE)
	{
		return;
	}

	fname = cff_gui_file_selection_new_gettext(gettext_nop("Open Video File"),
					NULL, mbar, kPropStrFileOpenPath);
	if (fname == NULL)
	{
		return;
	}

	cff_gui_file_close(NULL, NULL);

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) cff_gui_video_open, fname, NULL);

} /* static void cff_gui_file_open(GtkWidget*, GtkWidget**) */

static void cff_gui_file_save(GtkWidget *w, GtkWidget **mbar)
{
	gint				i, n_encoders, pos_old, *newpos_ptr, newpos_buf[10];
	gboolean			got_video, got_audio;
	gchar				*str, **ptr;
	GtkWidget			*button, *notebook, *bbox, *hbox, *vbox, *frame;
	cff_fmtlist_t			*format_info;
	cff_save_video_t		*vsave;
	cff_video_t			*video;
	cff_save_video_prev_options_t	*save_prev;
	gchar				*buf[512];
	cff_ListStore_t			*list;
	GtkTreeIter			iter;


	video = &gVideo;

	if (video->fname == NULL)
	{
		return;
	}

	vsave = g_malloc(sizeof(cff_save_video_t));

	vsave->common.mbar  = mbar;
	vsave->common.video = video;

	vsave->common.cancel.func = cff_gui_save_delevnt;
	vsave->common.cancel.data = vsave;

	vbox = cff_gui_save_build_common(&(vsave->common),
		gettext_nop("Save Video File"), kPropStrFileSavePath);

	cff_guitl_window_setup(vsave->common.window, kPropIntFileSaveWinWidth);

	hbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	GuiVBoxPackStart(vbox, frame, TRUE, TRUE, 4);

	save_prev = gSaveVideoPrevOptionsPtr;

	if (save_prev == NULL)
	{
		save_prev = g_malloc0(sizeof(cff_save_video_prev_options_t));
		gSaveVideoPrevOptionsPtr = save_prev;
	}


	/* List */

	cff_gui_build_list(gettext_nop("Available"), &(vsave->avail.list), hbox);


	/* Button */

	bbox = GuiVBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(bbox), 4);

	vsave->up =
		cff_guitl_vbox_add_button_gettext(bbox, gettext_nop("Up"),
				(GCallback) cff_gui_save_list_updown, vsave);
	vsave->avail.all =
		cff_guitl_vbox_add_button(bbox, ">>",
				(GCallback) cff_gui_save_list_leftright, vsave);
	vsave->avail.some =
		cff_guitl_vbox_add_button(bbox, ">",
				(GCallback) cff_gui_save_list_leftright, vsave);
	vsave->out.some =
		cff_guitl_vbox_add_button(bbox, "<",
				(GCallback) cff_gui_save_list_leftright, vsave);
	vsave->out.all =
		cff_guitl_vbox_add_button(bbox, "<<",
				(GCallback) cff_gui_save_list_leftright, vsave);
	vsave->down =
		cff_guitl_vbox_add_button_gettext(bbox, gettext_nop("Down"),
				(GCallback) cff_gui_save_list_updown, vsave);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, bbox);
	bbox = GuiVBoxNew(FALSE, 1);
	GuiVBoxPackStart(bbox, frame, FALSE, FALSE, 1);
	GuiHBoxPackStart(hbox, bbox, FALSE, FALSE, 4);


	/* List */

	cff_gui_build_list(gettext_nop("Output"), &(vsave->out.list), hbox);


	/* Check button */

	hbox = GuiHBoxNew(TRUE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 1);

	bbox = GuiVBoxNew(FALSE, 1);
	GuiHBoxPackStart(hbox, bbox, TRUE, FALSE, 1);

	vsave->shortest =
		cff_guitl_check_button_new_gettext(gettext_nop("Shortest duration"));
	GuiVBoxPackStart(bbox, vsave->shortest, FALSE, FALSE, 8);

	vsave->stream_sync =
		cff_guitl_check_button_new_gettext(gettext_nop("Stream sync"));
	GuiVBoxPackStart(bbox, vsave->stream_sync, FALSE, FALSE, 8);


	/* Softcut Option Menu */

	n_encoders = cff_ffmpeg_get_encoder_names(video, &(vsave->softcut.encoder_names));

	if (n_encoders <= 0)
	{
		/* No encoder found */
		vsave->softcut.encoder_names = NULL;
	}

	bbox = GuiVBoxNew(FALSE, 1);

	vsave->softcut.option =
			cff_gui_build_option_menu(hbox, bbox,
			gettext_nop("Soft cut"),
			gettext_nop("No"),
			(const gchar **) vsave->softcut.encoder_names);

	vsave->softcut.both_gops = cff_guitl_vbox_add_check_button_gettext(bbox,
					gettext_nop("Both GOPs"));

	gtk_widget_set_sensitive(vsave->softcut.both_gops, FALSE);

	if (vsave->softcut.encoder_names != NULL)
	{
		GuiSignalConnect(vsave->softcut.option, "changed",
			cff_gui_save_softcut_option_changed_cb, vsave->softcut.both_gops);
	}


	/* Extra */

	notebook = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
	gtk_container_set_border_width(GTK_CONTAINER(notebook), 4);

	frame = cff_guitl_frame_new_add_child_gettext(gettext_nop("Extra"), notebook);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 2);


	/* General-Tab */

	bbox = cff_gui_notebook_append_page(notebook, gettext_nop("General"));


	/* Check button */

	vsave->ts_keep_start = cff_guitl_vbox_add_check_button_gettext(bbox,
					gettext_nop("Keep start timestamps"));

	/* Option menu */

	vsave->common.format.u.list =
			cff_ffmpeg_get_format_names(vsave->common.video);

	ptr = buf;

	if ((format_info = vsave->common.format.u.list) != NULL)
	{
		while (format_info->name != NULL)
		{
			if (ptr - buf < (sizeof(buf) / sizeof(gchar*)) - 1)
			{
				*(ptr++) = g_strconcat(format_info->name, " ",
							format_info->codecs, NULL);
			}

			if (format_info->codecs != NULL)
			{
				/* malloc'ed */
				free(format_info->codecs);
				format_info->codecs = NULL;
			}

			format_info++;
		}

	} /* if ((format_info = vsave->common.format.u.list) != NULL) */

	*ptr = NULL;

	vsave->common.format.option =
			cff_gui_build_option_menu(bbox, NULL,
				gettext_nop("Format"),
				gettext_nop("Auto"),
				(const gchar **) buf);

	if (vsave->common.format.u.list != NULL)
	{
		GuiSignalConnect(vsave->common.format.option, "changed",
			cff_gui_save_optionmenu_callback, &(vsave->common));
	}


	/* Free strings in array */

	ptr = buf;

	while ((str = *(ptr++)) != NULL)
	{
		g_free(str);
	}


	/* Media-Tabs */

	cff_gui_build_media_option(notebook, vsave, kCFFMediaVideo);
	cff_gui_build_media_option(notebook, vsave, kCFFMediaAudio);
	cff_gui_build_media_option(notebook, vsave, kCFFMediaOther);


	/* Button */

	hbox = cff_guitl_hbutton_box_new(vbox, 8, GTK_BUTTONBOX_END);
	cff_guitl_button_box_add_button(hbox, gettext_nop("Cancel"),
				(GCallback) cff_gui_save_destroy, vsave);

	button = cff_guitl_button_box_add_button(hbox, gettext_nop("OK"),
				(GCallback) cff_gui_save_start, vsave);
	vsave->common.okay.widget = button;
	vsave->common.okay.bits   = 0;
	gtk_widget_grab_default(button);


	/* Last used streams (if any) */

	newpos_ptr = NULL;
	pos_old    = 0;

	if (save_prev->stream_idx != NULL && save_prev->nb_streams > 0)
	{
		newpos_ptr = newpos_buf;

		if (save_prev->nb_streams > sizeof(newpos_buf) / sizeof(gint))
		{
			newpos_ptr = g_malloc(sizeof(gint) * save_prev->nb_streams);
		}
	}


	/* Put streams in the available or output list */

	got_video = FALSE;
	got_audio = FALSE;

	i = 0;

	do
	{
		gint	media, stream_idx;
		gchar	*str, buf[256];

		if (cff_ffmpeg_get_stream_info(video, i,
					&stream_idx, buf, sizeof(buf)) < 0)
		{
			continue;
		}

		media	= kCFFMediaOther;
		list	= &(vsave->avail.list);
		str	= strchr(buf, ' ');

		if (str != NULL)
		{
			if (g_ascii_strncasecmp(&str[1], "video", 5) == 0)
			{
				media = kCFFMediaVideo;
			}

			else if (g_ascii_strncasecmp(&str[1], "audio", 5) == 0)
			{
				media = kCFFMediaAudio;
			}

			if (media == kCFFMediaVideo)
			{
				if (got_video == FALSE)
				{
					got_video = TRUE;
					list = &(vsave->out.list);
				}
			}

			else if (media == kCFFMediaAudio)
			{
				if (got_audio == FALSE)
				{
					got_audio = TRUE;
					list = &(vsave->out.list);
				}
			}

		} /* if (str != NULL) */

		if (newpos_ptr != NULL)
		{
			gint pos_new;

			list	= &(vsave->avail.list);
			pos_new	= 0;

			do
			{
				if (stream_idx == save_prev->stream_idx[pos_new])
				{
					/*
						Mapping the new position of
						each child to its old position
						before the re-ordering
					*/

					newpos_ptr[pos_new] = pos_old++;
					list = &(vsave->out.list);

					break;
				}
			}
			while (++pos_new < save_prev->nb_streams);

		} /* if (newpos_ptr != NULL) */

		gtk_list_store_append(list->store, &iter);

		str = cff_tl_replace_non_ascii_chars_in_str(buf);
		gtk_list_store_set(list->store, &iter,
				0, str, 1, stream_idx, 2, media, -1);
	}
	while(++i < video->nb_streams);


	/* Set last used options */

	if (newpos_ptr != NULL)
	{
		gint lastpos;

		list	= &(vsave->out.list);
		i	= 0;
		lastpos	= 0;

		do
		{
			gint pos =  newpos_ptr[i];

			if (lastpos > pos)
			{
				gtk_list_store_reorder(list->store, newpos_ptr);
				break;
			}

			lastpos = pos;
		}
		while(++i < save_prev->nb_streams);

		if (newpos_ptr != newpos_buf)
		{
			g_free(newpos_ptr);
		}

	} /* if (newpos_ptr != NULL) */


	if (save_prev->format_idx)
	{
		GuiOptionMenuSetActive(vsave->common.format.option, save_prev->format_idx);
	}

	if (vsave->softcut.encoder_names != NULL)
	{
		if (save_prev->encoder_idx)
		{
			GuiOptionMenuSetActive(vsave->softcut.option, save_prev->encoder_idx);

			if (save_prev->both_gops)
			{
				gtk_toggle_button_set_active(
					GTK_TOGGLE_BUTTON(vsave->softcut.both_gops), TRUE);
			}
		}

	} /* if (vsave->softcut.encoder_names != NULL) */

	if (save_prev->shortest)
	{
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(vsave->shortest), TRUE);
	}

	if (save_prev->stream_sync &&
			cff_ffmpeg_video_cut_list_distance(&gVideo) != 0)
	{
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(vsave->stream_sync), TRUE);
	}

	if (save_prev->ts_keep_start)
	{
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(vsave->ts_keep_start), TRUE);
	}

	i = kCFFMediaVideo;

	do
	{
		if (save_prev->media[i].bsfname_idx)
		{
			GuiOptionMenuSetActive(vsave->media[i].bsf_option,
						save_prev->media[i].bsfname_idx);
		}

		if (save_prev->media[i].ts_from_duration)
		{
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(vsave->media[i].ts_from_duration), TRUE);
		}
	}
	while (++i < kCFFMediaTotal);


	/* Select the first row, set signal func, set buttons (in)sensitive */

	cff_gui_save_list_selection_init(&(vsave->avail.list), vsave);
	cff_gui_save_list_selection_init(&(vsave->out.list), vsave);

	gtk_widget_show_all(vsave->common.window);

} /* static void cff_gui_file_save(GtkWidget*, GtkWidget**) */

static void cff_gui_file_image(GtkWidget *w, GtkWidget **mbar)
{
	gint			fmt_num;
	gchar			**long_names;
	GtkWidget		*button, *hbox, *vbox, *vbox1, *bbox, *fbox, *frame;
	cff_save_image_t	*vimage;
	cff_video_t		*video;

	video = &gVideo;

	if (video->fname == NULL)
	{
		return;
	}

	vimage = g_malloc(sizeof(cff_save_image_t));

	vimage->common.mbar  = mbar;
	vimage->common.video = video;

	vimage->common.cancel.func = cff_gui_image_delevnt;
	vimage->common.cancel.data = vimage;

	vbox = cff_gui_save_build_common(&(vimage->common),
		gettext_nop("Save Image"), kPropStrFileImagePath);

	cff_guitl_window_setup(vimage->common.window, kPropIntFileImageWinWidth);

	vbox1 = GuiVBoxNew(FALSE, 4);
	GuiVBoxPackStart(vbox, vbox1, FALSE, FALSE, 4);

	fbox = GuiVBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(fbox), 4);
	GuiVBoxPackStart(vbox, fbox, FALSE, FALSE, 4);

	/* Spin buttons */

	hbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	GuiVBoxPackStart(fbox, frame, FALSE, FALSE, 4);

	bbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(bbox), 4);
	GuiHBoxPackStart(hbox, bbox, TRUE, FALSE, 4);

	vimage->width.prev_value = video->vframe.frame_width;
	vimage->width.widget = cff_gui_image_spinbutton(bbox,
			video->vframe.frame_width, gettext_nop("Width"));

	vimage->height.prev_value = video->vframe.frame_height;
	vimage->height.widget = cff_gui_image_spinbutton(bbox,
			video->vframe.frame_height, gettext_nop("Height"));

	/* Check button */

	hbox = GuiHBoxNew(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 4);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, hbox);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
	GuiVBoxPackStart(fbox, frame, FALSE, FALSE, 4);

	bbox = GuiHBoxNew(FALSE, 4);
	GuiHBoxPackStart(hbox, bbox, TRUE, FALSE, 4);

	hbox = bbox;
	bbox = GuiVBoxNew(FALSE, 4);
	GuiHBoxPackStart(hbox, bbox, FALSE, FALSE, 4);

	gtk_container_set_border_width(GTK_CONTAINER(bbox), 4);

	vimage->keep_dimension =
		cff_guitl_check_button_new_gettext(gettext_nop("Keep dimension"));
	GuiVBoxPackStart(bbox, vimage->keep_dimension, FALSE, FALSE, 4);
	GuiSignalConnect(vimage->keep_dimension,
		"toggled", cff_gui_image_keep_dimension_toggled, vimage);

	vimage->keep_ratio =
		cff_guitl_check_button_new_gettext(gettext_nop("Keep aspect ratio"));
	GuiVBoxPackStart(bbox, vimage->keep_ratio, FALSE, FALSE, 4);

	vimage->gray = cff_guitl_check_button_new_gettext(gettext_nop("Gray"));
	GuiVBoxPackStart(bbox, vimage->gray, FALSE, FALSE, 4);
				
	/* Option menu */

	fmt_num = cff_ffmpeg_get_image_format_names(&long_names,
					&(vimage->common.format.u.names));

	vimage->common.format.option =
			cff_gui_build_option_menu(vbox, NULL,
				gettext_nop("Format"),
				gettext_nop("Auto"),
				(const gchar **) long_names);

	if (fmt_num > 0)
	{
		if (long_names != NULL)
		{
			/* malloc'ed */
			free(long_names);
		}

		GuiSignalConnect(vimage->common.format.option, "changed",
			cff_gui_save_optionmenu_callback, &(vimage->common));
	}

	/* Padding */

	vbox1 = GuiVBoxNew(FALSE, 4);
	GuiVBoxPackStart(vbox, vbox1, TRUE, FALSE, 4);

	/* Button */

	hbox = cff_guitl_hbutton_box_new(vbox, 8, GTK_BUTTONBOX_END);

	cff_guitl_button_box_add_button(hbox, gettext_nop("Cancel"),
				(GCallback) cff_gui_image_destroy, vimage);

	button = cff_guitl_button_box_add_button(hbox, gettext_nop("OK"),
				(GCallback) cff_gui_save_image, vimage);
	vimage->common.okay.widget = button;
	vimage->common.okay.bits   = 0;
	gtk_widget_grab_default(button);

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(vimage->keep_ratio), TRUE);

	gtk_widget_show_all(vimage->common.window);

} /* static void cff_gui_file_image(GtkWidget*, GtkWidget**) */

static void cff_gui_file_close(GtkWidget *w, GtkWidget **mbar)
{
	/* mbar == NULL => from 'cff_gui_file_open' */

	if (mbar != NULL && cff_gui_file_discard_changes(mbar) == FALSE)
	{
		return;
	}

	cff_ffmpeg_file_close(&gVideo);

	cff_gui_save_video_prev_settings_free();
	cff_gui_frame_info(&gCurrentVideoFrame, &gVideo);
	cff_gui_slider_adjustment_configure(0);
	cff_gui_video_button_set_sensitive(FALSE);

	cff_gui_widget_queue_draw_area(gCurrentVideoFrame.drawingArea);
	cff_gui_widget_queue_draw_area(gDrawingAreaBar);

	cff_guitl_window_set_title_filename(gWindata.mainwin, NULL);

} /* static void cff_gui_file_close(GtkWidget*, GtkWidget**) */

static void cff_gui_file_info(GtkWidget *w, GtkWidget **mbar)
{
	char		*str;
	GtkWidget	*ibox;

	ibox = GuiVBoxNew(FALSE, 2);
	str  = cff_ffmpeg_file_info(&gVideo);

	str = cff_tl_replace_non_ascii_chars_in_str(str);
	cff_gui_ffmpeg_info_label(NULL, str, ibox);

	if (str)
	{
		/* malloc'ed */
		free(str);
	}

	cff_gui_scrolled_info_window(gettext_nop("Video File Info"),
				ibox, mbar, kPropIntFileInfoWinWidth);

} /* static void cff_gui_file_info(GtkWidget*, GtkWidget**) */

static void cff_gui_file_save_rc(GtkWidget *w, gpointer data)
{
	cff_main_save_rc();

} /* static void cff_gui_file_save_rc(GtkWidget*, gpointer) */

static gboolean cff_gui_file_quit_delevnt(GtkWidget *w, GdkEvent *e, GtkWidget **mbar)
{
	cff_video_t	*video;
	cff_logwindow_t	*log;
	cff_thread_t	*thread;

	video = &gVideo;

	if (gIsBusy || gtk_main_level() > 1)
	{
		return TRUE;
	}

	if (cff_gui_file_discard_changes(mbar) == FALSE)
	{
		return TRUE;
	}

	/* 'cff_main_save_rc' called from 'cff_main_atexit' */

	log = &gLogWindow;

	if (log->window != NULL)
	{
		int		save_flag;
		gchar		*logname;
		GtkWidget	*window;

		if (cff_gprop_get_bool(kPropBoolOptSaveConfig,
					&save_flag) == FALSE || save_flag)
		{
			/* Get log-level, will be set to 'quiet' */
			logname = cff_gprop_get_str(kPropStrOptLoglevel);
		}

		else
		{
			logname = NULL;
		}

		window = log->window;
		cff_gui_logwindow_free(log->window, NULL, log);
		gtk_widget_destroy(window);

		if (logname != NULL)
		{
			/* Restore log-level */
			cff_gprop_set_str(kPropStrOptLoglevel, logname);
			g_free(logname);
		}

	} /* if (log->window != NULL) */

	cff_ffmpeg_file_close(video);

	thread = &gThread;

	if (thread->thread != NULL)
	{
		/* Terminate */
		thread->func = NULL;
		GuiMutexUnlock(&(thread->mutex));
		g_thread_join(thread->thread);
	}

#ifdef HAVE_CAIRO
	if (gCurrentVideoFrame.surfaceRGB24 != NULL)
	{
		cairo_surface_destroy(gCurrentVideoFrame.surfaceRGB24);
	}
#endif

	return FALSE;

} /* static gboolean cff_gui_file_quit_delevnt(GtkWidget*, GdkEvent*, GtkWidget**) */

static void cff_gui_file_quit(GtkWidget *w, GtkWidget **mbar)
{
	if (gWindata.mainwin != NULL)
	{
		if (cff_gui_file_quit_delevnt(w, NULL, mbar) == FALSE)
		{
			gtk_widget_destroy(gWindata.mainwin);
			gWindata.mainwin = NULL;
		}

	} /* if (gWindata.mainwin != NULL) */

} /* static void cff_gui_file_quit(GtkWidget*, GtkWidget**) */


/*
	-------------
	 Menu 'Edit'
	-------------
*/

static void cff_gui_edit_undo_redo(cff_video_t *video, int action)
{
	int retv;

	if (action == 0)
	{
		retv = cff_ffmpeg_video_undo(video);
	}

	else
	{
		retv = cff_ffmpeg_video_redo(video);
	}

	if (retv < 0)
	{
		cff_guitl_beep();
	}

	else
	{
		gboolean cut_add = TRUE;

		if (cff_ffmpeg_video_cut_find_current(video) >= 0)
		{
			cut_add = FALSE;
		}

		cff_gui_video_button_cut_set_sensitive(cut_add);
		cff_gui_widget_queue_draw_area(gDrawingAreaBar);
	}

} /* static void cff_gui_edit_undo_redo(cff_video_t*, int) */

static void cff_gui_edit_undo(GtkWidget *w, cff_video_t *video)
{
	cff_gui_edit_undo_redo(video, 0);

} /* static void cff_gui_edit_undo(GtkWidget*, cff_video_t*) */

static void cff_gui_edit_redo(GtkWidget *w, cff_video_t *video)
{
	cff_gui_edit_undo_redo(video, 1);

} /* static void cff_gui_edit_redo(GtkWidget*, cff_video_t*) */

static void cff_gui_edit_load(GtkWidget *w, GtkWidget **mbar)
{
	gint		n;
	gchar		*fname;
	cff_video_t	*video;

	video = &gVideo;

	if (video->dirty &&
		cff_guitl_dialog_warning_2btn(gettext_nop("Add list to existing cutpoints?"),
					gettext_nop("Cancel"),
					gettext_nop("Yes"),
					mbar) == GTK_RESPONSE_CANCEL)
	{
		return;
	}

	fname = cff_gui_file_selection_new_gettext(gettext_nop("Load Cut List"),
					NULL, mbar, kPropStrEditPath);
	if (fname == NULL)
	{
		return;
	}

	n = cff_ffmpeg_video_cut_list_load(&gVideo, fname);

	g_free(fname);

	if (n < 0)
	{
		cff_guitl_beep();
		cff_guitl_dialog_error(gettext_nop("Cannot load cut list!"), NULL);
	}

	else
	{
		int i;

		cff_gui_widget_queue_draw_area(gDrawingAreaBar);

		i = cff_ffmpeg_video_cut_find_current(video);

		cff_gui_video_button_cut_set_sensitive((i < 0) ? TRUE : FALSE);

		if (n > 0)
		{
			gchar *str = g_strdup_printf("%d %s",
				n, (n == 1) ?
				gettext("item was discarded!") :
					gettext("items were discarded!"));
			cff_guitl_dialog_warning(str, mbar);
			g_free(str);
		}
	}

} /* static void cff_gui_edit_load(GtkWidget*, GtkWidget**) */

static void cff_gui_edit_save(GtkWidget *w, GtkWidget **mbar)
{
	gint		retv;
	gchar		*fname, *dot, *str;
	cff_video_t	*video;

	video = &gVideo;

	if (video->fname == NULL)
	{
		return;
	}

	dot = strrchr(video->fname, '.');

	if (dot != NULL)
	{
		*dot = '\0';
	}

	str = g_strconcat(video->fname, ".cff", NULL);

	if (dot != NULL)
	{
		*dot = '.';
	}

	fname = cff_gui_file_selection_new_gettext(gettext_nop("Save Cut List"),
					str, mbar, kPropStrEditPath);
	g_free(str);

	if (fname == NULL)
	{
		return;
	}

#ifndef HAVE_GTKOVERWRITECONFIRMATION

	if (cff_gui_overwrite(fname, mbar) == FALSE)
	{
		g_free(fname);
		return;
	}
#endif

	/* Save cut list ... */

	retv = cff_ffmpeg_video_cut_list_save(&gVideo, fname);

	g_free(fname);

	if (retv < 0)
	{
		cff_guitl_beep();
		cff_guitl_dialog_error(gettext_nop("Cannot save cut list!"), NULL);
	}

} /* static void cff_gui_edit_save(GtkWidget*, GtkWidget**) */

static void cff_gui_edit_clear(GtkWidget *w, GtkWidget **mbar)
{
	cff_video_t *video = &gVideo;

	if (video->cut.num == 0 ||
		cff_guitl_dialog_warning_2btn(gettext_nop("Clear cut list?"),
					gettext_nop("Cancel"),
					gettext_nop("Yes"),
					mbar) == GTK_RESPONSE_CANCEL)
	{
		return;
	}

	cff_ffmpeg_video_cut_list_clear(video);

	cff_gui_video_button_cut_set_sensitive(TRUE);
	cff_gui_widget_queue_draw_area(gDrawingAreaBar);

} /* static void cff_gui_edit_clear(GtkWidget*, GtkWidget**) */

static void cff_gui_edit_add(GtkWidget *w, cff_video_t *video)
{
	if (cff_ffmpeg_video_cut_add(video) < 0)
	{
		cff_guitl_beep();
	}

	else
	{
		cff_gui_video_button_cut_set_sensitive(FALSE);
		cff_gui_widget_queue_draw_area(gDrawingAreaBar);
	}

} /* static void cff_gui_edit_add(GtkWidget*, cff_video_t*) */

static void cff_gui_edit_remove(GtkWidget *w, cff_video_t *video)
{
	if (cff_ffmpeg_video_cut_remove(video) < 0)
	{
		cff_guitl_beep();
	}

	else
	{
		cff_gui_video_button_cut_set_sensitive(TRUE);
		cff_gui_widget_queue_draw_area(gDrawingAreaBar);
	}

} /* static void cff_gui_edit_remove(GtkWidget*, cff_video_t*) */

static void cff_gui_edit_frame_display(GtkWidget *w, gpointer data)
{
	enum cff_VideoFrame	next;
	cff_frame_seek_t	frame_seek;

	next = GPOINTER_TO_INT(data);

	frame_seek.type		= kFrameSeekCutpoint;
	frame_seek.value.e_next	= next;
	cff_gui_video_frame(&frame_seek);

} /* static void cff_gui_edit_frame_display(GtkWidget*, gpointer) */

static void cff_gui_edit_program_open(GtkWidget *w, gpointer data)
{
	cff_video_t *video = &gVideo;

	if (video->dirty &&
		cff_guitl_dialog_warning_2btn(gettext_nop("Clear cut list?"),
			gettext_nop("Cancel"),
			gettext_nop("Yes"),
			&cff_gui_menu_bar[kMenuBarEdit]) == GTK_RESPONSE_CANCEL)
	{
		return;
	}

	cff_ffmpeg_video_cut_list_clear(video);

	cff_gui_save_video_prev_settings_free();
	cff_gui_video_button_set_sensitive(FALSE);
	cff_gui_widget_queue_draw_area(gDrawingAreaBar);

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) cff_gui_program_open, data, NULL);

} /* static void cff_gui_edit_program_open(GtkWidget*, gpointer) */


/*
	----------------
	 Menu 'Options'
	----------------
*/

static void cff_gui_opt_toggle_bool(GtkWidget *mi, gpointer data)
{
	static gboolean clicked = FALSE;

	int	active, prop;
	guint	n;

	if (clicked)
	{
		/* Already handled */
		clicked = FALSE;
		return;
	}

	active = GuiCheckMenuItemGetActive(GTK_CHECK_MENU_ITEM(mi));
	prop = GPOINTER_TO_INT(data);

	cff_gprop_set_bool(prop, active);

	n = prop - kPropBoolOptKeyFrame;

	if (n <= kPropBoolOptPFrame)
	{
		if (mi == cff_gui_checkmenu_frame_widget[n])
		{
			/* Menu clicked */
			mi = cff_gui_popup_checkmenu_frame_widget[n];
		}

		else
		{
			/* Popup-Menu clicked */
			mi = cff_gui_checkmenu_frame_widget[n];
		}

		/* Following function emits the activate-signal */
		clicked = TRUE;

		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi), active);
	}

} /* static void cff_gui_opt_toggle_bool(GtkWidget*, gpointer) */

static void cff_gui_opt_logwindow_menu_selected(GtkWidget *mi, gpointer data)
{
	gint		logindex;
	const gchar	*logname;

	logindex = GPOINTER_TO_INT(data);
	logname  = gtk_menu_item_get_label(GTK_MENU_ITEM(mi));

	if (logindex == 0)
	{
		cff_gui_logwindow_free(NULL, NULL, &gLogWindow);
		cff_gui_logwindow_set_title(gLogWindow.window, logname);
	}

	else
	{
		if (cff_gui_logwindow_open(logname) == TRUE)
		{
			cff_gprop_set_str(kPropStrOptLoglevel, logname);
		}
	}

} /* static void cff_gui_opt_logwindow_menu_selected(GtkWidget*, gpointer) */


/*
	-------------
	 Menu 'Help'
	-------------
*/

static void cff_gui_help_ffmpeg(GtkWidget *w, GtkWidget **mbar)
{
	int		i;
	char		*str;
	const char	*title;
	GtkWidget	*vbox;

	vbox	= GuiVBoxNew(FALSE, 2);
	i	= 0;

	do
	{
		switch (i)
		{
			case 0:
				str	= cff_ffmpeg_get_libs_info();
				title	= gettext_nop("Libs");
				break;
			case 1:
				str	= cff_ffmpeg_get_formats_info();
				title	= gettext_nop("Formats");
				break;
			case 2:
				str	= cff_ffmpeg_get_codecs_info();
				title	= gettext_nop("Codecs");
				break;
			default: /*case 3: */
				str	= cff_ffmpeg_get_bitstream_filters_info();
				title	= gettext_nop("Bitstream Filters");
				break;

		} /* switch (i) */

		cff_gui_ffmpeg_info_label(title, str, vbox);

		if (str)
		{
			/* malloc'ed */
			free(str);
		}
	}
	while (++i < 4);

	cff_gui_scrolled_info_window(gettext_nop("CutterFF: FFmpeg"),
				vbox, mbar, kPropIntHelpFFmpegWinWidth);

} /* static void cff_gui_help_ffmpeg(GtkWidget*, GtkWidget**) */

static void cff_gui_help_cutterff(GtkWidget *w, GtkWidget **mbar)
{
	static const gchar *cff_help_str[] =
	{
		gettext_nop("Navigation"),
		gettext_nop("Button '<' or Key 'Left' | Button '>' or Key 'Right'"),
		gettext_nop("Previous or next frame"),
		gettext_nop("Button '<<' or Key 'Down' | Button '>>' or Key 'Up'"),
		gettext_nop("Half second back or forward"),
		gettext_nop("Button '<<<' or Key 'Page Down' | Button '>>>' or Key 'Page Up'"),
		gettext_nop("Five seconds back or forward"),
		gettext_nop("Button '|<' or Key 'Home' | Button '>|' or Key 'End'"),
		gettext_nop("Previous or next cutpoint or start or end of video"),
		gettext_nop("With shift key or the right mouse button instead of the left"),
		gettext_nop("Button '<' or Key 'Left' | Button '>' or Key 'Right'"),
		gettext_nop("One minute back or forward"),
		gettext_nop("Button '<<' or Key 'Down' | Button '>>' or Key 'Up'"),
		gettext_nop("Five minutes back or forward"),
		gettext_nop("Button '<<<' or Key 'Page Down' | Button '>>>' or Key 'Page Up'"),
		gettext_nop("Ten minutes back or forward"),
		gettext_nop("Button '|<' or Key 'Home' | Button '>|' or Key 'End'"),
		gettext_nop("Start or end of video"),
		gettext_nop("A mouseclick on the position bar will jump to this postion."),
		gettext_nop("Note: the times above depends on the video and may differ!")
	};

	gint		i;
	gchar		*markup, **dst, *str[sizeof(cff_help_str) / sizeof(gchar*)];
	const gchar	**src;
	GtkWidget	*vbox, *label;

	vbox = GuiVBoxNew(FALSE, 2);

	label = gtk_label_new(NULL);
	GuiVBoxPackStart(vbox, label, TRUE, TRUE, 4);

	i = sizeof(cff_help_str) / sizeof(gchar*);

	src = cff_help_str;
	dst = str;

	do
	{
		*(dst++) = cff_charconv_utf8_gettext(*(src++));
	}
	while (--i > 0);

	markup = g_markup_printf_escaped(
		"<span size=\"xx-large\" weight=\"ultrabold\" stretch=\"ultraexpanded\"><i><u>%s</u></i></span>\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<u><b>%s</b></u>\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n\n"
		"%s",
		str[0],
		str[1],
		str[2],
		str[3],
		str[4],
		str[5],
		str[6],
		str[7],
		str[8],
		str[9],
		str[10],
		str[11],
		str[12],
		str[13],
		str[14],
		str[15],
		str[16],
		str[17],
		str[18],
		str[19]);

	gtk_label_set_markup(GTK_LABEL(label), markup);

	i = sizeof(cff_help_str) / sizeof(gchar*);

	src = cff_help_str;
	dst = str;

	do
	{
		gchar		*d;
		const gchar	*s;

		s = *(src++);
		d = *(dst++);
		CFF_CHARCONV_UTF8_FREE(d, s);
	}
	while (--i > 0);

	cff_gui_scrolled_info_window(gettext_nop("CutterFF: Help"),
				vbox, mbar, kPropIntHelpCutterFFWidth);
	
} /* static void cff_gui_help_cutterff(GtkWidget*, GtkWidget**) */

static void cff_gui_help_about(GtkWidget *w, GtkWidget **mbar)
{
	static const gchar cff_about_str_0[] =
			gettext_nop("A small programm for cuttting videos");
	static const gchar cff_about_str_1[] =
			gettext_nop("Distributed under GNU-GPL 2 or later");
	static const gchar cff_about_str_2[] = gettext_nop("Version");
	static const gchar cff_about_str_3[] = gettext_nop("Author:");

	/* main.c */
	extern const gchar *cutterff_version_str;

	GtkWidget	*window, *vbox, *hbox, *button, *frame, *label;
	gint		major, minor, micro;
	gchar		*markup, *str[4];

	vbox	= GuiVBoxNew(FALSE, 2);
	window	= cff_guitl_window_new_gettext(gettext_nop("CutterFF: About"), vbox, 4);
	cff_guitl_dialog_set_modal(window, gWindata.mainwin, mbar);
	gtk_window_set_resizable(GTK_WINDOW(window), FALSE);

	label = gtk_label_new(NULL);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, label);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 5);

	str[0] = cff_charconv_utf8_gettext(cff_about_str_0);
	str[1] = cff_charconv_utf8_gettext(cff_about_str_1);
	str[2] = cff_charconv_utf8_gettext(cff_about_str_2);
	str[3] = cff_charconv_utf8_gettext(cff_about_str_3);

#if GTK_MAJOR_VERSION == 2
	major = GTK_MAJOR_VERSION;
	minor = GTK_MINOR_VERSION;
	micro = GTK_MICRO_VERSION;
#else
	major = gtk_get_major_version();
	minor = gtk_get_minor_version();
	micro = gtk_get_micro_version();
#endif

	markup = g_markup_printf_escaped(
		"<span size=\"xx-large\" weight=\"ultrabold\" stretch=\"ultraexpanded\"><i><u>%s</u></i></span>\n\n"
		"<tt>%s\n%s <b>%s</b> %s <b>%s-%i.%i.%i</b></tt>\n\n"
		"%s\n\n"
		"%s %s (%u-bit)\n\n"
		"%s %s %s\n"
		"%s",

		"CutterFF",
		str[0],
		"using", "FFmpeg", "and", "GTK+", major, minor, micro,
		str[1],
		str[2], cutterff_version_str, (unsigned int) (8 * sizeof(char*)),
		str[3],
		/* Harald F&ouml;rster */
		"Harald F" "\xC3\xB6" "rster",
		"<harald_foerster@users.sourceforge.net>",
		"URL: https://cutterff.sourceforge.io/");

	gtk_label_set_markup(GTK_LABEL(label), markup);
	g_free(markup);

	CFF_CHARCONV_UTF8_FREE(str[3], cff_about_str_3);
	CFF_CHARCONV_UTF8_FREE(str[2], cff_about_str_2);
	CFF_CHARCONV_UTF8_FREE(str[1], cff_about_str_1);
	CFF_CHARCONV_UTF8_FREE(str[0], cff_about_str_0);

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	GuiMiscSetPadding(label, 15);

	hbox	= cff_guitl_hbutton_box_new(vbox, 8, GTK_BUTTONBOX_END);
	button	= cff_guitl_button_box_add_button(hbox,
					gettext_nop("OK"), NULL, NULL);
	GuiSignalConnectSwapped(button, "clicked", gtk_widget_destroy, window);

	gtk_widget_show_all(window);

} /* static void cff_gui_help_about(GtkWidget*, GtkWidget**) */


/*
	----------------
	 Menu Functions
	----------------
*/

void cff_gui_adjust_menu(GtkWidget *mni, gpointer data)
{
	guint		item;
	GtkWidget	**menu;
	cff_video_t	*video;
	gboolean	*ptr, menu_sensitive[kMenuItemsTotal],
			popup_sensitive[kPopupItemsTotal];

	menu = &cff_gui_menu_widget[kMenuItemsTotal];
	item = kMenuItemsTotal;

	video = data;

	if (video == NULL)
	{
		do
		{
			if (*(--menu) == mni)
			{
				break;
			}
		}
		while (--item > 0);

		if (item == 0)
		{
			return;
		}

		video = &gVideo;

	} /* if (video == NULL) */

	memset(menu_sensitive, '\0', sizeof(menu_sensitive));
	memset(popup_sensitive, '\0', sizeof(popup_sensitive));

	if (video->inFormatCtx != NULL)
	{
		menu_sensitive[kMenuFileClose] = TRUE;
		menu_sensitive[kMenuFileInfo]  = TRUE;
	}

	if (video->vframe.ic != NULL)
	{
		if (video->vframe.undo[0] != NULL)
		{
			menu_sensitive[kMenuEditUndo]   = TRUE;
			popup_sensitive[kPopupEditUndo] = TRUE;
		}

		if (video->vframe.redo[0] != NULL)
		{
			menu_sensitive[kMenuEditRedo]   = TRUE;
			popup_sensitive[kPopupEditRedo] = TRUE;
		}

		if (video->cut.list == NULL || video->cut.num != 1 ||
			video->cut.list[0].frame != video->vframe.dts_stop)
		{
			menu_sensitive[kMenuFileSave]  = TRUE;
			popup_sensitive[kPopupFileSave] = TRUE;
		}

		menu_sensitive[kMenuFileImage] = TRUE;

		menu_sensitive[kMenuEditLoad]  = TRUE;
		popup_sensitive[kPopupEditLoad] = TRUE;

		if (video->cut.num > 0)
		{
			menu_sensitive[kMenuEditSave]  = TRUE;
			menu_sensitive[kMenuEditClear] = TRUE;
			menu_sensitive[kMenuEditFrame] = TRUE;

			popup_sensitive[kPopupEditSave]  = TRUE;
			popup_sensitive[kPopupEditClear] = TRUE;
			popup_sensitive[kPopupEditFrame] = TRUE;

			if (cff_ffmpeg_video_cut_find_current(video) >= 0)
			{
				menu_sensitive[kMenuEditRemove] = TRUE;
			}
		}

		if (menu_sensitive[kMenuEditRemove] == FALSE)
		{
			menu_sensitive[kMenuEditAdd] = TRUE;
		}

	} /* if (video->ic != NULL) */

	if (video->inFormatCtx != NULL &&
			video->program.index != NULL &&
					video->program.num > 1)
	{
		menu_sensitive[kMenuEditProgram] = TRUE;
	}

	if (mni == cff_gui_popup_menu)
	{
		menu = &cff_gui_popup_widget[kPopupItemsTotal];
		ptr  = &popup_sensitive[kPopupItemsTotal];
		item = kPopupItemsTotal;
	}

	else
	{
		menu = &cff_gui_menu_widget[kMenuItemsTotal];
		ptr  = &menu_sensitive[kMenuItemsTotal];
		item = kMenuItemsTotal;
	}

	do
	{
		gtk_widget_set_sensitive(*(--menu), *(--ptr));
	}
	while (--item != 0);

} /* void cff_gui_adjust_menu(GtkWidget*, gpointer) */

static gboolean _cff_gui_destroy_submenu(GtkWidget *submenu)
{
	GList *children;

	while ((children = GuiMenuShellGetChildren(submenu)) != NULL)
	{
		gtk_widget_destroy(GTK_WIDGET(children->data));
	}

	return Gui_SOURCE_REMOVE;

} /* static gint _cff_gui_destroy_submenu(GtkWidget*) */

static void cff_gui_destroy_submenu(GtkWidget *submenu, gpointer data)
{
	/*
		Destroy a little bit later, else the
		'activate' - signal will not be emitted
	*/

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) _cff_gui_destroy_submenu, submenu, NULL);

} /* static void cff_gui_destroy_submenu(GtkWidget*, gpointer) */

static void cff_gui_build_frame_submenu(GtkWidget *submenu, gpointer data)
{
	unsigned int	i, current;
	cff_video_t	*video;

	video   = data;
	current = cff_ffmpeg_video_cut_find_current(video);

	i = 0;

	do
	{
		unsigned int	frame_number;
		GtkWidget	*mi;
		gchar		buf[64];

		frame_number = cff_ffmpeg_get_cut_frame_number(video, i);

		if (frame_number == (unsigned int) -1)
		{
			continue;
		}

		g_snprintf(buf, sizeof(buf), "%u", frame_number);

		mi = gtk_menu_item_new_with_label(buf);

		if (i == current)
		{
			gtk_widget_set_sensitive(mi, FALSE);
		}

		gtk_menu_shell_append(GTK_MENU_SHELL(submenu), mi);
		gtk_widget_show(mi);

		GuiSignalConnect(mi, "activate",
			cff_gui_edit_frame_display, GINT_TO_POINTER(i));
	}
	while (++i < video->cut.num);

} /* static void cff_gui_build_frame_submenu(GtkWidget*, gpointer) */

static void cff_gui_build_program_submenu(GtkWidget *submenu, gpointer data)
{
	unsigned int	i;
	cff_video_t	*video;

	video = data;

	i = 0;

	do
	{
		int		prgnum;
		const char	*name;
		GtkWidget	*mi;
		gchar		buf[64];

		name = cff_ffmpeg_get_program_name(video, i, &prgnum);

		if (name != NULL)
		{
			if (cff_tl_is_ascii_str(name) == FALSE)
			{
				if (strlen(name) >= sizeof(buf))
				{
					name = NULL;
				}

				else
				{
					strcpy(buf, name);
					name = cff_tl_replace_non_ascii_chars_in_str(buf);
				}
			}
		}

		if (name == NULL)
		{
			if (prgnum < 0)
			{
				prgnum = i + 1;
			}

			g_snprintf(buf, sizeof(buf), "%d", prgnum);
			name = buf;
		}

		if (i == video->program.selected)
		{
			mi = gtk_check_menu_item_new_with_label(name);
			gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi), TRUE);
		}

		else
		{
			mi = gtk_menu_item_new_with_label(name);
		}

		gtk_menu_shell_append(GTK_MENU_SHELL(submenu), mi);
		gtk_widget_show(mi);

		GuiSignalConnect(mi, "activate",
			cff_gui_edit_program_open, GINT_TO_POINTER(i));
	}
	while (++i < video->program.num);

} /* static void cff_gui_build_program_submenu(GtkWidget*, gpointer) */

static void cff_gui_build_loglevel_submenu(GtkWidget *submenu, gpointer data)
{
	int i, current;

	if (gLogWindow.window == NULL ||
		(current = cff_ffmpeg_get_current_loglevel_index()) < 0)
	{
		current = 0;
	}

	i = 0;

	for (;;)
	{
		GtkWidget	*mi;
		const gchar	*logname;

		logname = cff_ffmpeg_get_loglevel_name(i);

		if (logname == NULL)
		{
			break;
		}

		mi = gtk_menu_item_new_with_label(logname);

		if (i == current)
		{
			gtk_widget_set_sensitive(mi, FALSE);
		}

		gtk_menu_shell_append(GTK_MENU_SHELL(submenu), mi);
		gtk_widget_show(mi);

		GuiSignalConnect(mi, "activate",
			cff_gui_opt_logwindow_menu_selected, GINT_TO_POINTER(i));
		i++;

	} /* for (;;) */

} /* static void cff_gui_build_loglevel_submenu(GtkWidget*, gpointer) */

static gboolean cff_gui_mouse_button_pressed(GtkWidget *widget,
					GdkEvent *event, GtkWidget *popup_menu)
{
	GdkEventButton *bevent;

#if GTK_MAJOR_VERSION > 3

	bevent = (GdkEventButton *) event;

	if (bevent != NULL &&
		(bevent->type != GDK_BUTTON_PRESS || bevent->button != 3))
	{
		return FALSE;
	}

	gtk_menu_popup_at_pointer(GTK_MENU(popup_menu), event);
#else
	guint	button;
	guint32	time;

	bevent = (GdkEventButton *) event;

	if (bevent == NULL)
	{
		time	= GDK_CURRENT_TIME;
		button	= 0;
	}

	else
	{
		button = bevent->button;

		if (bevent->type != GDK_BUTTON_PRESS || button != 3)
		{
			return FALSE;
		}

		time = bevent->time;
	}

	gtk_menu_popup(GTK_MENU(popup_menu), NULL, NULL, NULL, NULL, button, time);

#endif /* #if GTK_MAJOR_VERSION > 3 .. #else */

	return TRUE;

} /* static gboolean cff_gui_mouse_button_pressed(GtkWidget*,
							GdkEvent*, GtkWidget*) */

static gboolean cff_gui_sf10_show_popup_menu(GtkWidget *widget, GtkWidget *popup_menu)
{
	return cff_gui_mouse_button_pressed(widget, NULL, popup_menu);

} /* static gboolean cff_gui_sf10_show_popup_menu(GtkWidget*, GtkWidget*) */

static void cff_gui_build_menu_item(const cff_guiMenuItem_t *item, GtkWidget *menu, GtkWidget *window)
{
	for (;;)
	{
		int		active, prop;
		GtkWidget	*mi, *submenu;

		switch (item->type)
		{
			case kMTypeItem:
			case kMTypeSubmenuCheckItem:
				mi = cff_guitl_menu_item_new_gettext(item->label);

				if (item->type == kMTypeItem)
				{
					break;
				}

				submenu = gtk_menu_new();
				gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), submenu);
				gtk_widget_realize(submenu);

				if (item->callback)
				{
					GuiSignalConnect(submenu, "show",
						item->callback, item->data);
					GuiSignalConnect(submenu, "hide",
						cff_gui_destroy_submenu, NULL);
				}

				break;

			case kMTypeCheckItem:
				mi = cff_guitl_check_menu_item_new_gettext(item->label);
				prop = GPOINTER_TO_INT(item->data);

				if (cff_gprop_get_bool(prop, &active) == FALSE)
				{
					if (prop == kPropBoolOptDelete ||
							prop == kPropBoolOptCursor ||
								prop == kPropBoolOptSaveConfig)
					{
						active = TRUE;
					}

					else
					{
						active = FALSE;
					}
				}

				gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi), active);
				break;

			case kMTypeSeparator:
				mi = GuiMenuSeparatorNew();
				break;

			default:
				return;
		}

		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
		gtk_widget_show(mi);

		if (item->accel_id != kGAccelKeyNone)
		{
			cff_gaccel_bind_widget(item->accel_id, item->label, mi, window);
		}

		if (item->callback && item->type != kMTypeSubmenuCheckItem)
		{
			GuiSignalConnect(mi, "activate", item->callback, item->data);
		}

		if (item->storage)
		{
			*(item->storage) = mi;
		}

		item++;

	} /* for (;;) */

} /* static void cff_gui_build_menu_item(const cff_guiMenuItem_t*, GtkWidget*, GtkWidget*) */

static GtkWidget *cff_gui_build_menu(GtkWidget *window)
{
	static const struct build_menu_array
	{
		const gchar		*label;
		const cff_guiMenuItem_t	*entries;
		gboolean		adjust_menu;

	} build_menu_array[] = {

		{gettext_nop("File"), cff_gui_menuFile, TRUE},
		{gettext_nop("Edit"), cff_gui_menuEdit, TRUE},
		{gettext_nop("Options"), cff_gui_menuOptions, FALSE},
		{gettext_nop("Help"), cff_gui_menuHelp, FALSE},
		{NULL}
	};

	GtkWidget *mbar, *mbb, **mbbptr, *menu;

	const struct build_menu_array *build_tab;

	mbar = gtk_menu_bar_new();

	cff_gui_signal_connect_key_or_mouse_button(mbar, kButtonPressEvent);

#if GTK_MAJOR_VERSION == 2
	GuiMenuBarDisableShadow(mbar);
#endif

	build_tab = build_menu_array;
	mbbptr = cff_gui_menu_bar;

	do
	{
		mbb = cff_guitl_menu_item_new_gettext(build_tab->label);
		GuiSignalConnect(mbb, "destroy", gtk_widget_destroyed, mbbptr);
		*(mbbptr++) = mbb;
		gtk_menu_shell_append(GTK_MENU_SHELL(mbar), mbb);

		if (build_tab->adjust_menu)
		{
			GuiSignalConnect(mbb, "activate",
					cff_gui_adjust_menu, &gVideo);
		}

		menu = gtk_menu_new();
		gtk_widget_realize(menu);

		gtk_menu_item_set_submenu(GTK_MENU_ITEM(mbb), menu);

		cff_gui_build_menu_item(build_tab->entries, menu, window);
	}
	while ((++build_tab)->label);


	/* Popup Menu */

	menu			= gtk_menu_new();
	cff_gui_popup_menu	= menu;

	GuiSignalConnect(menu, "show", cff_gui_adjust_menu, &gVideo);

	gtk_widget_realize(menu);

	cff_gui_build_menu_item(cff_gui_menuPopup, menu, NULL);

	gtk_widget_add_events(window, GDK_BUTTON_PRESS_MASK);

	GuiSignalConnect(window, "button-press-event",
				cff_gui_mouse_button_pressed, menu);
	GuiSignalConnect(window, "popup-menu",
				cff_gui_sf10_show_popup_menu, menu);

	return mbar;

} /* static GtkWidget *cff_gui_build_menu(GtkWidget*) */


/*
	------
	 INIT
	------
*/


static void cff_gui_noop(void)
{
	return;

} /* static void cff_gui_noop(void) */


#if ! defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2

static GdkGC *cff_gui_set_gdk_color(const GtkWidget *widget, const char *str)
{
	GdkGC    *context;
	GdkColor color;

	gdk_color_parse(str, &color);
	context	= gdk_gc_new((GdkDrawable *) widget->window);
	gdk_gc_set_rgb_fg_color(context, &color);

	return context;

} /* static GdkGC *cff_gui_set_gdk_color(const GtkWidget*, const char*) */

#endif /* ! defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2 */


static GtkWidget *cff_gui_hbuttonbox_new(GtkWidget *hbox)
{
	GtkWidget *bbox, *frame;

	bbox = GuiHBoxNew(TRUE, 3);
	gtk_container_set_border_width(GTK_CONTAINER(bbox), 3);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, bbox);
	GuiHBoxPackStart(hbox, frame, TRUE, FALSE, 3);

	return bbox;

} /* static GtkWidget *cff_gui_hbuttonbox_new(GtkWidget*) */

static gint cff_gui_get_scale_margin(GtkWidget *scale)
{
	gint width;

	gtk_widget_style_get(scale, "slider-length", &width, NULL);
	width /= 2;

	if (width <= 0)
	{
		/* Default values */
		width = 31 / 2;
	}

	return width;

} /* static gint cff_gui_get_scale_margin(GtkWidget*) */

void cff_gui_init(const gchar *fname, const char *loglevel)
{
	gint			i, width, new_width;
	gchar			*str;
	GtkWidget		*button, *drawbar, *frame, *menu_bar,
				*bbox, *dbox, *hbox, *sbox, *vbox, *v1box,
				*widget, *window;
	cff_slider_t		*slider;
	cff_video_frame_t	*current_frame;

#if GTK_MAJOR_VERSION == 2
#ifdef HAVE_CAIRO
	GdkColor	*color;
#else
	GdkGC		**gcontext;
#endif
#else
	GdkRGBA		*color;
#endif

#if GTK_MAJOR_VERSION == 3
	gboolean	minor_version_lt_20;
#endif

#ifndef BUILD_MINGW
	GdkPixbuf	*pixbuf;
#endif

	g_log_set_handler("Gdk", G_LOG_LEVEL_MASK, (GLogFunc) cff_gui_noop, NULL);

	/* Main window */

	gWindata.topwin = NULL;

	vbox = GuiVBoxNew(FALSE, 0);

	window = cff_guitl_window_new_gettext(NULL, vbox, 2);
	gWindata.mainwin = window;
	cff_guitl_window_setup(window, kPropIntMainWinWidth);

#ifdef BUILD_MINGW
	/* On Windows a new window seems always appear at top left */
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
#endif
	GuiSignalConnect(window, "delete-event",
				cff_gui_file_quit_delevnt, NULL);
	GuiSignalConnect(window, "destroy",
				gtk_main_quit, NULL);
	GuiSignalConnect(window, "screen-changed",
				cff_gui_mainwindow_cursor_new, &gCursor);

	cff_gui_signal_connect_key_or_mouse_button(window, kButtonPressEvent);
	cff_gui_signal_connect_key_or_mouse_button(window, kKeyPressEvent);

	gtk_widget_realize(window);

	cff_gui_mainwindow_cursor_new(window, NULL, &gCursor);

#ifndef BUILD_MINGW

	/* Application icon */

	pixbuf = gdk_pixbuf_new_from_xpm_data((const char **) cutterff_xpm);
	gtk_window_set_icon(GTK_WINDOW(window), pixbuf);
#endif


	/* Menu and popup */

	menu_bar = cff_gui_build_menu(window);
	GuiVBoxPackStart(vbox, menu_bar, FALSE, FALSE, 0);


#if GTK_MAJOR_VERSION == 3

	minor_version_lt_20 = FALSE;

	if (gtk_get_major_version() == 3 && gtk_get_minor_version() < 20)
	{
		minor_version_lt_20 = TRUE;

		cff_gui_css_load_from_data(menu_bar,
					"GtkMenuBar,"
					"menubar {"
					"  -GtkMenuBar-shadow-type: None;"
					"}",
					GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
	}

#endif /* GTK_MAJOR_VERSION == 3 */

	current_frame = &gCurrentVideoFrame;


	/* Video picture */

	widget = gtk_drawing_area_new();
	current_frame->drawingArea = widget;

	gtk_widget_set_size_request(widget, 400, 225);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, widget);
	GuiVBoxPackStart(vbox, frame, TRUE, TRUE, 5);
	GuiSignalConnect(widget, cff_expose_event,
				cff_gui_draw_frame, current_frame);
	GuiSignalConnect(widget, cff_configure_event,
				cff_gui_draw_frame_configure_cb, &gVideo);


	/* Info Bar */

	hbox = GuiHBoxNew(FALSE, 0);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 0);

	v1box = GuiVBoxNew(FALSE, 1);
	GuiHBoxPackStart(hbox, v1box, TRUE, TRUE, 8);

	hbox = GuiHBoxNew(TRUE, 2);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 1);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, hbox);
	GuiVBoxPackStart(v1box, frame, FALSE, FALSE, 2);

	widget = gtk_label_new(NULL);
	current_frame->info = GTK_LABEL(widget);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	widget = gtk_label_new(NULL);
	current_frame->time = GTK_LABEL(widget);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	cff_gui_frame_info(current_frame, &gVideo);


	/* Slider Bar */

	dbox = GuiVBoxNew(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(dbox), 5);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, dbox);
	GuiVBoxPackStart(v1box, frame, FALSE, FALSE, 2);

	drawbar = gtk_drawing_area_new();
	gDrawingAreaBar = drawbar;
	gtk_widget_add_events(drawbar, GDK_BUTTON_PRESS_MASK);

	cff_gui_signal_connect_key_or_mouse_button(drawbar, kButtonPressEvent);

	GuiSignalConnect(drawbar, "button-press-event",
					cff_gui_draw_bar_button, &gVideo);
	GuiSignalConnect(drawbar, cff_expose_event,
					cff_gui_draw_bar, &gVideo);

	slider = &gSlider;
	slider->scale = GuiHScaleNew(NULL);

	gtk_scale_set_draw_value(GTK_SCALE(slider->scale), FALSE);
	gtk_widget_set_direction(slider->scale, GTK_TEXT_DIR_LTR);

	gtk_range_set_restrict_to_fill_level(GTK_RANGE(slider->scale), FALSE);
	slider->adjust = gtk_range_get_adjustment(GTK_RANGE(slider->scale));


#if GTK_MAJOR_VERSION == 2

	width = cff_gui_get_scale_margin(slider->scale);

#elif GTK_MAJOR_VERSION == 3

	if (minor_version_lt_20)
	{
		width = cff_gui_get_scale_margin(slider->scale);
	}

	else
	{
#endif

#if GTK_MAJOR_VERSION >= 3

	/*
		Don't know how to get the style properties,
		so I create here my own scale slider!
		The center of the slider should match the
		drawbar (start, stop, cutpoints) above.
	*/

	width = 10;

	cff_gui_css_load_from_data(slider->scale,
				"scale,"
				"scale.horizontal {"
				"  min-height: 0px;"
				"  min-width:  0px;"
				"  border:  0px;"
				"  margin: 10px;"
				"  padding: 0px;"
				"}"
				"scale slider,"
				"scale.horizontal slider {"
				"  min-height: 20px;"
				"  min-width:  20px;"
				"}"
				"scale trough,"
				"scale.horizontal trough {"
				"  min-height: 6px;"
				"  border:  0px;"
				"  margin:  0px;"
				"  padding: 0px;"
				"}",
				GTK_STYLE_PROVIDER_PRIORITY_USER);

#if GTK_MAJOR_VERSION == 3
	} /* if (minor_version_lt_20) .. else */
#endif

#endif /* GTK_MAJOR_VERSION >= 3 */


	sbox = GuiHBoxNew(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(sbox), width / 2);

	gtk_widget_set_size_request(drawbar, -1, 16);
	GuiHBoxPackStart(sbox, drawbar, TRUE, TRUE, width / 2);

	GuiVBoxPackStart(dbox, sbox, FALSE, FALSE, 0);
	GuiVBoxPackStart(dbox, slider->scale, FALSE, FALSE, 0);

	cff_gui_slider_adjustment_configure(0);

	slider->button    = 0;
	slider->is_slider = FALSE;

	cff_gui_signal_connect_key_or_mouse_button(slider->scale,
							kButtonPressEvent);

	GuiSignalConnect(slider->scale, "button-press-event",
				cff_gui_slider_button, slider);
	GuiSignalConnect(slider->scale, "button-release-event",
				cff_gui_slider_button, slider);
	GuiSignalConnect(slider->scale, "change-value",
				cff_gui_video_frame_slider, slider);

	/*
		Colors for drawables
	
		kColorBlack,
		kColorDarkGrey,
		kColorGrey,
		kColorGreen,
		kColorRed
	*/

#if GTK_MAJOR_VERSION == 2

#ifdef HAVE_CAIRO
	color = &gColor[kColorBlack];
	gdk_color_parse("Black",     color++);
	gdk_color_parse("DarkGrey",  color++);
	gdk_color_parse("LightGrey", color++);
	gdk_color_parse("Green",     color++);
	gdk_color_parse("Red",       color);
#else
	gcontext = &gContext[kColorBlack];
	*(gcontext++) = cff_gui_set_gdk_color(window, "Black");
	*(gcontext++) = cff_gui_set_gdk_color(window, "DarkGrey");
	*(gcontext++) = cff_gui_set_gdk_color(window, "LightGrey");
	*(gcontext++) = cff_gui_set_gdk_color(window, "Green");
	*gcontext     = cff_gui_set_gdk_color(window, "Red");
#endif
#else
	color = &gRGBA[kColorBlack];
	gdk_rgba_parse(color++, "Black");
	gdk_rgba_parse(color++, "DarkGrey");
	gdk_rgba_parse(color++, "LightGrey");
	gdk_rgba_parse(color++, "Green");
	gdk_rgba_parse(color,   "Red");

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */


	/* Button Bar */

	hbox = GuiHBoxNew(FALSE, 0);
	GuiVBoxPackStart(v1box, hbox, FALSE, FALSE, 5);

	bbox = cff_gui_hbuttonbox_new(hbox);


	/* Goto frame */

	i = 0;

	do
	{
		button = gtk_button_new_with_label(gButtonVideoFrameLabel[i]);
		GuiHBoxPackStart(bbox, button, TRUE, TRUE, 1);
		GuiSignalConnect(button, "button-press-event",
					cff_gui_video_frame_button, GINT_TO_POINTER(i));
		gui_button_video_frame[i] = button;
	}
	while (++i < kButtonVideoFrameTotal);


	/* Padding */

	bbox = GuiHBoxNew(FALSE, 0);
	GuiHBoxPackStart(hbox, bbox, FALSE, FALSE, 5);

	/* Add or remove cutpoints */

	bbox = cff_gui_hbuttonbox_new(hbox);

	button = cff_guitl_hbox_add_button(bbox, "+",
				(GCallback) cff_gui_edit_add, &gVideo);
	gui_button_video_cut[kButtonVideoCutAdd] = button;

	button = cff_guitl_hbox_add_button(bbox, "-",
				(GCallback) cff_gui_edit_remove, &gVideo);
	gui_button_video_cut[kButtonVideoCutRemove] = button;


	/* Goto frame */

	GuiSignalConnect(window, "key-press-event",
				cff_gui_video_frame_key, NULL);


	/* Test filter extensions */

	cff_gui_file_filter_validate_prop(kPropStrFileVideoFilter);
	cff_gui_file_filter_validate_prop(kPropStrFileImageFilter);


	/* Open a thread */

	cff_gui_thread_new(NULL, NULL, NULL);


	/* Get the filename */

	if (fname == NULL || fname[0] == '\0')
	{
		str = NULL;
	}

	else
	{
		gchar *utf8 = cff_charconv_utf8_from_locale(fname, NULL);

#ifdef BUILD_MINGW
		str = (utf8 == fname) ? g_strdup(utf8) : utf8;
#else
		str = cff_charconv_utf8_to_filename(utf8);

		if (str == utf8 || str == fname)
		{
			if (str == utf8 && str == fname)
			{
				str = g_strdup(fname);
			}
		}

		else
		{
			CFF_CHARCONV_UTF8_FREE(utf8, fname);
		}

#endif /* #ifdef BUILD_MINGW .. #else */

	} /* if (fname == NULL || fname[0] == '\0') .. else */


	if (str != NULL)
	{
		/* Log window is on the top */
		gtk_widget_show_all(window);
	}


	/* Open the log window if level is not 'quiet' */

	if (loglevel == NULL)
	{
		gchar		*logname;
		const gchar	*logquiet;

		logquiet = cff_ffmpeg_get_loglevel_name(0);
		logname  = cff_gprop_get_str(kPropStrOptLoglevel);

		if (logname != NULL)
		{
			if (strcmp(logname, logquiet) != 0)
			{
				cff_gui_logwindow_open(logname);
			}

			g_free(logname);
		}

	} /* if (loglevel == NULL) */

	if (str == NULL)
	{
		/* Main window is on the top */
		gtk_widget_show_all(window);
		cff_gui_video_button_set_sensitive(FALSE);
		cff_guitl_window_set_title_filename(window, NULL);
	}

	else
	{
		/* Open the file */
		g_idle_add_full(G_PRIORITY_LOW,
			(GSourceFunc) cff_gui_video_open, str, NULL);
	}


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

#if GTK_MAJOR_VERSION == 3
	if (minor_version_lt_20)
	{
#endif

	new_width = cff_gui_get_scale_margin(slider->scale);

	if (new_width != width)
	{
		new_width /= 2;
		gtk_container_set_border_width(GTK_CONTAINER(sbox), new_width);
		GuiHBoxSetSpacing(sbox, new_width);
	}

#if GTK_MAJOR_VERSION == 3
	} /* if (minor_version_lt_20) */
#endif

#endif /* GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 */

} /* void cff_gui_init(const gchar*, const char*) */

/* EOF */
